#!/usr/bin/perl
##
## <bssp.pl> 
##
##  Add CB atom from original PDB file to DSSP file    
##      and assign CIS-conformation residue    
##
## <ADDING INFORMATION>
## (1) XYZ-coordinates of CB atoms
##     (if no CB atom, make pseudo CB atom).
## (2) Correctly assigned COMPND (name of molecules)
## (3) SSbond pairs 
##
##  [Usage] 
##    bssp.pl [dsspfile] [pdbfile] (-options)
##
## >> CAUTION << 
##  This script is only for "classical" dssp format.
##  If you use newer versions of DSSP (such as dsspcmbi),
##  make dsspfiles with "-c" option (dsspcmbi -c ..).
##
## [Major Updated History]
## * Apr 9,2003
##    Remove bugs in "MODRES".
## * August 27,2002
##    An option  for "MODRES".
## * April 27,2002
##    An option  "MSE".
## * Jul 12, 2001
##  An option '-rk [REMARK]' was added.
## * Jun 30, 2001
##  Variables (x,y,z) are changed to $P[0],$P[1],$P[2].  
##
## >> Variables (DSSP) <<
##  $DsspLine[$i] : All the Dssp Data
##  $CAD[$i][3]   : xyz of Calpha in DSSP :   $CAD[$i][3]
##  $INDEX{$numpdb}{$chain} : DSSP residue number from residue_string anc chainID;
## >> Variables (PDB) <<
##  $PdbResName[$i]   : 3-letter Residue Name
##  $C[$i][3], $CA[$i][3], $N[$i][3], $CB[$i][3], $CG[$i][3]
##    : xyz of C, CA, N, CB, SG in PDB : 
##


%CharAmino = ('ALA','A','ARG','R','ASN','N','ASP','D','CYS','C','GLN','Q',
          'GLU','E','GLY','G','HIS','H','ILE','I','LEU','L','LYS','K',
          'MET','M','PHE','F','PRO','P','SER','S','THR','T','TRP','W',
          'TYR','Y','VAL','V');

##############################
##### ARGUMENT HANDLING ######
##############################


$OPT{'SS'}    = 'T';
$OPT{'Omega'} = 'F';
$OPT{'rk'}    = "";
$OPT{'MSE'}   = "T";
$OPT{'MOD'}   = 'T';
$OPT{'abSS'}  = 'T';

$PDBDIR = "/DB/PDB";
$LastModDate = "Dec 2, 2003";

if (scalar(@ARGV)<2)
{ print "bssp.pl [dsspfile] [pdbfile] (-options)\n";
  print " coded by T.Kawabata. LastModDate : $LastModDate\n";
  print " <Options>\n";
  printf(" -SS    : output SSbond (TF) [%s]\n",$OPT{'SS'}); 
  printf(" -Omega : output Omega  (TF) [%s]\n",$OPT{'Omega'}); 
  printf(" -MSE   : Output 'm' for MSE (TF) [%s]\n",$OPT{'MSE'}); 
  printf(" -MOD   : Change MODRES to lowered original AA (ex. CSE->'c') [%s]\n",$OPT{'MOD'}); 
  printf(" -rk    : REMARK string [%s]\n",$OPT{'rk'}); 
  printf(" -abSS  : Change 'a','b','c'.. SS-bonded amino acids to 'C' [%s]\n",$OPT{'abSS'}); 
  exit(1); } 

$dsspfile = $dsspent =  $ARGV[0];
if (! -e $dsspfile) { $dsspfile = "/lab/DB1/DSSP/$dsspfile";}
$pdbfile   = $ARGV[1];

$k = 2; 
while ($k<scalar(@ARGV))
{
 if ($ARGV[$k]=~/^-/)
 { $OPT{substr($ARGV[$k],1)} = $ARGV[$k+1];
   ++$k; }
 ++$k;
} # while #

if ($pdbfile eq "")
{ $pdbdir = substr($dsspent,1,2);
  $pdbent = substr($dsspent,0,4);
  $pdbfile = "$PDBDIR/$pdbdir/pdb$pdbent.ent"; }

###############################
#### (1) READING DSSP FILE ####
###############################

open(D,$dsspfile) || die "Can't find dsspfile \"$dsspfile\"";
@d = split(/\//,$dsspfile);
$read_on = 0; $c = 0;
while (eof(D)!=1)
{
 $line = <D>;
 chop($line);
 if (length($line)<5) {$read_on = 0;}
 
 if ($read_on == 1)
  {
    $DsspLine[$c] = $line;
    $numpdb = substr($line,6,5);
  
    $aa = substr($line,13,1);
    if (($aa ne '!')&&($aa ne ' '))
    { 
     $chain  = substr($line,11,1);
     $INDEX{$numpdb}{$chain} = $c; 
     $ChainExist{$chain} = 1;
     $NUMPDB[$c] = $numpdb;
     $CHAIN[$c]  = $chain;
     $CAD[$c][0] = substr($line,107,7);  
     $CAD[$c][1] = substr($line,114,7);  
     $CAD[$c][2] = substr($line,121,7);  
     # printf("%s %s %s\n",$CAD[$c][0],$CAD[$c][1],$CAD[$c][2]);
    }  
    ++$c;  
 }

 if ($line=~/\#\s+RESIDUE AA STRUCTURE/) { $read_on = 1; }

 if ($read_on ==0)
 {
     if ($line=~/SECONDARY STRUCTURE DEFINITION/) 
    { ($head,$first_dssp,$tail) = split(/\*+\s+|\s+\*+/,$line); }
  elsif ($line=~/^REFERENCE/) {$ref_dssp = $line;}
  elsif ($line=~/^HEADER/)    {$header_dssp = $line;}
  elsif ($line=~/^COMPND/)    {$compnd_dssp = $line;}
  elsif ($line=~/^SOURCE/)    {$source_dssp = $line;}
  elsif ($line=~/^AUTHOR/)    {$author_dssp = $line;}
  elsif (length($line)>30)  
   {$other_dssp_line .= $line."\n";}
 }

} ## while ##
close(D);

@CHAINlist = keys(%ChainExist);
if (scalar(@CHAINlist) == 1) {
   $chain_dssp = $CHAINlist[0];} else {$chain_dssp = '-';}


##########################
## (2) READING PDB FILE ##
##########################

open(P,$pdbfile) || die "Can't find pdbfile \"$pdbfile\"\n";
$compndstr = $sourcestr = $author_pdb = "";   
while (eof(P)!=1)
{
 $line = <P>;
 if ($line=~/^COMPND/) {$compndstr  .= substr($line,10,62)."\n";} 
 if ($line=~/^SOURCE/) {$sourcestr  .= substr($line,10,62)."\n";} 
 if ($line=~/^AUTHOR/) {$author_pdb .= substr($line,10,62)."\n";} 
 if ($line=~/^MODRES/)
 {
  #MODRES 1AA1 KCX L  201  LYS  LYSINE NZ-CARBOXYLIC ACID
  #MODRES 1AA1 KCX B  201  LYS  LYSINE NZ-CARBOXYLIC ACID
  #MODRES 2AHJ CSD A  112  CYS  CYSTEINE SULFINIC ACID                             
  #MODRES 2AHJ CEA A  114  CYS  CYSTEINE SULFENIC ACID 
  #MODRES 1EUN PRO A  136                                <- nothing is done for this
  #MODRES 1914 CYS   4039  CYS  COVALENTLY LINKED TO BME <- nothing is done for this
  #MODRES 1DMT ASN A  144  ASN  GLYCOSYLATION SITE       <- nothing is done for this

  $modres = substr($line,12,3);
  $orires = substr($line,24,3);
 
  if (($CharAmino{$orires} ne "")&&($modres ne $orires))
  {
   $numpdb = substr($line,18,5);
   $chain  = substr($line,16,1);
   $osym = $CharAmino{$orires};
   $osym =~tr /A-Z/a-z/;
   if ($osym eq '') {$osym = 'x';} 
   $ind = $INDEX{$numpdb}{$chain}; 
   if ($ind ne "")
   { $ModResOrig[$ind] = $osym;  
     $ModResLine{$modres} = $line; }
  } 
 }
  
 
 if (($line=~/^ATOM/)||($line=~/^HETATM/)) 
  {
   $atom = substr($line,13,3);
   $chain = substr($line,21,1);
   
   if (($atom eq 'CB ')||($atom eq 'C  ')||($atom eq 'N  ')||
       ($atom eq 'CA ')||($atom eq 'SG ')) 
   { 
      $numpdb = substr($line,22,5); 
      $ind = $INDEX{$numpdb}{$chain};
      
      if ($ind ne "")
      { 
       $P[0] = substr($line,30,8);
       $P[1] = substr($line,38,8);
       $P[2] = substr($line,46,8);
       $PdbResName[$ind] = substr($line,17,3);
       if (($OPT{'MSE'} eq 'T') && ($PdbResName[$ind] eq 'MSE') && ($atom eq 'CA ')) 
        { ++$Nmse;}
       if (($OPT{'MOD'} eq 'T') && ($ModResOrig[$ind] ne '') && ($atom eq 'CA ')) 
        { ++$Nmod; push(@MODRNUM,$numpdb);}
       if (&DDIS(\@{$CAD[$ind]},\@P) < 100.0) ## exclude far atoms from DSSP C_alpha ## 
       { 
         if (($atom eq "CB ")&&($CB[$ind][0] eq "")) { &Equal_Vec(\@{$CB[$ind]},\@P); }
         if (($atom eq "N  ")&&($N[$ind][0] eq ""))  { &Equal_Vec(\@{$N[$ind]},\@P); }
         if (($atom eq "C  ")&&($C[$ind][0] eq ""))  { &Equal_Vec(\@{$C[$ind]},\@P); }
         if (($atom eq "CA ")&&($CA[$ind][0] eq "")) { &Equal_Vec(\@{$CA[$ind]},\@P); }
         if (($atom eq "SG ")&&($SG[$ind][0] eq "")) { &Equal_Vec(\@{$SG[$ind]},\@P); }
       } 
      }
    }
  }
} ## while ##
close(P);

&Translate_COMPND_SOURCE($compndstr,$sourcestr,$chain_dssp,$compnd_pdb,$source_pdb);


### Find SSbond ###
if ($OPT{'SS'} eq 'T') { &Find_SSbond(\@SSstr); }

#####################
## (3) OUTPUT BSSP ##
#####################

&Get_Date($date,$time);

## OUTPUT HEADERS ##

&Output_ConstantDotLine("**** $first_dssp **** $date [BSSP]",127);

print "$ref_dssp\n";
print "$header_dssp\n";

if ($compnd_pdb ne "")
{ &Output_ConstantDotLine("COMPND    $compnd_pdb",127); }
elsif ($compnd_dssp ne "") {print "$compnd_dssp\n"; }

if ($source_pdb ne "")
{ &Output_ConstantDotLine("SOURCE    $source_pdb",127); }
elsif ($source_dssp ne "") {print "$source_dssp\n"; }

if ($author_pdb ne "")
{ &Output_ConstantDotLine("AUTHOR    $author_pdb",127); }
elsif ($author_dssp ne "") {print "$author_dssp\n"; }

if ($OPT{'rk'} ne "")
{ &Output_ConstantDotLine("REMARK    $OPT{'rk'}",127); }

if (($OPT{'MSE'} eq 'T')&&($Nmse>0))
{ Output_ConstantDotLine("REMARK    NUMBER_OF_MSE(m) $Nmse",127); }

if (($OPT{'MOD'} eq 'T')&&($Nmod>0))
{ 
  $str = ""; foreach $a (@MODRNUM) {$str .= "$a ";}
  Output_ConstantDotLine("REMARK    NUMBER_OF_MODRES $Nmod \"$str\"",127); 
  foreach $m (keys(%ModResLine))
   { Output_ConstantDotLine("REMARK    $ModResLine{$m}",127); }
 }

print "$other_dssp_line";

if ($OPT{'SS'} eq 'T') 
 { foreach $x (@SSstr) 
   { &Output_ConstantDotLine($x,127); }
  }


## OUTPUT CONTENTS ##

print "#   RESIDUE AA STRUCTURE  BP1  BP2  ACC  N-H-->O   O-->H-N  N-H-->O  O-->H-N   X-CB  Y-CB  Z-CB  PHI  PSI     X-CA   Y-CA   Z-CA";
if ($OPT{'Omega'} eq 'T') { print "  OMEGA\n";} else {print "\n";}

$CisStr = "";

for ($i=0;$i<=$#DsspLine;++$i)
{ 
  $head  = substr($DsspLine[$i],0,75);
  $tail  = substr($DsspLine[$i],95,100);
  $rnum  = substr($DsspLine[$i],5,6);
  $chain = substr($DsspLine[$i],11,1);

  if (($OPT{'MSE'} eq 'T')&&($PdbResName[$i] eq 'MSE')){substr($head,13,1) = 'm'; }
  if (($OPT{'abSS'} eq 'T')&&($PdbResName[$i] eq 'CYS')){substr($head,13,1) = 'C'; }
  if (($OPT{'MOD'} eq 'T')&&($ModResOrig[$i] ne ''))   {substr($head,13,1) = $ModResOrig[$i]; }

  $CBtype = " ";
  if (($i>0)&&($C[$i-1][0] ne "")&&($N[$i][0] ne "")&&($CA[$i][0] ne ""))
  { $CisType = &CIS_or_not(\@{$CA[$i-1]},\@{$C[$i-1]},\@{$N[$i]},\@{$CA[$i]},$Omega);
    if ($CisType eq "C") {$CisStr .= sprintf("#CIS%s\n",substr($DsspLine[$i],0,14));}
   }
  else {$CisType = ' ';}

  if ($CB[$i][0] eq "")
  { 
   if (($N[$i][0] ne "")&&($C[$i][0] ne "")&&($CA[$i][0] ne ""))
    { &Make_CB(\@{$CA[$i]},\@{$N[$i]}, \@{$C[$i]},\@{$CB[$i]}); 
       $CBtype = 'P';  }
   elsif ($CA[$i][0] ne "") 
   { &Equal_Vec(\@{$CB[$i]},\@{$CA[$i]});
     $CBtype = 'A'; } 
   else 
    { $CBtype = '?';  
      &Equal_Vec(\@{$CB[$i]},\@{$CAD[$i]}); }
  } 

 $omegastr = "";
 if ($OPT{'Omega'} eq 'T') {$omegastr =  sprintf(" %6.1lf",$Omega); } 
  
 $cbstr = sprintf("%6.1f%6.1f%6.1f",$CB[$i][0],$CB[$i][1],$CB[$i][2]); 
 print"$head$CisType$CBtype$cbstr$tail$omegastr\n";
} 



###################
#### FUNCTIONS ####
###################

sub Make_CB{

 ## $_[0] : ref of \@CA (input)
 ## $_[1] : ref of \@N  (input)
 ## $_[2] : ref of \@C  (input)
 ## $_[3] : ref of \@CB (output)

 my($CAx) = $_[0]->[0]; my($CAy) = $_[0]->[1]; my($CAz) = $_[0]->[2];
 my($Nx)  = $_[1]->[0]; my($Ny)  = $_[1]->[1]; my($Nz)  = $_[1]->[2];
 my($Cx)  = $_[2]->[0]; my($Cy)  = $_[2]->[1]; my($Cz)  = $_[2]->[2];

 my($nx); my($ny); my($nz);
 my($cx); my($cy); my($cz);
 my($dx); my($dy); my($dz);
 my($rx); my($ry); my($rz);
 my($A); my($B); my($C);
 my($L);

 $R = 1.525;

 ## n = N - CA ## 
 $nx = $Nx - $CAx;  $ny = $Ny - $CAy;  $nz = $Nz - $CAz;
 $L = sqrt($nx*$nx+$ny*$ny+$nz*$nz);
 $nx /= $L; $ny /= $L; $nz /= $L; 
 
 ## c = C - CA ##  
 $cx = $Cx - $CAx;  $cy = $Cy - $CAy;  $cz = $Cz - $CAz;
 $L = sqrt($cx*$cx+$cy*$cy+$cz*$cz);
 $cx /= $L; $cy /= $L; $cz /= $L; 
 
 ## d = n x c ## 
 $dx = $ny*$cz - $nz*$cy;
 $dy = $nz*$cx - $nx*$cz;
 $dz = $nx*$cy - $ny*$cx; 
 $L = sqrt($dx*$dx+$dy*$dy+$dz*$dz);
 $dx /= $L; $dy /= $L; $dz /= $L; 

 ## r = A n + B c + C d  ##
 $A = $B = -0.5 *$R;
 $C = sqrt(2.0/3.0)*$R;

 $rx = $A *$nx + $B * $cx + $C * $dx; 
 $ry = $A *$ny + $B * $cy + $C * $dy; 
 $rz = $A *$nz + $B * $cz + $C * $dz; 

 $_[3]->[0] = $rx + $CAx; 
 $_[3]->[1] = $ry + $CAy; 
 $_[3]->[2] = $rz + $CAz; 

} ## end of  Make_CB()






sub CIS_or_not{
 ## $_[0] : ref of \@CA (input)
 ## $_[1] : ref of \@N  (input)
 ## $_[2] : ref of \@C  (input)
 ## $_[3] : ref of \@Ca (input)
 ## $_[4] : $Omega (output)

 my($CAx) = $_[0]->[0]; my($CAy) = $_[0]->[1]; my($CAz) = $_[0]->[2];
 my($Cx)  = $_[1]->[0]; my($Cy)  = $_[1]->[1]; my($Cz)  = $_[1]->[2];
 my($Nx)  = $_[2]->[0]; my($Ny)  = $_[2]->[1]; my($Nz)  = $_[2]->[2];
 my($Cax) = $_[3]->[0]; my($Cay) = $_[3]->[1]; my($Caz) = $_[3]->[2];

 ##        OMEGA(i-1) 
 ## CA - C  -----  N  - Ca
 ## i-1  i-1       i    i
 ##
 
 my($px); my($py); my($pz);
 my($qx); my($qy); my($qz);
 my($nx); my($ny); my($nz);
 my($mx); my($my); my($mz);
 my($Ln); my($Lm); my($cos); my($sin);
 my($dotprod);

 #printf("CA %.3f %.3f %.3f\n",$CAx,$CAy,$CAz);
 #printf("C  %.3f %.3f %.3f\n",$Cx,$Cy,$Cz);
 #printf("N  %.3f %.3f %.3f\n",$Nx,$Ny,$Nz);
 #printf("Ca %.3f %.3f %.3f\n",$Cax,$Cay,$Caz);
 
 ### n = (C-N) x (C-CA)  ##
 $px = $Nx -  $Cx; $py = $Ny -  $Cy; $pz = $Nz -  $Cz;
 $qx = $CAx - $Cx; $qy = $CAy - $Cy; $qz = $CAz - $Cz;
 $nx = $py*$qz - $pz*$qy;
 $ny = $pz*$qx - $px*$qz;
 $nz = $px*$qy - $py*$qx; 
 #print "px $px qx $qx nx $nx\n"; 

 ### m = (N-Ca) x (N-C)  ##
 $px = $Cax - $Nx; $py = $Cay - $Ny; $pz = $Caz - $Nz;
 $qx = $Cx  - $Nx; $qy = $Cy  - $Ny; $qz = $Cz  - $Nz;

 $mx = $py*$qz - $pz*$qy;
 $my = $pz*$qx - $px*$qz;
 $mz = $px*$qy - $py*$qx; 

 ###
 ### if (mn < 0) "TRANS", if (mn >=0) "CIS".   
 ###
 
 $dotprod = $nx*$mx + $ny*$my + $nz*$mz;

 if ($OPT{'Omega'} eq "T")
 { $Ln = sqrt($nx*$nx+$ny*$ny+$nz*$nz);
   $Lm = sqrt($mx*$mx+$my*$my+$mz*$mz);
   $cos = $dotprod/$Ln/$Lm;
   $sin = sqrt(1-$cos*$cos);
   $omega = atan2($sin,$cos) /3.14159 * 180.0; 
   $_[4] = $omega; 
   #printf("Omega %6.1lf DOTPROD %7.4lf\n",$omega,$dotprod); 
  }

 if ($dotprod < 0.0) {return(' ');} else {return('C');}

} ## end of  CIS_or_not() ##





sub Translate_COMPND_SOURCE{
 my($compnd_str)  = $_[0];
 my($source_str)  = $_[1];
 my($tar_chain)   = $_[2];
 ## $_[3] : return mol    name  
 ## $_[4] : return source name  

 my(@line);
 my($x); my($head); my($data); my($mol_id_cur);
 my(%CHAIN); my(%MOLECULE); my(%ORGANISM); my(@IDlist);
 my($first_molname);
 my($oldstyle) = 0; 
 
 $_[3] = $_[4] = "";

 ## FOR COMPND LINE ##
 @line = split(/\n/,$compnd_str);
 if (scalar(@line)==1) {$_[3] = $line[0];} ## FOR OLD STYLE ##
 foreach $x (@line)
 { $x =~s/^\s+//; $x =~s/\s+$//; $x =~s/;$//;
   ($head,$data) = split(/\s?:\s?/,$x);
   if ($head =~/MOL_ID/)    {$mol_id_cur = $data; push(@IDlist,$data);}
   if ($head =~/CHAIN/)     {$CHAIN{$mol_id_cur} = $data;}
   if ($head =~/MOLECULE/)  {$MOLECULE{$mol_id_cur} = $data;} }
 

 ## FOR SOURCE LINE ##
 @line = split(/\n/,$source_str);
 if (scalar(@line)==1) {$_[4] = $line[0];} ## FOR OLD STYLE ##
 foreach $x (@line)
 { $x =~s/^\s+//; $x =~s/\s+$//; $x =~s/;$//;
   ($head,$data) = split(/\s?:\s?/,$x);
   if ($head =~/MOL_ID/)    {$mol_id_cur = $data; }
   if ($head =~/ORGANISM/)  {$ORGANISM{$mol_id_cur} = $data;}
 }

 ## EXTRACT INFO OF TAR_CHAIN ##

 if (($tar_chain ne '')&& ($tar_chain ne '-') && ($tar_chain ne ' ')) 
 { 
  foreach $x (@IDlist)
  {
   if ($CHAIN{$x} =~ $tar_chain) 
    { $_[3] = $MOLECULE{$x}; $_[4] = $ORGANISM{$x};}
  }
 }

 if ($_[3] eq "") {$_[3] = $MOLECULE{$IDlist[0]};}
 if ($_[4] eq "") {$_[4] = $ORGANISM{$IDlist[0]};}

} ## end of Translate_COMPND_SOURCE() 



sub Get_Date{
 ## $_[0] : DATE (return)
 ## $_[1] : TIME (return)
 my(@Month) = ('JAN','FEB','MAR','APR','MAY','JUN','JUL','AUG','SEP','OCT','NOV','DEC');
 my(@date) = localtime(time());
 my($year)  = $date[5]+1900;
 my($month) = $Month[$date[4]];
 my($day)   = $date[3];
 $_[0] =  "$day-$month-$year";
 $_[1] =  "$date[2]:$date[1]:$date[0]";
}



sub Output_ConstantDotLine
{ my($line) = $_[0];
  my($Len)  = $_[1];
  my($len) = length($line);
  my($sp) = "";

  $line =~s/\s+$//;  
  if (length($line)>$Len) {$line = substr($line,0,$Len);} 
  if ($len<$Len) { $sp = ' ' x ($Len-length($line)); }
  print "$line$sp\.\n";  

} ## end of Output_ConstantDotLine() 






sub Find_SSbond{
 ## $_[0] : ref of @SSbond_string
 my($i); my($j); 
 my($dx); my($dy); my($dz); my($D); my($DD);
 my($str); my(@Slist) = ();
 my($Dthre) = 2.2; my($Nss) = 0;
 my($DDthre) = $Dthre*$Dthre;

 ## <Example of Distance>
 ## [5pti-]
 ##  SG   5 - SG  55 Dis 2.04
 ##  SG  14 - SG  38 Dis 2.03
 ##  SG  30 - SG  51 Dis 2.02
 ## [8fabA]
 ##  SG  22 - SG  87 Dis 2.04
 ##  SG 134 - SG 193 Dis 2.01
 ##
 ## <variables, which must be defined globally>
 ##  @DsspLine,@NUMPDB @CHAIN,@SG 

 for ($i=0;$i<=$#DsspLine;++$i)
 { if ($SG[$i][0] ne ""){ push(@Slist,$i); } } 

 @{$_[0]} = ();
 foreach $i (@Slist)
 {
  #print "i $i\n";
  foreach $j (@Slist)
  {
   #print "j $j\n";
   if ($j>$i)
   { 
    $dx = $SG[$i][0] - $SG[$j][0];
    $dy = $SG[$i][1] - $SG[$j][1];
    $dz = $SG[$i][2] - $SG[$j][2];
    $DD = $dx*$dx + $dy*$dy +$dz*$dz;
     if ($DD <= $DDthre)
     {
      $D = sqrt($DD);
      ++$Nss;
     
      $str = sprintf("SSBOND%4d CYS %1s %s   CYS %1s %s  DSG %.2lf A",
             $Nss,$CHAIN[$i],$NUMPDB[$i],$CHAIN[$j],$NUMPDB[$j],$D);
      push(@{$_[0]},$str);
     }
    }
  } ## $j
 } ## $i

} ## end of Find_SSbond()




sub DDIS{
 # $_[0] : ref of \@X
 # $_[1] : ref of \@Y

 ## return squared distance ##

 my($dx) = $_[0]->[0] - $_[1]->[0];
 my($dy) = $_[0]->[1] - $_[1]->[1];
 my($dz) = $_[0]->[2] - $_[1]->[2];

 return($dx*$dx+$dy*$dy*$dz*$dz);
} ## end of DDIS() ##



sub Equal_Vec{
 ## @X = @Y 
 ## $_[0]  ref of @X[3]
 ## $_[1]  ref of @Y[3]

 $_[0]->[0] = $_[1]->[0]; 
 $_[0]->[1] = $_[1]->[1]; 
 $_[0]->[2] = $_[1]->[2]; 

} ## end of Equal_Vec() ##
