#include <stdio.h>
#include "matrix_clapack.h"
#include "protein.h"


int searchpdbnum(char *name, char **Rnum, int size) {

	for (int j=0; j < size; j++)
		if (strcmp(name, Rnum[j]) == 0)	return j;

	printf("Error: %s  not found while searching for pdb number.\n");
	exit(0);
}

// Reads a match in format num1  num2 

// 12  23
// 13  -1
// 14  24
// 15  29


void readmatchingTMOrig( char *name, Vector<int> &match) {
        char *token, buff[200];
        int i, j;
        FILE* fp = fopen(name,"r"); 
        if(!fp) throw("Can't open file %s\n.", name);

        match = -1;
        while (fgets(buff, 100, fp)) {
                if (strlen(buff) >= 99) {
                printf("Warning: line longer than 100\n");
                printf("Increase line size or fix the matching format.\n");
                exit(1);
        }

        token = strtok(buff, " ");
        i=atoi(token); 
        token = strtok(NULL, " ");
        j = atoi(token);
        match(i) = j;
 
        }
    fclose(fp);

}

void readmatchingIdentity( char *name, Vector<int> &match) {
        char *token, buff[200];
        int i;
        int r=match.r;
        for (i=0;i<r;i++)
            match(i)=i;
}

void makematching(struct PROTEIN *prot1, struct PROTEIN *prot2, Vector<int> &match) 
{
int i, j, n1, n2;

n1=prot1->Naa;
n2=prot2->Naa;


match = -1;

for (i=1; i<=n1; i++)
 for (j=1; j<=n2; j++) {
    if( strcmp(prot1->Rnum[i],prot2->Rnum[j]) == 0 ) {
//      int tmp;
//      sscanf(prot1->Rnum[i],"%d",&tmp);
//      if (tmp > 16) 
        match(i)=j;
//      printf("%d %d \n",i,j);
    }
   }

}

// Reads a REF matching in .xml
void readmatching(struct PROTEIN *pA, struct PROTEIN *pB, char *name1, char *name2, Vector<int> &match) {
	char name[100] = "", buff[100];
	char *token;   
	int i, j;
	FILE* fp = fopen(name, "r");
	
    sprintf(name, "testSisy/REF/%s-%s.xml", name1, name2);        
	
	if(!fp) throw("Can't open file %s\n.", name);
    
	match = -1;
	while (fgets(buff, 100, fp)) {
		if (strlen(buff) >= 99) {
        	printf("Warning: line longer than 100\n");
        	printf("Increase line size.\n");
        	exit(1);
      	}
      	
      	token = strtok(buff, ": \t");
      	
      	if (strcmp(token, "<row><meq>") != 0)  continue;
      	
      	token = strtok(NULL, ": \t");
      	i = searchpdbnum(token, pA->Rnum, pA->Naa);
      	token = strtok(NULL, ": \t");
      	token = strtok(NULL, ": \t");
      	token = strtok(NULL, ": \t");
      	j = searchpdbnum(token, pB->Rnum, pB->Naa);
      	match(i) = j;  
	}

    fclose(fp);    
}



// Reads a REF matching in .xml but name2-name1 and load it backwards
void readmatchingback(struct PROTEIN *pA, struct PROTEIN *pB, char *name1, char *name2, Vector<int> &match) {
	char name[100] = "", buff[100];
	char *token;   
    int i, j;
    FILE* fp = fopen(name, "r");

	sprintf(name, "testSisy/REF/%s-%s.xml", name2, name1);        
	
	if(!fp) throw("Can't open file %s\n.", name);

    match = -1;
	while (fgets(buff, 100, fp)) {
		if (strlen(buff) >= 99) {
        	printf("Warning: line longer than 100\n");
        	printf("Increase line size.\n");
        	exit(1);
      	}
      	
      	token = strtok(buff, ": \t");
      
      	if (strcmp(token, "<row><meq>") != 0)  continue;
      
      	token = strtok(NULL, ": \t");
      	i = searchpdbnum(token, pB->Rnum, pB->Naa);
      	token = strtok(NULL, ": \t");
      	token = strtok(NULL, ": \t");
      	token = strtok(NULL, ": \t");
      	j = searchpdbnum(token, pA->Rnum, pA->Naa);
      	match(j) = i;  
   	}

    fclose(fp);    
}

float ProbTop(float score) {
  int ind ;
  float ProbF[51];
  
ProbF[0]=0;
ProbF[1]=0.000325278109033257;
ProbF[2]=0.000185279994821199;
ProbF[3]=0.000214964584595514;
ProbF[4]=0.000322971291894387;
ProbF[5]=0.000386926973625130;
ProbF[6]=0.000593319213603080;
ProbF[7]=0.000941951077903261;
ProbF[8]=0.00183338998932586;
ProbF[9]=0.00394437996677445;
ProbF[10]=0.00806569322261858;
ProbF[11]=0.0166468207167073;
ProbF[12]=0.0333223562401582;
ProbF[13]=0.0652411600073785;
ProbF[14]=0.124799331679338;
ProbF[15]=0.220816964579499;
ProbF[16]=0.355364699185388;
ProbF[17]=0.503868774850157;
ProbF[18]=0.64167444037161;
ProbF[19]=0.765585996923962;
ProbF[20]=0.83477032986369;
ProbF[21]=0.86414492445197;
ProbF[22]=0.906561009804603;
ProbF[23]=0.924896084473862;
ProbF[24]=0.923020679491624;
ProbF[25]=0.943893798179462;
ProbF[26]=0.93975001463199;
ProbF[27]=0.958711017906071;
ProbF[28]=0.967919760180003;
ProbF[29]=0.987466439051186;
ProbF[30]=0.986588889427008;
ProbF[31]=1;
ProbF[32]=1;
ProbF[33]=0.993639305328655;
ProbF[34]=0.992470268227436;
ProbF[35]=1;
ProbF[36]=1;
ProbF[37]=1;
ProbF[38]=1;
ProbF[39]=1;
ProbF[40]=1;
ProbF[41]=1;
ProbF[42]=1;
ProbF[43]=1;
ProbF[44]=1;
ProbF[45]=1;
ProbF[46]=1;
ProbF[47]=1;
ProbF[48]=1;
ProbF[49]=1;
ProbF[50]=1;

   ind = floor(score*50); 
   return ((ProbF[ind+1]-ProbF[ind])*(50*score-ind) + ProbF[ind]);
 
}

float ProbHom(float score) {
  int ind ;
  float ProbH[51];
ProbH[0]=0;
ProbH[1]=6.7568160183313e-05;
ProbH[2]=5.39100955212445e-05;
ProbH[3]=9.01000447286867e-05;
ProbH[4]=0.000148149438415461;
ProbH[5]=0.000236037067280919;
ProbH[6]=0.000418266060671194;
ProbH[7]=0.000771977250410268;
ProbH[8]=0.00135898038368965;
ProbH[9]=0.00279244305853413;
ProbH[10]=0.00480998934399694;
ProbH[11]=0.00802055024377918;
ProbH[12]=0.0143415897248563;
ProbH[13]=0.0248791853542652;
ProbH[14]=0.0456487238290666;
ProbH[15]=0.0707481696927087;
ProbH[16]=0.0982098225626232;
ProbH[17]=0.123114270020204;
ProbH[18]=0.142070597225834;
ProbH[19]=0.173832566315353;
ProbH[20]=0.196934229631177;
ProbH[21]=0.224186574515474;
ProbH[22]=0.257009082213106;
ProbH[23]=0.332249976494784;
ProbH[24]=0.412202449269025;
ProbH[25]=0.497344913337772;
ProbH[26]=0.61958587205732;
ProbH[27]=0.722753066058499;
ProbH[28]=0.783386202416814;
ProbH[29]=0.853407039237406;
ProbH[30]=0.868567351054528;
ProbH[31]=0.890579498125396;
ProbH[32]=0.89949958774336;
ProbH[33]=0.902911198705818;
ProbH[34]=0.914110099770055;
ProbH[35]=0.940647877938671;
ProbH[36]=0.906492115474211;
ProbH[37]=0.946444504006146;
ProbH[38]=0.95185937077628;
ProbH[39]=0.91433037173607;
ProbH[40]=0.919382287280062;
ProbH[41]=0.981112491788686;
ProbH[42]=0.953430875032616;
ProbH[43]=0.956300338303286;
ProbH[44]=0.963721756824878;
ProbH[45]=1;
ProbH[46]=0.95775792975913;
ProbH[47]=1;
ProbH[48]=1;
ProbH[49]=1;
ProbH[50]=1;
  ind = floor(score*50); 
   return ((ProbH[ind+1]-ProbH[ind])*(50*score-ind) + ProbH[ind]);
}

float PValueHom(float score) {
  int ind ;
  float PvalueH[51];
PvalueH[0]=1;
PvalueH[1]=0.997239434414794;
PvalueH[2]=0.973019440268869;
PvalueH[3]=0.910914431837788;
PvalueH[4]=0.813976024752483;
PvalueH[5]=0.690719086543841;
PvalueH[6]=0.555643201307961;
PvalueH[7]=0.423569320899764;
PvalueH[8]=0.307190644945954;
PvalueH[9]=0.214262995687719;
PvalueH[10]=0.144560368354702;
PvalueH[11]=0.0961341608415315;
PvalueH[12]=0.0640833770184017;
PvalueH[13]=0.0436338039193637;
PvalueH[14]=0.0306550665584049;
PvalueH[15]=0.0219081307145444;
PvalueH[16]=0.0157795428199111;
PvalueH[17]=0.0114694664799095;
PvalueH[18]=0.0082872090670909;
PvalueH[19]=0.00592187621443108;
PvalueH[20]=0.0041380922220734;
PvalueH[21]=0.002794741916333;
PvalueH[22]=0.00182567755954532;
PvalueH[23]=0.00119396666484761;
PvalueH[24]=0.000774479761824098;
PvalueH[25]=0.000492249414454693;
PvalueH[26]=0.000330187457176279;
PvalueH[27]=0.000235375699856877;
PvalueH[28]=0.000171984117928181;
PvalueH[29]=0.000136705324507047;
PvalueH[30]=0.000107490073705141;
PvalueH[31]=8.59920589640906e-05;
PvalueH[32]=7.00063556950958e-05;
PvalueH[33]=5.51231147205566e-05;
PvalueH[34]=4.40984917764453e-05;
PvalueH[35]=3.74837180099341e-05;
PvalueH[36]=2.81127885074506e-05;
PvalueH[37]=2.31517081826782e-05;
PvalueH[38]=1.92930901522503e-05;
PvalueH[39]=1.21270852385669e-05;
PvalueH[40]=6.6147737665112e-06;
PvalueH[41]=5.51231147205566e-06;
PvalueH[42]=3.3073868832556e-06;
PvalueH[43]=1.6536934416278e-06;
PvalueH[44]=5.5123114717226e-07;
PvalueH[45]=5.5123114717226e-07;
PvalueH[46]=0;
PvalueH[47]=0;
PvalueH[48]=0;
PvalueH[49]=0;
PvalueH[50]=0;
  ind = floor(score*50); 
   return ((PvalueH[ind+1]-PvalueH[ind])*(50*score-ind) + PvalueH[ind]);
}

float PValueTop(float score) {
  int ind ;
  float PvalueF[51];
PvalueF[0]=1;
PvalueF[1]=0.997126648112356;
PvalueF[2]=0.9719007903585;
PvalueF[3]=0.907219059613577;
PvalueF[4]=0.806294944601362;
PvalueF[5]=0.67797717406043;
PvalueF[6]=0.537421946639885;
PvalueF[7]=0.400081931609514;
PvalueF[8]=0.279381042887118;
PvalueF[9]=0.183584835412841;
PvalueF[10]=0.112727782514668;
PvalueF[11]=0.065017271137321;
PvalueF[12]=0.0352231228698356;
PvalueF[13]=0.0181571018502984;
PvalueF[14]=0.00908027459154848;
PvalueF[15]=0.00441373509315846;
PvalueF[16]=0.00213792089060694;
PvalueF[17]=0.00107729149956681;
PvalueF[18]=0.000573406355502737;
PvalueF[19]=0.000338987725197026;
PvalueF[20]=0.000214883744446892;
PvalueF[21]=0.000134445979145914;
PvalueF[22]=9.25034300961469e-05;
PvalueF[23]=6.60738786401049e-05;
PvalueF[24]=4.36662154491563e-05;
PvalueF[25]=2.98768842547092e-05;
PvalueF[26]=1.72366639931143e-05;
PvalueF[27]=9.76744292946474e-06;
PvalueF[28]=4.59644373151935e-06;
PvalueF[29]=2.87277733224123e-06;
PvalueF[30]=1.14911093285208e-06;
PvalueF[31]=1.14911093285208e-06;
PvalueF[32]=1.14911093285208e-06;
PvalueF[33]=5.74555466426041e-07;
PvalueF[34]=0;
PvalueF[35]=0;
PvalueF[36]=0;
PvalueF[37]=0;
PvalueF[38]=0;
PvalueF[39]=0;
PvalueF[40]=0;
PvalueF[41]=0;
PvalueF[42]=0;
PvalueF[43]=0;
PvalueF[44]=0;
PvalueF[45]=0;
PvalueF[46]=0;
PvalueF[47]=0;
PvalueF[48]=0;
PvalueF[49]=0;
PvalueF[50]=0;
  ind = floor(score*50); 
   return ((PvalueF[ind+1]-PvalueF[ind])*(50*score-ind) + PvalueF[ind]);
}

