/*
 
 <Matras.c>

 Matras version 1.2

 This program is for comparing protein tertiary structure
 ( "MATRAS" : MArkovian TRAnsition of Structure evolution)

 Matras contains three kind of structure feature comparison  

 (E) Environment score
 (D) Cbeta-Cbeta distance comparison 
 (S) Secondary structure element comparison 

 >> REFERENCE  <<

 Takeshi Kawabata and Ken Nishikawa.
 "Protein Structure Comparison Using the Markov Transition Model of Evolution".
  Proteins vol.41:108-122(2000).

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "IO.h"
#include "GotohDPdis.h"
#include "DiffStr.h"
#include "BetaTopo.h"
#include "ScoreENV.h"
#include "ScoreDIS.h"
#include "ScoreSSE.h"
#include "SecStrEle.h"
#include "RMS.h"
#include "OutAln.h"
#include "LlistIO.h"
#include "CalZsc.h"
#include "main.h"
#include "protein.h"
#include "LlistAlign.h"
#include "StrLibL.h"
#include "OutLibL.h"
#include "MergeSort.h"
#include "RegressL.h"
#include "BuildUpL.h"


/**** ALIGNMENT PARAMETER ****/ 
struct ALIGN_PARAMETER PAR;
char Version[256];
char LastModDate[256];

/*** PROTEIN ***/
static struct PROTEIN proA, proB;  

/*** DISTANCE MAP ***/
static struct MAP2D DmapA,DmapB;

char Vtype; /* if 'T', everything visible */

/**** BUFFER STRING FOR ALL-VS-ALL COMPARISON ****/
static char **buffer;
static FILE    *fpout;

/***** FUNCTIONS (LOCAL) ****/
int First_Appear();
void Output_Reference();
static void Evaluate_Similarity();
static void Output_for_Web();
static void Output_All_Alignments();


main(argc,argv)
 int argc;
 char **argv;
{ 
  char fname[128],OMolType,ORasType,ORasColorType,ORasColorNumType,OPdbType;
  char MODE[32],Otype,command[256],buff[256];
  char DisScTypeE,DisMesType,AliOutType,RankType,RnumPlainType;
  int AVAbunbo,AVAbunshi;  /* for all-vs-all division */
  FILE *fp;
  int Nsta,Nend;
  int i,j,k,m,p,q,ncomb,calok,r,L;
  int Nbuffer,Nlibrary,Nrank;
  float  Zthre,Rthre,DRMSthre,CRMSthre;
  double Sc;
  double A,B,S,R,Mean,SD; 
  char ARnumSta[6],ARnumEnd[6],BRnumSta[6],BRnumEnd[6];
  double Ga[3],Gb[3],Rmat[3][3];  /* <- For CRMS calculation */  
  char outfile[256];
  int MaxNaaA,MaxNaaB; 
  /**** Input / Input File *****/ 
  char QueFile[128],LibFile[128],proAfile[128],proBfile[128];
  char inssefile[128]; 
  char ssescfile[100],disscfile[100],envscfile[100];
  char inalifile[100],conalifile[100];
  char OutHead[128]; 
  
  /**** Output / Output File *****/ 
  char OAliType;      /* Output Alignment     ('T' or 'F') */ 
  char OSseAliType;   /* Output SSE Alignment ('T' or 'F') */ 
  char ORmatType;     /* Inverse Rotation Matrix ('T' or 'F') */ 
  char OheadStr[256];
  char outmtxfile[64],outclwfile[64],outonefile[64];
  
  /*** Structure of AlignNode (alignment) ***/ 
  struct AlignNode   AHead;       /* Residue-Residue Alignment  */ 
  struct AlignNode   SHead;       /* SSE Alignment */
  struct AlignNode   ALibHead;    /* For one-vs-library comparison  */ 
  struct AlignNode   *an,*sn,*bn; /* for temporary */

  /*** Structure of LibNode (structure library) ***/ 
  struct LibNode     LibHead;
  struct LibNode     *ln,*mn;     /* for temporary */

  sprintf(Version,"1.2");
  sprintf(LastModDate,"Apr 23, 2006");

  /**** DEFAULT PARAMETER ****/ 

  //Read_Matras_Environment_File();
  Set_AAnum(); 
  RnumPlainType = 'F';
  sprintf(ssescfile,"3U");
  sprintf(envscfile,"T10-3U.rom"); PAR.EnvScType = 'T';
  sprintf(disscfile,"3U");  PAR.DisScType = 'N';
  PAR.SseAli = 'T';   
  PAR.EnvAli =  PAR.DisAli = 'F';  /* Setting for PD*/
  PAR.AlgType  = 'L';
  PAR.GapExtE  = -6.0; PAR.GapExtD = -100.0;  
  PAR.GapOpenE = PAR.GapOpenD = 100;  /* Finaly set Beta = 5* GapExt */  
  PAR.GapExtS = -2.0;
  PAR.Nkeep = 35;  PAR.Nrepeat = 10; 
  PAR.Wmark = 4;  
  /*
  PAR.Wmark = 20;  
  */
 
  PAR.Rsse_sub = 10.0; 
  PAR.Rdis_sub = 10.0; 
  PAR.SseOffsetE = -1.0;
  PAR.SseOffsetD =  0.0; 
  PAR.ConOffset = 0.0;
  PAR.OtypeD = 'F';
  PAR.SeqAliType = 'F';

  PAR.MAXNEIGHBOURS = 46;    /* JR PD */
  PAR.squaredratio = 11.5;
  PAR.MAXDistExt = 9;        /* JR PD */
  PAR.OutPDBdeform = 0;

  Otype = '-';
  outfile[0] = '-'; DisScTypeE = '-'; DisMesType = 'C';
  ORasType = 'R'; 
  ORasColorType = 'R';
  ORasColorNumType   = 'C';
  OPdbType = 'B'; OMolType = '-'; 
  OAliType = OSseAliType = ORmatType = '-';
  PAR.SelfAvoidType = '-';  
  proAfile[0] = proBfile[0] = conalifile[0] = inalifile[0] = inssefile[0] = '\0'; 
  LibFile[0] = QueFile[0] = '\0'; 
  outmtxfile[0] = outclwfile[0] = outonefile[0] = '\0';   
  OheadStr[0] = '\0'; 
  ARnumSta[0] = ARnumEnd[0] = BRnumSta[0] = BRnumEnd[0] =  '\0';
  AliOutType = 'T'; RankType = 'D'; Zthre = 4.0; Rthre = 10.0;

  AVAbunshi = 0; AVAbunbo = 1; 
  DRMSthre = CRMSthre = -1.0;

   

  /**** OPTION EXPLANATION ****/ 
  
  if (argc>1) sprintf(MODE,"%s",argv[1]); 
  
  if ((argc<2)||(MODE[0]=='H')||(MODE[0]=='h')||(strcmp(MODE,"-h")==0))
   { printf("ProtDeform [P||H] <options>\n");
     printf(" a program for protein 3D structure comparison\n");
     printf(" uses Matras code by Takeshi Kawabata. ver %s. Last Modified : %s.\n",Version,LastModDate);
     printf(" [MODE] P:paii,  H: show help message in detail\n"); 
     printf(">>Typical Example<<\n");
     printf(" *Pairwise comparion\n");
     printf("   ProtDeform P -A 1mbd-.bssp -B 4hhbA.bssp\n"); 
/*
     printf(" *Libray-vs-query comparion\n");
     printf("   Matras L -Q 1mbd-.bssp -L listfile\n"); 
     printf(" *All-vs-all comparion\n");
     printf("   Matras A -L listfile\n"); 
*/
     printf(" -GAP : Gap Value [%f]\n",PAR.MAXDistExt);
     printf(" -NEIG : Neighbourhood size [%d]\n",PAR.MAXNEIGHBOURS);
     printf(">>Options<<\n"); 
     printf(" (-options selection [default])\n"); 
     printf(" -A : protein A file   -B :protein B file\n");
     printf(" -O : - Silent, 'N'ormal, 'P'ymol [%c]\n",Otype);
     printf(" -OA: Output Alignment (T|F) [%c]\n",OAliType);
     printf(" -ia: input alignment file [%s]   -is:input sse alignment file[%s]\n",
              inalifile,inssefile);    
     printf(" -ca: input constraint alignment file[%s]\n",conalifile);    
     printf(" -As: start resnum[%s] -Ae:end resnum[%s] (for proA)\n",ARnumSta,ARnumEnd);    
     printf(" -Bs: start resnum[%s] -Be:end resnum[%s] (for proB)\n",BRnumSta,BRnumEnd);    
     printf(" -Qs: start resnum[%s] -Qe:end resnum[%s] (for proQ)\n",ARnumSta,ARnumEnd);    
     printf(" -SO: SubOptimal Alignment(T|-)[%c]\n",PAR.SubOpt);    
     printf(" -SA: Self-Avoiding Alignment(T-)[%c]\n",PAR.SelfAvoidType);
     printf(">>Options for one-vs-library mode<<\n");    
     printf(" -Q : query bssp file   -L: library file\n");
     printf(" -zt: Threshold Zvalue [%f]\n",Zthre); 
     printf(" -rt: Threshold Rdis   [%f]\n",Rthre); 
     printf(" -AO: output alignment (T-) [%c]\n",AliOutType); 
     printf(" -R : ranking 'D'is_sc by Nc^2 fitZ 'S'sse_sc by rawZ 'R'dis [%c]\n",RankType); 
   
    /*** Display Detailed Option ***/
    if ((MODE[0]=='H')||(MODE[0]=='h')||(strcmp(MODE,"-h")==0))
    {
     printf(">> Alignment Parameters <<\n"); 
     printf(" -S : SSE         alignment type (T|F) [%c]\n",PAR.SseAli);
     printf(" -E : Environment alignment type (T|F) [%c]\n",PAR.EnvAli);
     printf(" -D : Distance    alignment type (T|F) [%c]\n",PAR.DisAli);
     printf(" -SQ: SequenceAlignment (T|F) [%c]\n",PAR.SeqAliType); 
     printf(" -sf: SseSc filename[3U]    -ef:  EnvSc file[T10-3U.rom]\n");    
     printf(" -df: DisSc filename[DR-3U.rom]\n");    
     printf(" -ge: ext gap for EnvAli[%.0lf] -gd: ext gap for DisAli[%.0lf] \n",
            PAR.GapExtE,PAR.GapExtD);    
     printf(" -gs: ext gap for SseAli[-2]\n");
     printf(" -ofe: SseOffset for Env  Alignment[%f]\n",PAR.SseOffsetE);
     printf(" -ofd: SseOffset for Dis  Alignment[%f]\n",PAR.SseOffsetD);
     printf("        if offset is set less than zero, then offset = |alpha+beta|\n");
     printf(" -nr: repeated number for DisSc align [%d]\n",PAR.Nrepeat); 
     printf(" -wm: Window_width of marking [%d]\n",PAR.Wmark); 
     printf(" -nk: keeping number for sse build [%d]\n",PAR.Nkeep);  
     printf(" -ed: Evaluation DisSc file\n");
     printf(" -dm: Distance Measure. 'C'rms 'D'rms [%c]\n",DisMesType);    
     printf(" -dr: DRMS filter threshold [%f]\n",DRMSthre);    
     printf(" -cr: CRMS filter threshold [%f]\n",CRMSthre);    
     printf(">> SubOptimal Alignment Parameter <<\n"); 
     printf(" -ss: Rsse for SubOptimal[%f]\n",PAR.Rsse_sub);    
     printf(" -ds: Rdis for SubOptimal[%f]\n",PAR.Rdis_sub);    
    
     printf(">> Output Options <<\n"); 
     printf(" -OA : Output Alignment (T|F) [%c]\n",OAliType); 
     printf(" -OS : Output SSE Alignment (T|F) [%c]\n",OSseAliType); 
     printf(" -OR : Output Rotation Matrix (T|F) [%c]\n",ORmatType); 
     printf(" -om : output 'M'olscript sctipt (T|F) [%c]\n",OMolType);    
     printf(" -or : output 'R'asmol script, 'E'ach rasmol script[%c]\n",ORasType);    
     printf(" -orc: output Rasmol/Molscript Color 'R'ainbow 'G'ray [%c]\n",ORasColorType);    
     printf(" -orn: output Rasmol/Molscript Color Number 'C':Ncomp, 'A':proA, 'B':proB [%c]\n",
                          ORasColorNumType);    
     printf(" -ox : output rotation matrix file[%s]\n",outmtxfile);    
     printf(" -ow : output ClustalW-style alignment[%s]\n",outclwfile);    
     printf(" -o1 : output oneline information[%s]\n",outonefile);    
     printf(" -oh : Output File Header\n");    
     printf(" -op : output superimposed str 'B'ssp,'P'db atoms [%c]\n",OPdbType);    
     printf(" -rp : output Rnum with Plain style[%c]\n",RnumPlainType);    

    }


    exit(1); }

/*
       if (MODE[0]=='P') {printf("#<MODE> P:PAIRWISE COMPARISON\n");}
  else if (MODE[0]=='L') {printf("#<MODE> L:ONE-VS-LIBRARY COMPARISON\n");}
  else if (MODE[0]=='A') {printf("#<MODE> A:ALL-VS-ALL COMPARISON\n");}
  else { fprintf(stderr,"#ERROR:MODE \"%s\" ?\n",MODE); exit(1);}
*/

  proA.chain = proB.chain = '-';


  /*************************/ 
  /**** READING OPTIONS ****/ 
  /*************************/ 
  
  k = 2;
  while (k<argc)
  {
    if (argv[k][0]=='-')
    { 
     L = strlen(argv[k]); 

          if ((argv[k][1]=='L')&&(L==2)) { ++k; sprintf(LibFile,"%s",argv[k]); 
                                           sprintf(proBfile,"%s",argv[k]);}
     else if ((argv[k][1]=='Q')&&(L==2)) { ++k; sprintf(QueFile,"%s",argv[k]); 
                                           sprintf(proAfile,"%s",argv[k]);}
     else if ((argv[k][1]=='A')&&(L==2)) { ++k; sprintf(proAfile,"%s",argv[k]); }
     else if ((argv[k][1]=='B')&&(L==2)) { ++k; sprintf(proBfile,"%s",argv[k]); }
     else if ((argv[k][1]=='V')&&(L==2)) { ++k; Vtype = argv[k][0]; }
     else if ((argv[k][1]=='O')&&(L==2)) { ++k; Otype = argv[k][0]; }
     else if ((argv[k][1]=='S')&&(L==2)) { ++k; PAR.SseAli = argv[k][0]; }
     else if ((argv[k][1]=='E')&&(L==2)) { ++k; PAR.EnvAli = argv[k][0]; }
     else if ((argv[k][1]=='D')&&(L==2)) { ++k; PAR.DisAli = argv[k][0]; }
     else if ((argv[k][1]=='R')&&(L==2)) { ++k; RankType = argv[k][0]; }
     else if ((argv[k][1]=='d')&&(argv[k][2]=='i'))
     { ++k;  
      L = strlen(argv[k]); r = First_Appear(argv[k],'/'); 
      if ((L>1)&&(r>0)) 
      { Get_Part_Of_Line(buff,argv[k],r+1,L-1); AVAbunbo = atoi(buff); 
        Get_Part_Of_Line(buff,argv[k],0,r-1);   AVAbunshi = atoi(buff);  } 
     }
     else if ((argv[k][1]=='G')&&(argv[k][2]=='A')&&(argv[k][3]=='P')) {++k; PAR.MAXDistExt = atof(argv[k]); }
     else if ((argv[k][1]=='N')&&(argv[k][2]=='E')&&(argv[k][3]=='I')&&(argv[k][4]=='G')) {++k; PAR.MAXNEIGHBOURS = atoi(argv[k]); }
     else if ((argv[k][1]=='S')&&(argv[k][2]=='O')) {++k; PAR.SubOpt = argv[k][0]; }
     else if ((argv[k][1]=='A')&&(argv[k][2]=='O')) {++k; AliOutType = argv[k][0]; }
     else if ((argv[k][1]=='S')&&(argv[k][2]=='A')) {++k; PAR.SelfAvoidType   = argv[k][0]; }
     else if ((argv[k][1]=='O')&&(argv[k][2]=='A')) {++k; OAliType    = argv[k][0]; }
     else if ((argv[k][1]=='O')&&(argv[k][2]=='S')) {++k; OSseAliType = argv[k][0]; }
     else if ((argv[k][1]=='O')&&(argv[k][2]=='R')) {++k; ORmatType = argv[k][0]; }
     else if ((argv[k][1]=='o')&&(argv[k][2]=='m')) {++k; OMolType = argv[k][0]; }
     else if ((argv[k][1]=='o')&&(argv[k][2]=='r')&&(L==3)) {++k; ORasType = argv[k][0]; }
     else if ((argv[k][1]=='o')&&(argv[k][2]=='r')&&(argv[k][3]=='c'))
                  {++k; ORasColorType = argv[k][0]; }
     else if ((argv[k][1]=='o')&&(argv[k][2]=='r')&&(argv[k][3]=='n')) 
                  {++k; ORasColorNumType = argv[k][0]; }
     else if ((argv[k][1]=='o')&&(argv[k][2]=='p')) {++k; OPdbType = argv[k][0]; }
     else if ((argv[k][1]=='a')&&(L==2)) { ++k; PAR.AlgType   = argv[k][0]; } 
     else if ((argv[k][1]=='s')&&(argv[k][2]=='f'))
      { ++k; sprintf(ssescfile,"%s",argv[k]);}
     else if ((argv[k][1]=='e')&&(argv[k][2]=='f'))
      { ++k; 
        if (strlen(argv[k])==1) 
        { PAR.EnvScType = argv[k][0]; envscfile[0] = '\0';
          if (PAR.EnvScType == 'B') {PAR.GapExtE = -2; }
        } 
        else sprintf(envscfile,"%s",argv[k]); }
     else if ((argv[k][1]=='d')&&(argv[k][2]=='f'))
     { ++k;
           if (strncmp(argv[k],"DALI",4)==0) { PAR.DisScType = 'D'; disscfile[0] = '\0';}  
      else if (strncmp(argv[k],"SSAP",4)==0) { PAR.DisScType = 'S'; disscfile[0] = '\0'; 
              Get_Part_Of_Line(buff,argv[k],4,strlen(argv[k])-1);
              PAR.Bssap = atof(buff); PAR.Assap = 1.0; } 
      else { sprintf(disscfile,"%s",argv[k]); PAR.DisScType = 'N';}
     }
     
     else if ((argv[k][1]=='e')&&(argv[k][2]=='d'))
     { ++k;
           if (strncmp(argv[k],"DALI",4)==0) { DisScTypeE = 'D';}  
      else if (strncmp(argv[k],"SSAP",4)==0) { DisScTypeE = 'S';  
              Get_Part_Of_Line(buff,argv[k],4,strlen(argv[k])-1);
              PAR.Bssap = atof(buff); PAR.Assap = 1.0; } 
      else { sprintf(disscfile,"%s",argv[k]); DisScTypeE = 'N';}
      }
     
    else if ((argv[k][1]=='g')&&(argv[k][2]=='e')) { ++k; PAR.GapExtE   = atof(argv[k]); }
    else if ((argv[k][1]=='g')&&(argv[k][2]=='d')) { ++k; PAR.GapExtD    = atof(argv[k]); }
    else if ((argv[k][1]=='g')&&(argv[k][2]=='s')) { ++k; PAR.GapExtE    = atof(argv[k]); }
    else if ((argv[k][1]=='r')&&(argv[k][2]=='t')) { ++k; Rthre = atof(argv[k]); }
    else if ((argv[k][1]=='z')&&(argv[k][2]=='t')) { ++k; Zthre = atof(argv[k]); }
    else if ((argv[k][1]=='d')&&(argv[k][2]=='m')) { ++k; DisMesType  = argv[k][0]; }
    else if ((argv[k][1]=='o')&&(argv[k][2]=='x')) { ++k; sprintf(outmtxfile,"%s",argv[k]);} 
    else if ((argv[k][1]=='o')&&(argv[k][2]=='h')) { ++k; sprintf(OheadStr,"%s",argv[k]);} 
    else if ((argv[k][1]=='o')&&(argv[k][2]=='w')) { ++k; sprintf(outclwfile,"%s",argv[k]);} 
    else if ((argv[k][1]=='o')&&(argv[k][2]=='1')) { ++k; sprintf(outonefile,"%s",argv[k]);} 
    else if ((argv[k][1]=='n')&&(argv[k][2]=='k')) { ++k; PAR.Nkeep  = atoi(argv[k]); }
    else if ((argv[k][1]=='n')&&(argv[k][2]=='r')) { ++k; PAR.Nrepeat = atoi(argv[k]); }
    else if ((argv[k][1]=='w')&&(argv[k][2]=='m')) { ++k; PAR.Wmark   = atoi(argv[k]); }
    else if ((argv[k][1]=='o')&&(argv[k][2]=='f')&&(argv[k][3]=='e')) 
                 { ++k; PAR.SseOffsetE  = atof(argv[k]); }
    else if ((argv[k][1]=='o')&&(argv[k][2]=='f')&&(argv[k][3]=='d')) 
                 { ++k; PAR.SseOffsetD  = atof(argv[k]); }
    else if ((argv[k][1]=='i')&&(argv[k][2]=='a')) 
         { ++k; sprintf(inalifile,"%s",argv[k]); 
           PAR.SubOpt = 'F';   /* Setting for PD for PDScore Only */
           PAR.SseAli = '-'; PAR.EnvAli = '-'; PAR.AlgType = 'L'; }
    else if ((argv[k][1]=='i')&&(argv[k][2]=='s')) 
         { ++k; sprintf(inssefile,"%s",argv[k]); }
    else if ((argv[k][1]=='c')&&(argv[k][2]=='a')) 
         { ++k; sprintf(conalifile,"%s",argv[k]); PAR.ConOffset = -PAR.GapExtD * 100;  }
    else if ((argv[k][1]=='c')&&(argv[k][2]=='o')) { ++k; PAR.ConOffset  = atof(argv[k]); }
    else if ((argv[k][1]=='A')&&(argv[k][2]=='s')) { ++k; sprintf(ARnumSta,"%s",argv[k]);} 
    else if ((argv[k][1]=='A')&&(argv[k][2]=='e')) { ++k; sprintf(ARnumEnd,"%s",argv[k]);} 
    else if ((argv[k][1]=='Q')&&(argv[k][2]=='s')) { ++k; sprintf(ARnumSta,"%s",argv[k]);} 
    else if ((argv[k][1]=='Q')&&(argv[k][2]=='e')) { ++k; sprintf(ARnumEnd,"%s",argv[k]);} 
    else if ((argv[k][1]=='B')&&(argv[k][2]=='s')) { ++k; sprintf(BRnumSta,"%s",argv[k]);} 
    else if ((argv[k][1]=='B')&&(argv[k][2]=='e')) { ++k; sprintf(BRnumEnd,"%s",argv[k]);} 
    else if ((argv[k][1]=='s')&&(argv[k][2]=='s')) { ++k; PAR.Rsse_sub = atof(argv[k]);} 
    else if ((argv[k][1]=='d')&&(argv[k][2]=='s')) { ++k; PAR.Rdis_sub = atof(argv[k]);} 
    else if ((argv[k][1]=='d')&&(argv[k][2]=='r')) { ++k; DRMSthre = atof(argv[k]);} 
    else if ((argv[k][1]=='c')&&(argv[k][2]=='r')) { ++k; CRMSthre = atof(argv[k]);} 
    else if ((argv[k][1]=='r')&&(argv[k][2]=='p')) { ++k; RnumPlainType = argv[k][0]; }
    else if ((argv[k][1]=='S')&&(argv[k][2]=='Q')) { ++k; PAR.SeqAliType = argv[k][0];
      if (PAR.SeqAliType == 'T')
      {
        printf("#SEQUENCE ALIGNMENT MODE\n");
        PAR.SseAli = 'F'; PAR.EnvAli = 'T'; PAR.DisAli = 'F';
        PAR.EnvScType = 'B'; /* BLOSUM62 */ 
        sprintf(envscfile,"");
        PAR.GapExtE = -1.0; PAR.GapOpenD  = -11.0; /* Blast Standard */ 
       }
     } 
 
    else {printf("#ERROR:Can't understand option %s\n",argv[k]); exit(1);}

    } /* if '-' */
 
     ++k;

  } /* while k */



  /**** READING SIMIRARITY SCORES ****/ 
//  if (strlen(envscfile)>0)     Read_Env_Score(envscfile); 
 // if (PAR.DisScType == 'N')    Read_All_Dis_Score(disscfile); 
  /* if (PAR.SseAli    == 'T')    Read_All_SSE_Score(ssescfile);
 */
 
  if (PAR.GapOpenE > 1) PAR.GapOpenE = PAR.GapExtE   * 5.0;    
  if (PAR.GapOpenD > 1) PAR.GapOpenD = PAR.GapExtD * 5.0;  

  if (PAR.SseAli == 'T') 
   { if (PAR.SseOffsetE < 0.0) PAR.SseOffsetE  = (float)fabs(PAR.GapOpenE); 
     if (PAR.SseOffsetD < 0.0) PAR.SseOffsetD  = (float)fabs(PAR.GapOpenD); } 
   else PAR.SseOffsetE = PAR.SseOffsetD = 0.0;

  if (PAR.SelfAvoidType == 'T') 
     PAR.NonSseOffsetE  = -PAR.SseOffsetE;  else PAR.NonSseOffsetE = 0;

  PAR.NonSseOffsetD = 0;

  /*** OUTPUT PARAMETER INFORMATION ***/
  
  if (Otype=='W') OPdbType = '-';
  if (Otype=='P') PAR.OutPDBdeform = 1;
  /* if (Otype=='-')*/  OPdbType = ORasType = OMolType = '-';
 if (Otype != '-') { /* JR */
  Output_Reference();
  PAR.OutAlignment = 1;
  sprintf(PAR.COMMAND,""); 
  for (i=0;i<argc;++i)  { strcat(PAR.COMMAND,argv[i]); strcat(PAR.COMMAND," ");} 
 
 
  printf("#<COMMAND> \"%s\"\n",PAR.COMMAND);
  printf("#<DATE> \"%s\"\n",Get_Date_String());
 
  printf("#<PARAMETERS>\n");
  if (MODE[0] == 'P')
   printf("# ProAFile \"%s\" ProBFile \"%s\"\n",proAfile,proBfile);
  else if (MODE[0] == 'L')
  printf("# QueFile \"%s\" LibFile \"%s\"\n",MODE,QueFile,LibFile);
  else if (MODE[0] == 'A')
  printf("# LibFile \"%s\"\n",LibFile);
  printf("# SseAliType %c EnvAliType %c AlgType %c\n",
            PAR.SseAli,PAR.EnvAli,PAR.AlgType);
 }
else PAR.OutAlignment = 0;
/*  printf("# ssescfile \"%s\" envscfile \"%s\" DisSc %c disscfile \"%s\" DisScE %c EnvType %c Nenvstate %d\n",
           ssescfile,envscfile,PAR.DisScType,disscfile,DisScTypeE,PAR.EnvScType,PAR.Nenvstate);
  printf("# GapExtE %.1lf GapExtD %.1lf Nkeep %d Nrep %d SseOffsetD %d\n",
          PAR.GapExtE,PAR.GapExtD,PAR.Nkeep,PAR.Nrepeat,PAR.SseOffsetD);
*/
  if ((PAR.SseAli == 'T')&&(PAR.EnvAli == 'X')) 
  { PAR.SseOffsetE    =  (int)fabs(PAR.GapExtE)*10; 
    PAR.NonSseOffsetE = -(int)fabs(PAR.GapExtE)*10;  }

  /******************************/
  /*  [P] PAIRWISE COMPARISON   */
  /******************************/


  if (MODE[0]=='P')  
  {
    /** READ PROTEIN A **/ 
    sprintf(proA.pname,"%s",proAfile); 
    Read_PDB_File(&proA,proA.pname); 
 //   Read_BSSP_File_LinearList(&proA,proA.pname,'A','W',ARnumSta,ARnumEnd); 

    if (Vtype == 'T') Output_Data_Horizontal(&proA); 
//    Malloc_MAP2D(&DmapA,proA.Naa);  
 
    /** READ PROTEIN B **/ 
    sprintf(proB.pname,"%s",proBfile); 
    Read_PDB_File(&proB,proB.pname);
//  Read_BSSP_File_LinearList(&proB,proB.pname,'A','W',BRnumSta,BRnumEnd); 
    
    if (Vtype == 'T') Output_Data_Horizontal(&proB); 
 //   Malloc_MAP2D(&DmapB,proB.Naa);  

    /*** Cal Distance Map **
    if (PAR.DisScType == 'D')
    { Cal_DmapCA(&proA,&DmapA); Cal_DmapCA(&proB,&DmapB); } 
    else 
    { Cal_DmapCB(&proA,&DmapA); Cal_DmapCB(&proB,&DmapB); }
 
    proA.SelfSdis = SelfScoreDistanceS(&proA,&DmapA);
    proB.SelfSdis = SelfScoreDistanceS(&proB,&DmapB);
    proA.SelfSsse = SelfScoreSSE(&proA);
    proB.SelfSsse = SelfScoreSSE(&proB);
   */
 
    /** Malloc Alignment Variables * 
    Malloc_GotohDP_Matrix(proA.Naa,proB.Naa);
   */ if (conalifile[0] != '\0')
    { if (PAR.ConOffset > 0.0) 
        Read_CASP_style_Constraint_Alignment(conalifile,&proA,&proB);  }

   /**** ALIGNMENT *****/
  // Pairwise_3D_Alignment(&proA,&proB,&DmapA,&DmapB,&AHead,&SHead,inalifile,inssefile);

   /* JR sn = SHead.next;   For SO SSE */
   
   mainPD(&proA, &proB, AHead.next, inalifile ); /* JR ProtDeform */
   /*** OUTPUT EACH ALIGNMENT ***/
 
exit(0);
 
   an = &AHead;
   while (an->next != NULL)
   {
    an = an->next; 
    sn = an->sn;     /* JR for SO SSE */
    if (DRMSthre>=0.0) DRMS_Filter(&proA,&proB,an,&DmapA,&DmapB,DRMSthre); 
    if (CRMSthre>=0.0) CRMS_Filter(&proA,&proB,an,'A',CRMSthre); 
    sprintf(OutHead,"%s%d",OheadStr,an->num);
   
   /*** Evaluate Similarity Score ***/
    PAR.EnvScType = 'B';
   /* JR for PD  an->sim->CRMS  =  CRMS(&proA,&proB,an,Ga,Gb,Rmat,'A','-'); 
    
    Evaluate_Similarity(&proA,&proB,&DmapA,&DmapB,an);
  */

    /* JR for PD   Output_Alignment_Information("-",&proA,&proB,an); */
    if (Otype != '-') Output_Path_Horizontal(&proA,&proB,an,"-");  

   
   /*** OUTPUT FOR WEB ***/ 
   if (Otype == 'W') Output_for_Web(&proA,&proB,an,OutHead,RnumPlainType);

   if (outclwfile[0]!='\0') 
    Output_Path_Horizontal_ClustalW(&proA,&proB,an,outclwfile);

   if (outonefile[0] != '\0')
    Output_Oneline_Information(outonefile,&proA,&proB,an);
  
   if (OAliType == 'T') 
    { sprintf(outfile,"%s.ali",OutHead);
      Write_Alignment_CASP_Style(&proA,&proB,an,outfile,RnumPlainType,Ga,Gb,Rmat); }

   if (OSseAliType == 'T')
    { sprintf(outfile,"%s.sal",OutHead);
      Write_SSE_Alignment(&proA,&proB,sn,outfile); 
      sprintf(outfile,"%s.salras",OutHead);
      Write_SSE_Alignment_Rasmol(&proA,&proB,sn,outfile); 
      sprintf(outfile,"%s.csalras",OutHead);
      Write_SSE_Alignment_Rasmol_Color(&proA,&proB,sn,outfile); 
      Output_SSE_Alignment_Horizontal("-",&proA,&proB,sn); 
    }

   if ((ORmatType == 'T')||(OPdbType == 'B')||(OPdbType=='P'))
    { an->sim->CRMS =  CRMS(&proA,&proB,an,Ga,Gb,Rmat,'A','-');
      sprintf(outfile,"%s.mat",OutHead);
      Write_Rotation_Matrix(outfile,&proA,&proB,an->Nali,Ga,Gb,Rmat,an->sim->CRMS); }
 
   /*** OUTPUT RASMOL SCRIPT ***/
   if (ORasType !='-') 
    { 
     sprintf(outfile,"%s.ras",OutHead);
     Write_Rasmol_Color_Alignment(outfile,&proA,&proB,an,ORasColorType,ORasColorNumType);
    }


   if (ORasType == 'E')
    { 
     Write_Rasmol_Color_Alignment_Each_Protein("A.ras",&proA,&proB,'A',an,ORasColorNumType,ORasColorType);
     Write_Rasmol_Color_Alignment_Each_Protein("B.ras",&proA,&proB,'B',an,ORasColorNumType,ORasColorType);
    } 
   if (OMolType !='-')
    {
    /*
       Write_Molscript_Alignment_Each_Protein("A.in",&proA,&proB,an,'X','a');
       Write_Molscript_Alignment_Each_Protein("B.in",&proB,&proA,an,'Y','a');
       Write_Molscript_Alignment_Each_Protein("Aa.in",&proA,&proB,an,'X','A');
       Write_Molscript_Alignment_Each_Protein("Ba.in",&proB,&proA,an,'Y','B'); 
    */
     if (ORasColorNumType == 'C')
      { Write_Molscript_Alignment_Each_Protein("A.in",&proA,&proB,'A',an,'A','a',ORasColorType);
        Write_Molscript_Alignment_Each_Protein("B.in",&proA,&proB,'B',an,'B','a',ORasColorType); }
     if (ORasColorNumType == 'A')
      { Write_Molscript_Alignment_Each_Protein("A.in",&proA,&proB,'A',an,'A','A',ORasColorType);
        Write_Molscript_Alignment_Each_Protein("B.in",&proA,&proB,'B',an,'B','A',ORasColorType); }
     if (ORasColorNumType == 'B')
      { Write_Molscript_Alignment_Each_Protein("A.in",&proA,&proB,'A',an,'A','B',ORasColorType);
        Write_Molscript_Alignment_Each_Protein("B.in",&proA,&proB,'B',an,'B','B',ORasColorType); }
    } 

   /*** OUTPUT SUPERIMPOSED STRUCTURE ***/
    if (OPdbType == 'B')
     { sprintf(command,"perl %s/rotSTR.pl %s.mat %s -nc A -S B > %s.pdb",
             BASE_DIR,OutHead,proAfile,OutHead);
       printf("#%s\n",command); system(command); 
       sprintf(command,"perl %s/rotSTR.pl - %s -nc B -S B >> %s.pdb",BASE_DIR,proBfile,OutHead);
       printf("#%s\n",command); system(command); } 
   
    if (OPdbType == 'P')
     { sprintf(command,"perl %s/rotSTR.pl %s.mat %s -nc A -S P -ch %c -H W > %s.pdb",
           BASE_DIR,OutHead,proAfile,proA.chain,OutHead);
       printf("%s\n",command); system(command); 
       sprintf(command,"perl %s/rotSTR.pl - %s -nc B -S P -ch %c -H W>> %s.pdb",
         BASE_DIR,proBfile,proB.chain,OutHead);
       printf("%s\n",command); system(command); } 


    } /* while (an) */ 
   
//    Free_AlignNode_Llist(&SHead);
//    Free_AlignNode_Llist(&AHead);

 } /* 'P' */


  /**********************************/
  /* [L] ONE-VS-LIBRARY COMPARISON  */
  /**********************************/

  else if ((MODE[0]=='S')||(MODE[0]=='L'))  
  { 
 
   AHead.next = NULL;
   /** Read Query Protein **/ 
   sprintf(proA.pname,"%s",proAfile); 
   Read_BSSP_File_LinearList(&proA,proA.pname,'A','W',ARnumSta,ARnumEnd); 
   Malloc_MAP2D(&DmapA,proA.Naa);  
   Cal_DmapCB(&proA,&DmapA);
   printf("%-16s %s\n","[QUERY_PROTEIN]",proA.pname); 
   printf("%-16s %s\n","[QUERY_COMPND]", proA.compnd); 
   if ((ARnumSta[0] != '\0')&&(ARnumEnd[0] != '\0'))
    { printf("%-16s %s\n","[START_RNUM]", ARnumSta); 
      printf("%-16s %s\n","[END_RNUM]",   ARnumEnd);  }
   printf("%-16s Naa %d Nsse %d\n","[QUERY_SIZE]", proA.Naa,proA.Nsse); 
   proA.SelfSdis = SelfScoreDistanceS(&proA,&DmapA);
   proA.SelfSsse = SelfScoreSSE(&proA);

   /** Read Library **/
   Read_Library_List(LibFile,&LibHead,&Nlibrary,&MaxNaaB);
   if (MaxNaaB==0) Check_Library_List_MaxNaa(&LibHead,&MaxNaaB); 
   
   printf("%-16s %d\n","[MAX_NAAB]",MaxNaaB);
   printf("%-16s %s\n","[LIBRARY_FILE]",LibFile); 
   printf("%-16s %d\n","[LIBRARY_SIZE]",Nlibrary);
   Malloc_Protein(&proB,MaxNaaB,'A','W'); 
   Malloc_SSE(&proB,MaxNaaB/10); 
   Malloc_MAP2D(&DmapB,MaxNaaB);  
   Malloc_GotohDP_Matrix(proA.Naa,MaxNaaB);
 
   AHead.next = NULL; 
   ALibHead.next = NULL; 
   
   ln = &LibHead;
    
   while (ln->next !=NULL)
    {
      ln = ln->next;
      if (Vtype == 'T') 
       printf("#[%d] %s %s\n",ln->num,ln->Name, ln->Tax);  
      sprintf(proB.pname,"%s.bssp",ln->Name); 
      Get_Part_Of_Line(proB.Tax,ln->Tax,0,127); 
      Read_BSSP_File(&proB,proB.pname); 
      Cal_DmapCB(&proB,&DmapB);  
      proB.SelfSsse = SelfScoreSSE(&proB);
      proB.SelfSdis = SelfScoreDistanceS(&proB,&DmapB);
      Pairwise_3D_Alignment(&proA,&proB,&DmapA,&DmapB,&AHead,&SHead,inalifile,inssefile);
    
     /* Scan all the SubOpt alignments */ 
      an = &AHead; 
      while (an->next !=NULL)
     { 
       an = an->next;
       an->sim->Name = (char *)malloc(sizeof(char)*(strlen(ln->Name)+3));
       an->sim->Tax  = (char *)malloc(sizeof(char)*(strlen(ln->Tax )+3));
       an->sim->mal_nametax = 1;
       sprintf(an->sim->Name,"%s",ln->Name);
       sprintf(an->sim->Tax,"%s",ln->Tax);
    
       an->sim->Naa = proB.Naa; an->sim->Nsse = proB.Nsse;
       Evaluate_Similarity(&proA,&proB,&DmapA,&DmapB,an);
       an->sim->CRMS = CRMS(&proA,&proB,an,Ga,Gb,Rmat,'A','-');

       Add_AlignNode_To_Tail(&ALibHead,an);
     } 
 
   Free_AlignNode_Llist(&SHead);
   Free_AlignNode_Llist(&AHead);
 
  } /* while (ln) */
 
  /** ScEnv Raw Zscore Ranking **/ 
   if (RankType=='E') 
   { 
     Set_Value_AlignNode(&ALibHead,'E');
     Cal_Raw_Zscore_Llist(&ALibHead,&Mean,&SD);
     printf("[WAY_OF_RANKING] EnvSc raw Zscore (Mean %.3lf SD %.3lf)\n",Mean,SD); 
     Set_Value_AlignNode(&ALibHead,'Z');
     Merge_Sort_Double_Linked_List(&ALibHead,'R'); 
     printf("%-16s %f\n","[Z_THRESHOLD]",Zthre); 
     printf("%-16s %d\n","[NRANK]",Nrank); }

   /** Sse Raw-Zscore Ranking **/ 
   if (RankType=='S') 
   { 
     Set_Value_AlignNode(&ALibHead,'S');
     Cal_Raw_Zscore_Llist(&ALibHead,&Mean,&SD);
     printf("[WAY_OF_RANKING] SseSc raw Zscore (Mean %.3lf SD %.3lf)\n",Mean,SD); 
     Set_Value_AlignNode(&ALibHead,'Z');
     Merge_Sort_Double_Linked_List(&ALibHead,'R'); 
     Count_Thre_Value_AlignNode(&ALibHead,Zthre,'M',&Nrank);
     printf("%-16s %f\n","[Z_THRESHOLD]",Zthre); 
     printf("%-16s %d\n","[NRANK]",Nrank); }

   /** DisSc XX-fit Zscore Ranking **/ 
   if (RankType=='D') 
   { 
     Set_Value_AlignNode(&ALibHead,'D');
     Set_SubValue_AlignNode(&ALibHead,'C','Q');
     TwoStep_Linear_Fitting_Llist(&ALibHead,&A,&B,&SD);
     Set_Value_AlignNode(&ALibHead,'Z');
     Merge_Sort_Double_Linked_List(&ALibHead,'R'); 
     Count_Thre_Value_AlignNode(&ALibHead,Zthre,'M',&Nrank);
     printf("[WAY_OF_RANKING] DisSc Ncmp^2-fitting Zsc ( %lf *x*x + %.3lf SD %.3lf )\n",A,B,SD);
     printf("%-16s %f\n","[Z_THRESHOLD]",Zthre); 
     printf("%-16s %d\n","[NRANK]",Nrank); }

   /** Raw Rdis Ranking **/ 
   if (RankType=='R') 
   { 
     Set_Value_AlignNode(&ALibHead,'D');
     Set_SubValue_AlignNode(&ALibHead,'C','Q');
     Cal_Raw_MeanSD_Llist(&ALibHead,&Mean,&SD);
     printf("[WAY_OF_RANKING] Rdis ranking (Mean %.3lf SD %.3lf)\n",Mean,SD);
     TwoStep_Linear_Fitting_Llist(&ALibHead,&A,&B,&Mean,&SD);
     Set_Value_AlignNode(&ALibHead,'R');
     Merge_Sort_Double_Linked_List(&ALibHead,'R'); 
     Count_Thre_Value_AlignNode(&ALibHead,Rthre,'M',&Nrank);
     printf("%-16s %f\n","[R_THRESHOLD]",Rthre); 
     printf("%-16s %d\n","[NRANK]",Nrank); 
   }

    
    /*** Re-read best library structure ***/ 
    bn = &ALibHead;
    p = 0;
    while ((bn->next !=NULL)&&(p<Nrank))
    { 
      bn = bn->next; 
      sprintf(proB.pname,"%s.bssp",bn->sim->Name); 
      Read_BSSP_File(&proB,proB.pname);
      bn->pro = (struct PROTEIN*)malloc(sizeof(struct PROTEIN));
      bn->mal_pro = 1;
      Malloc_Protein(bn->pro,proB.Naa,'A','W');  
      Malloc_SSE(bn->pro,proB.Nsse); 
      Equal_Protein(bn->pro,&proB);  
    } 
    
    printf("\n"); 
    Output_Library_with_Compnd(&ALibHead,Nrank); 
    printf("\n"); 
    Output_Library_Detailed(&ALibHead,Nrank,RankType); 
    printf("\n"); 
    Output_Library_Sec_Symbol(&proA,&ALibHead,Nrank); 
    printf("\n"); 
    Output_Library_ClustalW(&proA,&ALibHead,Nrank); 

    /*** Output All data ***/ 
   if (outfile[0]!='-')  
    { 
      fp = fopen(outfile,"w");
      bn = &ALibHead;

      while (bn->next != NULL)
      {
        bn = bn->next;     
        fprintf(fp,"%d %-6s %-10s d %7.1f s %7.1f e %7.1f a %5.0lf Na %d c %d s %d c %d id %.1f rm %.2lf\n",
         bn->num,bn->sim->Name,bn->sim->Tax, bn->sim->ScDis,bn->sim->ScSSE,
         bn->sim->ScEnv,bn->sim->ScAA,
         bn->sim->Naa,bn->sim->Ncomp,bn->sim->Nsse,bn->sim->Ncomp_sse,bn->sim->SeqID,bn->sim->DRMS);
      }
      fclose(fp); 
    }
 
   if (AliOutType == 'T') Output_All_Alignments(&proA,&ALibHead,Nrank);

   } /* 'S' */





  /******************************/
  /* [A] ALL-VS-ALL COMPARISON  */
  /******************************/
  
  else if (MODE[0] == 'A') 
  {
   /*** Setting Output ***/ 
   if (outfile[0] == '-') { fpout = stdout; }
   else 
    { fpout = fopen(outfile,"w");
      if (fpout == NULL) {printf("#ERROR:Can't write to \"%s\"\n",outfile); exit(1);} }
   fprintf(fpout,"#[%s]\n",PAR.COMMAND); fflush(stdout);

   /** Read Library **/
   Read_Library_List(LibFile,&LibHead,&Nlibrary,&MaxNaaB);
   Check_Library_List(&LibHead);
   printf("#Nlibrary %d MaxNaaLib %d\n",Nlibrary,MaxNaaB); 
   if (Nlibrary<=1) { printf("#ERROR:Nlibrary %d is too small\n",Nlibrary); exit(1); }
 
   /* Read and malloc library */
   fprintf(fpout,"#READ ALL THE STRUCTURE\n"); fflush(stdout);
   ln = &LibHead;
   while (ln->next !=NULL)
   { 
     ln = ln->next;
     if (Vtype=='T') printf("#Read [%d]:%s\n",ln->num,ln->Name); 
     ln->pro = (struct PROTEIN *)malloc(sizeof(struct PROTEIN)); 
     sprintf(ln->pro->pname,"%s.bssp",ln->Name); 
     Read_BSSP_File_LinearList(ln->pro,ln->pro->pname,'A','W',ARnumSta,ARnumEnd); 
     ln->pro->SelfSsse = SelfScoreSSE(ln->pro); 
   }
  
  /** SET Nsta and Nend **/ 
  Cal_NstaNend_CompTime(&LibHead,&Nsta,&Nend,AVAbunshi,AVAbunbo);
  printf("#AVAbunshi/bunbo  [%d]/[%d]\n",AVAbunshi,AVAbunbo);
  printf("#Npair_start  %d Npair_end %d Npair_to_be_calculated %d\n",Nsta,Nend,Nend-Nsta-1);
  Cal_MaxNaa_from_NstaNend(&LibHead,Nsta,Nend,&MaxNaaA,&MaxNaaB);
  printf("#MALLOC FOR DP:MaxNaaA %d MaxNaaB %d\n",MaxNaaA,MaxNaaB);
 
  /** Malloc Matrix **/ 
  Malloc_GotohDP_Matrix(MaxNaaA,MaxNaaB);
  Malloc_MAP2D(&DmapA,MaxNaaA); 
  Malloc_MAP2D(&DmapB,MaxNaaB); 
 
  /* Malloc buffer string */ 
  Nbuffer = 400; 
  buffer = (char **)malloc(sizeof(char *)*Nbuffer);
  for (i=0;i<Nbuffer;++i) buffer[i] = (char *)malloc(sizeof(char)*200);

  /****  ALL vs ALL comparison ****/
  printf("#COLDEF [proA] [proB] [NaaA] [NaaB] [Ncomp] [ScSSE] [ScEnv] [ScDis] [SqID] [CRMS] [Rdis] [Rsse]\n");
  ncomb = k = calok = 0;
  ln = &LibHead;  
  AHead.next = NULL;
  
  while (ln->next != NULL)  
   { 
     mn = ln;  
     ln = ln->next;
     if (calok==1)
     { Cal_DmapCB(ln->pro,&DmapA); ln->mark = 1; 
       if (ln->pro->SelfSdis==0.0) ln->pro->SelfSdis = SelfScoreDistanceS(ln->pro,&DmapA); 
     } 
       else ln->mark = 0;
      
     while (mn->next != NULL)  
     {
      mn = mn->next; 
      if ((calok==0)&&(ncomb==Nsta)) calok = 1;
      
      /* if (ln == mn) calok = 1; else calok = 0; */
     
      if (calok == 1)
       {
        if (ln->mark==0) 
        {
          Cal_DmapCB(ln->pro,&DmapA);  ln->mark = 1;
          if (ln->pro->SelfSdis==0.0) ln->pro->SelfSdis = SelfScoreDistanceS(ln->pro,&DmapA); 
         }
 
        if (Vtype=='T') 
        { printf("#CAL_PAIR %d %s %d %s \n",ln->num,ln->pro->pname,mn->num,mn->pro->pname); 
          fflush(stdout); }

        Cal_DmapCB(mn->pro,&DmapB);
        if (mn->pro->SelfSdis == 0.0) mn->pro->SelfSdis = SelfScoreDistanceS(mn->pro,&DmapB); 

        Pairwise_3D_Alignment(ln->pro,mn->pro,&DmapA,&DmapB,&AHead,&SHead,inalifile,inssefile);
        an = AHead.next; 
        Evaluate_Similarity(ln->pro,mn->pro,&DmapA,&DmapB,an);
        if (DisMesType=='C') an->sim->CRMS = CRMS(ln->pro,mn->pro,an,Ga,Gb,Rmat,'A','-');
 
        sprintf(buffer[k],"%5s %5s %4d %4d %4d %.1f %.1f %.1f %5.2lf %5.2f %5.2f %5.2f\n",
        ln->pro->cname, mn->pro->cname, ln->pro->Naa, mn->pro->Naa, an->sim->Ncomp,  
        an->sim->ScSSE,an->sim->ScEnv,an->sim->ScDis, an->sim->SeqID, an->sim->CRMS,
        an->sim->Rdis,an->sim->Rsse);

        ++k;
        if (k>(Nbuffer-1)) 
         { for (m=0;m<k;++m) {fprintf(fpout,"%s",buffer[m]);  fflush(fpout);}
          k = 0; }

        Free_AlignNode_Llist(&AHead); 
        Free_AlignNode_Llist(&SHead);
     } 

      ++ncomb; 
      if ((calok==1)&&(ncomb==Nend)) calok = 0; 
 
     } /* mn loop */

     Free_Protein(ln->pro);
     Free_SSE(ln->pro);
     free(ln->pro); 
     ln->pro = NULL;

    } /* ln loop */ 

    for (m=0;m<k;++m) { fprintf(fpout,"%s",buffer[m]); fflush(stdout);}

   fclose(fpout);

  } /* 'A' */


} /* end of main() */




/******************/
/*** FUNCTIONS ****/
/******************/


int First_Appear(line,sym)
 char *line,sym;
{
 int L,i,hit;
 L = strlen(line); 
 hit = i = 0; 
 while ((i<L)&&(hit==0))
  if (line[i]==sym) hit = 1; else ++i; 
 if (hit==0) return(-1); else return(i); 

} /* end of First_Appear() */



void Output_Reference()
{
 printf("# ProtDeform VER %s: PROGRAM FOR PROTEIN 3D STRUCTURE COMPARISON ####\n","1.0"); 
 printf("# coded by J. Rocha, J. Segura, R. Wilson. Last Modified : %s\n","July 1st, 2010"); 
 printf("#<REFERENCE>\n"); 
 printf("# J. Rocha, et al.\n");
 printf("# \"Structural Protein Alignment by a sequence of local transformations\".\n");
 printf("#   Boinformatics vol.25,1625-1631, 2009.\n");

 printf("#uses the first classifier of MATRAS VER %s: PROGRAM FOR PROTEIN 3D STRUCTURE COMPARISON ####\n",Version); 
 printf("# coded by Takeshi Kawabata. Last Modified : %s\n",LastModDate); 
 printf("#<REFERENCE>\n"); 
 printf("# Takeshi Kawabata and Ken Nishikawa.\n");
 printf("# \"Protein Structure Comparison Using the Markov Transition Model of Evolution\".\n");
 printf("#   Proteins vol.41:108-122(2000).\n");

} /* end of Output_Reference() */
 




void Evaluate_Similarity(pA,pB,DmapA,DmapB,an)
 struct PROTEIN *pA,*pB;
 struct MAP2D   *DmapA,*DmapB;
 struct AlignNode  *an;
{

 an->sim->SeqID = SeqID(pA,pB,an); 
 an->sim->SecID = Sec3ID(pA,pB,an); 
 an->sim->ExpID = ExpID(pA,pB,an); 
 an->sim->Ncomp = Ncompare(an);
 an->sim->DRMS  = DRMS(pA,pB,an,DmapA,DmapB);
 an->sim->ScDis =(float)ScoreDistanceS(pA,pB,an,DmapA,DmapB) + 
                 (float)GapPenalty(an,PAR.GapExtD,PAR.GapOpenD); 

 if ((pA->SelfSdis!=0.0)&&(pB->SelfSdis!=0.0))
  { 
   an->sim->Rdis = 100.0*an->sim->ScDis/(0.5*(pA->SelfSdis + pB->SelfSdis));
   an->sim->RelS = Reliability_Rdis(an->sim->Rdis,'S');
   an->sim->RelO = Reliability_Rdis(an->sim->Rdis,'O');
  } 
 else an->sim->Rdis = an->sim->RelS = an->sim->RelO = 0.0;

 if ((pA->SelfSsse!=0.0)&&(pB->SelfSsse!=0.0))
 { 
   an->sim->Rsse = (float)(100.0*an->sim->ScSSE/(0.5*(pA->SelfSsse + pB->SelfSsse)));
  }
 else an->sim->Rsse = 0.0;

} /* end of Evaluate_Similarity() */


void Output_for_Web(pA,pB,an,outhead,RnumPlainType)
 struct PROTEIN   *pA,*pB;
 struct AlignNode *an;
 char *outhead;
 char RnumPlainType;
{
  char buff[256];
  double Ga[3],Gb[3],Rmat[3][3];  

  sprintf(buff,"%s.info",outhead);
  Output_Alignment_Information(buff,pA,pB,an);
  sprintf(buff,"%s.hol",outhead);
   Output_Path_Horizontal_HTMLcolor(pA,pB,an,buff);
  sprintf(buff,"%s.clw",outhead);
   Output_Path_Horizontal_ClustalW(pA,pB,an,buff);
  sprintf(buff,"%s.cs",outhead);
   Write_JavaView_Color_Alignment(buff,pA,pB,an);
  sprintf(buff,"%s.ras",outhead);
   Write_Rasmol_Color_Alignment(buff,pA,pB,an,'R','C');
  sprintf(buff,"%s.ver",outhead);
   Write_Alignment_CASP_Style(pA,pB,an,buff,RnumPlainType,Ga,Gb,Rmat);
 
  an->sim->CRMS =  CRMS(pA,pB,an,Ga,Gb,Rmat,'A','-');
  sprintf(buff,"%s.mat",outhead);
   Write_Rotation_Matrix(buff,pA,pB,an->Nali,Ga,Gb,Rmat,an->sim->CRMS);

  sprintf(buff,"%s.sse",outhead);
   Output_SSE_Alignment_Horizontal(buff,pA,pB,an->sn);
  
  exit(1); 

} /* end of  Output_for_Web() */



void Output_All_Alignments(pQ,head,Nrank)
 struct PROTEIN *pQ;    /* Query protein */
 struct AlignNode *head;
 int Nrank;
{
 struct AlignNode *bn;
 int p;

 printf("\n[ALIGNMENTS]\n\n");
 bn = head;
 p = 0;
 while ((bn->next !=NULL)&&(p<Nrank))
  {
    ++p;
    bn = bn->next;
    printf(">%d %s [%s] \"%s\"\n",p,bn->sim->Name,bn->sim->Tax,bn->pro->compnd);
    printf("#Naa %3d start %s end %s SqID %3.0lf %% crms %4.1f Ssse %5.0lf Sdis %6.0lf Rdis %5.1lf Z %.2lf\n\n",
      bn->sim->Naa,bn->pro->Rnum[bn->staB],bn->pro->Rnum[bn->endB], 
      bn->sim->SeqID, bn->sim->CRMS,bn->sim->ScSSE,bn->sim->ScDis,bn->sim->Rdis ,bn->Zsc);
       Output_Path_Horizontal(pQ,bn->pro,bn,"-");
       printf("//\n\n");
 } /* while (bn) */

} /* end of Output_All_Alignments() */
