// Version December 2009 by segments
// Version September 3 2008
// Multiple solutions from Matras
// version for parameter gathering July 2008
// nodes sorted by distance in the neighbourhood
// No H with E 
// only 3 values for the environment
// Can read xml matchings  June 2008
// statistical classification of environmental features May 2008
// Use Bayes
// It uses pdb numbering for the aminoacids

/* Matching by a Sequence of Transformations */
#include <stdio.h> 	
#include <math.h> 
#include <list>
#include "matrix_clapack.h"
#include "protein.h"
// #include "stdlib.h"
#include "LlistAlign.h"
#include "vector.h"	

#include "Geometry.h"
#include "WriteDeformPDB.h"
#include "Evaluation.h"


//////////////////////////////////////////
//										//
//     Adjustable parameters			//
//      set in config file				//
//										//
//////////////////////////////////////////

// output (0,1,2,3)						
int VERBOSE = 1;
int STATISTICS = 0;

//MatrasPGM
bool MATRASPGM = false;	

// Type of Geometry to use
enum { RIGID, AFFINE } GEOMETRY = RIGID;

// Number of nearest neighbours in graph

int	minNEIGHBOURS = 8;

double GAPOPEN;// = exp(-MAXDistExt*MAXDistExt/squaredratio);
double GAPEXTN;//  = GAPOPEN;

// Total iterations
int ITERATIONS = 5;


//////////////////////////////////
//								//
// 	 Global variables			//
//								//
//////////////////////////////////

struct AlignNode *AlignCurrent;


//DP matrix
Matrix<double> H;
Matrix<int> NumAssign;
Matrix<int> Hx;
Matrix<int> Hy;



Matrix<double> probenv;


//////////////////////////////////////////
//										//
// 	    	New Structures				//
//										//
//////////////////////////////////////////

struct nodedist { 						
		  int node; 					
		  double distsq; 				
};							
typedef struct nodedist *Nodedist;	




//////////////////////////////////////////
//										//
// 	     New Functions					//
//										//
//////////////////////////////////////////

#define Sqr(a) ((a)*(a))									

// Write a matrix to a pgm file											
void MatrixPgm(Matrix<double>& A, char *fname) {																		
	int i, j;
	FILE* fp = fopen(fname, "wb");										
	
	if (!fp) return;													
																		
	fprintf(fp, "P5\n%d %d\n255\n", A.c, A.r);				
	for (i=0; i < A.r; i++)											
		for (j=0; j < A.c; j++)
			fputc(max(0 , (int)min(((A(i,j)) * 256) , 255)) , fp);		
	fclose(fp);															
}																		


void Find_Chains(struct PROTEIN *pr) {
int i, chain = 0;

pr->Chains= (int *) malloc(sizeof(int) * (pr->Naa+1));


pr->Chains[0] = chain;
	for (i=1; i<=pr->Naa; i++) {
		if (Sqr(pr->PosCA[i][0] - pr->PosCA[i-1][0]) + 
   			Sqr(pr->PosCA[i][1] - pr->PosCA[i-1][1]) + 
   			Sqr(pr->PosCA[i][2] - pr->PosCA[i-1][2]) > 20 )  // Two adjacent CA are at about 3.7 A from each other in the same chain
			chain++;
   		pr->Chains[i] = chain;
	}
}


void Find_Partition(struct PROTEIN *pr) {
int i, j, segment = 0;

for (i=0;i<=pr->Naa;i++)
   pr->Sec[i] = typeSSE(pr->Sec[i]);   /* Caution: Too early may afect the SSE management*/

pr->Partition = (NodeNeigh) malloc(sizeof(struct nodeneighbour)*(pr->Naa/2 + 5) );


i=1;
/* 
  i represents the first element of the new segment. */
while (i<=pr->Naa) {
   j=i+1;
   while (j-i < 9 && j<=pr->Naa && pr->Sec[j] == pr->Sec[j-1] && pr->Chains[j] == pr->Chains[j-1]) j++;
   if (j-i==9) 
       j=i+7;
   /* j is the first element that should not be in the segment 
      Partition[segment].ending is the last element of the segment */
   pr->Partition[segment++].ending= j-1;   
   i=j;
}
pr->Partition[segment].ending=-1;
}



//////////////////////////////////////////
//				      					//
//   Compute nearest neighbours graph  	//
//				
//    Compute the partition into segments and the nighbours of the segments      // 					//
//////////////////////////////////////////

extern "C" void Set_Neighbours(struct PROTEIN *pr, int n) {															
	double d;						
	int i, j, k, ktmp, s;
	nodedist nodelist[100];	

        Find_Chains(pr);
        Find_Partition(pr);								
		
	if (n > 100) throw("Too many neighbours");
	
        i=1; 
	for (s=0; pr->Partition[s].ending != -1; s++) {
         for (j=0; j <= n; j++) {	
			nodelist[j].distsq = 1000000;					
			nodelist[j].node = -1;						
	 }
	 for (; i <= pr->Partition[s].ending; i++) {							
		for (j=1; j <= pr->Naa; ++j) 
			if (abs(j-i) > 2) {
				d = Sqr(pr->PosCA[i][0] - pr->PosCA[j][0]) + 
				    Sqr(pr->PosCA[i][1] - pr->PosCA[j][1]) +	
					Sqr(pr->PosCA[i][2] - pr->PosCA[j][2]);		
				ktmp = n;
				
				for (k=0; k < n; k++)	 
		// Search for a place according to the distance or if the index already appears			
					if (d < nodelist[k].distsq || j==nodelist[k].node) {
						ktmp=k;
                        			break;
                 			}
                                if (k<n && j==nodelist[k].node) { //if the index already appears
                                  if (d<= nodelist[k].distsq) 
					nodelist[k].distsq=d;
  				  
         			  continue;	
				  }
      
                                //ktmp is the place to insert it but the same index can appear later 
 
				if (k< n-1) k++;
 				while (k< n-1 && nodelist[k].node != j ) k++;
				k--;                              

                		for (; k >= ktmp; k--) {
					nodelist[k+1].distsq = nodelist[k].distsq;
					nodelist[k+1].node = nodelist[k].node;
				}
				
				if (ktmp < n) {
                			nodelist[ktmp].distsq = d;
                    			nodelist[ktmp].node = j;
   				}
			}
	}

		// Insert the neigbours in the protein segment s, sorted by the index
		
        int minval,val,r;
        pr->Partition[s].list = (int *) malloc((n+1) * sizeof(int));
	for(r=0; r<n && (minval=nodelist[r].node) != -1; r++) {
		  int indexmin=r;
                  for (k=r+1; k<n; k++) 
                       if ((val=nodelist[k].node) != -1 && val < minval) {
                            minval = val;
                            indexmin=k;
                       }
		   pr->Partition[s].list[r] = minval;
                   nodelist[indexmin].node = nodelist[r].node;
         }
         pr->Partition[s].list[r] = -1;
	
       }
}															


void copy_alignment(Vector<int> &match, int pos, Matrix<int> &matchings) {
   int max = match.r;
   for (int i=0; i < max; i++)
      matchings(i, pos) = match(i);
}


void printAllDistances(struct PROTEIN *pA, struct PROTEIN *pB, Matrix<double> &SS) {
	int i, image;
	double x;
	
	for (i=0; i <= pA->Naa; ++i)
		for (image=0; image <= pB->Naa; ++image)
			if ((x = SS(i, image)) > 0)
             	printf("%s %s %3.2lf \n", pA->Rnum[i], pB->Rnum[image], sqrt(-log(x)*PAR.squaredratio));
}


// Dynamic programing over score matrix gmatch
void Align_PD(struct PROTEIN *proA, struct PROTEIN *proB, Vector<int> &match, Matrix<double> &gmatch, int iter, double *score) {																					
	int i, j;								
	double max;
	int h_i, h_j, h_tmp_i, h_tmp_j, kk;

	double gap_open = GAPOPEN;

	double gap_extension = GAPEXTN;	
																				
	H(0,0) = 0;
   	NumAssign(0,0) = 0;  // Number of consecutive alignments						
        Hx(0,0) = -10;
        Hy(0,0) = -10;	

																				
	//START GAP
	for (i=1; i <= proA->Naa; i++) {
												   
		H(i,0) = (i-1) * gap_extension + gap_open;								
        Hx(i,0) = -10;
															
        Hy(i,0) = -10;
															
        NumAssign(i, 0) = 0;
														
	}																				
						
	for (j=1; j <= proB->Naa; j++) {
												
	H(0,j) = (j-1) * gap_extension + gap_open;								
        Hx(0,j) = -10;	
														
        Hy(0,j) = -10;	
														
        NumAssign(0, j) = 0;	
													
	}																				
																					
	//DP MATRIX
	for (i=1; i <= proA->Naa; i++) {												
		for (j=1; j <= proB->Naa; j++) {
			if (((proA->Sec[i] == 'H') || (proA->Sec[i]=='E')) && 
				((proB->Sec[j] == 'H') || (proB->Sec[j]=='E')) && (proA->Sec[i] != proB->Sec[j]))
					H(i,j) = H(i-1,j-1) - 1;
			else
				H(i,j) = H(i-1,j-1) + gmatch(i,j);
									
			
			Hx(i,j) = i-1;			
												
			Hy(i,j) = j-1;
															
            NumAssign(i, j) = NumAssign(i-1, j-1) + 1;	
											
            if (NumAssign(i-1, j) > 5)
												
        		if ((j < proB->Naa) && (i> 1) && ((proA->Sec[i] == 'H') || (proA->Sec[i]=='E')) && (proA->Sec[i] == proB->Sec[j+1])		
	 
						&& (proA->Sec[i-1] == proB->Sec[j]) && (proA->Sec[i] == proA->Sec[i-1]))
            		max = -10000; //Don't cut SSE matching in the path (i-1,j)
              								
				else
					max = H(i-1,j) + gap_open;	
									
			else if (NumAssign(i-1, j) > 0)
				max = -10000;  // Avoid isolated matchings
							
			else
				max = H(i-1,j) + gap_extension;	
									
																					
			if (max > H(i,j)) {													
				H(i,j) = max;	
													
				Hx(i,j) = i-1;
														
				Hy(i,j) = j;
                		NumAssign(i, j) = 0;	
												
			}																		
																					
            if (NumAssign(i, j-1) > 5 )
				if ((i < proA->Naa) && (j > 1) && ((proA->Sec[i+1] == 'H') || (proA->Sec[i+1]=='E')) && (proA->Sec[i+1] == proB->Sec[j])	   
								&& (proA->Sec[i] == proB->Sec[j-1]) && (proA->Sec[i+1] == proA->Sec[i]))
					max = -10000; //Don't cut SSE in the path (i,j-1)
				else
					max = H(i,j-1) + gap_open;
										
			else if (NumAssign(i, j-1) > 0)
										
				max = -10000;  // Avoid isolated matchings
							
			else
				max = H(i,j-1) + gap_extension;
										
																					
			if (max > H(i,j)) {													
				H(i,j) = max;
														
				Hx(i,j) = i;
														
				Hy(i,j) = j-1;
                		NumAssign(i, j) = 0;	
												
			}																		
		}																			
	}																				
/*
	//END_GAP													
	for (i=1; i < proA->Naa; i++) {
		H(i,proB->Naa) = H(i,proB->Naa) + (proA->Naa - i - 1) * gap_extension + gap_open;
	}																				
	for (j=1; j < proB->Naa; j++) {
		H(proA->Naa,j) = H(proA->Naa,j) + (proB->Naa - j - 1) * gap_extension + gap_open;
	}																				
																					
	//BEST SCORE																	
	max = -1000000000;	
															
	for (i=1; i <= proA->Naa; i++) {													   
		if (H(i,proB->Naa) > max) {													
			h_i = i;										
			h_j = proB->Naa;	
														
			max = H(i,proB->Naa);	
												
		}																			
	}																				
																					
	for (j=1; j <= proB->Naa; j++) {													
		if (H(proA->Naa,j) > max) {													
			h_i = proA->Naa;	
														
			h_j = j;
														
			max = H(proA->Naa,j);
													
		}																			
	}																				
   	*score = max;	
																
*/
	*score= H(proA->Naa,proB->Naa);																					
	//PATH																	
	h_i=proA->Naa;
	h_j=proB->Naa;

	match = -1;
	while ((h_i >= 1) && (h_j >= 1)) {	
		h_tmp_i=h_i;
		h_tmp_j=h_j;
		h_i=Hx(h_tmp_i,h_tmp_j);
		h_j=Hy(h_tmp_i,h_tmp_j);
		if(h_i==h_tmp_i-1 && h_j==h_tmp_j-1){
			match(h_tmp_i)=h_tmp_j;		
		}																			
	}																				
																				
	if (VERBOSE > 1) {																
		Matrix<double> ali (proA->Naa, proB->Naa);
									
		ali = 0;				
		for (i=1; i <= proA->Naa; i++) {
			if( match(i) > -1 )	
													
				ali(i,match(i)) =1;
												
		}																			
		char fname[30];	
															
		sprintf(fname, "align%i.pgm", iter);	
									
		MatrixPgm(ali, fname);	
													
	}
}	


// Initialise score matrix
Matrix<double> SSE_matrix(struct PROTEIN *pA, struct PROTEIN *pB,  struct AlignNode *sn) {
	int i, j, n, k, x, y, z;
	int nA, nB;
	Matrix<double> SSE(pA->Naa+1, pB->Naa+1);
	
	SSE = 2.1 * GAPEXTN;
	////Don't align aa of diferent ss type Alpha-Beta 
	for(i=1;i<=pA->Naa;i++){
		for(j=1;j<=pB->Naa;j++){
			if(pA->Sec[i]!=pB->Sec[j] && ((pA->Sec[i] == 'H') || (pA->Sec[i]=='E')) && ((pB->Sec[j] == 'H') || (pB->Sec[j]=='E'))){
				SSE(i,j)=0;
			}
		}
	}
	
	for (k=0; k<=sn->Nali-1; ++k) {
		z = 0;
		nA = sn->nalA[k];
		nB = sn->nalB[k];
		x = pA->SSE[nA].Rend - pA->SSE[nA].Rsta;
		y = pB->SSE[nB].Rend - pB->SSE[nB].Rsta;
		n = abs(x - y) + 1;		// number of complete diagonals
		if (x <= y)
			for (i=pA->SSE[nA].Rsta; i <= pA->SSE[nA].Rend; i++) {
				for (j=pB->SSE[nB].Rsta; j < pB->SSE[nB].Rsta+n; j++) {
					SSE(i,j + z) = 1;
				}
				z++;
			}
		else
			for (j=pB->SSE[nB].Rsta; j <= pB->SSE[nB].Rend; j++) {
				for (i=pA->SSE[nA].Rsta; i < pA->SSE[nA].Rsta+n; i++) {
					SSE(i + z,j) = 1;
				}
				z++;
			}
	}
	if (VERBOSE > 1) MatrixPgm(SSE, "Initial.pgm");
	return SSE;
}


//Compute score matrix of Alignment  
void ScoreMatrix(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> &Alignment,  Matrix<double> &score) {
	int fixedmatch[1006][2], fpos;
	Matrix<double> T;
	score=0;
	
	for(int a=0;pA->Partition[a].ending!=-1;a++)
           for(int b=0;pB->Partition[b].ending!=-1;b++)

            {
                /* change from site to segment */
                /*  For each neighbour index A of segment a of pA, we see if its image is a neighbour of
		    segment b of pB. If so, the pair (indexA,image) will be in the set to get T.
                    Then, all the amino acid pairs of the segments get applied the transformation T.
                */
		int indexA, image, indexB, k2;
		k2 = 0;
		
		fpos=0;
                // Assume that the neighbours are sorted by the index
  		for (int k=0; (indexA=pA->Partition[a].list[k]) != -1; k++) 
			if ((image=Alignment(indexA)) > -1) {
					while (((indexB=pB->Partition[b].list[k2]) != -1) && image > indexB) k2++;
                                        if (indexB==-1) break;
                                        if (image < indexB) continue; 

 					
					fixedmatch[fpos][0]=indexA;				
					fixedmatch[fpos++][1]=image;
			}
			

		fixedmatch[fpos][0]=-1;
 
		if(fpos>minNEIGHBOURS){
				// if( GEOMETRY==RIGID ) 
				T=RigidTransform(pA,pB,fixedmatch);
				//if( GEOMETRY==AFFINE )T=AffineTransform(pA,pB,fixedmatch);
                                
                                int lasta=pA->Partition[a].ending;
				int lastb=pB->Partition[b].ending;
				for (int i=(a==0?1:pA->Partition[a-1].ending);i<=lasta;i++)
				  for (int j=(b==0?1:pB->Partition[b-1].ending);j<=lastb;j++)
				    score(i,j)= max(score(i,j),Consistency(pA,pB,i,j,T));				
		}                

	   }
	/*
	if (VERBOSE > 1) {
		char fname[30];
//		sprintf(fname, "score%i.pgm", iter);
		MatrixPgm(score, fname);
	}*/
}


void FMATRASPGM(int A, int B) {
	int n, x, y;
	char fname[30];
	Matrix<double>gmatch(A, B);
	
	FILE* fp = fopen("1.ali.coo", "r");

	if (!fp) throw("Can't open file");
	
	n = fscanf(fp, "%i %i", &x, &y);
	gmatch = 0;
	while (n == 2) {	
		gmatch(x,y) = 1;
		n = fscanf(fp, "%i %i", &x, &y);
	}
	fclose(fp);
	
	sprintf(fname, "matras_align.pgm");
	MatrixPgm(gmatch, fname);
}


//////////////////////////////////////////
//										//
//           Main procedure 			//
//                      				//
//										//
//////////////////////////////////////////

extern "C" int mainPD(struct PROTEIN *proA, struct PROTEIN *proB, struct AlignNode *Ahead,char *inalifile) {
	int i;
	int numsolSSE;
	double bestscore, onescore;

       GAPOPEN  = exp(-PAR.MAXDistExt*PAR.MAXDistExt/PAR.squaredratio);
       GAPEXTN  = GAPOPEN;
       int READMATCHING = 1; // (inalifile[0] != '\0') ;	
    // Long test
if (PAR.OutAlignment && !READMATCHING && !STATISTICS) 
    	printf("# Iterations=%d Neighbours=%d minNeighbours=%d squaredratio=%lf MAXDistExt=%lf GAPEXTN=%lf\n", ITERATIONS, PAR.MAXNEIGHBOURS, minNEIGHBOURS, PAR.squaredratio, PAR.MAXDistExt, GAPEXTN);
	
	


	Vector<int> match(proA->Naa+1), match_ref(proA->Naa+1), match_tmp(proA->Naa+1);
    Vector<int> bestmatch(proA->Naa+1);
//    Matrix<int> matchings(proA->Naa+1, ITERATIONS+1);

	//Score DP matrix
	Matrix<double> SS(proA->Naa+1, proB->Naa+1);
//	Matrix<double> SS_tmp(proA->Naa+1, proB->Naa+1);	
	
    H.resize(proA->Naa+1, proB->Naa+1);
    NumAssign.resize(proA->Naa+1, proB->Naa+1);
    Hx.resize(proA->Naa+1, proB->Naa+1);
    Hy.resize(proA->Naa+1, proB->Naa+1);
     
 	if (!STATISTICS && !READMATCHING) {
    	bestscore = -1;
    	int MaxBadIterations = ITERATIONS/10;
    	double MinScore = 0.2 * max(proA->Naa, proB->Naa);
		
    	for (numsolSSE=1; numsolSSE <= 4 && Ahead!=NULL ; (numsolSSE++, Ahead=Ahead->next)) {
//                      AlignCurrent = Ahead;
			SS = SSE_matrix(proA, proB, Ahead->sn);		//Initialize score matrix for DP
			Align_PD(proA, proB, match, SS, -1, &onescore); 	//Alignment of DP
//            copy_alignment(match, 0, matchings);
			match_tmp = -1;
//			SS_tmp = -1;
			i = 0;
			while(match!=match_tmp && i<ITERATIONS) {
				match_tmp = match;
//				SS_tmp = SS;
				ScoreMatrix(proA, proB, match,SS);				//Get score matrix from aligment match
				Align_PD(proA, proB, match, SS, i, &onescore);		//Get alignment from score matrix
				i++;
 //           			copy_alignment(match, i, matchings);
         		}
			
			if (onescore > bestscore) {
				bestscore = onescore;
			    bestmatch = match;
			}
     	}   
   	}  
    
	match = bestmatch;
	
    if (READMATCHING) {
        makematching(proA,proB,match_ref);
	//readmatchingTM( inalifile, match_ref);
        if (PAR.OutAlignment)
         printf("Prot1  Prot2  Nalign  PDscore rawscore\n");
        CalcPDscore(proA,proB,match_ref,GAPEXTN);
        // lTMscore(proA, proB, match, bestscore, 2, match_ref);
	}
    else 
    	if (STATISTICS) {
        	readmatchingback(proA, proB, proA->cname, proB->cname, match_ref);
        	ScoreMatrix(proA, proB, match_ref,SS);
        	printAllDistances(proA, proB, SS);
   		}
    	else if (PAR.OutAlignment)
			lTMscore(proA, proB, match, bestscore, 1, match_ref); 	// 3d argument is not used
             else  
                        CalcPDscore(proA,proB,match,GAPEXTN);
  
	if ( PAR.OutPDBdeform) {   
           printAlignment(proA,proB,match_ref);
           DeformPDB(proA, proB, match_ref);

        }
 
	if (VERBOSE > 1) {
		Matrix<double> gmatch(proA->Naa, proB->Naa);
		gmatch = 0;
		for (i=1; i < proA->Naa+1; i++) {
			if (match(i) > -1)
				gmatch(i, match(i)) = 1;
		}
		char fname[30];
		sprintf(fname, "lRMS_align.pgm");
		MatrixPgm(gmatch, fname);
	}
	if (MATRASPGM)	FMATRASPGM(proA->Naa, proB->Naa);	
	return 0;
}
