#include <stdio.h> 	
#include <math.h> 
#include <list>
#include "matrix_clapack.h"
#include "protein.h"
#include "vector.h"	

#include "Geometry.h"

#define Sqr(a) ((a)*(a))


extern int minNEIGHBOURS;


extern enum { RIGID, AFFINE } GEOMETRY;


// Positions of aminoacids
Matrix<double> Rpoint; 	//local rotation coordinates


//List of Transformations
struct interval {
    int start, end;
    int starti, endi;  // amino acid image index
    double T[4][4];
};
list <struct interval> TransL;


bool bystart(struct interval a, struct interval b) {
 return(a.start < b.start);
}


double cRMS(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> &Alignment, Matrix<double> &T){
	
	double pes;	
	int fixedmatch[1006][2], fpos = 0, i, p, j;
	int ALsize = 0;
	double score = 0;

	pes = 0;
	for (i=1; i <= pA->Naa; i++) {
		if(Alignment(i) > -1) {
			fixedmatch[fpos][0] = i;
			fixedmatch[fpos++][1] = Alignment(i);
			ALsize++;
		}	
	}
 	if(ALsize > 3) {
		fixedmatch[fpos][0] = -1;
		if(GEOMETRY == RIGID) 
			T = RigidTransform(pA, pB, fixedmatch);
	 	if(GEOMETRY == AFFINE)
	 		T = AffineTransform(pA, pB, fixedmatch);
	 	for(j=0; fixedmatch[j][0] != -1; j++) {
		pes=Consistency_l(pA, pB, fixedmatch[j][0],fixedmatch[j][1],T);
		score = score + pes;
	 }

	 score = score/ALsize;
	 score = sqrt(score);
	 return(score);	
		
	}else{
		return(1000000);
	}
}


/* Thsi version desn't work now */
double lRMS(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int>Alignment){
	int fixedmatch[1006][2],fpos=0,i,p,j, k;
	
	double pes;
	double pes0;

	Matrix<double> T;

	double score=0;
	double score0=0;
	pes=0;
	pes0=0;
	int ALsize=0;
	for (i=1; i <= pA->Naa; i++) {
		if(Alignment(i)>-1)
		{
			ALsize++;
			fpos=0;
			for(p=-2;p<=2;p++)
			if(i+p>=1 && i+p<=pA->Naa && Alignment(i+p)>-1) {
				fixedmatch[fpos][0]=i+p;
				fixedmatch[fpos++][1]=Alignment(i+p);
			}
			for (k=0; pA->Adj[i][k] != -1; k++) {
				if(Alignment(pA->Adj[i][k])>-1){
					fixedmatch[fpos][0]=pA->Adj[i][k];
					fixedmatch[fpos++][1]=Alignment(pA->Adj[i][k]);
				}
			}
			fixedmatch[fpos][0]=-1;
			if(fpos>minNEIGHBOURS){
				if( GEOMETRY==RIGID )T=RigidTransform(pA, pB, fixedmatch);
				if( GEOMETRY==AFFINE )T=AffineTransform(pA, pB, fixedmatch);
				pes=0;
				for(j=0;fixedmatch[j][0]!=-1;j++)
				{
					pes=pes+Consistency_l(pA, pB, fixedmatch[j][0], fixedmatch[j][1], T);
				}
				pes=pes/j;
				pes0=Consistency_l(pA, pB, i, Alignment(i), T);
			}
		}
		score=score+pes;
		score0=score0+pes0;	
	}
	if(ALsize>0){
		score=score/ALsize;
		score0=score0/ALsize;
		score0=sqrt(score0);
		score=sqrt(score);
		return(score0);
	}else{
		return(1000000);
	}
}


// It finds the transformation associated to an interval where  a  is.
Matrix<double> SearchListTrans(int a) {
	list <struct interval>::iterator interv;
	Matrix<double> Trans(4,4);

	for (interv=TransL.begin(); interv != TransL.end(); interv++) 
		if ((interv->start <= a) && (interv->end >= a)) {
			int i,j;
			for (i=0; i < 4; i++)
				for (j=0; j < 4; j++)
					Trans(i,j) = interv->T[i][j];
			
			return Trans;
   		}

	throw("Error finding the transformation for the last part.\n");   
}


void AddInterval(Matrix <double> &Trans, int a, int b, int ai, int bi) {
	struct interval x;
	int i,j;
	
	for (i=0; i < 4; i++)
  		for (j=0; j < 4; j++)
    		x.T[i][j] = Trans(i,j);
	
	x.start = a;
	x.end = b;
	x.starti = ai;
	x.endi = bi;
	TransL.push_front(x);
}

int find_segment(struct PROTEIN *p, int i) {
 int k,x;
 for (k=0; (x=p->Partition[k].ending) != -1; k++)
    if (i <= x) return(k);
 
}

int find_closest_pos_neigh(struct PROTEIN *p, int seg, int i) { 
  int k,x;
  float dist;
  int closest;
  dist = 1000000;
  for (k=0;(x=p->Partition[seg].list[k])!=-1; k++) {
    if (x > i) break;
    if (i - x < dist) {
       closest= k;
       dist = i - x;
    }
  }
  return(k);
}

// Find a transformation from the segment [jinit,jfin] in M to  [iinit,ifin] in D.
// Finds a rotation and translation for the first point of the chains c and for the other finds
// rotations, in order to preserve the length in the transformations of the segments. 
// The transformed coordenates are stored in R.
// It goes 7 by 7.
void findtransformationsegment(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int>Alignment, int iinit, int ifin, int jinit, int jfin) {
	int fixedmatch[1006][2], fpos=0, p, padd, *k;
	int i, j, centerA, centerB;
	Matrix<double> T;
	Vector<double> m(4);
	struct interval x;

	i = iinit; 
	j = jinit;
	fpos = 0;
	for (p=0; (i+p <= ifin) && ((ifin-i-p < 3) || (p < 7) ||      // Do not break SSE
				(i+p < ifin) && ((pA->Sec[i+p] == 'H') || (pA->Sec[i+p]=='E'))  && 
				(pA->Sec[i+p] == pB->Sec[j+p]) && 
				(pA->Sec[i+p+1] == pB->Sec[j+p+1]) &&
				(pA->Sec[i+p] == pA->Sec[i+p+1])); p++) {
 	   fixedmatch[fpos][0] = i+p;
	   fixedmatch[fpos++][1] = j+p;          
	}
	
  	int pmax = p;
  	if (fpos <= 3) {
	// Consider the neighbors too
                int segment,pos_neigh,neigh;
                segment = find_segment(pA,i);
                pos_neigh = find_closest_pos_neigh(pA,segment,i);
         	for (;(pos_neigh>=0 && fpos < 8);pos_neigh--) {
				if (Alignment((neigh= pA->Partition[segment].list[pos_neigh]) ) > -1) {
					fixedmatch[fpos][0] = neigh;
					fixedmatch[fpos++][1] = Alignment(neigh);
				}
		}
	}
	fixedmatch[fpos][0] = -1;

	if (fpos <= 3) 
   // Not enough points: do not consider them anymore
		for (i=iinit; i <= ifin; i++)
			Alignment(i) = -1;
	else {
  		T = RigidTransform(pA, pB, fixedmatch);
		for (p=0; p < pmax; p++) {
			m(0) = pB->PosCA[j+p][0];
			m(1) = pB->PosCA[j+p][1];
			m(2) = pB->PosCA[j+p][2];
			m(3) = 1;
			m = T*m;
			Rpoint(0,j+p) = m(0);
			Rpoint(1,j+p) = m(1);
			Rpoint(2,j+p) = m(2);
		}
        AddInterval(T, jinit, jinit+p-1, iinit, iinit+p-1);
	}     
    
  	i += p; 
  	j += p;
  	centerA = i-1;  
  	centerB = j-1;
  	while (i <= ifin) {
		fpos = 0;
        for (p=0; (i+p <= ifin) && ((ifin-i-p < 3) || (p < 7) ||	// Do not break SSE
				(i+p < ifin) && ((pA->Sec[i+p] == 'H') || (pA->Sec[i+p]=='E')) && 
				(pA->Sec[i+p] == pB->Sec[j+p]) && 
				(pA->Sec[i+p+1] == pA->Sec[i+p+1]) &&
				(pA->Sec[i+p] == pA->Sec[i+p+1])); p++) {
			fixedmatch[fpos][0]=i+p;
			fixedmatch[fpos++][1]=j+p;          
		}
        
        pmax = p;
        if (p < 6) {
			padd = 6-p;
	   		for (p=1; (p <= padd) && (i-p >= iinit); p++) {
	   			fixedmatch[fpos][0] = i-p;
	   			fixedmatch[fpos++][1] = j-p;
	   		}
        }
        fixedmatch[fpos][0] = -1;
        if (fpos > 3){  											// Always true
			T = Rotation(pA, pB, fixedmatch, centerA, centerB, Rpoint);
			for (p=0; p < pmax; p++) {
				m(0) = pB->PosCA[j+p][0];
				m(1) = pB->PosCA[j+p][1];
				m(2) = pB->PosCA[j+p][2];
				m(3) = 1;
				m = T*m;
				Rpoint(0,j+p) = m(0);
				Rpoint(1,j+p) = m(1);
				Rpoint(2,j+p) = m(2);
				/* if (p==0) {
                   double xx=Sqr(Rpoint(0,j)-Rpoint(0,j-1))+Sqr(Rpoint(1,j)-Rpoint(1,j-1))+Sqr(Rpoint(2,j)-Rpoint(2,j-1));
                   double yy=Sqr(Mpoint(0,j)-Mpoint(0,j-1))+Sqr(Mpoint(1,j)-Mpoint(1,j-1))+Sqr(Mpoint(2,j)-Mpoint(2,j-1));
                   printf("%d %lf %lf\n",j,xx,yy);
               	} */                        
			}
            AddInterval(T, j, j+p-1, i, i+p-1);
        }       
        i += pmax; 
        j += pmax;
        centerA = i-1;
        centerB = j-1;
  	}
}


void find_transformation_gap_begin(struct PROTEIN *pB, int prev,  int sig) {
	Matrix<double> T(4,4);
	Vector<double> m(4);
	int p;
	
	T = SearchListTrans(prev);
	for (p=prev+1; p < sig; p++) {
		m(0) = pB->PosCA[p][0];
		m(1) = pB->PosCA[p][1];
		m(2) = pB->PosCA[p][2];
		m(3) = 1;
		m = T*m;
		Rpoint(0,p) = m(0);
		Rpoint(1,p) = m(1);
		Rpoint(2,p) = m(2);
	}
	AddInterval(T, prev+1, sig-1, -1, -1);
}


void find_transformation_gap_end(struct PROTEIN *pB, int prev, int sig) {
	Matrix<double> T(4,4);
	Vector<double> m(4);
	int p;
	
	T = SearchListTrans(sig);
	for (p=prev+1; p < sig; p++) {
		m(0) = pB->PosCA[p][0]; 
		m(1) = pB->PosCA[p][1];
		m(2) = pB->PosCA[p][2];
		m(3) = 1;
		m = T*m;
		Rpoint(0,p) = m(0);
		Rpoint(1,p) = m(1);
		Rpoint(2,p) = m(2);
	}
	AddInterval(T, prev+1, sig-1, -1, -1);
}


void find_transformation_gap_free(struct PROTEIN *pB, int prev, int sig) {
	Vector<double>m(4);
	int p;
	
	for (p=prev+1; p < sig; p++){
		m(0) = pB->PosCA[p][0]; 
		m(1) = pB->PosCA[p][1];
		m(2) = pB->PosCA[p][2];
		Rpoint(0,p) = m(0);
		Rpoint(1,p) = m(1);
		Rpoint(2,p) = m(2);
	}
	// AddInterval(Id,prev+1,sig-1,-1,-1);
}


void find_transformation_gap(struct PROTEIN *pB, int prev, int sig) {
	Matrix<double> Rot(3,3); 
	Matrix<double> T(4,4); 
	Vector<double> v(3), w(3), r(3), axis(3), m(4), pos(3);
	double costheta;
	double wnorm, posnorm;

	v(0) = pB->PosCA[sig][0] - pB->PosCA[prev][0];
	v(1) = pB->PosCA[sig][1] - pB->PosCA[prev][1];
	v(2) = pB->PosCA[sig][2] - pB->PosCA[prev][2];
	w(0) = Rpoint(0,sig) - Rpoint(0,prev);
	w(1) = Rpoint(1,sig) - Rpoint(1,prev);
	w(2) = Rpoint(2,sig) - Rpoint(2,prev);

	wnorm = norm_vect(w) - 3.7 * 3.7;

	axis = cross_product(v, w);
	axis = normalize_vect(axis);
	costheta = cosangle(v, w);
	Rot = rodrigues_formula(axis, costheta);

	// Rot has the rotation matrix from v to w
	T = 0;
	int i, j;
	for (i=0; i < 3; i++)
		for (j=0; j < 3; j++)
			T(i,j) = Rot(i,j);

    T(3,3) = 1.0;
    T(0,3) = Rpoint(0,prev) - T(0,0) * pB->PosCA[prev][0] - T(0,1) * pB->PosCA[prev][1] - T(0,2) * pB->PosCA[prev][2];
    T(1,3) = Rpoint(1,prev) - T(1,0) * pB->PosCA[prev][0] - T(1,1) * pB->PosCA[prev][1] - T(1,2) * pB->PosCA[prev][2];
    T(2,3) = Rpoint(2,prev) - T(2,0) * pB->PosCA[prev][0] - T(2,1) * pB->PosCA[prev][1] - T(2,2) * pB->PosCA[prev][2];
    int lastpos = sig-1;
    for (i=prev+1; i < sig; i++) {
		m(0) = pB->PosCA[i][0];  
		m(1) = pB->PosCA[i][1];
		m(2) = pB->PosCA[i][2];
		m(3) = 1;
		pos(0) = m(0) - pB->PosCA[prev][0];
      	pos(1) = m(1) - pB->PosCA[prev][1];
		pos(2) = m(2) - pB->PosCA[prev][2];
		posnorm = norm_vect(pos);
		
		if ((lastpos == sig-1) && (posnorm > wnorm)) lastpos = i-1;
		
		m = T * m;
		Rpoint(0,i) = m(0);
		Rpoint(1,i) = m(1);
		Rpoint(2,i) = m(2);
	}

    if (lastpos >= prev+1) AddInterval(T, prev+1, lastpos, -1, -1);

    if (sig-1 >= lastpos+1) AddInterval(T, lastpos+1, sig-1, -2, -2);
}


void PrintIntervals(struct PROTEIN *pA, struct PROTEIN *pB, Matrix<double> &TRigid) {
	FILE *f;
	list <struct interval>::iterator interv, interv2;
	int i, j, r;
        char nameint[100];
	
        sprintf(nameint,"%s.%s.%s","intervals",pA->cname,pB->cname);
	f = fopen(nameint, "w");

	fprintf(f, "GLOBAL TRANSFORMATION \n");   
	fprintf(f, "%s %s\n", pB->Rnum[1], pB->Rnum[pB->Naa]);
	fprintf(f, "%s %s\n", pA->Rnum[1], pA->Rnum[pA->Naa]);
	for (i=0; i < 4; i++) {
		for (j=0; j < 4; j++)
			fprintf(f, "%2.3f ", TRigid(i,j));
	fprintf(f, "\n");
	}      

	for (interv=TransL.begin(), r=0; interv != TransL.end(); interv++, r++) {
		fprintf(f, "SEGMENT %d\n", r);   
		// fprintf(f,"%d %d\n",interv->start,interv->end);
		// fprintf(f,"%d %d\n",interv->starti,interv->endi);
		fprintf(f, "%s %s\n", pB->Rnum[interv->start], pB->Rnum[interv->end]);
		
		if (interv->starti > 0)
			fprintf(f, "%s %s\n", pA->Rnum[interv->starti], pA->Rnum[interv->endi]);
		else
			fprintf(f, "%d %d\n", interv->starti, interv->endi);
		
		for (i=0; i < 4; i++) {
			for (j=0; j < 4; j++)
				fprintf(f, "%2.3f ", interv->T[i][j]);
		fprintf(f, "\n");
		}
	}
     
	// Print the dotted lines
	for (interv=interv2=TransL.begin(), r=0; interv != TransL.end(); interv++) {
	   if (interv != interv2) {                
		i = interv2->end;
       		j = interv->start; 
       		if (interv->starti == -2) continue;  // skip deleted ones 
       		if (Sqr(Rpoint(0,i) - Rpoint(0,j)) + Sqr(Rpoint(1,i) - Rpoint(1,j)) + Sqr(Rpoint(2,i) - Rpoint(2,j)) > 16) { // Two adjacent CA are at about 3.7 A from each other in the same chain
				fprintf(f, "DOTTED LINE %d\n", r++);   
				fprintf(f, "%s %s\n", pB->Rnum[i], pB->Rnum[j]);
       			// fprintf(f,"%2.3lf %2.3lf %2.3lf %2.3lf %2.3lf %2.3lf\n",Rpoint(0,i),Rpoint(1,i),Rpoint(2,i),
				// Rpoint(0,j),Rpoint(1,j),Rpoint(2,j));
                }
    	   }    	
    	   interv2 = interv;
	}
  	fclose(f);
}


void DeformPDB(struct PROTEIN *proA, struct PROTEIN *proB, Vector<int> Alignment) {
	int  i, p, j;


	//FILE* fp1; 	// for original D   D= ProtA  M = ProtB
	//FILE* fp2;	// for R (loc. Trans of M into D)
	// FILE* fp3; 	// for rigid transform of M

	Rpoint.resize(3,proB->Naa+1);


 	//C-alpha coordinates for pdb's rotation files
	//fp1 = fopen("1.coo", "wb");	// for original D
	//fp2 = fopen("2.coo", "wb"); // for R (loc. Trans of M into D)
	//fp3 = fopen("3.coo", "wb"); // for rigid transform of M
	Vector<double> m(4), r(3);        

	// Find the best rigid transformation from M to D
	Matrix<double> TRigid(4,4);
	int TransCenter, RigidTransform;

	// Transformation  from M to D
 	TRigid = RigidTransformBasic(proA, proB, Alignment, &TransCenter);
	// RigidTransform=(TransCenter!=-1);
	if (RigidTransform) {      
		for (j=1; j <= proB->Naa; j++) {
			m(0) = proB->PosCA[j][0];
			m(1) = proB->PosCA[j][1];
			m(2) = proB->PosCA[j][2];
			m(3) = 1;
			m = TRigid * m;
 			proB->PosCA[j][0] = m(0);
			proB->PosCA[j][1] = m(1);
			proB->PosCA[j][2] = m(2);                       
			//fprintf(fp3, "%i %f %f %f\n", j, m(0), m(1), m(2));
		}
	}

	int matchedsegment = 0;
	int jinit, jfin, iinit, ifin;
	int antj;
    i = 1; 
	while (i <= proA->Naa) {
		if (!matchedsegment) {
			if ((j=Alignment(i)) > -1) {  // Start a new matched segment
				jinit = j;
				iinit = i;
             	matchedsegment = 1;
             	antj = j;
			}
		}
        else 
			if (((j=Alignment(i)) > -1) && (j == antj +1) && 
					(proB->Chains[j] == proB->Chains[jinit]) && (proA->Chains[i] == proA->Chains[iinit])) {
				antj = j;
         	}
         	else {
            	jfin = antj; 
            	ifin = i-1;     // end a matched segment
            	matchedsegment = 0;
            	findtransformationsegment(proA, proB, Alignment, iinit, ifin, jinit, jfin);
             	if (Alignment(i) > -1) i--;
         	}
      	i++;
	}
                                     // For the last matched segment : 
	if (matchedsegment) {
		jfin = antj; 
		ifin = i-1;
		findtransformationsegment(proA, proB, Alignment, iinit, ifin, jinit, jfin);
	}   

	// The segments not matched are transformed according to their matched endings in the same chain
	antj = 0;
    for (i=1; i <= proA->Naa; i++)
    	if (Alignment(i) > -1 ) {
			j = Alignment(i);
			if (j > antj+1) {   	// There is a gap between antj and j, both excluded.
				if (antj >0 ) {   	// There is a previous matching
					// Calculate the transformation from the vector (M(antj),M(j)) into the vector (R(antj),R(j)) 
					// Transform all the gap using the transformation if it fits, otherwise paint a dotted line.
                 	if (proB->Chains[antj] == proB->Chains[j])
						find_transformation_gap(proB, antj, j);
					else if (proB->Chains[antj] == proB->Chains[antj+1])
						find_transformation_gap_begin(proB, antj, j);
                 	else if (proB->Chains[j-1] == proB->Chains[j])
                   		find_transformation_gap_end(proB, antj, j);
					else if (RigidTransform) 
						find_transformation_gap_free(proB, antj, j);
				}
                else { 	// No previous matching
					if (proB->Chains[j-1] == proB->Chains[j])
						find_transformation_gap_end(proB, antj, j);
                 	else if (RigidTransform) 
                 		find_transformation_gap_free(proB, antj, j);
				}
			}
			antj = j;
		}
      
		// For the last part of B not matched
		if (antj < proB->Naa)
			if ((antj != -1) && (proB->Chains[antj] == proB->Chains[antj+1]))
				find_transformation_gap_begin(proB, antj, proB->Naa+1);
			else if (RigidTransform) 
				find_transformation_gap_free(proB, antj, proB->Naa+1);

	/*	// Print the coordenates in Rpoint
   		if (RigidTransform)
   			for (j=1; j <= proB->Naa; j++)
	  			fprintf(fp2, "%i %f %f %f\n", j, Rpoint(0,j), Rpoint(1,j), Rpoint(2,j));
        */
		TransL.sort(bystart);
		PrintIntervals(proA, proB, TRigid);
	//	fclose(fp1);
	//	fclose(fp2);
	//	fclose(fp3); 
}


char typeSSE(char x) {
	switch (x) {
		case 'E': return('E');		
		case ' ': return('-');	
		case '-': return('-');		
		case 'H': return('H');		
		case 'T': return('-');		
		case 'S': return('-');		
		case 'G': return('-');		
		case 66: return('-');		
	}
}
