/***

<protein.h>

Definition of 
 (i)    Structure for protein (struct PROTEIN)
 (ii)   Structure for SSE     (struct SecStrEle)
 (iii)  Structure for 2D-MAP  (struct MAP2D)

***/

/*JR */

struct nodeneighbour {
			int ending; //Partition[segment].ending is the last element of the segment
			int *list;
};
typedef struct nodeneighbour *NodeNeigh;


/*************************************/
/**** SECONDARY STRUCTURE ELEMENT ****/
/*************************************/

struct SecStrEle{
  float   G[3];     /* Coordinate of mass center */
  float   V[3];     /* Direcrion Vector */
  float   Pb[3];    /* Beginning Point  */
  float   Pe[3];    /* End Point           */
  float   Pefar[3]; /* Far away End Point  */
  float   Veb[3];   /* Pe - Pb             */
  float   M[3];     /* Mid point (Pb+Pe/2) */
  int     Rsta;     /* Start Residue Num */
  int     Rend;     /* End   Residue Num */
  int     L;        /* Rend - Rsta + 1  */
  int     num;      /* Helix Number, Sheet Number  */
  char    type;     /* 'H'elix or 'E'xtend */
};


/*************************************/
/************** PROTEIN **************/
/*************************************/

struct PROTEIN 
{
 char   fname[128];    /* file name ("/DB/BSSP/ti/1timA.bssp") */
 char   pname[128];     /* protein name ("1timA.bssp")          */
 char   cname[32];     /* Cutted filename ("1timA")            */
 char   FileType;      /* 'B'ssp, 'D'ssp, 'P'db'               */  
 char   Tax[128];      /* Structural Taxonomy                  */  
 char   *header;       /* "HEADER" of PDB                      */
 char   *compnd;       /* "COMPND" of PDB                      */
 int    Naa;           /* Number of Amino Acids                */
 int    MaxNaa;        /* Malloced Residue Number              */ 
 int    MaxNsse;       /* Malloced SSE     Number              */ 
 char   AngMallocType;
 char   XyzMallocType;
 char   chain;      
 int    Nbreak;
 float  SelfSdis;      /* Self Potential Value of Sdis */
 float  SelfSsse;      /* Self Potential Value of Ssse */

 /* Array of [NumSeq] is defined as range from '1' to 'Naa'. */  
 char   **Rnum;   /* [NumSeq][6]  */
 char   *Seq;     /* [NumSeq]     */
 char   *Sec;     /* [NumSeq]     */
 char   *Exp;     /* [NumSeq]     */
 char   *Top;     /* [NumSeq]     */
 char   *Break;   /* [NumSeq]     */   

 int	**Adj;    /* [NumSeq]     Adjacents   PD JR*/
 int    *ACC;     /* [NumSeq]     */ 
 int    *BP1;     /* [NumSeq]     */
 int    *BP2;     /* [NumSeq]     */

 /*** ANGLES ****/
 float  *Phi;     /* [NumSeq]     */
 float  *Psi;     /* [NumSeq]     */
 float  *Omega;   /* [NumSeq]     */


 /*** XYZ COORDINATES  ****/
 float **PosCA;   /* [NumSeq][3] */
 float **PosCB;   /* [NumSeq][3] */
 
 float **PosN;    /* [NumSeq][3] */
 float **PosC;    /* [NumSeq][3] */
 float **PosO;    /* [NumSeq][3] */
 
 /**** Secondary Structure Element Data ****/
 /* Arrray of [NumSSE] is defined as range from '0' to 'Nsse-1'. */  
 int              Nsse;       /* Number of SEE */
 struct SecStrEle *SSE;       /* [NumSSE] */
 char             *SseSeq;    /* [NumSeq] */
 float           **ssePh;     /* ssePh[NumSSE][NumSSE] Phi */
 float           **sseT1;     /* sseT1[NumSSE][NumSSE] theta1 */
 float           **sseT2;     /* sseT2[NumSSE][NumSSE] theta2 */
 float           **sseDi;     /* sseD1[NumSSE][NumSSE] Distance */

 int       *Chains;       /* JR  the Chain to which each aminoacid belongs */
 NodeNeigh Partition;     /* JR  list of segments and their neighbours */
};




/************************************/
/**** TWO-DIMENTIONAL FLOAT MAP *****/
/************************************/

struct MAP2D{
 int   N;      /* Malloced Number        */
 float **map;  /* 2D valiable: map[N][N] */ 
};



/********************************************/
/*** PARAMETERS FOR 3D PAIRWISE ALIGNMENT  **/
/********************************************/


struct ALIGN_PARAMETER{
 char COMMAND[1024];
 char SseAli;     /*  SSEscore Alignment ('T' or 'F') */
 char EnvAli;     /*  ENVscore Alignment ('T' or 'F') */
 char DisAli;     /*  DISscore Alignment ('T' or 'F') */
 char SelfAvoidType;  /* 'T' or 'F' */
 char AlgType;        /* Algorithm of DP: 'L'ocal,'N'ormal,'B'oth-edge-gap-free */
 char SubOpt;         /* 'T' or 'F' */
 char SeqAliType;     /* 'T' or 'F' */ 
 /*** Constraint ***/
 int *ConAtoB;      /* Constraint [NaaA] (malloc later)  */ 
 float ConOffset;  /* Offset score value for constraint */

 /*** 1st PHASE  :BuildUp using SseScoer ***/
 int  Nkeep;           /* Number of candidates kept in memory */
 float  GapExtS;       /* Gap Penalty for SSE */
 char   OtypeS;

 /***  2nd PHASE  : DP using EnvScore ***/
 float GapExtE;       /* Gap penalty = GapExtE*(k-1) + GapOpenE */
 float GapOpenE;      /* Gap penalty = GapExtE*(k-1) + GapOpenE */
 char   EnvScType;    /* Type of EnvScore */ 
 int    Nenvstate;    /* Number of States for EnvScore */ 
 float SseOffsetE;    /* SseOffset Score for matched region (default is |alpha+beta|) */   
 float NonSseOffsetE; /* SseOffset Score for non-matched region (only for self-avoiding)*/   
 float Assap,Bssap;   /* Parameter for SSAP like score : A/(|Da-Db|+B) */
 /***  3rd PHASE  : Iterative DP using DisScore ***/
 float GapExtD;       /* Gap penalty = GapExtD*(k-1) + GapOpenD */
 float GapOpenD;      /* Gap penalty = GapExtD*(k-1) + GapOpenD */
 int    Nrepeat;      /* Number of Maximum iteration    */ 
 char   DisScType;    /* Type of DisScore */ 
 float SseOffsetD;    /* SseOffset Score for matched region (default is zero) */   
 float NonSseOffsetD; /* SseOffset Score for non-matched region (not used) */   
 char   OtypeD;
 int    Wmark;        /* Width of masqued windows for pre-alignment */
 /***   ProtDeform PHASE JR ***/
 int    MAXNEIGHBOURS;   /* Number of nearest neighbours for each site */
 double squaredratio;    /* standard Gaussian ratio for converting distances to score */
 double MAXDistExt;      /* After this distance, Gap is maybe preferred */
 int OutAlignment;       /* Flag for alignment output */
 int OutPDBdeform;       /* Flag for PDB deformation output */
 /** for SubOptimal SSE Alignment **/
 char    mal_alissemap;  /* 0(no-malloc) or 1(malloc) */
 char    **AliSSEmap;   /* [NsseA][NsseB] (malloc later), 0(non-avoid) or 1(avoid) */
 float  Rsse_sub;      /* Threshold Value for Rsse */
 
 /** for SubOptimal Residue Alignment **/
 char   mal_alimap;    /* 0(no-malloc) or 1(malloc) */
 char    **Alimap;     /* [NaaA][NaaB] (malloc later), 0(non-avoid) or 1(avoid)  */
 float  Rdis_sub;     /* Threshold Value for Rdis */ 
};

extern struct ALIGN_PARAMETER PAR;  /* Globally defined alignment parameter */
