/*

 <BuildUpL.c>

 Functions for Building Up SSE elements
 to construct SSE-SSE alignment.
 Alignment data is coded using linked-list style. 

 coded by Takeshi Kawabata.

 >> Data Structure of Nkeep alignments << 

                      <---- Nkeep alignments ----> 
 head -> [Nkeep+1] -> [Nkeep] -> .... -> [2] -> [1] 
          node for          
          new alignment

  * Number of nodes does not change during the BuildUp calculations. 


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "ScoreSSE.h"
#include "SecStrEle.h"
#include "main.h"
#include "protein.h"
#include "LlistAlign.h"


/***** Function (GLOBAL) ****/
void Make_AlignNode_LList();
void Make_Initial_SSE_Alignments();
void Build_Up_LL();
void Write_SSE_Alignment();
void Write_SSE_Alignment_Rasmol();
void Write_SSE_Alignment_Rasmol_Color();
void Read_SSE_Alignment();
void Show_SSE_Ali();

/***** Function (LOCAL) ****/
static void Set_NewNode_On_Proper_Position();
static void Show_Nodes();
static void Set_useA_useB();
static int SeqOrderCheck_AddSSE();
static int SelfAvoidCheck_AddSSE();
static void Make_Added_SSE_Align();
static int Redundancy_Check();
static void Equal_AlignNode_Llist();
static int SSE_Number();



void Make_AlignNode_LList(head,Nnode,MaxNali)
 struct AlignNode *head;
 int Nnode;
 int MaxNali; 
{

 /**
   For making [Nnode] AlignNodes in linked list form.
 **/

 float inf;
 int n;
 struct AlignNode *bn;

 inf = 1000.0;

 head->value = head->subvalue = -inf; 
 head->num = head->MaxNali = head->mal_sim = head->mal_pro = 0;

 bn = head;
 for (n=0;n<Nnode;++n)
 {
   bn->next = (struct AlignNode *)malloc(sizeof(struct AlignNode)); 
   bn->next->prev  = bn; 
   bn->next->next  = NULL;

   bn->next->nalA = (int *)malloc(sizeof(int)*(MaxNali+1));
   bn->next->nalB = (int *)malloc(sizeof(int)*(MaxNali+1));
   bn->next->scores = (float *)malloc(sizeof(float)*(MaxNali+1)); /*JR */
   bn->next->MaxNali = MaxNali;
   bn->next->Nali    = 0;
   bn->next->value = bn->subvalue = -inf;
   bn->next->num = n + 1; 
   bn->next->mark = 0;   
   bn->next->mal_sim = 0;  
   bn->next->mal_pro = 0;
  
   bn = bn->next;
  }

} /* end of Make_AlignNode_LList() */




void Make_Initial_SSE_Alignments(head,pA,pB,Nkeep)
 struct AlignNode *head;
 struct PROTEIN *pA,*pB;
 int Nkeep;
{ int i,j,x,y;
  struct AlignNode *sn;
  float WorstSc;
 
  sn = head->next;

  for (i=0;i<pA->Nsse;++i)
   for (j=0;j<pB->Nsse;++j)
    if (pA->SSE[i].type == pB->SSE[j].type)
    { 
     for (x=i+1;x<pA->Nsse;++x)
      for (y=j+1;y<pB->Nsse;++y)
       if (pA->SSE[x].type == pB->SSE[y].type)
        if ((PAR.SelfAvoidType != 'T')|| ((i<j)&&(x<y)))
          if ((PAR.mal_alissemap == 0) || 
              ((PAR.AliSSEmap[i][j] == 0) && (PAR.AliSSEmap[x][y] == 0) )  )
        {    
          sn = head->next;
          sn->Nali = 2; 
          sn->nalA[0] = i; sn->nalA[1] = x;
          sn->nalB[0] = j; sn->nalB[1] = y;
 
          sn->value = ScoreSSE(pA,pB,sn) + SSE_Gap_Penalty(pA,pB,sn);
                
          /* printf("i %d j %d  x %d y %d Nali %d/%d Sc %f WorstSc %f\n",
               i,j,x,y,sn->Nali,sn->MaxNali,sn->value,sn->next->value); 
              */
 
        if (sn->value > sn->next->value) Set_NewNode_On_Proper_Position(head);

       } /* if x.type == y.type */  

    } /* if i.type == j.type */


} /* end of Make_Initial_SSE_alignments() */






void Build_Up_LL(pA,pB,head)
 struct PROTEIN *pA,*pB;
 struct AlignNode *head; 
{
  int i,x,j,y,k,p,Nadd,Ncount;
  int OneAdded;  /* Used to check if a matching is extended on not JR*/
  char *useA,*useB;
  float BestSc,BestSc_pre,ScAdd,WorstSc;
  struct AlignNode PreHead,CurHead;
  struct AlignNode *bn,*pn,*qn,*nn,*cn;

  useA = (char *)malloc(sizeof(char)*(pA->Nsse+1));
  useB = (char *)malloc(sizeof(char)*(pB->Nsse+1));

  Make_AlignNode_LList(&PreHead,(PAR.Nkeep+1),pA->Nsse+pB->Nsse);
  Make_AlignNode_LList(&CurHead,(PAR.Nkeep+1),pA->Nsse+pB->Nsse);
 
  Make_Initial_SSE_Alignments(&PreHead,pA,pB,PAR.Nkeep);
  Ncount = 1; 

  BestSc  = Tail_Value_AlignNode(&PreHead);
  WorstSc =  PreHead.next->next->value;
  if (PAR.OtypeS == 'O') printf("Ncount %d BestSc %f WorstSc %f\n",Ncount,BestSc,WorstSc);
 
  /*** BUILD UP MAIN LOOP ***/
  do{
  
  BestSc_pre = BestSc;

  pn = PreHead.next;
  Nadd = 0; 

  while (pn->next != NULL) 
  {
    pn = pn->next; 
    Set_useA_useB(pA,pB,pn,useA,useB);
    OneAdded = 0;
    if (pn->Nali > 0) 
    for (i=0;i<pA->Nsse;++i)
     if (useA[i]==0) 
     for (j=0;j<pB->Nsse;++j)
      if (useB[j]==0) 
       if (pA->SSE[i].type == pB->SSE[j].type)    
        if ((PAR.SelfAvoidType != 'T')|| (SelfAvoidCheck_AddSSE(pn,i,j)==1) )
        if (SeqOrderCheck_AddSSE(pn,i,j) == 1)
          if ((PAR.mal_alissemap == 0) || (PAR.AliSSEmap[i][j] == 0))   
         {
          nn = CurHead.next;
          Make_Added_SSE_Align(nn,pn,i,j);
 
         if (Redundancy_Check(CurHead.next,nn) == 1)
          {
           ScAdd = ScoreSSE_AddPair(pA,pB,pn,i,j)
                  - SSE_Gap_Penalty(pA,pB,pn) + SSE_Gap_Penalty(pA,pB,nn);
          nn->value = pn->value + ScAdd;
    /*
          nn->value = ScoreSSE(pA,pB,nn) + SSE_Gap_Penalty(pA,pB,nn);
   */

          if (nn->value > nn->next->value) 
           { Set_NewNode_On_Proper_Position(&CurHead);
             OneAdded=1;
             ++Nadd; }
 
         } /* Redundant = ok */

      } /* i,j loop*/ 

      if (OneAdded == 0 && pn->Nali > 0) {  /* JR pn was not extended: copy it. */
           nn = CurHead.next;
           pn->Nali--;
           Make_Added_SSE_Align(nn,pn,pn->nalA[pn->Nali],pn->nalB[pn->Nali]);
        if (Redundancy_Check(CurHead.next,nn) == 1)
          {
          
          nn->value = pn->value;
 

          if (nn->value > nn->next->value) 
           { Set_NewNode_On_Proper_Position(&CurHead);
            }
 
         } /* Redundant = ok */



      } /* If pn not extended JR*/
  } /* while(pn) */

   ++Ncount;

   BestSc = Tail_Value_AlignNode(&CurHead);
   WorstSc = CurHead.next->next->value;
   if (PAR.OtypeS=='O') printf("##Ncount %d BestSc %f WorstSc %f\n",Ncount,BestSc,WorstSc);
   bn = Tail_AlignNode(&CurHead); 
   /* Show_SSE_Ali(bn);   JR */
  
   Equal_AlignNode_Llist(&PreHead,&CurHead);
 
  
 } while ((Nadd>0)&&(BestSc!=BestSc_pre)); 


 /*** ENDING PROCEDURES ***/ 
 
 /* PreHead->next becomes head->next. */ 
 head->next = PreHead.next->next; 
 Merge_Sort_Double_Linked_List(head,'R'); 
 
 Free_AlignNode(PreHead.next);

 /* Free CurHead */ 
 Free_AlignNode_Llist(&CurHead);
 free(useA); useA = NULL;
 free(useB); useB = NULL;

} /* end of Build_Up_LL() */








void Set_NewNode_On_Proper_Position(head)
 struct AlignNode *head;
{
 /* >>Example<<  
 before : [head] -> [new(3)] -> [p(1)] -> [b(2)]   -> [a(4)] -> [e(5)]
 after  : [head] -> [p(1)]   -> [b(2)] -> [new(3)] -> [a(4)] -> [e(5)]
 */
 struct AlignNode *new,*bn;
 int hit; 

 new = head->next;
 
 /** Find Position To Be Set **/ 
 bn = new->next;
 hit = 0;
 while ((bn->next !=NULL)&&(hit==0))
 {
   bn = bn->next;
   if (bn->value <= new->value)
   {      if (bn->next == NULL) hit = 1;
     else if (new->value <= bn->next->value) hit = 1; }
 }

 /** Set **/ 
 if (hit==1)
 { 
  /* before: head -> new   ->  newnew  
     after : head -> newnew           */ 
    
   head->next->next->prev = head;
   head->next = head->next->next;
 
  /* before: bn -> an  
     after : bn -> new -> an */ 
  
   if (bn->next == NULL) new->next = NULL;
    else {  new->next = bn->next; 
            bn->next->prev = new; }

   bn->next  = new;
   new->prev = bn;
 }

} /* end of Set_Head_On_Proper_Position() */




void Show_Nodes(head)
 struct AlignNode *head;
{
  struct AlignNode *bn;
  int n;

 printf("[BEGIN]\n"); fflush(stdout);
  bn = head;
  n = 0;
  while (bn->next!=NULL)
  {
    bn = bn->next; 
    printf("[%d] num %d prev %d Nali %d val %f %d %d\n",n,
            bn->num,bn->prev->num,bn->Nali,bn->value,bn->mal_sim,bn->mal_pro);
    ++n;
  }

 printf("[END]\n"); fflush(stdout);

} /* end of Show_Nodes() */


void Show_SSE_Ali(bn)
 struct AlignNode *bn;
{
 int i;
 printf(">Nali %d\n",bn->Nali);
 for (i=0;i<bn->Nali;++i)
  printf("A %d B %d\n",bn->nalA[i],bn->nalB[i]);
 printf("//\n");

} /* end of Show_SSE_Ali() */



void Set_useA_useB(pA,pB,sn,useA,useB)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn;
 char *useA,*useB;
{
  int p;
  for (p=0;p<pA->Nsse;++p) {useA[p] = 0;}
  for (p=0;p<pB->Nsse;++p) {useB[p] = 0;}

  for (p=0;p<sn->Nali;++p)  
    { useA[sn->nalA[p]] = 1; useB[sn->nalB[p]] = 1; } 
 
} /* end of  Set_useA_useB() */


int SeqOrderCheck_AddSSE(sn,s,t)
 struct AlignNode *sn;  /* previous alignment */
 int s,t;               /* new SSEpairs to be added */
{ int i,j;

  /* if Seq_Order is OK, return(1)  else return(0).  */

  for (i=0;i<sn->Nali;++i)
   if (((sn->nalA[i]>s)&&(sn->nalB[i]<t)) ||
       ((sn->nalA[i]<s)&&(sn->nalB[i]>t)))  return(0);
 return(1);

} /* end of SeqOrderCheck_AddSSE() */



int SelfAvoidCheck_AddSSE(sn,s,t)
 struct AlignNode *sn;  /* previous alignment */
 int s,t;               /* new SSEpairs to be added */
{ int i,j;
  int minA,maxA,minB,maxB;

  /* if Self_Avoid is OK, return(1)  else return(0).  */
  /* SelfAvoid Condition bwn s->nalA[] and sn->nalB  */ 
  /*   minA--maxA...minB..maxB                       */
  /*            or                                   */
  /*   minB--maxB...minA..maxA                       */
   
  if (s==t) return(0);
 
  minA = sn->nalA[0]; 
  maxA = sn->nalA[sn->Nali-1]; 
  minB = sn->nalB[0]; 
  maxB = sn->nalB[sn->Nali-1]; 

  if (s < minA) minA = s;
  if (s > maxA) maxA = s;
  if (t < minB) minB = s;
  if (t > maxB) maxB = s;

  if ((maxA < minB) || (maxB < minA)) return(1); 
                                else  return(0);

} /* end of SelfAvoidCheck_AddSSE() */





void Make_Added_SSE_Align(nn,pn,new_i,new_j)
 struct AlignNode *nn,*pn;
 int new_i,new_j;
{
 int i,p,rank,ok;

 /** Add new_i and new_j into "proper" position  **/

 ok = 0; i = 0;
 while ((ok==0)&&(i<pn->Nali)) 
 { if (new_i < pn->nalA[i]) { ok = 1; rank = i;}
   ++i;  }
 if (ok==0) rank = pn->Nali;

 for (i= pn->Nali; i>rank;--i) 
  { nn->nalA[i] = pn->nalA[i-1]; 
    nn->nalB[i] = pn->nalB[i-1]; }

 nn->nalA[rank] = new_i;  
 nn->nalB[rank] = new_j;  
 
 for (i=(rank-1);i>=0;--i) 
  { nn->nalA[i] = pn->nalA[i]; 
    nn->nalB[i] = pn->nalB[i]; }

  nn->Nali = pn->Nali + 1;

} /* end of Make_Added_SSE_Align() */




int Redundancy_Check(head,new)
 struct AlignNode *head,*new;
{
 int i,miss;
 struct AlignNode *bn;

 /*
 if new is unique is among @head, return(1); else return(0);
 */

 bn = head;
 while (bn->next != NULL)
 {
  bn = bn->next; 
  if (bn->Nali == new->Nali) 
  { miss = 0;
   for (i=0;i<bn->Nali;++i)
    if ((bn->nalA[i]!= new->nalA[i])||(bn->nalB[i]!=new->nalB[i])) ++miss;
   if (miss==0) return(0);
  }
 }

 return(1);

} /* end of Redundancy_Check() */



void Equal_AlignNode_Llist(ahead,bhead) /* a := b; */
 struct AlignNode *ahead,*bhead;
{
 struct AlignNode *a,*b;
 int i;

 a = ahead; b = bhead;
 while ((a->next !=NULL) && (b->next != NULL))
 {
   a = a->next;
   b = b->next;
   if (a->MaxNali != b->MaxNali)
   {
     printf("#ERROR (in Equal_AlignNode_Llist): different MaxNali\n");
     exit(1);
    }
   a->MaxNali  = b->MaxNali;
   for (i=0;i<b->Nali;++i) 
    { a->nalA[i] = b->nalA[i]; a->nalB[i] = b->nalB[i]; }
   a->Nali   = b->Nali;
   a->num    = b->num;
   a->value  = b->value;
   a->mal_sim  = b->mal_sim;
   a->mal_pro  = b->mal_pro;
  }

} /* end of Equal_AlignNode_Llist() */


void Write_SSE_Alignment(pA,pB,sn,fname)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn; 
 char *fname;
{
 FILE *fp;
 int i,k;

 printf("#Write_SSE_Alignment() --> \"%s\"\n",fname);
 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#ERROR:Can't write to ssefile \"%s\"\n",fname); exit(1);}
 fprintf(fp,"#PRO1 %s\n",pA->pname);
 for (i=0;i<pA->Nsse;++i)
   fprintf(fp,"#SSE %c%d %s - %s\n",
    pA->SSE[i].type, pA->SSE[i].num,
    pA->Rnum[pA->SSE[i].Rsta], pA->Rnum[pA->SSE[i].Rend]);
 fprintf(fp,"#PRO2 %s\n",pB->pname);
 for (i=0;i<pB->Nsse;++i)
   fprintf(fp,"#SSE %c%d %s - %s\n",
    pB->SSE[i].type, pB->SSE[i].num,
    pB->Rnum[pB->SSE[i].Rsta], pB->Rnum[pB->SSE[i].Rend]);

 fprintf(fp,"#Ncomp_sse %d\n",sn->Nali);
 fprintf(fp,"#ScSSE %f\n",ScoreSSE(pA,pB,sn));
 fprintf(fp,"SSE-ALIGNMENT\n");
 for (k=0;k<sn->Nali;++k)
  if ((sn->nalA[k]>=0)&&(sn->nalB[k]>=0))
    fprintf(fp,"%c%d %c%d\n",
     pA->SSE[sn->nalA[k]].type, pA->SSE[sn->nalA[k]].num,
     pB->SSE[sn->nalB[k]].type, pB->SSE[sn->nalB[k]].num );

 fprintf(fp,"END\n");
 fclose(fp);

} /* end of Write_SSE_Alignment() */



void Write_SSE_Alignment_Rasmol(pA,pB,sn,fname)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn; 
 char *fname;
{
 FILE *fp;
 int k;
 char first;

 printf("#Write_SSE_Alignment_Rasmol() --> \"%s\"\n",fname);
 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#ERROR:Can't write to sserasfile \"%s\"\n",fname); exit(1);}
 /** for protein A **/
 fprintf(fp,"select (*:A)\ncolor white\n");
 fprintf(fp,"select (*:A)&&(");
 first = 1;
 for (k=0;k<sn->Nali;++k)
  if ((sn->nalA[k]>=0)&&(sn->nalB[k]>=0))
  { if (first==1) first = 0;  else fprintf(fp,",");
    fprintf(fp,"%s-%s",
    pA->Rnum[pA->SSE[sn->nalA[k]].Rsta], pA->Rnum[pA->SSE[sn->nalA[k]].Rend]); } 
 fprintf(fp,")\n");
 fprintf(fp,"color gray\n");
 
 /** for protein B **/
 first = 1;
 fprintf(fp,"select (*:B)\ncolor white\n");
 fprintf(fp,"select (*:B)&&(");
 for (k=0;k<sn->Nali;++k)
  if ((sn->nalA[k]>=0)&&(sn->nalB[k]>=0))
  {  if (first==1) first = 0; else fprintf(fp,",");
     fprintf(fp,"%s-%s",
     pB->Rnum[pB->SSE[sn->nalB[k]].Rsta], pB->Rnum[pB->SSE[sn->nalB[k]].Rend]);  }
 fprintf(fp,")\n");
 fprintf(fp,"color gray\n");
 fprintf(fp,"select\n");

 fclose(fp);

} /* end of Write_SSE_Alignment_Rasmol() */



void Write_SSE_Alignment_Rasmol_Color(pA,pB,sn,fname)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn; 
 char *fname;
{
 FILE *fp;
 int k;
 char first;
 static char *color[] = {"blue","cyan","green","yellow","orange","redorange","red"};

 printf("#Write_SSE_Alignment_Rasmol() --> \"%s\"\n",fname);
 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#ERROR:Can't write to sserasfile \"%s\"\n",fname); exit(1);}
 /** for protein A **/
 fprintf(fp,"select (*:A)\ncolor white\n");
 first = 1;
 for (k=0;k<sn->Nali;++k)
  if ((sn->nalA[k]>=0)&&(sn->nalB[k]>=0))
  {  fprintf(fp,"select (*:A) && %s-%s\n",
     pA->Rnum[pA->SSE[sn->nalA[k]].Rsta], pA->Rnum[pA->SSE[sn->nalA[k]].Rend]);  
     fprintf(fp,"color %s\n",color[(int)floor(7.0*k/sn->Nali)]); }
 
 /** for protein B **/
 first = 1;
 fprintf(fp,"select (*:B)\ncolor white\n");
 for (k=0;k<sn->Nali;++k)
  if ((sn->nalA[k]>=0)&&(sn->nalB[k]>=0))
   {  fprintf(fp,"select (*:B)&&%s-%s\n",
      pB->Rnum[pB->SSE[sn->nalB[k]].Rsta], pB->Rnum[pB->SSE[sn->nalB[k]].Rend]); 
      fprintf(fp,"color %s\n",color[(int)floor(7.0*k/sn->Nali)]); }
 fprintf(fp,"select\n");

 fclose(fp);

} /* end of Write_SSE_Alignment_Rasmol_Color() */








void Read_SSE_Alignment(pA,pB,sn,fname)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn; 
 char *fname;
{
 FILE *fp;
 int i,L;
 char start,line[256],word1[16],word2[16],num1[16],num2[16]; 
 /*
 [FILE FORMAT]
 #COMMENT
 SSE-ALIGNMENT
 E2 E2
 E3 E5
 E4 E6
 E5 E7
 E6 E8
 E7 E9
 E8 E10
 E9 E11
 E10 E12
 H3 H2
 E11 E13
 END
 */

 fp = fopen(fname,"r");
 if (fp==NULL)
 {
  printf("#ERROR:Can't read ssealifile \"%s\"\n",fname);
  exit(1);
  }
 start = 0; sn->Nali = 0;
 while (feof(fp)==0)
 {
  line[0] = '\0';
  fgets(line,255,fp);
  L = strlen(line);
  if (strncmp(line,"END",3)==0) start = 0;
  if ((start==1)&&(L>5))
  {
   if (line[L-1] == '\n') line[L-1] = '\0';
   sscanf(line,"%s %s",word1,word2);
   Get_Part_Of_Line(num1,word1,1,strlen(word1)); 
   sn->nalA[sn->Nali] = SSE_Number(pA,word1[0],atoi(num1));
   Get_Part_Of_Line(num2,word2,1,strlen(word1)); 
   sn->nalB[sn->Nali] = SSE_Number(pB,word2[0],atoi(num2));
  

 
   ++(sn->Nali);
   }
  if (strncmp(line,"SSE-ALIGNMENT",13)==0) start = 1;
 } 
 fclose(fp);

} /* end of Read_SSE_Alignment() */


int SSE_Number(pr,type,num)
 struct PROTEIN *pr;
 char type,num;
{
 int i;
 i = 0; 
 while (i<pr->Nsse)
 {
  
  if ((pr->SSE[i].type == type) &&
      (pr->SSE[i].num==num) ) return(i);
  ++i;
  }

 printf("#ERROR:Can't find SSE(type %c num %d)\n",type,num);
 exit(1); 
 return(-1);
  
} /* end of SSE_Number() */

