/*

<CalZsc.c>

 for calculating Z-score 
 using the pre-calculated default distribution

 coded by Takeshi Kawabata.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "main.h"
#include "protein.h"
#include "Amino.h"
#include "ScoreDIS.h"


/*** LOCAL VARIABLES ***/
static double DisFRQ[21][50];  /* Pre-calculated default distribution */


/*** GLOBAL FUNCTIONS ***/
void Read_DisFRQ_Each_Separation();
void Read_All_Dis_Frequency();
void Cal_Mean_SD_Dis_Score();





void Read_All_Dis_Frequency(indir)
 char *indir;
{
 char fname[256];
 int k,s;

 /** Initialize **/
 for (k=0;k<=20;++k)
  for (s=0;s<50;++s) DisFRQ[k][s] = 0.0;

 /** Read Files  **/
 for (k=0;k<=20;++k)
 { sprintf(fname,"%s/dis%d.frq",indir,k);
   /* printf("Try to read %s\n",fname); */
   Read_DisFRQ_Each_Separation(fname); }

} /* end of Read_All_Dis_Score() */




void Read_DisFRQ_Each_Separation(fname)
 char *fname;
{
 FILE *fp;
 int s,L,separation;
 char buffer[128],buff1[64],buff2[64];
 char read_on;
 double val;

 /* <FORMAT>
 @CLASS DISTANCE
 @TYPE B
 @SEPARATION 5
 @NSTATE 50
 @FREQUENCY
 0 0.0000000000
 1 0.0000088034
 2 0.0002959763
 :
 #MakeMat ../OCC-00Oct6/dis5.occ 1 -ff ../FRQ-00Oct6/dis5.frq -sg 1000
 
 */

 fp = fopen(fname,"r");
 if (fp==NULL) { printf("#ERROR:Can't find disscorefile \"%s\"\n",fname); exit(1);}
 read_on = 0;
 while (feof(fp)==0)
 {
  buffer[0] = '\0';
  fgets(buffer,128,fp);
  L = strlen(buffer);
  if (buffer[L-1]=='\n') buffer[L-1]= '\0';
  if ((buffer[0]!='#')&&(L>1))
  {
   if (buffer[0]=='@') read_on = 0;

   if (read_on==1)
   { sscanf(buffer,"%d %lf",&s,&val);
     DisFRQ[separation][s] =  val;
     /* 
     printf("sep %d %d %lf\n",separation,s,DisFRQ[separation][s]);  
     */
    }

    if (strncmp(buffer,"@SEPARATION",11)==0) sscanf(buffer,"%s %d",buff1,&separation);
    if (strncmp(buffer,"@FREQUENCY",10)==0) read_on = 1;
   }

  } /* while */
 /* printf("separation %d\n",separation); */
 fclose(fp);

} /* end of Read_DisFRQ_Each_Sepatation() */





void Cal_Mean_SD_Dis_Score(pA,pB,Nali,nalA,nalB,DmapA,DmapB)
 struct PROTEIN *pA,*pB;
 int Nali,*nalA,*nalB;
 struct MAP2D *DmapA,*DmapB;
{
 int i,j,k,kA,kB,dA,dB,dI,dJ;
 double Sc,Mean,Var,s,ss,var,Zsc,SD,len,I;
 double posfrq[50];

 Sc = Mean = Var = 0.0;
 for (i=1;i<=Nali;++i)
   if ((nalA[i] > 0)&&(nalB[i] >0))
    for (j=i+1;j<=Nali;++j)
     if ((nalA[j] > 0)&&(nalB[j] >0))
     {
          kA = nalA[j] - nalA[i];
          kB = nalB[j] - nalB[i];
          if (kA>kB) k = kA; else k = kB;
          if (k>=1)
          {
          if (21<=k) k = 0;
          dA = (int)DmapA->map[nalA[i]][nalA[j]];
          dB = (int)DmapB->map[nalB[i]][nalB[j]];
          if ((dA<50)&&(dB<50)) 
          { 
            Sc += DisROM[k][dA][dB];
            s = ss = 0.0; len = 0.0;
         
            for (dI=0;dI<50;++dI)
             { s  += DisFRQ[k][dI] * DisROM[k][dA][dI];  
               /* 
               printf("%d %d %d %lf %lf %lf\n",k,dI,dA,DisFRQ[k][dI],DisROM[k][dA][dI],s);
               */
               ss += DisFRQ[k][dI] * DisROM[k][dA][dI] * DisROM[k][dA][dI]; } 
          
            /* 
            I = 0.0;
            for (dI=0;dI<50;++dI)
             for (dJ=0;dJ<50;++dJ)
              I  += DisFRQ[k][dI] * DisFRQ[k][dJ] * DisROM[k][dI][dJ];  
             */

            var = ss - s*s;
            /* printf("k %d mean %lf var %lf I %lf\n",k,s,var,I);  */
            Mean += s;
            Var  += var;
           }
          
         } /* if k>= 1 */ 

     } /* if nalA >0 and nalB > 0 */

  SD = sqrt(Var);
  Zsc = (Sc - Mean)/SD;
  printf("%s %s Nc %d Sc %.1lf Mean %.1lf SD %.1lf Zsc %.1lf\n",
   pA->pname,pB->pname,Nali,Sc,Mean,SD,Zsc);


} /* end of Cal_Mean_SD_Dis_Score() */


