/**

<DiffStr.c>

coded by Takeshi Kawabata.

**/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h" 
#include "protein.h" 
#include "Amino.h" 
#include "Vector.h" 
#include "ScoreENV.h" 
#include "IO.h" 
#include "LlistAlign.h" 

/** FUNCTIONS (GLOBAL) **/
float SeqID();
int Ncompare();
float SecID();
float ExpID();
float Sec3ID();
float RexpD();
float DRMS();
float InDel();
void   Cal_DmapCA();
void   Cal_DmapCB();
void DRMS_Filter();

/** FUNCTIONS (LOCAL) **/
static int Number_Sec();

float SeqID(pA,pB,a)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{ int i,Ncomp,Nsame;
  float SeqID; 
 
  Ncomp = Nsame =  0;
  for (i=1;i<= a->Nali ; ++i)
   if ((a->nalA[i]!= -1)&&(a->nalB[i] != -1))
    { ++Ncomp;
      if (Number_Amino(pA->Seq[a->nalA[i]])==Number_Amino(pB->Seq[a->nalB[i]])) ++Nsame; }
  
  if (Ncomp>0) SeqID = (float)(100.0*Nsame/(float)Ncomp); else SeqID = 0.0;
  return(SeqID); 

} /* end of SeqID() */


int Ncompare(a)
 struct AlignNode *a;
{ int i,Ncomp;
  
  Ncomp  =  0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i]!= -1)&&(a->nalB[i] != -1)) ++Ncomp;

  return(Ncomp);

} /* end of Ncompare() */



float InDel(pA,pB,a)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{ int i,end,Nindel,Eindel;
  
  Nindel =  0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i]== -1)||(a->nalB[i] == -1)) ++Nindel;

  Eindel = 0; 

  end = 0; i = 1;
  while ((i<=a->Nali)&&(end==0)) 
   if (a->nalA[i]== -1 ) { Eindel += 1; ++i;} else {end = 1;} 

  end = 0; i = 1;
  while ((i<=a->Nali)&&(end==0)) 
   if (a->nalB[i]== -1 ) { Eindel += 1; ++i;} else {end = 1;} 

  end = 0; i = a->Nali;
  while ((i>=1)&&(end==0)) 
   if (a->nalA[i]== -1 ) { Eindel += 1; --i;} else {end = 1;} 
  
  end = 0; i = a->Nali;
  while ((i>=1)&&(end==0)) 
   if (a->nalB[i]== -1 ) { Eindel += 1; --i;} else {end = 1;} 

  return(100.0*(Nindel-Eindel)/(a->Nali-Eindel));

} /* end of InDel() */




float SecID(pA,pB,a)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{ int i,Ncomp,Nsame;
  
  Ncomp = Nsame =  0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i]!= -1)&&(a->nalB[i] != -1))
    { ++Ncomp;
      if (pA->Sec[a->nalA[i]]==pB->Sec[a->nalB[i]]) ++Nsame; }
  if (Ncomp>0) return(100.0*Nsame/Ncomp);
   else        return(0.0);

} /* end of SecID() */


float Sec3ID(pA,pB,a)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{ int i,Ncomp,Nsame;
  
  Ncomp = Nsame =  0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i]!= -1)&&(a->nalB[i] != -1))
    { ++Ncomp;
      if (Number_Sec(pA->Sec[a->nalA[i]],3)==Number_Sec(pB->Sec[a->nalB[i]],3)) 
          ++Nsame; 
     }
  if (Ncomp>0) return(100.0*Nsame/Ncomp);
   else        return(0.0);

} /* end of Sec3ID() */



float ExpID(pA,pB,a)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{ int i,Ncomp,Nsame;

  Ncomp = Nsame =  0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i]!= -1)&&(a->nalB[i] != -1))
    { ++Ncomp;
      if (pA->Exp[a->nalA[i]]==pB->Exp[a->nalB[i]]) ++Nsame; }
  if (Ncomp>0) return(100.0*Nsame/Ncomp);
   else        return(0.0);

} /* end of ExpID() */



float RexpD(pA,pB,a)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{ int i,Ncomp;
  float RexpA,RexpB,Diff;
  
  Ncomp =  0; Diff = 0.0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i]!= -1)&&(a->nalB[i] != -1))
    { ++Ncomp;
      RexpA = 100.0 * pA->ACC[a->nalA[i]]/(float)Standard_ACC(pA->Seq[a->nalA[i]]); 
      RexpB = 100.0 * pB->ACC[a->nalB[i]]/(float)Standard_ACC(pB->Seq[a->nalB[i]]); 
      Diff += fabs(RexpA-RexpB); 
     }
   
  return(Diff/Ncomp);

} /* end of RexpD() */




float DRMS(pA,pB,a,dA,dB)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a; 
 struct MAP2D *dA,*dB; 
{ int i,j,Ncomp;
  float diff,Diff;
 
  Ncomp =  0; Diff = 0.0;
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i] > 0)&&(a->nalB[i] >0))
   for (j=1;j<=a->Nali;++j)
    if ((a->nalA[j] > 0)&&(a->nalB[j] >0))
     if (i!=j)
     { ++Ncomp;
       diff = dA->map[a->nalA[i]][a->nalA[j]] -  dB->map[a->nalB[i]][a->nalB[j]]; 
       Diff += diff*diff; 
     }
  
   if (Diff>0.0) Diff = sqrt(Diff/Ncomp);
            else Diff = 0.0;
   return(Diff); 

} /* end of  DRMS() */


int Number_Sec(sym,Nstr)
 int Nstr;
{
 if (Nstr == 3)
 switch(sym)
  {
    case 'H': return(0); break;
    case 'E': return(1); break;
    default : return(2); break;
   }

} /* end of Number_Sec() */


void Cal_DmapCA(pr,D)
 struct PROTEIN *pr;
 struct MAP2D *D;
{ int i,j;
  float X[3],dd;
 
 if (D->N < pr->Naa) 
 { printf("#ERROR(Cal_DmapCA):Matrix %d is less than Naa %d (%s)\n",
          D->N,pr->Naa,pr->pname);
   exit(1); }

  for (i=1;i<= pr->Naa ;++i)
   for (j=i+1;j<= pr->Naa ;++j)
     { X[0] = pr->PosCA[i][0] - pr->PosCA[j][0];      
       X[1] = pr->PosCA[i][1] - pr->PosCA[j][1];      
       X[2] = pr->PosCA[i][2] - pr->PosCA[j][2];      
       dd = X[0]*X[0] + X[1]*X[1] + X[2]*X[2];
       D->map[i][j] = D->map[j][i] = (float)sqrt(dd); 
     }
} /* end of Cal_DmapCA() */


void Cal_DmapCB(pr,D)
 struct PROTEIN *pr;
 struct MAP2D   *D;
{ int i,j;
  float X[3],dd;

 if (D->N < pr->Naa) 
 { printf("#ERROR(Cal_DmapCB):Matrix %d is less than Naa %d (%s)\n",
          D->N,pr->Naa,pr->pname);
   exit(1); }

  for (i=1;i<= pr->Naa ;++i)
   for (j=i+1;j<= pr->Naa ;++j)
     { X[0] = pr->PosCB[i][0] - pr->PosCB[j][0];      
       X[1] = pr->PosCB[i][1] - pr->PosCB[j][1];      
       X[2] = pr->PosCB[i][2] - pr->PosCB[j][2];      
       dd = X[0]*X[0] + X[1]*X[1] + X[2]*X[2];
       D->map[i][j] = D->map[j][i] = (float)sqrt(dd); 
     }

} /* end of Cal_DmapCB() */


void DRMS_Filter(pA,pB,a,dA,dB,Dthre)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a; 
 struct MAP2D *dA,*dB; 
 float Dthre;
{ int i,j,Ncomp,maxi;
  float dd,diff,DD,DDthre,maxE;
  float *E;      /* Error  (for alignment number) */
  char   *mark;  /* compared residue (for alignement number )*/

  E    = (float *)malloc(sizeof(float)*(a->Nali+1));
  mark = (char   *)malloc(sizeof(char  )*(a->Nali+1));

  DDthre = Dthre*Dthre; 
  Ncomp =  0; DD = 0.0;
  for (i=1;i<=a->Nali;++i) E[i] = 0.0;
  for (i=1;i<=a->Nali;++i) 
   if ((a->nalA[i] > 0)&&(a->nalB[i] >0)) { ++Ncomp; mark[i] = 1;} else mark[i] = 0;
  
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i] > 0)&&(a->nalB[i] >0))
   for (j=i+1;j<=a->Nali;++j)
    if ((a->nalA[j] > 0)&&(a->nalB[j] >0))
     { diff = dA->map[a->nalA[i]][a->nalA[j]] -  dB->map[a->nalB[i]][a->nalB[j]]; 
       dd   = diff*diff; 
       DD += dd; E[i] += dd; E[j] += dd; }
  
  DD = DD*2.0/Ncomp/(Ncomp-1);
  printf("Ncomp %d Diff %lf DRMS %lf\n",Ncomp,DD,sqrt(DD)); 
  
  /*** Repeating Procedure */
  while (DD > DDthre)
  {
   printf("Ncomp %d DD %lf DRMS %lf\n",Ncomp,DD,sqrt(DD)); 
   maxE = 0.0;
   for (i=1;i<=a->Nali;++i) 
   if (mark[i]==1) 
    { if (E[i]>maxE) { maxE = E[i]; maxi = i; } } 

   DD = (DD*Ncomp*(Ncomp-1)/2.0 - E[maxi]) * 2.0/(Ncomp-1.0)/(Ncomp-2.0);
   Ncomp = Ncomp - 1; 
   for (i=1;i<=a->Nali;++i)  
    if (mark[i]==1) 
     {
       diff = dA->map[a->nalA[i]][a->nalA[maxi]] - dB->map[a->nalB[i]][a->nalB[maxi]]; 
       E[i] -= diff * diff; 
      } 
   mark[maxi] = 0; a->nalA[maxi] = a->nalB[maxi] = -1;
   } 
 
  free(E);  free(mark);  

} /* end of  DRMS_Filter() */


