#include <stdio.h> 	
#include <math.h> 
#include <list>
#include "matrix_clapack.h" 
#include "protein.h"
#include "LlistAlign.h"
#include "vector.h"	

#include "WriteDeformPDB.h"
#include "Evaluation.h"


extern struct AlignNode *AlignCurrent;

extern "C" {
 double CRMS(struct PROTEIN *pA,struct PROTEIN *pB,struct AlignNode *an,double *g1,double *g2, double Rmat[3][3], char AtomType, char InvType);


}

//////////////////////////////////
//								//
// 	    Global variables		//
//	   ( in TransSeq.cpp )		//
//								//
//////////////////////////////////

extern int minNEIGHBOURS;

extern enum { RIGID, AFFINE } GEOMETRY;

extern Matrix<double> probenv;

void printmat(Matrix<double>& A) //,char* fname)
{
//	FILE* fp=fopen(fname,"r");
//	if( !fp ) {
//             printf("Error loading %s.\n", fname);
//             return;
//        }

	int i,j;
	for(i=0;i<A.r;i++){
		for(j=0;j<A.c;j++){
                      printf("%lf ",A(i,j));
                }
        printf("\n");
        }
//	fclose(fp);
}




Matrix<double> LeadingEigenvector(Matrix<double> &A) {
	if (A.c != A.r)
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<double> C = A;
	Vector<double> v(A.r);

	integer lwork, info = 0, cols = A.c, rows = A.r;
	integer m;
	double ev, tol = 0.0, vi = 1.0;
	
	lwork = 32;
	lwork = (lwork + 2) * A.c;
	
	double* workspace = new double[lwork];
	integer *iwork = new integer[A.r * 5 + 1];
	integer *ifail = new integer[A.r + 1];
       
	//cout<<"starting"<<endl;
	dsyevx_("V", "I", "U", &cols, C.A[0], &rows, &vi, &vi, &rows, &rows, &tol, &m, &ev, v.A[0], &cols, workspace, &lwork, iwork, ifail, &info);
        //printf("%s %3.3lf %3.3lf %3.3lf %3.3lf \n","000", v(0),v(1),v(2),v(3));
	//cout<<"working"<<endl;	
	delete[] workspace;
	delete[] iwork;
	delete[] ifail;

	return v;
}


// Compute the Rigid transform from the covariance matrix and centres
// Uses the Horn method
// J. Opt Soc. Amer. A v4 n4 p629-542 1987
// Finds the best rotation centered at dmean and mmean
Matrix<double> SigmaToRigidTransform(Matrix<double> &Sigma, double dmean[3], double mmean[3]) {
	double trace;
	Matrix<double> Q(4,4),Q1(4,4);
	Matrix<double> T(4,4);
	Matrix<double> q;

	trace = Sigma(0,0) + Sigma(1,1) + Sigma(2,2);
	Q(0,0) = trace;
	Q(0,1) = Q(1,0) = Sigma(1,2) - Sigma(2,1);
	Q(0,2) = Q(2,0) = Sigma(2,0) - Sigma(0,2);
	Q(0,3) = Q(3,0) = Sigma(0,1) - Sigma(1,0);
	Q(1,1) = 2 * Sigma(0,0) - trace;
	Q(2,2) = 2 * Sigma(1,1) - trace;
	Q(3,3) = 2 * Sigma(2,2) - trace;
	Q(1,2) = Q(2,1) = Sigma(0,1) + Sigma(1,0);
	Q(1,3) = Q(3,1) = Sigma(0,2) + Sigma(2,0);
	Q(2,3) = Q(3,2) = Sigma(1,2) + Sigma(2,1);
        Q1=Q;
	q = LeadingEigenvector(Q);
	//q = LeadingEigenvector_old(Q1);
	T = 0.0;
	T(0,0) = q(0) * q(0) + q(1) * q(1) - q(2) * q(2) - q(3) * q(3);
	T(0,1) = 2 * (q(1) * q(2) - q(0) * q(3));
	T(0,2) = 2 * (q(1) * q(3) + q(0) * q(2));
	T(1,0) = 2 * (q(1) * q(2) + q(0) * q(3));
	T(1,1) = q(0) * q(0) + q(2) * q(2) - q(1) * q(1) - q(3) * q(3);
	T(1,2) = 2 * (q(2) * q(3) - q(0) * q(1));
	T(2,0) = 2 * (q(1) * q(3) - q(0) * q(2));
	T(2,1) = 2 * (q(2) * q(3) + q(0) * q(1));
	T(2,2) = q(0) * q(0) + q(3) * q(3) - q(1) * q(1) - q(2) * q(2);
	T(3,3) = 1.0;
	T(0,3) = dmean[0] - T(0,0) * mmean[0] - T(0,1) * mmean[1] - T(0,2) * mmean[2];
	T(1,3) = dmean[1] - T(1,0) * mmean[0] - T(1,1) * mmean[1] - T(1,2) * mmean[2];
	T(2,3) = dmean[2] - T(2,0) * mmean[0] - T(2,1) * mmean[1] - T(2,2) * mmean[2];

	return T;
}


// Compute the Rigid transform using matches
void RigidMean(double m[3], float **points, int fixedmatch[1006][2], int idx) {
	int k, l;
	m[0] = m[1] = m[2] = 0.0;

	// This computes the real centre 
	l=0;
	for (k=0; fixedmatch[k][0] != -1; k++) {
		m[0] += points[fixedmatch[k][idx]][0];
		m[1] += points[fixedmatch[k][idx]][1];
		m[2] += points[fixedmatch[k][idx]][2];
		l++;
	}
	
	m[0] /= l; 
	m[1] /= l;
	m[2] /= l;
}


double Consistency(struct PROTEIN *pA, struct PROTEIN *pB, int i, int j, Matrix<double> &T) {
	Vector<double> m(4), r(3);
	m(0) = pB->PosCA[j][0];
	m(1) = pB->PosCA[j][1];
	m(2) = pB->PosCA[j][2];
	m(3) = 1.0;
	m = T * m;
	r(0) = m(0) - pA->PosCA[i][0];
	r(1) = m(1) - pA->PosCA[i][1];
	r(2) = m(2) - pA->PosCA[i][2];
	return (exp(-(r(0) * r(0) + r(1) * r(1) + r(2) * r(2)) / (PAR.squaredratio)));
}


double Consistency_l(struct PROTEIN *pA, struct PROTEIN *pB, int i, int j, Matrix<double> &T) {
	Vector<double> m(4), r(3);
	m(0) = pB->PosCA[j][0];
	m(1) = pB->PosCA[j][1];
	m(2) = pB->PosCA[j][2];
	m(3) = 1.0;
	m = T * m;
	r(0) = m(0) - pA->PosCA[i][0];
	r(1) = m(1) - pA->PosCA[i][1];
	r(2) = m(2) - pA->PosCA[i][2];
	return (r(0) * r(0) + r(1) * r(1) + r(2) * r(2));
}


Matrix<double> RigidTransformBasic(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> Alignment, int *bestpos) {
        Matrix<double> T(4,4);
        double RMS;
        
        RMS = cRMS(pA, pB, Alignment, T);

        int it = 0;
        while ((it < 5) && (RMS > 2)) {
          double tRMS = RMS * 1.2 * RMS * 1.2;
          for (int i=1; i <= pA->Naa; i++)
             if (Alignment(i) > -1) 
               if (Consistency_l(pA, pB, i, Alignment(i), T) > tRMS)
                 Alignment(i) = -1;
               
          RMS = cRMS(pA, pB, Alignment, T);
          //printf("ItPrint=%d RMS=%lf\n",it,RMS);
          it++;
		} 
        
        if (RMS > 1000) *bestpos = -1;
        return (T);
}
    

Matrix<double> RigidTransform(struct PROTEIN *pA, struct PROTEIN *pB, int fixedmatch[1006][2]) {
	Matrix<double> Sigma(3,3);
	Matrix<double> T(4,4);
	double dmean[3], mmean[3], w, wght;
	int k, l, m, n;
	
	Sigma = 0;
		
	// compute the means
	RigidMean(dmean, pA->PosCA, fixedmatch, 0);
	RigidMean(mmean, pB->PosCA, fixedmatch, 1);	
	
	// covariance from fixed matches
	wght = 0;
	for (k=0; fixedmatch[k][0] != -1; k++) {
		wght++;
		for (m=0; m < 3; m++)
			for (n=0; n < 3; n++)
				Sigma(n,m) = Sigma(n,m) + (pA->PosCA[fixedmatch[k][0]][m] - dmean[m]) * (pB->PosCA[fixedmatch[k][1]][n] - mmean[n]);
	}

	Sigma = (1.0 / wght) * Sigma;
	T = SigmaToRigidTransform(Sigma, dmean, mmean);
	
	return T;
}

Matrix<double> RigidTransformMatras(struct PROTEIN *pA, struct PROTEIN *pB, int fixedmatch[1006][2]) {

	Matrix<double> T(4,4);
	double meanA[3], meanB[3], w, wght;
	int k;
	
        double Rmat[3][3];


		
	for (k=0; fixedmatch[k][0] != -1; k++) {
		AlignCurrent->nalA[k+1]=fixedmatch[k][1];
		AlignCurrent->nalB[k+1]=fixedmatch[k][0];
        }
	AlignCurrent->Nali=k;
 	w = CRMS(pB,pA,AlignCurrent,meanB,meanA, Rmat, 'A','-'); 
	int i,j;
	for (i=0; i < 3; i++)
		for (j=0; j < 3; j++)
			T(i,j) = Rmat[i][j];
        
	T(3,3) = 1.0;
	T(0,3) = meanA[0] - T(0,0) * meanB[0] - T(0,1) * meanB[1] - T(0,2) * meanB[2];
	T(1,3) = meanA[1] - T(1,0) * meanB[0] - T(1,1) * meanB[1] - T(1,2) * meanB[2];
	T(2,3) = meanA[2] - T(2,0) * meanB[0] - T(2,1) * meanB[1] - T(2,2) * meanB[2];	
	return T;
}


//Finds the best rotation of the M points centered at mcenter that fits the points
//in D after translation from mcenter to dcenter.
Matrix<double> Rotation(struct PROTEIN *pA, struct PROTEIN *pB, int fixedmatch[1006][2], int dcenter, int mcenter, Matrix<double> Rpoint) {
	Matrix<double> Sigma(3,3);
	Matrix<double> T(4,4);
	int k, l, m, n;
	double w, wght, c1[3], c2[3];
	
	Sigma = 0;
	c2[0] = pB->PosCA[mcenter][0];
	c2[1] = pB->PosCA[mcenter][1];
	c2[2] = pB->PosCA[mcenter][2];
	c1[0] = Rpoint(0, mcenter);  // The point mcenter already has the coordinates fixed
	c1[1] = Rpoint(1, mcenter);
	c1[2] = Rpoint(2, mcenter);
//	c1[0]=Dpoint(0,dcenter);
//	c1[1]=Dpoint(1,dcenter);
//	c1[2]=Dpoint(2,dcenter);

	// covariance from fixed matches
	wght = 0;
	for (k=0; fixedmatch[k][0] != -1; k++) {
		wght++;
		for (m=0; m < 3; m++)
			for (n=0; n < 3; n++)
				Sigma(n,m) = Sigma(n,m) + (pA->PosCA[fixedmatch[k][0]][m] - c1[m]) * (pB->PosCA[fixedmatch[k][1]][n] - c2[n]);
	}

	Sigma = (1.0 / wght) * Sigma;
	T = SigmaToRigidTransform(Sigma, c1, c2);
	return T;
}


// Finds a rigid transformation from M to R (only over a reduced matching)
// Compute the affine transform using matches
Matrix<double> AffineTransform(struct PROTEIN *pA, struct PROTEIN *pB, int fixedmatch[1006][2]) {
	Matrix<double> X(4,4), Y(4,4), Inv, Phi;
	int k, l, m, n;
	double w, mp, dp;
	
	X = 0;
	Y = 0;
	for (k=0; fixedmatch[k][0] != -1; k++)
		for (m=0; m < 4; m++)
			for (n=0; n < 4; n++) {
				if (m == 3) dp = 1.0; else dp = pA->PosCA[fixedmatch[k][0]][m];
				if (n == 3) mp = 1.0; else mp = pB->PosCA[fixedmatch[k][1]][n];
				X(m,n) = X(m, n) + dp * mp;
				if (m == 3) dp = 1.0; else dp = pB->PosCA[fixedmatch[k][1]][m];
				Y(m,n) = Y(m,n) + dp * mp;
			}

	Inv = InverseSymmetricPositiveDefinite(Y);
	Phi = X * Inv;
	return Phi;
}

void PDscore(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> &Alignment, double bestscore ) {
        int ALsize = 0;
        for (int i=1; i <= pA->Naa; i++) {
                if ( Alignment(i) > -1) {
                        ALsize++;
                }
        }
        double scorefinal = bestscore/(pA->Naa+pB->Naa-ALsize) ;  
        printf("%s %s %d %1.4lf %1.3lf\n",pA->cname,pB->cname,ALsize,scorefinal,bestscore);
}
        
          
double lTMscore(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> &Alignment, double bestscore, int Print, Vector<int> &Ref) {
	int fixedmatch[1006][2], fpos=0, i, j, p, k;
	Matrix<double> T(4,4);
	Vector<double> m(4),r(3);
	double pes0, pesref, scorefinal, score0, score = 0;
	int ALsize = 0;
	if (Print == 2) {
		printf("probgeo probenv rsa     polar   Hstr    Bstr    S  numB ");
		printf("numA  S ");
		printf("numB  S probgeo probenv rsa    polar    Hstr   Bstr  \n");
	}
	
        int a,b, a_prev, b_prev;
        a=0; b=0; 
	for (i=1; i <= pA->Naa; i++) {
		pes0 = 0;
                j= Alignment(i);
		if ( j > -1) {
			ALsize++;
                        a_prev=a; b_prev=b;
			while (pA->Partition[a].ending < i) a++;
			while (pB->Partition[b].ending < j) b++;
	 		if (a_prev == a && b_prev==b && (a>0 || b >0) && fpos>0) {
                            pes0=Consistency(pA,pB,i,j,T);
			      // if the pair (i,j) is in the same previuous segments, 
                                       // use the previous T
                        }
                        else {  //Calculation of T

 
		int indexA, image, indexB, k2;
		k2 = 0;
		
		fpos=0;
                // Assume that the neighbours are sorted by the index
  		for (int k=0; (indexA=pA->Partition[a].list[k]) != -1; k++) 
			if ((image=Alignment(indexA)) > -1) {
					while (((indexB=pB->Partition[b].list[k2]) != -1) && image > indexB) k2++;
                                        if (indexB==-1) break;
                                        if (image < indexB) continue; 

 					
					fixedmatch[fpos][0]=indexA;				
					fixedmatch[fpos++][1]=image;
			}
			

		fixedmatch[fpos][0]=-1;
 
		if(fpos>4){
				// if( GEOMETRY==RIGID ) 
				T=RigidTransform(pA,pB,fixedmatch);
				//if( GEOMETRY==AFFINE )T=AffineTransform(pA,pB,fixedmatch);
                                
				pes0=Consistency(pA,pB,i,j,T);				
		}     
                } // Calculation of T                             

	       }  // if Alignment(i) > -1

		score0 = score0 + pes0;	

		// For the reference alignment
		if (Print == 2) {
			pesref = 0;
			if (Ref(i) > -1) {
				fpos = 0;
					for (k=1; pA->Adj[i][k] != -1; k++) {
						if (Ref(pA->Adj[i][k]) > -1) {
							fixedmatch[fpos][0] = pA->Adj[i][k];
							fixedmatch[fpos++][1] = Ref(pA->Adj[i][k]);
						}
					}
				fixedmatch[fpos][0] = -1;
				fixedmatch[minNEIGHBOURS][0] = -1;
				if (fpos > 3) {
					if(GEOMETRY == RIGID)	T = RigidTransform (pA, pB, fixedmatch);
					if(GEOMETRY == AFFINE)	T = AffineTransform(pA, pB, fixedmatch);
					// Transformation from M to D
					pesref = Consistency(pA, pB, i, Ref(i), T);
				}
			}
		}
		
		int image;
		if (Print == 1) {
			if (Alignment(i) > -1) {
				cout<<" "<<typeSSE(pA->Sec[i])<<" " << pA->Rnum[i]<<" "<< pB->Rnum[Alignment(i)]<<" "<< typeSSE(pB->Sec[Alignment(i)])<<" "<<"\t\t"<<pes0<<"\t\t"<<endl;
			}
			else { 
				cout<<" "<<typeSSE(pA->Sec[i])<<" "<<pA->Rnum[i]<<" "<<-100<< " - \t\t - \t\t"<< endl;
			}
		}
		else if (Print == 2) {
				if (Ref(i) > -1) {
					image = Ref(i);
					printf("%1.5lf %1.5lf %c %5s ",pesref, probenv(i,image), typeSSE(pB->Sec[image]), pB->Rnum[image]);
					
				}
				else	printf("-       -       -       -       -       -       -     - ");

				printf("%5s %c ",pA->Rnum[i], typeSSE(pA->Sec[i]));
				if (Alignment(i) > -1) {
					image = Alignment(i);
					printf("%5s %c %1.5lf %1.5lf \n", pB->Rnum[image], typeSSE(pB->Sec[image]), pes0, probenv(i,image));
				}
				else	printf("    - - -       -       -       -       -       -   \n");
		}
	}
        scorefinal = bestscore/(pA->Naa+pB->Naa-ALsize) ; // score0/....; 

	if (Print > 0) {
		cout<<pA->cname<<" "<<pB->cname<<" Nassig="<<ALsize<<" "<<"Asize="<<pA->Naa<<" "<<"Bsize="<<pB->Naa<<
			" PDscore="<<scorefinal<<" rawscore="<<bestscore<<" RMSD="<<cRMS(pA, pB, Alignment, T)<<" ProbSameTopology="<< ProbTop(scorefinal) <<" ProbSameHomology="<< ProbHom(scorefinal) <<" PValueTopology="<< PValueTop(scorefinal) <<" PvalueHomology="<< PValueHom(scorefinal)<<endl ; 
	}
	return(scorefinal);
}



void CalcPDscore(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> &Alignment, double GAP) {
	int fixedmatch[1006][2], fpos=0, i, j, p, k;
	Matrix<double> T(4,4);
	Vector<double> m(4),r(3);
	double pes0, pesref, scorefinal, score0 = 0;
	int ALsize = 0;

        int a,b, a_prev, b_prev;
        a=0; b=0; 
	for (i=1; i <= pA->Naa; i++) {
		pes0 = 0;
                j= Alignment(i);
		if ( j > -1) {
			ALsize++;
                        a_prev=a; b_prev=b;
			while (pA->Partition[a].ending < i) a++;
			while (pB->Partition[b].ending < j) b++;

	 		if (a_prev == a && b_prev==b && (a>0 || b >0) && fpos>4) {
                            pes0=Consistency(pA,pB,i,j,T);
			      // if the pair (i,j) is in the same previuous segments, 
                                       // use the previous T
                        }
                        else {  //Calculation of T

 
		int indexA, image, indexB, k2;
		k2 = 0;
		
		fpos=0;
                // Assume that the neighbours are sorted by the index
  		for (int k=0; (indexA=pA->Partition[a].list[k]) != -1; k++) 

			if ((image=Alignment(indexA)) > -1) {
					while (((indexB=pB->Partition[b].list[k2]) != -1) && image > indexB) k2++;
                                        if (indexB==-1) break;
                                        if (image < indexB) continue; 

 					
					fixedmatch[fpos][0]=indexA;				
					fixedmatch[fpos++][1]=image;
			}

			

		fixedmatch[fpos][0]=-1;
 

		if(fpos>4){

				// if( GEOMETRY==RIGID ) 
				T=RigidTransform(pA,pB,fixedmatch);

				//if( GEOMETRY==AFFINE )T=AffineTransform(pA,pB,fixedmatch);
                                
				pes0=Consistency(pA,pB,i,j,T);				

		}     
                } // Calculation of T                             


	       }  // if Alignment(i) > -1


		score0 = score0 + pes0;	

	}
        scorefinal = (score0 + (pA->Naa+pB->Naa-2*ALsize)*GAP)/(pA->Naa+pB->Naa-ALsize) ; 

     
        printf("%s %s %d %1.4lf %1.3lf\n",pA->cname,pB->cname,ALsize,scorefinal,score0);


}

void printAlignment(struct PROTEIN *pA, struct PROTEIN *pB, Vector<int> Alignment) {

	int j, antj;
        FILE *alignfile;
        char name[100];
	
	int  fpos;
	Matrix<double> T(4,4);

        sprintf(name,"%s.%s.%s","ali",pA->cname,pB->pname);
	alignfile= fopen(name,"w");
        antj=0;
	int a, a_prev, b, b_prev;
        a=0; b=0;
	for (int i=1; i <= pA->Naa; i++) {	
		if (Alignment(i) > -1) {
			j=Alignment(i);
                        if (j>antj+1) 
                               for (++antj;antj<j; antj++)
 				fprintf(alignfile,"-\t-\t%s\t%c\n", pB->Rnum[antj], typeSSE(pB->Sec[antj]) );
// Calculation of quality of alignment
		        float pes0 = 0;
 
                        a_prev=a; b_prev=b;
			while (pA->Partition[a].ending < i) a++;
			while (pB->Partition[b].ending < j) b++;
	 		if (a_prev == a && b_prev==b && (a>0 || b >0) && fpos>0) {
                            pes0=Consistency(pA,pB,i,j,T);
			      // if the pair (i,j) is in the same previuous segments, 
                                       // use the previous T
                        }
                        else { //New T

                /* change from site to segment */
                /*  For each neighbour index A of segment a of pA, we see if its image is a neighbour of
		    segment b of pB. If so, the pair (indexA,image) will be in the set to get T.
                    Then, all the amino acid pairs of the segments get applied the transformation T.
                */
		int indexA, image, indexB, k2;
		k2 = 0;
		int fixedmatch[1006][2],
		fpos=0;
                // Assume that the neighbours are sorted by the index
  		for (int k=0; (indexA=pA->Partition[a].list[k]) != -1; k++) 
			if ((image=Alignment(indexA)) > -1) {
					while (((indexB=pB->Partition[b].list[k2]) != -1) && image > indexB) k2++;
                                        if (indexB==-1) break;
                                        if (image < indexB) continue; 

 					
					fixedmatch[fpos][0]=indexA;				
					fixedmatch[fpos++][1]=image;
			}
			

		fixedmatch[fpos][0]=-1;
 
		if(fpos>4){
				// if( GEOMETRY==RIGID ) 
				T=RigidTransform(pA,pB,fixedmatch);
				//if( GEOMETRY==AFFINE )T=AffineTransform(pA,pB,fixedmatch);
                                
				pes0=Consistency(pA,pB,i,j,T);				
		}                                  
                   } // enf of New T
// End of Calculation
			fprintf(alignfile,"%c\t%s\t%s\t%c\t%1.5f\n",typeSSE(pA->Sec[i]),  pA->Rnum[i], pB->Rnum[Alignment(i)], typeSSE(pB->Sec[Alignment(i)]), pes0);
			antj=j;
		}
		else 
			fprintf(alignfile,"%c\t%s\t-\t-\n",typeSSE(pA->Sec[i]),  pA->Rnum[i]);
		
			
	}
	for (++antj; antj <= pB->Naa; antj++) 
		fprintf(alignfile,"-\t-\t%s\t%c\n", pB->Rnum[antj], typeSSE(pB->Sec[antj]) );

fclose(alignfile);

}
