/*

 <IO.c>

 coded by Takeshi Kawabata.

*/

#include <stdio.h> 
#include <stdlib.h> 
#include <string.h> 
#include <math.h> 
#include <time.h> 
#include "main.h" 
#include "protein.h" 
#include "Amino.h" 
#include "Vector.h" 
#include "BetaTopo.h"
#include "OutAln.h"


/**** BASIC DIRECTORY DEFINED IN ".mastras" FILE ****/
char BASE_DIR[128];
char DSSP_DIR[128];
char BSSP_DIR[128];
char PDB_DIR[128];
char SCORE_DIR[128];


/** FUNCTIONS (GLOBAL) **/
void Get_Part_Of_Line();
void Cut_Filename();
int Standard_ACC();
void Read_Matras_Environment_File();
void Read_PDB_File();
void Write_PDB_File();
void Output_Data_Horizontal();
void Read_DSSP_File();
void Read_Sequence();
void Read_BSSP_File();
void Read_BSSP_File_Range();
void Malloc_Protein();
void Free_Protein();
void Malloc_SSE();
void Free_SSE();
void Read_CASP_style_Alignment();
void Read_CASP_style_Constraint_Alignment();
void Remove_Space();
void Malloc_MAP2D();
void Get_First_Field();
void Find_Split_Point();
void Remove_Head_Tail_Space();
char *Get_Date_String();

/** FUNCTIONS (LOCAL) **/
static char Expose();
static int Find_Rnumber();



void Get_Part_Of_Line(char *part,char *line,int s,int e)
{
 int i,E,L;
 L = strlen(line)-1;
 if (line[L] == '\n') L -= 1;
 if (e>L) E = L; else E = e;
 for (i=s;i<=E;++i) part[i-s] = line[i];
 part[E-s+1] = '\0';

} /* end of Get_Part_of_Line() */




void Read_Matras_Environment_File()
{
 FILE *fp;
 char homedir[128],matfile[64],Matfile[128],ln[200],head[32],data[128];
 int L;

 /** TRY TO FIND ".matras" in [currentdir] or [homedir] **/
 sprintf(matfile,".matras"); 
 fp = fopen(matfile,"r");
 if (fp==NULL)
 {
  sprintf(homedir,"%s",getenv("HOME"));
  sprintf(Matfile,"%s/%s",homedir,matfile);
  fp = fopen(Matfile,"r");
  if (fp==NULL)
  {
   printf("#ERROR:Not found \"%s\", \"%s\".\n",matfile,Matfile);
   exit(1);  
  } 
  
 }

 /**
 [FILE FORMAT]
 BASE_DIR  /export/home1/takawaba/work/Matras10
 SCORE_DIR /export/home1/takawaba/work/Matras10/ROM-00Oct6
 BSSP_DIR  /export/home5/DB/BSSP
 PDB_DIR   /lab/PDB
 **/

 BASE_DIR[0] = SCORE_DIR[0] = BSSP_DIR[0] = PDB_DIR[0] = '\0';
 
 while (feof(fp)==0)
 {
  ln[0] = '\0';
  fgets(ln,190,fp);
  L = strlen(ln);
  if ((ln[0]!='#')&&(L>5))
  {
   if (ln[L-1] == '\n') ln[L-1] = '\0';
   sscanf(ln,"%s %s\n",head,data);
        if (strcmp(head,"BASE_DIR")==0)  sprintf(BASE_DIR,"%s",data);
   else if (strcmp(head,"SCORE_DIR")==0) sprintf(SCORE_DIR,"%s",data);
   else if (strcmp(head,"BSSP_DIR")==0)  sprintf(BSSP_DIR,"%s",data);
   else if (strcmp(head,"PDB_DIR")==0)   sprintf(PDB_DIR,"%s",data);
  }
 } /* while */

 fclose(fp);

} /* end of Read_Matras_Environment_File() */ 




void Malloc_MAP2D(M,nmalloc)
 struct MAP2D *M;
 int nmalloc;
{ int i,Nmalloc;
     
  M->N = nmalloc;
  Nmalloc = nmalloc + 1;
  

  M->map = (float **)malloc(sizeof(float *)*Nmalloc); 
  for (i=0;i<Nmalloc;++i) 
    M->map[i] = (float *)malloc(sizeof(float)*Nmalloc); 
 

} /* end of Malloc_MAP2D() */




void Malloc_Protein(pr,Naa,ANGtype,XYZtype)
 struct PROTEIN *pr;
 int Naa;
 char ANGtype; /* '-':Angle Variable Not Malloc, otherwise malloc */
 char XYZtype; /* XYZ variables malloc. 'A':Calpha, 'B':Cbeta, 'W':both */ 
{ int i;

  pr->MaxNaa  = Naa;
  pr->AngMallocType = ANGtype;
  pr->XyzMallocType = XYZtype;

  pr->header = (char *)malloc(sizeof(char)*150);
  pr->compnd = (char *)malloc(sizeof(char)*150);

  pr->Seq  = (char *) malloc(sizeof(char)*(Naa+1));
  pr->Sec  = (char *) malloc(sizeof(char)*(Naa+1));
  pr->Exp  = (char *) malloc(sizeof(char)*(Naa+1));
  pr->Top  = (char *) malloc(sizeof(char)*(Naa+1));
  pr->Rnum  = (char **) malloc(sizeof(char *)*(Naa+1));
  pr->SseSeq  = (char *) malloc(sizeof(char)*(Naa+1));
  for (i=0;i<=Naa;++i) 
    pr->Rnum[i] = (char *) malloc(sizeof(char)*6);

  /** MALLOC ANGLE DATA **/
 
  if (ANGtype!='-')
  {
   pr->Break  = (char *) malloc(sizeof(char)*(Naa+1));
   pr->ACC    = (int *) malloc(sizeof(int)*(Naa+1));

   pr->BP1  = (int *) malloc(sizeof(int)*(Naa+1));
   pr->BP2  = (int *) malloc(sizeof(int)*(Naa+1));
 
   pr->Phi  = (float *) malloc(sizeof(float)*(Naa+1));
   pr->Psi  = (float *) malloc(sizeof(float)*(Naa+1));
  } 

  /** MALLOC XYZ DATA **/

  if ((XYZtype=='A')||(XYZtype=='W'))
  { 
    pr->PosCA = (float **) malloc(sizeof(float *)*(Naa+1));
    
    for (i=0;i<=Naa;++i)
      pr->PosCA[i] = (float *) malloc(sizeof(float)*3); 
  }


  if ((XYZtype=='B')||(XYZtype=='W'))
   { pr->PosCB = (float **) malloc(sizeof(float *)*(Naa+1));
     for (i=0;i<=Naa;++i)
      pr->PosCB[i] = (float *) malloc(sizeof(float)*3);  }



 pr->SelfSdis = pr->SelfSsse = 0.0;

} /* end of Malloc_Protein() */




void Free_Protein(pr)
 struct PROTEIN *pr;
{
  int i;

  if ((pr->XyzMallocType=='B')||(pr->XyzMallocType=='W'))
  {
    for (i=0;i<=pr->MaxNaa;++i) { free(pr->PosCB[i]); pr->PosCB[i] = NULL;}
    free(pr->PosCB);
    pr->PosCB = NULL;
  }
 
  if ((pr->XyzMallocType=='A')||(pr->XyzMallocType=='W'))
  {
    for (i=0;i<=pr->MaxNaa;++i) { free(pr->PosCA[i]); pr->PosCA[i] = NULL;}
    free(pr->PosCA);
    pr->PosCA = NULL;
   }


  if (pr->AngMallocType!='-')
  {
   free(pr->Psi); free(pr->Phi);
   free(pr->BP2); free(pr->BP1);
   free(pr->ACC); free(pr->Break);
   pr->Psi = NULL; pr->Phi = NULL; 
   pr->BP2 = NULL; pr->BP1 = NULL;
   pr->ACC = NULL; pr->Break = NULL;  
  }
 
  for (i=0;i<=pr->MaxNaa;++i) { free(pr->Rnum[i]); pr->Rnum[i] = NULL;}
  free(pr->Rnum); pr->Rnum = NULL;
 
  free(pr->SseSeq); free(pr->Top);
  free(pr->Exp); free(pr->Sec);
  free(pr->Seq); 
  free(pr->compnd); free(pr->header);
  pr->SseSeq = NULL; pr->Top = NULL;
  pr->Exp = NULL;    pr->Sec = NULL;
  pr->Seq = NULL;    pr->compnd = NULL;
  pr->header = NULL;

} /* end of Free_Protein() */



void Malloc_SSE(pr,Nsse)
 struct PROTEIN *pr;
 int Nsse;
{
  int i;
 
  pr->MaxNsse = Nsse+1;
 
  if (pr->MaxNsse >= 1) 
  { 
   pr->SSE = (struct SecStrEle *) malloc(sizeof(struct SecStrEle) * pr->MaxNsse);
  }
 
  if (pr->MaxNsse >= 2)
  { 
   pr->ssePh = (float **) malloc(sizeof(float *)*pr->MaxNsse);
    for (i=0;i<pr->MaxNsse;++i) pr->ssePh[i] = (float *) malloc(sizeof(float)*pr->MaxNsse);

   pr->sseT1 = (float **) malloc(sizeof(float *)*pr->MaxNsse);
    for (i=0;i<pr->MaxNsse;++i) pr->sseT1[i] = (float *) malloc(sizeof(float)*pr->MaxNsse);

   pr->sseT2 = (float **) malloc(sizeof(float *)*pr->MaxNsse);
    for (i=0;i<pr->MaxNsse;++i) pr->sseT2[i] = (float *) malloc(sizeof(float)*pr->MaxNsse);
 
   pr->sseDi = (float **) malloc(sizeof(float *)*pr->MaxNsse);
    for (i=0;i<pr->MaxNsse;++i) pr->sseDi[i] = (float *) malloc(sizeof(float)*pr->MaxNsse);
   
  }

} /* end of Malloc_SSE() */


void Free_SSE(pr)
 struct PROTEIN *pr;
{
  int i;
  
 if (pr->MaxNsse >= 2) 
  { 
   for (i=0;i<pr->MaxNsse;++i) { free(pr->sseDi[i]); pr->sseDi[i] = NULL; }
   free(pr->sseDi); pr->sseDi = NULL;
   
   for (i=0;i<pr->MaxNsse;++i) { free(pr->sseT2[i]); pr->sseT2[i] = NULL;}
   free(pr->sseT2); pr->sseT2 = NULL;
   
   for (i=0;i<pr->MaxNsse;++i) { free(pr->sseT1[i]); pr->sseT1[i] = NULL;}
   free(pr->sseT1); pr->sseT1 = NULL;
   
   for (i=0;i<pr->MaxNsse;++i) { free(pr->ssePh[i]); pr->ssePh[i] = NULL;}
   free(pr->ssePh); pr->ssePh = NULL;
 }    
  
  free(pr->SSE); pr->SSE = NULL;

} /* end of Free_SSE() */


void Equal_Protein(proA,proB)  /* proA := proB */
 struct PROTEIN *proA,*proB;
{ int i,j;

  proA->Naa  = proB->Naa;
  proA->Nsse = proB->Nsse;
  sprintf(proA->pname,"%s",proB->pname);
  sprintf(proA->cname,"%s",proB->cname);
  sprintf(proA->compnd,"%s",proB->compnd);
  sprintf(proA->header,"%s",proB->header);
  proA->SelfSdis = proB->SelfSdis; 
  proA->SelfSsse = proB->SelfSsse; 

  for (i=1;i<=proB->Naa;++i)
   { proA->Seq[i] = proB->Seq[i];
     proA->Sec[i] = proB->Sec[i];
     proA->Exp[i] = proB->Exp[i];
     proA->Top[i] = proB->Top[i];
    sprintf(proA->Rnum[i],"%s",proB->Rnum[i]);
     proA->SseSeq[i] = proB->SseSeq[i]; 
     if( ((proA->XyzMallocType=='A')||(proA->XyzMallocType=='W')) &&
         ((proB->XyzMallocType=='A')||(proB->XyzMallocType=='W')) )
      { proA->PosCA[i][0] = proB->PosCA[i][0];
        proA->PosCA[i][1] = proB->PosCA[i][1];
        proA->PosCA[i][2] = proB->PosCA[i][2];  }
     if( ((proA->XyzMallocType=='B')||(proA->XyzMallocType=='W')) &&
         ((proB->XyzMallocType=='B')||(proB->XyzMallocType=='W')) )
     { proA->PosCB[i][0] = proB->PosCB[i][0];
       proA->PosCB[i][1] = proB->PosCB[i][1];
       proA->PosCB[i][2] = proB->PosCB[i][2];  }
   }

  for (i=0;i<proB->Nsse;++i)
  {
     proA->SSE[i].L = proB->SSE[i].L;
     proA->SSE[i].Rsta = proB->SSE[i].Rsta;
     proA->SSE[i].Rend = proB->SSE[i].Rend;
     proA->SSE[i].type = proB->SSE[i].type;
  }
  for (i=0;i<proB->Nsse;++i)
   for (j=0;j<proB->Nsse;++j)
   {
     proA->sseDi[i][j] = proB->sseDi[i][j];
     proA->ssePh[i][j] = proB->ssePh[i][j];
     proA->sseT1[i][j] = proB->sseT1[i][j];
     proA->sseT2[i][j] = proB->sseT2[i][j];
   }

} /* end of Equal_Protein() */





void Read_PDB_File(pr,fname)
 struct PROTEIN *pr; 
 char *fname;
{
 FILE *fp;
 char line[100],B[16],rname[4],rnum0[6],rnum1[6],aname[4],dir[4],Fname[128];
 int end,i;
 float X[3];

 Get_Part_Of_Line(dir,fname,4,5);
 sprintf(Fname,"%s/%s/%s",PDB_DIR,dir,fname); 
 fp = fopen(Fname,"r");
 
 sprintf(pr->pname,"%s",fname); 
 Cut_Filename(pr->pname,pr->cname);
 if (fp==NULL) { 
  printf("#Can't find pdbfile \"%s\". Try %s\n",Fname,fname); 
  fp = fopen(fname,"r");
  if (fp ==NULL) {printf("#ERROR:Can't find pdbfile \"%s\"\n",fname); exit(1);} 
  }

 pr->chain = '-'; 

 for (i=0;i<pr->MaxNaa;++i) pr->PosCB[i][0] = pr->PosCB[i][1] = pr->PosCB[i][2] = 0.0; 

 end = 0; i = 0;
 while (end==0)
  {
   line[0] = '\0';
   fgets(line,80,fp);
   if ((feof(fp)!=0)||((line[0]=='E')&&(line[1]=='N'))) end = 1;
   else if ( (line[0]=='A')&&(line[1]=='T')&&(line[2]=='O')
             && ((pr->chain == '-') || (pr->chain==line[21])) ) 
    {
      Get_Part_Of_Line(B,line,30,37); X[0] = atof(B);
      Get_Part_Of_Line(B,line,38,45); X[1] = atof(B);
      Get_Part_Of_Line(B,line,47,53); X[2] = atof(B);
      Get_Part_Of_Line(rnum1,line,22,26);
      Get_Part_Of_Line(rname,line,17,19);
      Get_Part_Of_Line(aname,line,13,15);
      if (strcmp(rnum0,rnum1)!= 0)
        { ++i;  pr->Seq[i] = Amino_Name(rname);  
                strcpy(pr->Rnum[i],rnum1); 
            }

      if (strncmp("CA ",aname,3)==0)  { Equal_Vec(3,pr->PosCA[i],X); }
      if (strncmp("CB ",aname,3)==0)  { Equal_Vec(3,pr->PosCB[i],X); }
      
      /* 
      if (strncmp("N  ",aname,3)==0)  {Equal_Vec(3,pr->PosN[i] ,X); }
      if (strncmp("O  ",aname,3)==0)  {Equal_Vec(3,pr->PosO[i] ,X); }
      if (strncmp("C  ",aname,3)==0)  {Equal_Vec(3,pr->PosC[i] ,X); }
      */
 
      strcpy(rnum0,rnum1);
      }

   } /* while */

  pr->Naa = i;
  fclose(fp);

  /*
  for (i=1;i<=pr->Naa;++i) 
   printf("%3d %c %s %f %f\n",i,pr->Seq[i],pr->Rnum[i],pr->PosCA[i][0],pr->PosCB[i][0]); 
  */

  /* SET CB ATOM */ 
  for (i=1;i<=pr->Naa;++i) 
   if ((pr->PosCB[i][0]==0)&&(pr->PosCB[i][1]==0.0)&&(pr->PosCB[i][2]==0.0)) 
    { Equal_Vec(3,pr->PosCB[i],pr->PosCA[i]);   }
  
 /* Initialize Sec and SseSeq */ 
 for (i=1;i<=pr->Naa;++i)  pr->Sec[i] = pr->SseSeq[i] = ' ';

}/* end of Read_PDB_File()  */




void Write_PDB_File(pro,fname)
 struct PROTEIN *pro;
 char *fname;
{
 FILE *fp;
 int i,a;
 char rname[4],ch;
 double d,RG;

 fp = fopen(fname,"w");
 fprintf(fp,"REMARK   Chain %c Naa %d\n",pro->chain,pro->Naa);
 if (pro->chain=='-') ch = ' '; else ch = pro->chain;  
 a = 1; 
 for (i=1;i<= pro->Naa ;++i)
  {
    Get_TriAmino(pro->Seq[i],rname);
    fprintf(fp,"ATOM   %4d  N   %3s %c%5s   %8.3lf%8.3lf%8.3lf\n",
          a,rname,ch,pro->Rnum[i],pro->PosN[i][0],pro->PosN[i][1],pro->PosN[i][2]);
     ++a;
    fprintf(fp,"ATOM   %4d  CA  %3s %c%5s   %8.3lf%8.3lf%8.3lf",
          a,rname,ch,pro->Rnum[i],pro->PosCA[i][0],pro->PosCA[i][1],pro->PosCA[i][2]);
    fprintf(fp," %.0lf  %.0lf\n",pro->Phi[i],pro->Psi[i]);
 
    ++a;
    fprintf(fp,"ATOM   %4d  C   %3s %c%5s   %8.3lf%8.3lf%8.3lf\n",
          a,rname,ch,pro->Rnum[i],pro->PosC[i][0],pro->PosC[i][1],pro->PosC[i][2]);
    ++a;
    fprintf(fp,"ATOM   %4d  O   %3s %c%5s   %8.3lf%8.3lf%8.3lf\n",
          a,rname,ch,pro->Rnum[i],pro->PosO[i][0],pro->PosO[i][1],pro->PosO[i][2]);
    ++a;
    fprintf(fp,"ATOM   %4d  CB  %3s %c%5s   %8.3lf%8.3lf%8.3lf\n",
          a,rname,ch,pro->Rnum[i],pro->PosCB[i][0],pro->PosCB[i][1],pro->PosCB[i][2]);
    ++a;
  }
 
 fprintf(fp,"END\n");
 fclose(fp);
} /* end of Output_Pos_PDB() */



void Output_Data_Horizontal(pr)
 struct PROTEIN *pr;
{
 int i,j,k;
 int max;

 for (i=1;i<=pr->Naa;++i) 
  if ((pr->Sec[i]!='H')&&(pr->Sec[i]!='E')) pr->Sec[i] = 'C';
 
 max = ( pr->Naa - 1)/60 + 1;
 printf("\nNaa %d\n",pr->Naa);
 for (i=0;i<max;++i)
 {
   k = i%5;

        if (k==0)
     printf("             1         2         3         4         5         6 \n");
   else if (k==1)
     printf("             7         8         9         0         1         2 \n");
   else if (k==2)
     printf("             3         4         5         6         7         8 \n");
   else if (k==3)
     printf("             9         0         1         2         3         4 \n");
   else if (k==4)
     printf("             5         6         7         8         9         0 \n");
 
   printf("Seq:"); 
   for (j=1;j<=60;++j) 
    if ((60*i+j)<=pr->Naa) printf("%c",pr->Seq[60*i+j]);   printf("\n");

   printf("Sec:"); 
   for (j=1;j<=60;++j) 
    if ((60*i+j)<=pr->Naa) printf("%c",pr->Sec[60*i+j]); printf("\n");
   
   printf("Top:"); 
   for (j=1;j<=60;++j) 
    if ((60*i+j)<=pr->Naa) printf("%c",pr->Top[60*i+j]); printf("\n");
   
   printf("Exp:"); 
   for (j=1;j<=60;++j) 
    if ((60*i+j)<=pr->Naa) printf("%c",pr->Exp[60*i+j]); printf("\n");

   printf("\n");
  }

} /* end of Output_Data_Horizontal() */







void Read_DSSP_File(pr,fname)
 struct PROTEIN *pr;
 char *fname;
{
 FILE *fp; 
 char Line[200],buff[10],Fname[256];
 int  i,ok,end,count,acc;
 int READDATA,Rnum;
 
 Cut_Filename(pr->pname,pr->cname);
 sprintf(Fname,"%s/%s",DSSP_DIR,fname); 
 fp = fopen(Fname,"r");
 
 if (fp==NULL) {printf("#ERROR:Can't find dsspfile \"%s\"\n",Fname); exit(1); }
 pr->chain = '-'; 

 READDATA = end = 0;  count = 1;
 for (i=0;i<pr->MaxNaa;++i) pr->Seq[i] = pr->Sec[i] = pr->Exp[i] = ' ';
 Rnum = 0; 
 while (end ==0)
  { 
    Line[0] = '\0';
    if (fgets(Line,199,fp)==NULL) {end = 1; READDATA = 0;} 
 
    if ((READDATA==1)&&((pr->chain=='-')||(pr->chain==Line[11])))    
       { 
        if ((Line[13]!='!')&&(Rnum<(pr->MaxNaa-1))) 
         { 
           ++Rnum; 
           Get_Part_Of_Line(pr->Rnum[Rnum],Line,6,10); pr->Rnum[Rnum][5] = '\0';  
           pr->Seq[Rnum] = Line[13];
           pr->Sec[Rnum] = Line[16];
           Get_Part_Of_Line(buff,Line,34,37);    acc = atoi(buff); 
           if  ( ((double)acc/(double)Standard_ACC(pr->Seq[Rnum]))>0.2)
             pr->Exp[Rnum] = 'e'; else pr->Exp[Rnum] = 'b'; 
           Get_Part_Of_Line(buff,Line,26,28); pr->BP1[Rnum] = atoi(buff); 
           Get_Part_Of_Line(buff,Line,30,32); pr->BP2[Rnum] = atoi(buff);
 
           Get_Part_Of_Line(buff,Line, 95,100);  pr->Phi[Rnum] = atof(buff);
           Get_Part_Of_Line(buff,Line,101,106);  pr->Psi[Rnum] = atof(buff);
           Get_Part_Of_Line(buff,Line,107,113);  pr->PosCA[Rnum][0] = atof(buff);
           Get_Part_Of_Line(buff,Line,114,120);  pr->PosCA[Rnum][1] = atof(buff);
           Get_Part_Of_Line(buff,Line,121,127);  pr->PosCA[Rnum][2] = atof(buff);
           
           pr->Naa = Rnum;   }
       } 

      if ((Line[4]=='R')&&(Line[5]=='E')&&(Line[6]=='S'))
       READDATA = 1;


   } /* while */

  fclose(fp);
  Set_Beta_Topology(pr);
  Set_Coil_Type(pr);

}  /* end of Read_DSSP_File() */







void Read_BSSP_File(pr,fname)
 struct PROTEIN *pr;
 char *fname;
{
 FILE *fp; 
 char Line[200],buff[10],Fname[256],dir[8],RnumStr[8],headstr[32];
 int  i,ok,end,count,acc,L;
 int READDATA,Rnum;
 
 /*JR  Get_Part_Of_Line(dir,fname,2,3);
  */
 Get_Part_Of_Line(dir,fname,1,2);

 sprintf(Fname,"%s/%s/%s",BSSP_DIR,dir,fname); 
 fp = fopen(Fname,"r");
 
 sprintf(pr->pname,"%s",fname); 
 Cut_Filename(pr->pname,pr->cname);
 if (fp==NULL) { 
  printf("#Can't find bsspfile \"%s\". Try %s\n",Fname,fname); 
  fp = fopen(fname,"r");
  if (fp ==NULL) {printf("#ERROR:Can't find bsspfile \"%s\"\n",fname); exit(1);} 
  }

 pr->chain = '-'; 

 READDATA = end = 0;  
 count = 1; 
 /* pr->header[0] = '\0'; */
 /* pr->compnd[0] = '\0'; */
 for (i=0;i<pr->MaxNaa;++i) pr->Seq[i] = pr->Sec[i] = pr->Exp[i] = ' ';
 pr->Naa = 0; pr->Nsse = 0;

 Rnum = 0; 
 while (end ==0)
  { 
    Line[0] = '\0';
    if (fgets(Line,199,fp)==NULL) {end = 1; READDATA = 0;} 
      
    L = strlen(Line);
    if (Line[L-1] == '\n') Line[L-1] = '\0';
   
    if (strncmp(Line,"HEADER",5)==0)
     { Get_Part_Of_Line(pr->header,Line,10,126); 
       Remove_Head_Tail_Space(pr->header); }
    if (strncmp(Line,"COMPND",5)==0)
     { Get_Part_Of_Line(pr->compnd,Line,10,126); 
       Remove_Head_Tail_Space(pr->compnd); }
 
    if ((READDATA==1)&&((pr->chain=='-')||(pr->chain==Line[11])))    
       { 
        if ((Line[13]!='!')&&(Rnum<(pr->MaxNaa-1))) 
         { 
           /* pr->chain = Line[11];  */
           ++Rnum; 
           if (Rnum >= pr->MaxNaa) 
              { printf("#ERROR: bsspfile \"%s\" and \"%s\" is over MaxNaa %d",
                        fname,Fname,pr->MaxNaa); exit(1);}

           Get_Part_Of_Line(RnumStr,Line,6,10);  
           Remove_Space(pr->Rnum[Rnum],RnumStr);
           pr->Seq[Rnum] = Line[13];
           pr->Sec[Rnum] = Line[16];
           if (Line[75]=='C') pr->Seq[Rnum] += 32; /* for CIS conf */ 
           Get_Part_Of_Line(buff,Line,34,37);    acc = atoi(buff); 
           if  ( ((double)acc/(double)Standard_ACC(pr->Seq[Rnum]))>0.2)
             pr->Exp[Rnum] = 'e'; else pr->Exp[Rnum] = 'b'; 

           if (pr->AngMallocType != '-')
           { pr->ACC[Rnum] = acc; 
             Get_Part_Of_Line(buff,Line,26,28); pr->BP1[Rnum] = atoi(buff); 
             Get_Part_Of_Line(buff,Line,30,32); pr->BP2[Rnum] = atoi(buff);
 
             Get_Part_Of_Line(buff,Line, 95,100);  pr->Phi[Rnum] = atof(buff);
             Get_Part_Of_Line(buff,Line,101,106);  pr->Psi[Rnum] = atof(buff); }

           if ((pr->XyzMallocType == 'A')||(pr->XyzMallocType == 'W'))
           { Get_Part_Of_Line(buff,Line,107,113);  pr->PosCA[Rnum][0] = atof(buff);
             Get_Part_Of_Line(buff,Line,114,120);  pr->PosCA[Rnum][1] = atof(buff);
             Get_Part_Of_Line(buff,Line,121,127);  pr->PosCA[Rnum][2] = atof(buff); }
 
           if ((pr->XyzMallocType == 'B')||(pr->XyzMallocType == 'W'))
           { Get_Part_Of_Line(buff,Line,77,82);  pr->PosCB[Rnum][0] = atof(buff);
             Get_Part_Of_Line(buff,Line,83,88);  pr->PosCB[Rnum][1] = atof(buff);
             Get_Part_Of_Line(buff,Line,89,94);  pr->PosCB[Rnum][2] = atof(buff); }

           pr->Naa = Rnum;   }
       } 

      Get_Part_Of_Line(headstr,Line,0,13);
      if (strncmp(headstr,"#   RESIDUE AA",14)==0) READDATA = 1;


   } /* while */

  fclose(fp);
  Set_Beta_Topology(pr);
  Set_Coil_Type(pr);
  Set_SSE(pr); 
  Set_Dis_bwn_SSE(pr,'C');

}  /* end of Read_BSSP_File() */


void Read_BSSP_File_Range(pr,fname,RnumSta,RnumEnd)
 struct PROTEIN *pr;
 char *fname,*RnumSta,*RnumEnd;
{
 FILE *fp; 
 char Line[200],buff[10],Fname[256],dir[8],RnumStr[8],RnumStrRs[8],headstr[32];
 int  i,ok,end,count,acc;
 int READDATA,Rnum,RangeIn;
 
 Get_Part_Of_Line(dir,fname,1,2);
 sprintf(Fname,"%s/%s/%s",BSSP_DIR,dir,fname); 
 fp = fopen(Fname,"r");
 
 sprintf(pr->pname,"%s",fname); 
 Cut_Filename(pr->pname,pr->cname);
 if (fp==NULL) { 
  printf("# Can't find bsspfile \"%s\". Try %s\n",Fname,fname); 
  fp = fopen(fname,"r");
  if (fp ==NULL) {printf("#ERROR:Can't find bsspfile \"%s\"\n",fname); exit(1);} 
  }
 pr->chain = '-'; 

 READDATA = end = 0;  count = 1;
 for (i=0;i<pr->MaxNaa;++i) pr->Seq[i] = pr->Sec[i] = pr->Exp[i] = ' ';
 Rnum = 0; RangeIn = 0;

 while (end ==0)
  { 
     Line[0] = '\0';
     if (fgets(Line,199,fp)==NULL) {end = 1; READDATA = 0;} 
 
     if ((READDATA==1)&&((pr->chain=='-')||(pr->chain==Line[11]))&&
         (Line[13]!='!')&&(Rnum<(pr->MaxNaa-1))) 
         { 
           pr->chain = Line[11]; 
           Get_Part_Of_Line(RnumStr,Line,6,10); 
           Remove_Space(RnumStrRs,RnumStr);         
           if (strcmp(RnumStrRs,RnumSta)==0) RangeIn = 1;

           if (RangeIn == 1)
           { 
           ++Rnum; 
           Get_Part_Of_Line(pr->Rnum[Rnum],Line,6,10); pr->Rnum[Rnum][5] = '\0'; 
           pr->Seq[Rnum] = Line[13];
           pr->Sec[Rnum] = Line[16];
           if (Line[75]=='C') pr->Seq[Rnum] += 32; /* for CIS conf */ 
           Get_Part_Of_Line(buff,Line,34,37);    acc = atoi(buff); 
           if  ( ((double)acc/(double)Standard_ACC(pr->Seq[Rnum]))>0.2)
             pr->Exp[Rnum] = 'e'; else pr->Exp[Rnum] = 'b'; 

           if (pr->AngMallocType != '-')
           { pr->ACC[Rnum] = acc; 
             Get_Part_Of_Line(buff,Line,26,28); pr->BP1[Rnum] = atoi(buff); 
             Get_Part_Of_Line(buff,Line,30,32); pr->BP2[Rnum] = atoi(buff);
 
             Get_Part_Of_Line(buff,Line, 95,100);  pr->Phi[Rnum] = atof(buff);
             Get_Part_Of_Line(buff,Line,101,106);  pr->Psi[Rnum] = atof(buff); }

           if ((pr->XyzMallocType == 'A')||(pr->XyzMallocType == 'W'))
           { Get_Part_Of_Line(buff,Line,107,113);  pr->PosCA[Rnum][0] = atof(buff);
             Get_Part_Of_Line(buff,Line,114,120);  pr->PosCA[Rnum][1] = atof(buff);
             Get_Part_Of_Line(buff,Line,121,127);  pr->PosCA[Rnum][2] = atof(buff); }
 
           if ((pr->XyzMallocType == 'B')||(pr->XyzMallocType == 'W'))
           { Get_Part_Of_Line(buff,Line,77,82);  pr->PosCB[Rnum][0] = atof(buff);
             Get_Part_Of_Line(buff,Line,83,88);  pr->PosCB[Rnum][1] = atof(buff);
             Get_Part_Of_Line(buff,Line,89,94);  pr->PosCB[Rnum][2] = atof(buff); }
          
          pr->Naa = Rnum; 

          } /* if RangeIn */
         
        if (strcmp(RnumStrRs,RnumEnd)==0) RangeIn = 0;

       } /* if */ 

      Get_Part_Of_Line(headstr,Line,0,13);
      if (strncmp(headstr,"#   RESIDUE AA",14)==0) READDATA = 1;
 
   } /* while */

  fclose(fp);
  Set_Beta_Topology(pr);
  Set_Coil_Type(pr);

}  /* end of Read_BSSP_File_Range() */







char Expose(acc,rsym)
 int acc;
 char rsym;
{ double Rexp;

  Rexp = (double)acc/Standard_ACC(rsym);
  /*
       if (Rexp<=0.06) return('b');
  else if (Rexp<=0.36) return('m'); 
  else return('e'); 
  */
       if (Rexp<=0.20) return('b');
  else return('e'); 
}



int Standard_ACC(rsym)
 char rsym;
{
  /*
   These values are calculated
   by G-X-G fully extended conformation(phi=psi=180).
  */
  
  switch(rsym)
   { case 'A': return(112); break;
     case 'R': return(253); break;
     case 'N': return(165); break;
     case 'D': return(162); break;
     case 'C': return(150); break;
     case 'Q': return(196); break;
     case 'E': return(199); break;
     case 'G': return( 84); break;
     case 'H': return(191); break;
     case 'I': return(171); break;
     case 'L': return(178); break;
     case 'K': return(212); break;
     case 'M': return(207); break;
     case 'F': return(209); break;
     case 'P': return(129); break;
     case 'S': return(128); break;
     case 'T': return(154); break;
     case 'W': return(251); break;
     case 'Y': return(230); break;
     case 'V': return(150); break;
     default : return(112); 
  }

} /* end of Standard_ACC() */




void Cut_Filename(fname,cname)
 char *fname,*cname;
{
  int s,e;
  s = e = strlen(fname);
  while ((s>=0)&&(fname[s]!='/')) --s;
  while ((e>=0)&&(fname[e]!='.')) --e;
  if (e == -1) e = strlen(fname);
  Get_Part_Of_Line(cname,fname,s+1,e-1);
}


void Read_Sequence(pro,fname)
 struct PROTEIN *pro;
 char *fname;
{
 FILE *fp;
 int ok,i,count;
 char sym, buff[32],name[128];

 if ((fp =fopen(fname,"r"))==NULL)
  { printf(" I can't find %s.\n",fname); exit(1);}

 count = 0;
 while (feof(fp)==0)
  {
   sym = fgetc(fp); 
   if (sym=='>') { name[0] = '\0'; fgets(name,127,fp); name[strlen(name)-1] = '\0';}
   else if (isalpha(sym)!=0)
   { ++count; pro->Seq[count] = sym; }
  }
 fclose(fp);
 
 pro->Naa = count;

 for (i=1;i<=pro->Naa;++i) { pro->Sec[i] = 'X'; pro->Exp[i] = 'X';}
 sprintf(pro->pname,"%s",fname);
 Cut_Filename(pro->pname,pro->cname);

for (i=1;i<=pro->Naa;++i) sprintf(pro->Rnum[i],"%4d",i); 
 sprintf(pro->Tax,"******");  
  
} /* end of Read_Sequence() */




void Read_CASP_style_Alignment(fname,pA,pB,Nali,nalA,nalB)
 char *fname;
 struct PROTEIN *pA,*pB;
 int *Nali,*nalA,*nalB;
{
 FILE *fp;
 int read_on,rA,rB,dA,dB,rA0,rB0,j;
 char ln[100],aA[8],nA[8],aB[8],nB[8],buff1[64],buff2[64];
 char RnumPlain;
 /** 
 [FILE FORMAT EXAMPLE] 
 NPRO 2
 PRO1 1timA.bssp
 PRO2 1gca-.bssp
 COMMENT Naa1  247 Naa2  309
 RNUMPLAIN                   <- Add this line, MATRAS use the plain rnum number, 
 ALIGNMENT                      not use PDB rnum string.
 G 42 N 39
 A 43 D 40
 P 44 S 41
 - -1 Q 42
 - -1 N 43
 - -1 D 44
 S 45 Q 45
 END
 **/

 RnumPlain = 'F';
 fp = fopen(fname,"r");
 if (fp==NULL){ printf("#ERROR:Can't find alignment_file \"%s\"\n",fname); exit(1); }
 read_on = 0;
 *Nali = 0;
 while (feof(fp)==0)
 {
  ln[0] = '\0';
  fgets(ln,80,fp);
  if ((ln[0]=='E')&&(ln[1]=='N')&&(ln[2]=='D')) read_on = 0;
  if (strncmp(ln,"RNUMPLAIN",8)==0) RnumPlain = 'T'; 

  if (read_on == 1)
  {
   sscanf(ln,"%s %s %s %s\n",aA,nA,aB,nB);
   if (aA[0]=='-') rA = -1;
   else { if (RnumPlain=='T') rA = atoi(nA);
                         else rA = Find_Rnumber(pA,aA,nA); }
   if (aB[0]=='-') rB = -1;
   else { if (RnumPlain=='T') rB = atoi(nB);
                         else rB = Find_Rnumber(pB,aB,nB); }

   if ((rA>=-1)&&(rB>=-1)) {
    
      rA0 =  nalA[*Nali]; rB0 =  nalB[*Nali];
      dA = rA - rA0; dB = rB - rB0;
    
      if ((dA>1)&&(rA>=1)&&(rA0>=1)&&(*Nali>0))
       for (j=1;j<dA;++j) {++(*Nali); nalA[*Nali] = j + rA0; nalB[*Nali] = -1;}       
      
      if ((dB>1)&&(rB>=1)&&(rB0>=1)&&(*Nali>0))
       for (j=1;j<dB;++j) {++(*Nali); nalA[*Nali] = -1; nalB[*Nali] = j + rB0;}       

      ++(*Nali); nalA[*Nali] = rA; nalB[*Nali] = rB;
     }
    else printf("Error %s %s %s %s\n",aA,nA,aB,nB);
   }    
   if ((ln[0]=='A')&&(ln[1]=='L')&&(ln[2]=='I')) read_on = 1;
  }


 fclose(fp);

} /* end of Read_CASP_style_Alignment() */



void Read_CASP_style_Constraint_Alignment(fname,pA,pB)
 char *fname;
 struct PROTEIN *pA,*pB;
{
 FILE *fp;
 int read_on,rA,rB,i;
 char ln[100],aA[8],nA[8],aB[8],nB[8];

 printf("##Read Constraint\n");

 PAR.ConAtoB = (int *)malloc(sizeof(int)*(pA->Naa+1));

 fp = fopen(fname,"r"); 
 if (fp==NULL){ printf("#ERROR:Can't find constraint file \"%s\"\n",fname); exit(1); }
 read_on = 0;
 for (i=0;i<=pA->Naa;++i) PAR.ConAtoB[i] = -1;
 while (feof(fp)==0)
 {
  ln[0] = '\0';
  fgets(ln,80,fp);
  if ((ln[0]=='E')&&(ln[1]=='N')&&(ln[2]=='D')) read_on = 0;

  if (read_on == 1)
  {
   sscanf(ln,"%s %s %s %s\n",aA,nA,aB,nB);
   if (aA[0]=='-') rA = -1;
            else   rA = Find_Rnumber(pA,aA,nA);
   if (aB[0]=='-') rB = -1;
            else   rB = Find_Rnumber(pB,aB,nB);
   if ((rA>=1)&&(rB>=1)) { PAR.ConAtoB[rA] = rB; }
  }   
 
   if ((ln[0]=='A')&&(ln[1]=='L')&&(ln[2]=='I')) read_on = 1;
  }

 fclose(fp);

} /* end of Read_CASP_style_Constraint_Alignment() */




int Find_Rnumber(pr,amino,num)
 struct PROTEIN *pr;
 char *amino,*num;
{
 int i,hit;
 char rnumstr[16];
 hit = 0;
 i = 1;
 while ((hit==0)&&(i<=pr->Naa))
 {
  Remove_Space(rnumstr,pr->Rnum[i]);
  /*
  if ((strcmp(rnumstr,num)==0)&&(pr->Seq[i]==amino[0])) hit = 1;
  */ 
  if ((strcmp(rnumstr,num)==0)&&(numAA[pr->Seq[i]]==numAA[amino[0]])) hit = 1;
  else ++i;
  }
 if (hit==1) return(i); 
  else 
  {
   printf("#Find_Rnumber : Can't find \"%s\" \"%s\"\n",amino,num);  
   return(-2);
   } 

} /* end of Find_Rnumber() */



void Remove_Space(rsp,orig)
 char *rsp,*orig;
{
 int i,j;

 if (orig[0]!='\0')
 {
 j =0;
 for (i=0;i<strlen(orig);++i)
  {
   if (orig[i]!=' ') { rsp[j] = orig[i];  ++j;}
  }
 rsp[j] = '\0';
 }

} /* end of Remove_Space() */



void Set_Break(pr)
 struct PROTEIN *pr;
{
 int i;
 float X[3];
 double D,DD;

 pr->Nbreak = 0;
 for (i=1;i<pr->Naa;++i)
 {
  X[0] = pr->PosCA[i][0] - pr->PosCA[i+1][0];  
  X[1] = pr->PosCA[i][1] - pr->PosCA[i+1][1];  
  X[2] = pr->PosCA[i][2] - pr->PosCA[i+1][2];  
  DD = X[0]*X[0] + X[1]*X[1] + X[2]*X[2];
  if (DD>0.0) {D  = sqrt(DD);} else {D = 0.0;}

  if ((D>4.5)||(D<2.7)) { pr->Break[i] = 1;  ++(pr->Nbreak); }
                   else pr->Break[i] = 0;
  /*
  if (pr->Break[i]==1)
   printf("BREAK %s %d %f\n",pr->pname,i,pr->DmapCA[i][i+1]);
  */
 }

} /* end of Set_Break() */




void Remove_Head_Tail_Space(string)
 char *string;
{
 char *buff;
 int i,L,st,en,ok;

 L = strlen(string);
 buff = (char *)malloc(sizeof(char )*(L+1));
 sprintf(buff,"%s",string);

 /*** START NUM ***/
 i = 0; ok = 0;
 while ((ok==0)&&(i<L))
 { if (buff[i]!=' ') ok = 1; else ++i; }
 st = i;

 if (ok==0) {string[0] = '\0';}
 else
 {
 /*** END NUM ***/
  i = L-1; ok = 0;
  while ((ok==0)&&(i>=0))
  { if ((buff[i]!=' ')&&(buff[i]!='\n')) ok = 1; else --i; }
  en = i;
  Get_Part_Of_Line(string,buff,st,en);
 }

 free(buff);

} /* end of Remove_Head_Tail_Space() */



void Get_First_Field(first, rest, str)
 char *first,  *rest,  *str;
{
 int s,t,L;
 /***
  <EXAMPLE>
  str  = " ABCDE  FGH ";
     then
  first = "ABCDE";
  rest = "FGH ";
 ***/

 /** First **/
 s = 0; L = strlen(str);
 while (((str[s]==' ')||(str[s]=='|'))&&(s<L)) ++s;

 t = 0;
 while ((str[s]!=' ')&&(str[s]!='|')&&(s<L))
 { first[t] = str[s];
   ++s; ++t;}
  first[t] = '\0';

 /** Rest **/
 while (((str[s]==' ')||(str[s]=='|'))&&(s<L)) ++s;
 t = 0;
 while (s<L)
 { rest[t] = str[s];
  ++t; ++s; }
 rest[t] = '\0';

} /* end of Get_First_Field() */



void Find_Split_Point(str,splsym,Asplpos,Nspl,Nsplmax)
 char *str;          /* Input String */
 char splsym;        /* Symbol for split */
 int *Asplpos;       /* Array of position of splsym */
 int *Nspl;          /* Number of splsym */
 int Nsplmax;        /* malloced number of Asplpos  */
{
 int i,L;
 char pre,post;
  /*
  Asplpos[0]    is always equal to  -1,
  Asplpos[Nspl] is always equal to  lengh(str).
 
  [Example 1]
   *str = "ab/cd/ef";  splsym = '/';
   Asplpos -> {-1,2,5,8}; Nspl = 3;
  [Example 2]
   *str = "ab/cd////ef";  splsym = '/';
   Asplpos -> {-1,2,5,8,11}; Nspl = 4;

  */

 L = strlen(str); *Nspl = 1;
 Asplpos[0] = -1;

 for (i=0;i<L;++i)
 {
  if (i>0)     pre  = str[i-1]; else pre = '\0';
  if (i<(L-1)) post = str[i+1]; else post = '\0';

  if ( (str[i]==splsym)&&((*Nspl)<Nsplmax)&&
       ((pre != splsym) || (post != splsym)) )
   {
     Asplpos[*Nspl] = i;
     ++(*Nspl);
     /* printf("split %d %c %d\n",nspl,str[i],i); */
   }
  pre = str[i];
 }
 Asplpos[*Nspl] = L;
} /* end of Find_Split_Point() */



char *Get_Date_String()
{
 time_t      now_t;
 struct tm  *loc_t;
 static char Mon[][4] = {"Jan","Feb","Mar","Apr","May","Jun",
                         "Jul","Aug","Sep","Oct","Nov","Dec"};
 static char string[64]; 
 now_t = time(NULL);
 loc_t = localtime(&now_t);

 sprintf(string,"%s %d,%d %d:%d:%d",
  Mon[loc_t->tm_mon],loc_t->tm_mday,loc_t->tm_year+1900,
  loc_t->tm_hour,loc_t->tm_min,loc_t->tm_sec);

 return(string);

} /* end of Get_Date_String() */

