#define _CRT_SECURE_NO_DEPRECATE		// turn off deprecate warning in vc++ 
#include <stdio.h>
#include <string.h>
#include <ctype.h>

////////////////////////////////////////////////////////////////////
//
// Adjustable parameters
// set in config file

// Total iterations
extern	int		ITERATIONS;
// Type of Geometry to use
extern	enum { RIGID,AFFINE } GEOMETRY;

// Number of nearest neighbours in graph
extern	int		NEIGHBOURS;

// output (0,1,2,3)
extern	int VERBOSE;

//Dynamic Programing
extern bool DYNAMICPROGRAMMING;

//Mtars Initialization
extern bool MATRASINIT;

//MatrasPGM
extern	bool MATRASPGM;

// filenames
extern	char P1FILE[256];
extern	char P2FILE[256];
extern	char ALFILE[256];
extern	char MATCHFILE[256];
extern	char SALFILE[256];
extern	char SSE_A[256];
extern	char SSE_B[256];
extern  char PMATCHSCORE[256];

// mode
extern	enum { MATCHMODE,PARAMMODE } MODE;


//gaps
extern	double GAPOPEN;
extern	double GAPEXTN;
extern	double GAPINIT;
extern	double GAPEND;

/************************************************************************/
/*	int geti(FILE *)						*/
/*	Input: 		file to scan					*/
/*	Output: 	integer from file 				*/
/*	Calls: 		-						*/
/*	Includes:	stdio.h						*/
/*	Written by:	R.C. Wilson 2/10/95				*/
/************************************************************************/

int geti(FILE *fp)
{
	int i;

	fscanf(fp,"%d",&i);
	return i;
}

/************************************************************************/
/*	int getf(FILE *)						*/
/*	Input: 		file to scan					*/
/*	Output: 	floating point from file 			*/
/*	Calls: 		-						*/
/*	Includes:	stdio.h						*/
/*	Written by:	R.C. Wilson 2/10/95				*/
/************************************************************************/

float getf(FILE *fp)
{
	float f;

	fscanf(fp,"%f",&f);
	return f;
}

double getd(FILE *fp)
{
	double f;

	fscanf(fp,"%lf",&f);
	return f;
}

/************************************************************************/
/*	int Gets(FILE *,char *)						*/
/*	Input: 		file to scan					*/
/*	Output: 	string from file 				*/
/*	Calls: 		-						*/
/*	Includes:	stdio.h						*/
/*	Written by:	R.C. Wilson 2/10/95				*/
/************************************************************************/

void Gets(FILE *fp,char *s)
{
	fscanf(fp,"%s",s);
}



/* Token interpreting functions						*/

int Interpret_paramfile(char *s,FILE *fp)
{
	char dummy[256];

	if( !strcmp("Iterations",s) ) 			{ITERATIONS=geti(fp);return true;}
	if( !strcmp("NearestNeighbours",s) ) 	{NEIGHBOURS=geti(fp);return true;}
	if( !strcmp("MatrasInit",s) )			{MATRASINIT=(getd(fp)!=0);return true;}
	if( !strcmp("MatrasPGM",s) )			{MATRASPGM=(getd(fp)!=0);return true;}
	if( !strcmp("Geometry",s) )				{
		Gets(fp,dummy);
		if( !strcmp("Rigid",dummy) ) GEOMETRY=RIGID;
		else GEOMETRY=AFFINE;
		return true;
	}
	if( !strcmp("Verbose",s) )				{VERBOSE=geti(fp);return true;}
	if( !strcmp("Protein1",s) )				{Gets(fp,P1FILE);return true;}
	if( !strcmp("Protein2",s) )				{Gets(fp,P2FILE);return true;}
	if( !strcmp("Align",s) )				{Gets(fp,ALFILE);return true;}
	if( !strcmp("Match",s) )				{Gets(fp,MATCHFILE);return true;}
	if( !strcmp("Align",s) )				{Gets(fp,ALFILE);return true;}
	if( !strcmp("SALfile",s) )				{Gets(fp,SALFILE);return true;}
	if( !strcmp("SSE_Afile",s) )				{Gets(fp,SSE_A);return true;}
	if( !strcmp("SSE_Bfile",s) )				{Gets(fp,SSE_B);return true;}
	if( !strcmp("PMATCHscore",s) )				{Gets(fp,PMATCHSCORE);return true;}
	if( !strcmp("GAPopen",s) )				{GAPOPEN=getd(fp);return true;}
	if( !strcmp("GAPextension",s) )				{GAPEXTN=getd(fp);return true;}
	if( !strcmp("GAPinitial",s) )				{GAPINIT=getd(fp);return true;}
	if( !strcmp("GAPend",s) )				{GAPEND=getd(fp);return true;}
	Gets(fp,dummy);
	return false;
}

/* This program scans the start of a data file for tokens and values
//  of the form  <token>=value
// Reads file up to first non-alphabetic character as first character of the
// line
*/
/************************************************************************/
/*	void scanfile(FILE *)						*/
/*	Input: 		file to scan					*/
/*	Output: 	none						*/
/*	Calls: 		(Interpret)					*/
/*	Includes:	stdio.h ctype.h					*/
/*	Written by:	R.C. Wilson 2/10/95				*/
/************************************************************************/


void LoadParamFile(char* fname)
{
	FILE* fp=fopen(fname,"r");
	if( fp==NULL ) throw("Can't open parameter file.");

	char s[128];

	while(true) {
		fscanf(fp,"%s ",s);
		if( s[0]=='#' ) fgets(s,127,fp);	/* If its a comment, skip over    */
		else {
			if( !isalpha(s[0]) || feof(fp) ) {		/* if not an alphabetic, end of tokens */
				if( !isdigit(s[0]) && !feof(fp) ) {  	/* so it must be a number	  */
					fprintf(stderr,"Spurious data in file:%s\n",s);
				}
				fclose(fp);
				return;
			}
			fscanf(fp," = ");		/* Skip over = if it is there	*/
			if( !Interpret_paramfile(s,fp) ) {	/* Process string		*/
				fprintf(stderr,"Unknown token:%s\n",s);
			}
		}
	}
	fclose(fp);
}



