/*

 <Jacobi3.c>

 Jacobi Method for 3-dimentional matrix ([3][3])

 coded by Takeshi Kawabata.

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define NN 3 

static void equal_matrix();
static void prod_matrix();
static void print_matrix();
static void find_max_abs(); 

void Jacobi_Wilkinson3(A,U)
double A[NN][NN],U[NN][NN];
{
 double R[NN][NN],TR[NN][NN],BUFF[NN][NN];
 double max,p,q,co,si,co2,si2,eps;
 double wa,sa,r,sqrt2;
 int i,j,mi,mj,c,cmax;

 sqrt2 = sqrt(2.0);
 eps = 0.00001;

   /* --------SETTING U[][] = I --------*/

   for (i=0;i<NN;++i)
     for (j=0;j<NN;++j) if (i==j) U[i][j] = 1.0;  else U[i][j] = 0.0;

    find_max_abs(&mi,&mj,&max,A);
    c = 0;

    while ((max>eps)&&(c<100))
    {
      /*--------- SET of cos sin -----------*/
      

      wa = (A[mi][mi] + A[mj][mj])/2;
      sa = (A[mi][mi] - A[mj][mj])/2;
      r = sqrt(sa*sa + A[mi][mj]*A[mi][mj]);
      co =  sqrt(1.0+sa/r)/sqrt2;  si =  A[mi][mj]/2.0/r/co;
   
      if (sa>0.0)
        { co =  sqrt(1.0+sa/r)/sqrt2;  si =  A[mi][mj]/2.0/r/co; }
      else
        { co =   sqrt(1.0-sa/r)/sqrt2; si = - A[mi][mj]/2.0/r/co; }

      /* -------- SET of Rot Matrix R and TR----- */
      for (i=0;i<NN;++i)
        for (j=0;j<NN;++j) if (i==j)  R[i][j] = 1.0;  else R[i][j] =0.0;

       R[mi][mi] =   co; R[mi][mj] = -si;
       R[mj][mi] =   si; R[mj][mj] = co;

      for (i=0;i<NN;++i)
        for (j=0;j<NN;++j) TR[j][i] = R[i][j];

       prod_matrix(A,R,BUFF);
       prod_matrix(TR,BUFF,A);
       prod_matrix(U,R,BUFF);
       equal_matrix(U,BUFF);

       find_max_abs(&mi,&mj,&max,A);
       ++c;
       
      } /* end of while */

}/* end of Jacobi_Wilkinson3() */




void prod_matrix(A,B,C) /* C = A * B */
 double A[NN][NN],B[NN][NN],C[NN][NN];
{ int i,j,k;
  double sum;
  for (i=0;i<NN;++i)
    for (j=0;j<NN;++j)
     { sum =0;
       for (k=0;k<NN;++k) { sum = sum + A[i][k] * B[k][j];}
       C[i][j] = sum; }
 } 


void equal_matrix(A,B)    /* A = B */
 double A[NN][NN],B[NN][NN];
{ int i,j;
  for (i=0;i<NN;++i)
    for (j=0;j<NN;++j) A[i][j] = B[i][j];
}



void find_max_abs(mi,mj,max,A) /* Max_Abs element A[mi][mj] = max */
 int *mi,*mj;
 double *max,A[NN][NN];
{ int i,j,p,q;

  *max = 0.0;
  for (i=0;i<NN;++i)
    for (j=0;j<NN;++j)
      if (i!=j)
       if (fabs(A[i][j])> *max)
              { *max = fabs(A[i][j]);     p = i; q= j;}
   *mi =p; *mj = q;
 } /* end of find_max_abs() */


void print_matrix(A)
 double A[NN][NN];
{ int i,j;
  printf("\n");
  for (i=0;i<NN;++i)
    for (j=0;j<NN;++j)
     { printf(" [%d%d]:%3.8lf",i,j,A[i][j]);
       if (j==(NN-1)) { printf("\n");}          }
  printf("\n");
}/* end of print_matrix() */




