/*

 <LlistAlign.c>
 
  Functions for LinearList data for alignment 

 coded by Takeshi Kawabata.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <math.h>
#include "main.h" 
#include "protein.h" 
#include "IO.h" 
#include "LlistAlign.h" 

/****
  >>  Definition of "LineNode" linearlist structure  <<

  The end node is marked by next==NULL.

****/


/*** FUNCTIONS (GLOBAL) ***/

void Malloc_AlignNode();
void Show_AlignNode();
void Add_AlignNode_To_Tail();
void Free_AlignNode_Llist();
void Free_AlignNode();
struct AlignNode *Tail_AlignNode();
float Tail_Value_AlignNode();
int Number_Of_Nodes();
void Joint_AlignList();
void Set_Sta_End_Of_Alignment();
void Equal_AlignNode();


/*** FUNCTIONS (LOCAL) ***/
static int Overlap_Path();
static void Equal_Similarity();


void Show_AlignNode(head)
 struct AlignNode *head;
{
 struct AlignNode *bn;
 char pt;

 bn = head;
 while (bn->next != NULL)
 { bn = bn->next;
   if (bn->next == NULL) pt = 'N'; else pt = '*';
   printf("num %d Nali %d nalA %d nalB %d mark %d %c\n",
    bn->num,bn->Nali,bn->nalA[1],bn->nalB[1],bn->mark,pt); 
  }

} /* end of Show_AlignNode() */






void Joint_AlignList(head,tail)
 struct AlignNode *head,*tail;
{
 /* 
  head = head + tail
  tail := NULL 
 */


 struct AlignNode *bn;

 /* Find End Node */
 bn = head;
 head->num = -1; 
 while (bn->next != NULL) bn = bn->next;

 bn->next = tail->next;
 tail->next->prev = bn;

 /* Renumbering */ 
 while (bn->next != NULL) 
  { bn->next->num = bn->num +1;
    bn = bn->next;
   }

 tail->next = NULL;

} /* end of Joint_AlignList() */



void Add_AlignNode_To_Tail(head,add)
 struct AlignNode *head;  /* Head of Linked List */
 struct AlignNode *add;   /* Novel node to be added */
{
 struct AlignNode *bn;
 int L;
 
 /* Find Tail Node */
 bn = head;
 head->num = -1; 
 while (bn->next != NULL) bn = bn->next;

 /* Malloc New Node */
 bn->next = (struct AlignNode *)malloc(sizeof(struct AlignNode));
 
 bn->next->prev = bn;
 bn->next->num = bn->num + 1;
 bn->next->next = NULL;
 bn = bn->next;
 
 /* Malloc nalA and nalB for NEW node */
 bn->nalA = (int *)malloc(sizeof(int)*(add->Nali+1));
 bn->nalB = (int *)malloc(sizeof(int)*(add->Nali+1));
 bn->scores = (float *)malloc(sizeof(float)*(add->Nali+1));  /* JR*/


 Equal_AlignNode(bn,add);

 /** Set vaious values ***/

 if (add->mal_sim == 1)
 { bn->sim = (struct SIMILARITY*)malloc(sizeof(struct SIMILARITY));   
   bn->mal_sim = 1;
   bn->sim->mal_nametax = 0; 
   Equal_Similarity(bn->sim,add->sim); } 
 else bn->mal_sim = 0;

 if (add->mal_pro == 1)
 { 
   Malloc_Protein(bn->pro,add->pro->Naa,'A','W'); 
   bn->mal_pro = 1;
   Equal_Protein(bn->pro,add->pro); }
 else bn->mal_pro = 0;
 
} /* end of Add_AlignNode_To_Tail() */




void Free_AlignNode_Llist(head)
 struct AlignNode *head;
{
 struct AlignNode *bn;
 int Nfree,ret;


 /* Shift to End Node */ 
 bn = head; 
 while (bn->next !=NULL) { bn = bn->next; }

 /* Free AlignNode */
 Nfree = 0;
 while (bn != head)
 {
   bn = bn->prev;
   Free_AlignNode(bn->next);  
   free(bn->next); bn->next = NULL;
  }

 /* printf("Nfree %d\n",Nfree); */

} /* end of Free_AlignNode_Llist() */





void Set_Sta_End_Of_Alignment(Nali,nal,sta,end)
 int Nali,*nal,*sta,*end;
{
 int i;
 
 *sta = *end = 0;
 for (i=1;i<=Nali;++i)
 {
  if (nal[i]>0)
  { if (*sta == 0) *sta = nal[i];
    *end = nal[i]; }
 }

} /* end of Set_Sta_End_Of_Alignment() */


void Malloc_AlignNode(an,MaxNali,SimMalType)
 struct AlignNode *an;
 int MaxNali;
 char SimMalType; /* if 'S', sim is malloced. */ 
{
 an->MaxNali = MaxNali;
 an->nalA = (int *)malloc(sizeof(int)*(MaxNali+1));
 an->nalB = (int *)malloc(sizeof(int)*(MaxNali+1));
 an->scores = (float *)malloc(sizeof(float)*(MaxNali+1));  /*JR*/

 if (SimMalType == 'S')
 { an->sim = (struct SIMILARITY*)malloc(sizeof(struct SIMILARITY));
   an->sim->mal_nametax = 0; 
   an->sim->SeqID = an->sim->SecID = an->sim->ExpID = 0.0; 
   an->sim->ScEnv = an->sim->ScDis = an->sim->ScSSE = an->sim->ScAA = 0.0;
   an->sim->Rdis = an->sim->Rsse  =  an->sim->CRMS = an->sim->DRMS = 0.0;
   an->sim->RelO = an->sim->RelS = 0.0;

   an->mal_sim = 1; }
 else an->mal_sim = 0;

 an->mal_pro = 0;

} /* end of  Malloc_AlignNode() */


void Free_AlignNode(an)
 struct AlignNode *an;
{
 int ret;
 /*
 printf("Free_AlignNode Nali %d mal_sim %d\n",an->Nali,an->mal_sim);
 */ 

 if (an->mal_sim == 1)
 {
  if (an->sim->mal_nametax == 1)
   { free(an->sim->Name); an->sim->Name = NULL;
     free(an->sim->Tax);  an->sim->Tax = NULL;
    }
   free(an->sim); an->sim = NULL;
  }

 if (an->MaxNali>0) 
  { free(an->nalB); an->nalB = NULL; 
    free(an->nalA); an->nalA = NULL; 
    free(an->scores); an->scores = NULL;    /*JR */
    an->MaxNali = 0;} 

 if (an->mal_pro == 1) 
  { free(an->pro); an->pro = NULL; }

} /* end of Free_AlignNode() */




struct AlignNode *Tail_AlignNode(head)
 struct AlignNode *head;
{
 struct AlignNode *bn;

 bn = head; 
 while (bn->next !=NULL) bn = bn->next;
 return(bn);

} /* end of Tail_AlignNode() */

float Tail_Value_AlignNode(head)
 struct AlignNode *head;
{
 struct AlignNode *bn;

 bn = head; 
 while (bn->next !=NULL) bn = bn->next;
 return(bn->value);

} /* end of Tail_Value_AlignNode() */







struct AlignNode *Find_Overlap_Path(head,Nali,nalA,nalB)
 struct AlignNode *head;
 int Nali,*nalA,*nalB;  /* Alignment to be checked */
{
 struct AlignNode *bn,*ovnode;
 int hit,ov;

 bn = head; ovnode = NULL;
 hit = 0; 
 
 while ((bn->next !=NULL)&&(hit==0))
 { bn = bn->next;
   if (Overlap_Path(Nali,nalA,nalB,bn->Nali,bn->nalA,bn->nalB)==1) 
     { hit = 1; ovnode = bn; }  
  }

  return(ovnode);

} /* end of Find_Overlap_Path() */







int Overlap_Path(nali,nala,nalb,Nali,nalA,nalB)
 int nali,*nala,*nalb;
 int Nali,*nalA,*nalB;
{
 int i,ovA,ovB;
     
 /* Both regions for A and regions for B are overlapped,
    return 1, otherwise  return 0 */
      
 ovA = ovB = 0;
 if ( ((nala[1]<=nalA[1])&&(nalA[1]<=nala[nali]))  ||
      ((nalA[1]<=nala[1])&&(nala[1]<=nalA[Nali]))  ) ovA = 1;
      
 if ( ((nalb[1]<=nalB[1])&&(nalB[1]<=nala[nali]))  ||
      ((nalB[1]<=nalb[1])&&(nalb[1]<=nalB[Nali]))  ) ovB = 1;
        

 if ((ovA==1)&&(ovB==1)) return(1); else return(0);

} /* end of Overlap_Path() */


int Number_Of_Nodes(head)
 struct AlignNode *head;
{
 struct AlignNode *bn;
 int N;

 bn = head;
 N = 0;
 bn->num = -1;
 while (bn->next !=NULL)
 {
  bn->next->num = bn->num + 1;  
  bn = bn->next;
  ++N;
 } 

 return(N);

} /* end of Number_Of_Nodes() */


void Equal_AlignNode(a,b)  /* a: = b; */
 struct AlignNode *a,*b;
{
 int i;

 a->Nali = b->Nali;
 a->staA = b->staA;
 a->endA = b->endA;
 a->staB = b->staB;
 a->endB = b->endB;
 a->sn = b->sn; /* JR for SO SSE */
 a->value = b->value; /* JR */
 for (i=0;i<=a->Nali;++i)
  { a->nalA[i] = b->nalA[i]; 
    a->nalB[i] = b->nalB[i];  }

} /* end of Equal_AlignNode() */




void Equal_Similarity(p,q)  /* p: = q; */
 struct SIMILARITY *p,*q;
{
 
 if (q->mal_nametax == 1)
 {
  p->Name = (char *)malloc(sizeof(char)*(strlen(q->Name)+1));
  sprintf(p->Name,"%s",q->Name); 
  p->Tax = (char *)malloc(sizeof(char)*(strlen(q->Tax)+1));
  sprintf(p->Tax,"%s",q->Tax); 
  p->mal_nametax = 1;
  }

 p->Naa     = q->Naa;
 p->Nsse    = q->Nsse;
 p->Ncomp   = q->Ncomp;
 p->Ncomp_sse = q->Ncomp_sse;
 p->SeqID = q->SeqID;
 p->SecID = q->SecID;
 p->ScEnv = q->ScEnv;
 p->ScDis = q->ScDis;
 p->ScSSE = q->ScSSE;
 p->ScAA  = q->ScAA;
 p->Rdis  = q->Rdis;
 p->Rsse  = q->Rsse;
 p->CRMS  = q->CRMS;
 p->DRMS  = q->DRMS;
 p->RelS  = q->RelS;
 p->RelO  = q->RelO;
 p->mark  = q->mark;

} /* end of Equal_Similarity() */
