/*

 <LlistIO.c>
 
  Functions for LinearList-style data IO.
 
  coded by Takeshi Kawabata.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "main.h" 
#include "protein.h" 
#include "Amino.h" 
#include "Vector.h" 
#include "OutAln.h" 
#include "BetaTopo.h" 
#include "main.h" 
#include "IO.h" 
#include "LlistIO.h" 


/****
  >>  Definition of "LineNode" linearlist structure  <<

  This Linear List is different from othre ones.
  (1)  End node is marked by type='E', not by next==NULL.
  (2)  New node does not always needs memory allocation.
       Sometimes the new nodes are already allocated.
  (3)  Number of data "line" is fixed at 150 characters.

****/

struct LineNode{
  char line[150];
  int  Nline;
  char type;   /* 'E'nd */
  struct LineNode *next;
};



/*** LOCAL VARIABLES  ***/
static struct LineNode HeadLineNode;


/*** FUNCTIONS (GLOBAL) ***/
void Read_LinearList_Data();
void Read_BSSP_File_LinearList();
extern void Set_Neighbours();

/*** FUNCTIONS (LOCAL) ***/
static void Show_LineNodes();
static int Estimate_SSE_Number();
static void Translate_BSSP_File_LinearList();



void Read_LinearList_Data(fname, NLINE)
char *fname;
int *NLINE;
{
 FILE *fp;
 char str[150];
 struct LineNode *curnode,*endnode;
 int L, nline;

 HeadLineNode.type = 'E';
 curnode = &HeadLineNode;
 
 fp = fopen(fname,"r");
 if (fp==NULL) {fprintf(stderr,"#ERROR:Can't find Llistfile \"%s\"\n",fname); exit(1);}

 nline = 0;
 while (feof(fp)==0)
 {
  str[0] = '\0'; 
  fgets(str,149,fp);
  L = strlen(str);
  if (L>0)
  {
   if (str[L-1]=='\n') str[L-1] = '\0';

   if (curnode->next==NULL)
   { curnode->next = (struct LineNode *)malloc(sizeof(struct LineNode)); 
     curnode->next->next = NULL; }
   curnode->type = '-';
   curnode = curnode->next;
   sprintf(curnode->line,"%s",str);
   curnode->type = 'E';
   curnode->Nline = nline;
   ++nline;
  } 
 } /* while */

 *NLINE = nline; 
 fclose(fp);

} /* end of Read_LinearList_Data() */




void Show_LineNodes(head)
 struct LineNode *head;
{
 struct LineNode *bn;
 char pt;

 bn = head;
 while (bn->type != 'E')
 {
  bn = bn->next;
  if (bn->next == NULL) pt = 'N'; else pt = '*';
  printf("%d %s %c\n",bn->Nline,bn->line,pt);
  }

} /* end of Show_LineNodes() */




void Read_BSSP_File_LinearList(pr,fname,angtype,xyztype,RnumSta,RnumEnd)
 struct PROTEIN *pr;
 char *fname;
 char angtype,xyztype;    
 char *RnumSta,*RnumEnd;  /* Start and End Residue Number */
{
 FILE *fp;
 char Fname[256],dir[8];
 int Nline,Nsse;

 /*** SET FILE NAME ***/ 
 sprintf(Fname, "%s", fname);
 fp = fopen(Fname, "r");
 Cut_Filename(pr->pname,pr->cname);

 if (fp==NULL) {
  Get_Part_Of_Line(dir,fname,1,2);
  sprintf(Fname,"%s/%s/%s",BSSP_DIR,dir,fname);
  fp = fopen(Fname,"r");
  if (fp ==NULL) {fprintf(stderr,"#ERROR: Can't find bsspfile \"%s\"\n",fname); exit(1);}
  }
 fclose(fp);
 sprintf(pr->pname,"%s",fname);

 /**** READING LINEAR LIST AND MALLOC PROTEIN ***/
 Read_LinearList_Data(Fname,&Nline);
 Malloc_Protein(pr,Nline,angtype,xyztype);
  
 /**** TRANSLATING LINEAR LIST AND MALLOC SSE ***/
 Translate_BSSP_File_LinearList(pr,RnumSta,RnumEnd);
 Nsse = Estimate_SSE_Number(pr);
 Malloc_SSE(pr,Nsse);
 
 Set_Beta_Topology(pr);
 Set_Coil_Type(pr);
 Set_SSE(pr); 
 Set_Dis_bwn_SSE(pr,'C');
 Set_Neighbours(pr, PAR.MAXNEIGHBOURS);     /* PD */
	
}  /* end of Read_BSSP_File_LinearList() */







void Translate_BSSP_File_LinearList(pr,RnumSta,RnumEnd)
 struct PROTEIN *pr;
 char *RnumSta,*RnumEnd;
{
 int  i,ok,end,count,acc;
 int READDATA,Rnum,Nline,Nsse,RangeIn;
 char buff[10],RnumStr[8],chain_appear,headstr[32];
 struct LineNode *X;
 
 X = &HeadLineNode;
 pr->chain = chain_appear = '-';
 pr->header[0] = pr->compnd[0] = '\0';

 READDATA = end = 0;  count = 1;
 for (i=0;i<pr->MaxNaa;++i) pr->Seq[i] = pr->Sec[i] = pr->Exp[i] = ' ';
 Rnum = 0; 

 if ((RnumSta[0]=='-')||(RnumSta[0]=='\0')) RangeIn = 1;

 while ((X->type != 'E')&&(X->next != NULL))
  {
    X = X->next;
    /* printf("%s\n",X->line); */
    if (strncmp(X->line,"HEADER",5)==0) 
     { Get_Part_Of_Line(pr->header,X->line,10,126); Remove_Head_Tail_Space(pr->header); }
    if (strncmp(X->line,"COMPND",5)==0) 
     { Get_Part_Of_Line(pr->compnd,X->line,10,126); Remove_Head_Tail_Space(pr->compnd); }

    if ((READDATA==1)&&((pr->chain=='-')||(pr->chain == X->line[11])))
       {
         
        if ((X->line[13]!='!')&&(Rnum<(pr->MaxNaa-1)))
         {
           Get_Part_Of_Line(buff,X->line,6,10);  
           Remove_Space(RnumStr,buff); 
            
           if (strcmp(RnumStr,RnumSta)==0) RangeIn = 1;
 
           if (RangeIn == 1)  
           {
            chain_appear = X->line[11];
            ++Rnum;
            sprintf(pr->Rnum[Rnum],"%s",RnumStr);
            pr->Seq[Rnum] = X->line[13];
            pr->Sec[Rnum] = X->line[16];
            if (X->line[75]=='C') pr->Seq[Rnum] = tolower(pr->Seq[Rnum]);  /* for CIS conf */
            Get_Part_Of_Line(buff,X->line,34,37);    acc = atoi(buff);
            if  ( ((double)acc/(double)Standard_ACC(pr->Seq[Rnum]))>0.2)
              pr->Exp[Rnum] = 'e'; else pr->Exp[Rnum] = 'b';

            if (pr->AngMallocType != '-')
            { pr->ACC[Rnum] = acc;
              Get_Part_Of_Line(buff,X->line,26,28); pr->BP1[Rnum] = atoi(buff);
              Get_Part_Of_Line(buff,X->line,30,32); pr->BP2[Rnum] = atoi(buff);

              Get_Part_Of_Line(buff,X->line, 95,100);  pr->Phi[Rnum] = atof(buff);
              Get_Part_Of_Line(buff,X->line,101,106);  pr->Psi[Rnum] = atof(buff); }

            if ((pr->XyzMallocType == 'A')||(pr->XyzMallocType == 'W'))
            { Get_Part_Of_Line(buff,X->line,107,113);  pr->PosCA[Rnum][0] = atof(buff);
              Get_Part_Of_Line(buff,X->line,114,120);  pr->PosCA[Rnum][1] = atof(buff);
              Get_Part_Of_Line(buff,X->line,121,127);  pr->PosCA[Rnum][2] = atof(buff); }

           if ((pr->XyzMallocType == 'B')||(pr->XyzMallocType == 'W'))
            { Get_Part_Of_Line(buff,X->line,77,82);  pr->PosCB[Rnum][0] = atof(buff);
              Get_Part_Of_Line(buff,X->line,83,88);  pr->PosCB[Rnum][1] = atof(buff);
              Get_Part_Of_Line(buff,X->line,89,94);  pr->PosCB[Rnum][2] = atof(buff); }
        
            pr->Naa = Rnum;   
      
          } /* RangeIn */ 

          if (strcmp(RnumStr,RnumEnd)==0) RangeIn = 0;
         
      } /* if not ! */
 
      } /* if  READDATA */

      Get_Part_Of_Line(headstr,X->line,0,13);
      if (strncmp(headstr,"#   RESIDUE AA",14)==0) READDATA = 1;

   } /* while */

  if (chain_appear != ' ') pr->chain = chain_appear;

}  /* end of Translate_BSSP_File_LinearList() */



int Estimate_SSE_Number(pr)
 struct PROTEIN *pr;
{
 int i,hon,eon,Nsse;

 hon = eon = 0; Nsse = 0; 
 for (i=1;i<=pr->Naa;++i)
 {
  if ((pr->Sec[i]=='H')&&(hon==0)) { hon=1; ++Nsse; }
  if ((pr->Sec[i]!='H')&&(hon==1)) { hon=0; }

  if ((pr->Sec[i]=='E')&&(eon==0)) { eon=1; ++Nsse; }
  if ((pr->Sec[i]!='E')&&(eon==1)) { eon=0; }
 
  /* printf("%d %c %d %d\n",i,pr->Sec[i],hon,eon); */
 }

 /* printf("Nsse %d\n",Nsse); */
 return(Nsse);

} /* end of Estimate_SSE_Number() */

