/**

<OutAln.c>

Functions for output alignment information

 coded by Takeshi Kawabata.

**/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h" 
#include "protein.h" 
#include "Amino.h" 
#include "ScoreDIS.h" 
#include "ScoreSSE.h" 
#include "SecStrEle.h" 
#include "LlistAlign.h" 



/** Function (GLOBAL) **/
int  Output_SSE_Alignment_Horizontal();
int  Output_SSE_Alignment_Vertical();
int  Output_Path_Horizontal();
int  Output_Path_Horizontal_ClustalW();
int  Output_Path_Horizontal_HTMLcolor();
int  Output_Alignment_Information();
int  Output_Oneline_Information();
int  Write_JavaView_Color_Alignment();
int  Write_Alignment_CASP_Style();
int  Write_Rasmol_Color_Alignment();
int  Write_Rasmol_Color_Alignment_Each_Protein();
int  Write_Molscript_Alignment_Each_Protein();

/** Function (LOCAL) **/
static void Make_All_SSE_Align();
static void Short_SSE_Extention();
static void Short_SSE_Deletion();
static void PrintAlignedString();
static void PrintAlignedString_Sec();
static void Set_SameSeq();




void Make_All_SSE_Align(pA,pB,Nsal,salA,salB,Nsaln,salAn,salBn)
  struct PROTEIN *pA,*pB;
  int Nsal,*salA,*salB;
  int *Nsaln,*salAn,*salBn;
{ int i,j,k,m,a,b,da,db;

  /********** DEAL WITH BEGIN EDGE ************/ 
  *Nsaln = 0; k = 0;
  da = salA[k]; a = 0;
  db = salB[k]; b = 0;
  if (da>0)
      for (m=0;m<da;++m) { salAn[*Nsaln] = a+m; salBn[*Nsaln] = -1; ++(*Nsaln);}
  if (db>0)
     for (m=0;m<db;++m) { salBn[*Nsaln] = b+m; salAn[*Nsaln] = -1; ++(*Nsaln);}

  /********** DEAL WITH MIDDLE  ************/ 
  
  while (k<Nsal)
  {
   salAn[*Nsaln] = salA[k];
   salBn[*Nsaln] = salB[k];
   ++k; ++(*Nsaln);
   if (k<Nsal)
   { 
    da = salA[k]-salA[k-1]; a = salA[k-1];
    db = salB[k]-salB[k-1]; b = salB[k-1];
    if (da>1)
      for (m=1;m<da;++m) { salAn[*Nsaln] = a+m; salBn[*Nsaln] = -1; ++(*Nsaln);}
    if (db>1)
      for (m=1;m<db;++m) { salBn[*Nsaln] = b+m; salAn[*Nsaln] = -1; ++(*Nsaln);}
    } 
   }

  /********** DEAL WITH END EDGE  ************/ 
  da = pA->Nsse - 1 - salA[Nsal-1]; a = salA[Nsal-1];
  db = pB->Nsse - 1 - salB[Nsal-1]; b = salB[Nsal-1];
  if (da>0)
      for (m=1;m<da;++m) { salAn[*Nsaln] = a+m; salBn[*Nsaln] = -1; ++(*Nsaln);}
  if (db>0)
     for (m=1;m<db;++m) { salBn[*Nsaln] = b+m; salAn[*Nsaln] = -1; ++(*Nsaln);}

} /* end of Make_All_SSE_Align() */ 

  



int Output_SSE_Alignment_Horizontal(fname,pA,pB,sn)
  char *fname;  /* if fname = "" or fname = "-", output stdout */
  struct PROTEIN *pA,*pB;
  struct AlignNode *sn; /* SSE Alignment */ 
{ int i;
  FILE *fp;
  int Nsaln,*salAn,*salBn;


  salAn = (int *)malloc(sizeof(int) * (pA->Nsse+pB->Nsse+1));
  salBn = (int *)malloc(sizeof(int) * (pA->Nsse+pB->Nsse+1));


  /*** FILE OPEN ***/ 
  if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } }

  /*** Make gapped SSE alignment ***/ 
  Make_All_SSE_Align(pA,pB,sn->Nali,sn->nalA,sn->nalB,&Nsaln,salAn,salBn);

  fprintf(fp,"[BEGIN SSE ALIGNMENT]\n");
  /*** HORIZONTAL ALIGNMENT ***/
  fprintf(fp,"proA:");
  for (i=0;i<Nsaln;++i) if (salAn[i]>=0) fprintf(fp,"%c",AlphSSE[pA->SSE[salAn[i]].num]); 
                        else fprintf(fp,"-");
  fprintf(fp,"\n");
  fprintf(fp,"    :");
  for (i=0;i<Nsaln;++i) if (salAn[i]>=0) fprintf(fp,"%c",pA->SSE[salAn[i]].type); 
                        else fprintf(fp,"-");
  fprintf(fp,"\n");

  fprintf(fp,"    :");
  for (i=0;i<Nsaln;++i) if ((salAn[i]>=0)&&(salBn[i]>=0)) fprintf(fp,"*"); else fprintf(fp," ");
  fprintf(fp,"\n");

  fprintf(fp,"    :");
  for (i=0;i<Nsaln;++i) if (salBn[i]>=0) fprintf(fp,"%c",pB->SSE[salBn[i]].type); 
                        else fprintf(fp,"-");
  fprintf(fp,"\n");
  fprintf(fp,"proB:");
  for (i=0;i<Nsaln;++i) if (salBn[i]>=0) fprintf(fp,"%c",AlphSSE[pB->SSE[salBn[i]].num]); 
                        else fprintf(fp,"-");
  fprintf(fp,"\n");
  fprintf(fp,"[END SSE ALIGNMENT]\n");
 
  if (fp!=stdout) fclose(fp); 

  free(salAn);
  free(salBn);

  return(1);

} /* end of Output_SSE_Alignment_Horizontal() */



int Output_SSE_Alignment_Vertical(pA,pB,sn,fname)
  struct PROTEIN *pA,*pB;
  struct AlignNode *sn; /* SSE Alignment */ 
  char *fname;  /* if fname = "" or fname = "-", output stdout */
{ int i;
  FILE *fp;
  int Nsaln,*salAn,*salBn;
  
  /*** FILE OPEN ***/ 
  if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } }

  /*** Make gapped SSE alignment ***/ 
  salAn = (int *)malloc(sizeof(int) * (pA->Nsse+pB->Nsse+1));
  salBn = (int *)malloc(sizeof(int) * (pA->Nsse+pB->Nsse+1));
  
  Make_All_SSE_Align(pA,pB,sn->Nali,sn->nalA,sn->nalB,&Nsaln,salAn,salBn);

  fprintf(fp,"[BEGIN SSE ALIGNMENT]\n");

  for (i=0;i<Nsaln;++i)
  {
    if (salAn[i]>=0)
      fprintf(fp,"(%c %2d %6s - %6s)",
        pA->SSE[salAn[i]].type, pA->SSE[salAn[i]].num,
        pA->Rnum[pA->SSE[salAn[i]].Rsta], pA->Rnum[pA->SSE[salAn[i]].Rend]);
    else
      fprintf(fp,"(-  -    -   -    -  )");

    fprintf(fp," vs ");
    if (salBn[i]>=0)
      fprintf(fp,"(%c %2d %6s - %6s)\n",
        pB->SSE[salBn[i]].type, pB->SSE[salBn[i]].num,
        pB->Rnum[pB->SSE[salBn[i]].Rsta], pB->Rnum[pB->SSE[salBn[i]].Rend]);
    else
      fprintf(fp,"(-  -    -   -    -  )\n");
  
   }
 fprintf(fp,"[END SSE ALIGNMENT]\n");

  return(1);

} /* end of Output_SSE_Alignment_Vertical() */ 







int Write_JavaView_Color_Alignment(fname,pA,pB,a)
 char *fname;
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{
 FILE *fp;
 int i,j,k,Nc;
 char col[6],*cseq;

 
 col[0] = 'b'; col[1] = 'c'; col[2] = 'g'; 
 col[3] = 'y'; col[4] = 'o'; col[5] = 'r';

 Nc = 0;
 for (i=1;i<=a->Nali;++i) if ((a->nalA[i]>0)&&(a->nalB[i]>0)) ++Nc; 

 cseq = (char *)malloc(sizeof(char)*(pA->Naa+pB->Naa+2)); 
 for (i=1;i<=pA->Naa;++i) cseq[i] = 'l';  
 for (i=1;i<=pB->Naa;++i) cseq[i+pA->Naa] = 'd';  

 k = 0;
 for (i=1;i<=a->Nali;++i)
 if ((a->nalA[i]>0)&&(a->nalB[i]>0))  
   { 
    cseq[a->nalA[i]]         = col[(6*k)/(Nc+1)]; 
    cseq[a->nalB[i]+pA->Naa] = col[(6*k)/(Nc+1)]; 
     ++k; 
    }

 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } 
 for (i=1;i<=(pA->Naa+pB->Naa);++i) fprintf(fp,"%c",cseq[i]); 
 fclose(fp);
 free(cseq);
 
 return(1);

} /* end of Write_JavaView_Color_Alignment() */



int Write_Alignment_CASP_Style(pA,pB,a,fname,RnumPlainType,Ga,Gb,Rmat)
  struct PROTEIN *pA,*pB;
  struct AlignNode *a;  
  char *fname;
  char RnumPlainType;  /* if 'T', output PlainRnum stype */
  double Ga[3],Gb[3],Rmat[3][3];  /* Parameters for Superimposing */
{ int max,i,j,end;
  FILE *fp;
  char A,B;
  char rA[8],rB[8];

  printf("#Write_Alignment_CASP_Style()-->\"%s\"\n",fname);
  fp = fopen(fname,"w");
  if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } 

  fprintf(fp,"NPRO 2\n");
  fprintf(fp,"PRO1 %s \n",pA->pname);
  fprintf(fp,"PRO2 %s \n",pB->pname);
  fprintf(fp,"COMMENT Naa1 %4d Naa2 %4d\n",pA->Naa,pB->Naa);  
  fprintf(fp,"COMMENT Ncomp %d SqID %.1f RMS %.3f DRMS %.3f\n",
          a->sim->Ncomp,a->sim->SeqID,a->sim->CRMS,a->sim->DRMS);
  fprintf(fp,"COMMENT ScDis %.1f Rdis %.1f\n",a->sim->ScDis,a->sim->Rdis);
  if (RnumPlainType=='T') fprintf(fp,"RNUMPLAIN\n");
  fprintf(fp,"PARAM_FOR_SUPERIMPOSING\n"); 
  fprintf(fp,"#Afit=R*(A-Ga)+Gb\n");
  fprintf(fp,"Ga %8.5lf %8.5lf %8.5lf\n",Ga[0],Ga[1],Ga[2]);
  fprintf(fp,"Gb %8.5lf %8.5lf %8.5lf\n",Gb[0],Gb[1],Gb[2]);
  fprintf(fp,"R0 %8.5lf %8.5lf %8.5lf\n",Rmat[0][0],Rmat[0][1],Rmat[0][2]);
  fprintf(fp,"R1 %8.5lf %8.5lf %8.5lf\n",Rmat[1][0],Rmat[1][1],Rmat[1][2]);
  fprintf(fp,"R2 %8.5lf %8.5lf %8.5lf\n",Rmat[2][0],Rmat[2][1],Rmat[2][2]);
  fprintf(fp,"ALIGNMENT\n");
  for (i=1;i<=a->Nali;++i)
  {
    if (a->nalA[i]>=0)
       { A = pA->Seq[a->nalA[i]]; 
         if (RnumPlainType=='T') sprintf(rA,"%d",a->nalA[i]);
                           else  sprintf(rA,"%s",pA->Rnum[a->nalA[i]]); } 
    else   {A = '-'; sprintf(rA,"-1");}

    if (a->nalB[i]>=0)
       { B = pB->Seq[a->nalB[i]]; 
         if (RnumPlainType=='T') sprintf(rB,"%d",a->nalB[i]);
                           else  sprintf(rB,"%s",pB->Rnum[a->nalB[i]]); } 
    else   {B = '-'; sprintf(rB,"-1");}

    fprintf(fp,"%c %s %c %s %3.0f\n",A,rA,B,rB,100*a->scores[i]);
   }
  fprintf(fp,"END\n");
  fclose(fp);
  
  return(1);

} /* end of Write_Alignment_CASP_style() */






int Output_Path_Horizontal(pA,pB,a,fname)
  struct PROTEIN *pA,*pB;
  struct AlignNode *a; 
  char *fname;  /* if fname = "" or fname = "-", output stdout */
{ int max, i,j;
  int end,hon,eon,Rend,Ncol;
  char *sameseq,sym;
  FILE *fp;

  Ncol = 60; 
  if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } }

  /** Set sameseq **/
  sameseq = (char *)malloc(sizeof(char)*(a->Nali+Ncol+1));
  Set_SameSeq(pA,pB,a,sameseq,Ncol);
 
  max = (int)ceil(a->Nali/(double)Ncol);
 
 /** OUTPUT **/

 for (i=0;i<max;++i)
  {
  
  /*** FOR PROTEIN A ***/
  fprintf(fp,"     :"); 
   PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->SseSeq,&Rend);
  fprintf(fp,"\n");
 
  fprintf(fp,"SecA :");
   PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->Sec,&Rend);
  fprintf(fp,"\n");

 /*
  fprintf(fp,"ExBuA:");
   PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->Exp,&Rend);
  fprintf(fp,"\n");
 */
 
  if (a->nalA[Ncol*i+1]>0) fprintf(fp,"%5s:",pA->Rnum[a->nalA[Ncol*i+1]]);
  else                   fprintf(fp,"     :");
  
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->Seq,&Rend);
  if (Rend >=1) fprintf(fp,":%s\n",pA->Rnum[Rend]); else fprintf(fp,":\n");

  fprintf(fp,"      ");
  for (j=1;j<=Ncol;++j)  if ((Ncol*i+j)<=a->Nali)fprintf(fp,"%c",sameseq[Ncol*i+j]);
  fprintf(fp,"\n");

  /*** FOR PROTEIN B ***/
  if (a->nalB[Ncol*i+1]>=1) fprintf(fp,"%5s:",pB->Rnum[a->nalB[Ncol*i+1]]);
  else                    fprintf(fp,"     :");

  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->Seq,&Rend);
  if (Rend >=1) fprintf(fp,":%s\n",pB->Rnum[Rend]); else fprintf(fp,":\n");

  fprintf(fp,"SecB :");
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->Sec,&Rend);
  fprintf(fp,"\n");

/*
  fprintf(fp,"ExBuB:");
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->Exp,&Rend);
  fprintf(fp,"\n");
*/ 
  fprintf(fp,"     :");
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->SseSeq,&Rend);
  fprintf(fp,"\n");

  fprintf(fp,"\n");
 }

 if (fp!=stdout) fclose(fp);
 free(sameseq);
 return(1);
} /* end of Output_Path_Horaizontal() */




int Output_Path_Horizontal_ClustalW(pA,pB,a,fname)
  struct PROTEIN *pA,*pB;
  struct AlignNode *a; 
  char *fname;  /* if fname = "" or fname = "-", output stdout */
{ int max, i,j;
  int end,hon,eon,Rend;
  int Ncol; 
  char *sameseq,sym;
  FILE *fp;
  
  Ncol = 60; 
  if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } }

  fprintf(fp,"CLUSTAL W (1.82) multiple sequence alignment\n");
  fprintf(fp,"\n\n");

  /** Set sameseq **/
  sameseq = (char *)malloc(sizeof(char)*(a->Nali+Ncol+1));
  Set_SameSeq(pA,pB,a,sameseq,Ncol);
 
  max = (int)ceil(a->Nali/(double)Ncol);
 
 /** OUTPUT **/

 for (i=0;i<max;++i)
  {
  /*** FOR PROTEIN A ***/
  fprintf(fp,"%-16s",pA->cname); 
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->Seq,&Rend);
  fprintf(fp,"\n");
 
  /*** FOR PROTEIN B ***/
  fprintf(fp,"%-16s",pB->cname); 
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->Seq,&Rend);
  fprintf(fp,"\n");

  fprintf(fp,"%-16s"," ");
  for (j=1;j<=Ncol;++j)  if ((Ncol*i+j)<=a->Nali) fprintf(fp,"%c",sameseq[Ncol*i+j]);
  fprintf(fp,"\n\n");

 } /* i */

 if (fp!=stdout) fclose(fp);
 free(sameseq);
 return(1);

} /* end of Output_Path_Horaizontal_ClustalW() */









int Output_Path_Horizontal_HTMLcolor(pA,pB,a,fname)
  struct PROTEIN *pA,*pB;
  struct AlignNode *a;   
  char *fname;  /* if fname = "" or fname = "-", output stdout */
{ int max, i,j;
  int end,hon,eon,Rend,Ncol;
  char *sameseq,sym;
  FILE *fp;

  Ncol = 60; 
  if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } }

  fprintf(fp,"[BEGIN ALIGNMENT]\n");
  /** Set sameseq **/
  sameseq = (char *)malloc(sizeof(char)*(a->Nali+Ncol+1));
  Set_SameSeq(pA,pB,a,sameseq,Ncol);

  max = (int)ceil(a->Nali/(double)Ncol);
 
 /** OUTPUT **/

 for (i=0;i<max;++i)
  {
    /*** FOR PROTEIN A ***/
    fprintf(fp,"     :"); PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->SseSeq,&Rend);
    fprintf(fp,"\n");
  
    fprintf(fp,"SecA :");
    PrintAlignedString_Sec(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->Sec);
    fprintf(fp,"\n");
 
  if (a->nalA[Ncol*i+1]>0) fprintf(fp,"%5s:",pA->Rnum[a->nalA[Ncol*i+1]]);
  else                fprintf(fp,"     :");
  
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalA,pA->Naa,pA->Seq,&Rend);
  if (Rend >=1) fprintf(fp,":%s\n",pA->Rnum[Rend]); else fprintf(fp,":\n");

  fprintf(fp,"      ");
  for (j=1;j<=Ncol;++j)  if ((Ncol*i+j)<=a->Nali)fprintf(fp,"%c",sameseq[Ncol*i+j]);
  fprintf(fp,"\n");

  /*** FOR PROTEIN B ***/
  if (a->nalB[Ncol*i+1]>=1) fprintf(fp,"%5s:",pB->Rnum[a->nalB[Ncol*i+1]]);
  else                 fprintf(fp,"     :");

  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->Seq,&Rend);
  if (Rend >=1) fprintf(fp,":%s\n",pB->Rnum[Rend]); else fprintf(fp,":\n");
 
  fprintf(fp,"SecB :");
  PrintAlignedString_Sec(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->Sec);
  fprintf(fp,"\n");
   
  fprintf(fp,"     :");
  PrintAlignedString(fp,Ncol,i,a->Nali,a->nalB,pB->Naa,pB->SseSeq,&Rend);
  fprintf(fp,"\n");

  fprintf(fp,"\n");
 }

 fprintf(fp,"[END ALIGNMENT]\n");

 if (fp!=stdout) fclose(fp);
 free(sameseq);
 return(1);
} /* end of Output_Path_Horaizontal_HTMLcolor() */



 


int Write_Rasmol_Color_Alignment(fname,pA,pB,a,ColorType,ColorNumType)
 char *fname;
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
 char   ColorType; /* 'G'ray, 'R'ainbow */
 char   ColorNumType; /* 'C':Ncomp, 'A':Rnum of proA, 'B' Rnum of proB 'a'lignment_number */
{
 int i,j,k,Nsta,NstaA,NstaB,Nseg;
 int Ncomp,ncomp;
 char type; /* 'F'irst, 'C'ontinue, change 'L'ine */  
 FILE *fp;
 signed char *colseq;
 int  Ncolor; 
 static char *rainbowcolor[] = {"blue","cyan","green","yellow","orange","redorange","red"};
 static char *graycolor[] = {"[64,64,64]","[96,96,96]","[128,128,128]","[160,160,160]"};


 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#WARNING:Can't write to hori-alifile \"%s\"\n",fname); return(1); } 
 
 /** SET COLSEQ (if a is aligned, colseq[a] = 0..7,  otherwise coseq[a] = -1 **/ 
 if (ColorType=='G') Ncolor = 4; else Ncolor = 7;

 colseq = (signed char *)malloc(sizeof(signed char)*(a->Nali+2));
 Ncomp = 0;
 for (i=1;i<=a->Nali;++i) if ((a->nalA[i]>0)&&(a->nalB[i]>0)) ++Ncomp;

 ncomp = 0;
 for (i=1;i<=a->Nali;++i) 
  { if ((a->nalA[i]>0)&&(a->nalB[i]>0)) 
      { 
        if (ColorNumType == 'C') colseq[i] = (int)floor(Ncolor*ncomp/Ncomp); 
        if (ColorNumType == 'a') colseq[i] = (int)floor(Ncolor*(i-1)/a->Nali); 
        if (ColorNumType == 'A') colseq[i] = (int)floor(Ncolor*(a->nalA[i]-1)/pA->Naa); 
        if (ColorNumType == 'B') colseq[i] = (int)floor(Ncolor*(a->nalB[i]-1)/pB->Naa); 
        ++ncomp; }
    else colseq[i] = -1; }

 colseq[a->Nali+1] = -1;

 /*** OUTPUT HEADER  ***/
 fprintf(fp,"# COMMAND   '%s'\n",PAR.COMMAND); 
 fprintf(fp,"# DATE      '%s'\n",Get_Date_String()); 
 fprintf(fp,"# Filename  '%s'\n",fname);
 /*** OUTPUT ECHO HEADER  ***/
 fprintf(fp,"echo \"*** Pairwise 3D Alignment using Matras %s : %s vs %s  ***\"\n",
            Version,pA->cname,pB->cname);
 fprintf(fp,"echo \"  chain 'A' -> %s,  chain 'B' -> %s\"\n",pA->cname,pB->cname);
 fprintf(fp,"echo \"  alignment range : '%s -> %s' vs '%s -> %s'\"\n",
   pA->Rnum[a->staA],pA->Rnum[a->endA],pB->Rnum[a->staB],pB->Rnum[a->endB]);
 fprintf(fp,"echo \"  Ncomp %d CRMS %f\"\n",a->sim->Ncomp,a->sim->CRMS);
 
 fprintf(fp,"echo \"< SPECIAL SET TO BE SELECTED >\"\n");
 fprintf(fp,"echo \" select ALIGN     : select an aligned region\"\n");
 fprintf(fp,"echo \" select IDENTICAL : select identical sites in the aligned region\"\n");


 /*** DEFINE ALIGN REGION ***/

 fprintf(fp,"#DEFINE ALIGN REGION\n");
 NstaA = NstaB = -1; type = 'F'; Nseg = 0;
 for (i=1;i<=a->Nali;++i)
 { 
    if ((NstaA==-1)&&(colseq[i]>=0)) {NstaA = a->nalA[i]; NstaB = a->nalB[i];}
    if ((NstaA>= 0)&&(colseq[i+1]<0)) 
    { 
            if (type=='F') { fprintf(fp,"define ALIGN (");        type = 'C';}
       else if (type=='L') { fprintf(fp,"define ALIGN ALIGN||("); type = 'C';}
       else fprintf(fp,","); 
      
       if (NstaA== a->nalA[i]) fprintf(fp,"%d:A,",atoi(pA->Rnum[NstaA])); 
       else fprintf(fp,"%d-%d:A,",atoi(pA->Rnum[NstaA]),atoi(pA->Rnum[a->nalA[i]])); 
   
       if (NstaB == a->nalB[i]) fprintf(fp,"%d:B",atoi(pB->Rnum[NstaB])); 
       else fprintf(fp,"%d-%d:B",atoi(pB->Rnum[NstaB]),atoi(pB->Rnum[a->nalB[i]])); 
       NstaA = -1;
      
       ++Nseg;
       if (Nseg > 5) { fprintf(fp,")\n"); Nseg = 0; type = 'L'; } 
     }
 }
 
 if (type == 'C') fprintf(fp,")\n"); 


 /*** DEFINE IDENTICAL REGION ***/

 fprintf(fp,"#DEFINE IDENTICAL REGION\n");
 NstaA = -1; type = 'F'; Nseg = 0;
 for (i=1;i<=a->Nali;++i)
 { 
    if ((NstaA==-1)&&(colseq[i]>=0)&&(pA->Seq[a->nalA[i]]==pB->Seq[a->nalB[i]])) 
      { NstaA = a->nalA[i]; NstaB = a->nalB[i];}
    if ((NstaA>= 0)&&((colseq[i+1]<0)||(pA->Seq[a->nalA[i+1]]!=pB->Seq[a->nalB[i+1]])) ) 
     { 
            if (type == 'F') {fprintf(fp,"define IDENTICAL (");  type = 'C';} 
       else if (type == 'L') {fprintf(fp,"define IDENTICAL IDENTICAL || (");  type = 'C';} 
       else fprintf(fp,","); 
       
       if (NstaA==a->nalA[i]) fprintf(fp,"%d:A,",atoi(pA->Rnum[NstaA])); 
       else fprintf(fp,"%d-%d:A,",atoi(pA->Rnum[NstaA]),atoi(pA->Rnum[a->nalA[i]])); 
       if (NstaB==a->nalB[i]) fprintf(fp,"%d:B",atoi(pB->Rnum[NstaB])); 
       else fprintf(fp,"%d-%d:B" ,atoi(pB->Rnum[NstaB]),atoi(pB->Rnum[a->nalB[i]]));  
       
       NstaA = -1;

      ++Nseg;
   
      if (Nseg > 5) 
       { fprintf(fp,")\n"); Nseg = 0; type = 'L'; } 
     }
 }

 if (type=='C') fprintf(fp,")\n"); 


 /** Header **/
 fprintf(fp,"#DEFAULT DISPLAY MODE\n");
 fprintf(fp,"select\n");
 fprintf(fp,"wireframe false\n");
 fprintf(fp,"spacefill false\n");
 fprintf(fp,"backbone 50\n");

 fprintf(fp,"#COLORING ALIGNED REGION\n");
 if (ColorType == 'G')
 { fprintf(fp,"select :A\ncolor white\n");
   fprintf(fp,"select :B\ncolor white\n"); 
   fprintf(fp,"background white\n"); 
  }
 else
 { fprintf(fp,"select :A\ncolor white\n");
   fprintf(fp,"select :B\ncolor gray\n"); }
 NstaA = -1; 
 for (i=1;i<=a->Nali;++i)
 {
    if ((NstaA==-1)&&(colseq[i]>=0)) {NstaA = a->nalA[i]; NstaB = a->nalB[i];}

    if ((NstaA >=1)&&(colseq[i]!=colseq[i+1])) 
     { fprintf(fp,"select %d-%d:A,%d-%d:B\n",
                   atoi(pA->Rnum[NstaA]),atoi(pA->Rnum[a->nalA[i]]),
                   atoi(pB->Rnum[NstaB]),atoi(pB->Rnum[a->nalB[i]]));
       if (ColorType == 'G') fprintf(fp,"color %s\n",graycolor[colseq[i]]); 
                        else fprintf(fp,"color %s\n",rainbowcolor[colseq[i]]); 
      NstaA = -1; }
  }

 fprintf(fp,"select all\n");
 fclose(fp);
 free(colseq);
 
 return(1);

} /* end of Write_Rasmol_Color_Alignment() */







int Write_Rasmol_Color_Alignment_Each_Protein(fname,pA,pB,OutProteinType,a,Ntype,ColorType)
 char  *fname;
 struct PROTEIN *pA,*pB;
 char OutProteinType;  /* 'A' output pA, 'B' output pB */
 struct AlignNode *a;
 char  Ntype;     /* Coloring Number:  'A':nalA, 'B':nalB,  'C':Ncomp, 'a':alignment number */
 char  ColorType; /* 'R'ainbow, 'G'ray */
{
 int i,j,k,beg,Nsta,Nend,nsta,mi,Nall,t,Ncomp;
 float T;
 FILE *fp;
 struct PROTEIN *pO; /* Protein to be outputted */

 int Ncolor; 
 char color[10][32];
 static char *rainbowcolor[] = {"blue","cyan","green","yellow","orange","redorange","red"};
 static char *graycolor[] = {"[32,32,32]","[64,64,64]","[96,96,96]",
                             "[128,128,128]","[160,160,160]","[192,192,192]"};

      if (OutProteinType == 'A') pO = pA;
 else if (OutProteinType == 'B') pO = pB;
 else pO = pA;
                                                                                
 if (ColorType=='G') 
 { Ncolor = 5;
   for (i=0;i<Ncolor;++i) sprintf(color[i],"%s",graycolor[i]); } 
 else 
 { Ncolor = 7;
   for (i=0;i<Ncolor;++i) sprintf(color[i],"%s",rainbowcolor[i]); } 

 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#WARNING:Can't write to \"%s\"\n",fname); return(1); } 
 fprintf(fp,"# COMMAND   %s\n",PAR.COMMAND); 
 fprintf(fp,"# DATE      %s\n",Get_Date_String()); 
 fprintf(fp,"# Filename  %s\n",fname);
 fprintf(fp,"# OutProteinType %c\n",OutProteinType);
 fprintf(fp,"# Ntype     %c\n",Ntype);
 fprintf(fp,"# ColorType %c\n",ColorType);
 
 fprintf(fp,"select all\n");
 fprintf(fp,"color white\n");
 fprintf(fp,"wireframe false\n");
 fprintf(fp,"background white\n");
 if (OutProteinType=='A') 
  fprintf(fp,"select *:A\nbackbone 100\nselect *:B\nbackbone false\n");
 if (OutProteinType=='B') 
  fprintf(fp,"select *:B\nbackbone 100\nselect *:A\nbackbone false\n");

 Nsta = -1;
 Ncomp = Ncompare(a); 
 fprintf(fp,"echo \"*** Pairwise 3D Alignment using Matras %s : %s vs %s  ***\"\n",
            Version,pA->cname,pB->cname);
 if (OutProteinType=='A') fprintf(fp,"echo \"  show only chain 'A' -> %s\"\n",pA->cname);
 if (OutProteinType=='B') fprintf(fp,"echo \"  show only chain 'B' -> %s\"\n",pB->cname);
 fprintf(fp,"echo \" color %c coloring_number %c\"\n",ColorType,Ntype);

 t = 0; T = 0.0;
 for (i=1;i<=a->Nali;++i)
 {
  if ((a->nalA[i]>=1)&&(a->nalB[i]>=1))
  {
   if (OutProteinType=='A') fprintf(fp,"select %d:A\n",atoi(pA->Rnum[a->nalA[i]]));
   if (OutProteinType=='B') fprintf(fp,"select %d:B\n",atoi(pB->Rnum[a->nalB[i]]));
   if (Ntype=='A') T = (float)(a->nalA[i]-1)/(float)pA->Naa; 
   if (Ntype=='B') T = (float)(a->nalB[i]-1)/(float)pB->Naa; 
   if (Ntype=='C') T = (float)t/(float)Ncomp; 
   if (Ntype=='a') T = (float)(i-1)/(float)a->Nali; 
   fprintf(fp,"color %s\n",color[(int)floor(Ncolor*T)]); 
   ++t;
   } 
  } /* i */

 fprintf(fp,"select all\n");
 fclose(fp);
 
 return(1); 
} /* end of  Write_Rasmol_Color_Alignment_Each_Protein() */





void Short_SSE_Extention(pr)
 struct PROTEIN *pr;
{
 /* 
  This is for Molscript.
  Molscript can't draw 2-residue strands.
 */
 int i,j,Npre,Npos; 
 
 for (i=0;i<pr->Nsse;++i)
 if (pr->SSE[i].L < 3) 
 {
  Npre = pr->SSE[i].Rsta - 1;
  Npos = pr->SSE[i].Rend + 1;
  if ((Npre>=1)&&(pr->Sec[Npre]!='H')&&(pr->Sec[Npre]!='E')) 
   { pr->Sec[Npre] = pr->SSE[i].type;
     pr->SSE[i].Rsta = Npre; 
     pr->SSE[i].L = pr->SSE[i].Rend - pr->SSE[i].Rsta +1; }
  else if ((Npos<=pr->Naa)&&(pr->Sec[Npos]!='H')&&(pr->Sec[Npos]!='E')) 
   { pr->Sec[Npos] = pr->SSE[i].type;
     pr->SSE[i].Rend = Npos; 
     pr->SSE[i].L = pr->SSE[i].Rend - pr->SSE[i].Rsta +1; }
  }

} /* end of Short_SSE_Extention() */


void Short_SSE_Deletion(pr)
 struct PROTEIN *pr;
{
 /* 
  This is for Molscript.
  Molscript can't draw 2-residue strands.
 */
 int i,j,Ndel; 
 struct SecStrEle  *tmpSSE; 

 tmpSSE = (struct SecStrEle*)malloc(sizeof(struct SecStrEle)*(pr->Nsse+1));

 /* Check deletion */
 Ndel = 0; 
 for (i=0;i<pr->Nsse;++i)
 {
   tmpSSE[i].Rsta = pr->SSE[i].Rsta;   
   tmpSSE[i].Rend = pr->SSE[i].Rend;   
   tmpSSE[i].L    = pr->SSE[i].L;
   tmpSSE[i].type = pr->SSE[i].type;

   if (pr->SSE[i].L < 3) 
   {
    for (j=pr->SSE[i].Rsta;j<= pr->SSE[i].Rend;++j) pr->Sec[j] = ' ';
    tmpSSE[i].type = 'd'; 
    ++Ndel; 
   }
 }

 /* Delete SSEs to be deleted */
 i = 0; j = 0;
 while (i<pr->Nsse)
 {
  if (tmpSSE[i].type != 'd')
  {
   pr->SSE[j].Rsta = tmpSSE[i].Rsta;   
   pr->SSE[j].Rend = tmpSSE[i].Rend;   
   pr->SSE[j].L    = tmpSSE[i].L;
   pr->SSE[j].type = tmpSSE[i].type;
   ++j; 
  }
  ++i;
 }

 free(tmpSSE);
 pr->Nsse = j;

} /* end of Short_SSE_Deletion() */





int Write_Molscript_Alignment_Each_Protein(fname,pA,pB,OutProteinType,a,Chain,Ntype,ColorType)
 char *fname;
 struct PROTEIN *pA,*pB;
 char OutProteinType;  /* 'A' output pA, 'B' output pB */
 struct AlignNode *a; 
 char Chain;
 char Ntype; /* A : alignment number colour B : nalB number color  a:alignment number */
 char ColorType; /* 'R'ainbow, 'G'ray */
{
 int i,j,k,l,hit,pos,Nhit,Ncolall,c,Ncomp;
 double t,tt,v,s,e;
 FILE *fp;
 char ch[2],rsta[8],rend[8];
 struct PROTEIN *pO; /* Protein to be outputted */


 /*
 Short_SSE_Extention(pA); 
 Short_SSE_Extention(pB); 
 */
 Short_SSE_Deletion(pA); 
 Short_SSE_Deletion(pB); 

      if (OutProteinType == 'A') pO = pA;
 else if (OutProteinType == 'B') pO = pB;
 else pO = pA;

 if (Chain == '-') {ch[0] ='\0';} else {ch[0] = Chain; ch[1] = '\0';}

 s = 0.7; e = 0.8;
 
 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#ERROR:Can't write to molscriptfile \"%s\"\n",fname); 
                return(0);}

 fprintf(fp,"! COMMAND   %s\n",PAR.COMMAND); 
 fprintf(fp,"! DATE      %s\n",Get_Date_String()); 
 fprintf(fp,"! Filename  %s\n",fname);
 fprintf(fp,"! Chain     %c\n",Chain);
 fprintf(fp,"! Ntype     %c\n",Ntype);
 fprintf(fp,"! ColorType %c\n",ColorType);
 fprintf(fp,"plot\n noframe\n\n");
 fprintf(fp,"read mol \"1.pdb\";\n");;
 fprintf(fp," transform atom *\n");
 fprintf(fp,"    by centre position atom *\n");
 fprintf(fp,"    by rotation x 180.0;\n\n");

 Ncomp = Ncompare(a); c = 0;
 /***** Write Color Information ****/
 for (i=1;i<=a->Nali;++i)
 if ((a->nalA[i]>0)&&(a->nalB[i]>0)) 
 {
  if (Ntype=='A') t = (double)a->nalA[i]/pA->Naa;
  if (Ntype=='B') t = (double)a->nalB[i]/pB->Naa;
  if (Ntype=='C') t = (double)c/Ncomp;
  if (Ntype=='a') t = (double)(i-1)/a->Nali;
  ++c;
  v = (s - (1-e+s)*t);
  if (v<0.0) v = - (1-e+s)*(t-1)+e;
  /* printf("a->nalA %d a->nalB %d'%s'\n",a->nalA[i],a->nalB[i],pA->Rnum[a->nalA[i]]); */

       if (OutProteinType == 'A') Remove_Space(rsta,pA->Rnum[a->nalA[i]]);
  else if (OutProteinType == 'B') Remove_Space(rsta,pB->Rnum[a->nalB[i]]);
 
  fprintf(fp,"set colourparts on,\n");
  if (ColorType == 'G') 
  { 
     tt =  t * 0.8 + 0.2;  
    /* tt = t * 0.8 + 0.2;  */
    fprintf(fp,"residuecolour residue %s%s rgb %.2lf %.2lf %.2lf;\n",ch,rsta,tt,tt,tt); }
  else 
   fprintf(fp,"residuecolour residue %s%s hsb %.2lf 1.0 1.0;\n",ch,rsta,v); 
 }
 /***** WRITE SSE INFORMATION  *****/ 
 Remove_Space(rsta,pO->Rnum[1]);
 Remove_Space(rend,pO->Rnum[pO->SSE[0].Rsta]);
 fprintf(fp,"  coil from %s%s to %s%s;\n",ch,rsta,ch,rend);

 for (i=0;i<pO->Nsse;++i)
 {
   /*********** SSE ***********/
   Remove_Space(rsta,pO->Rnum[pO->SSE[i].Rsta]);
   Remove_Space(rend,pO->Rnum[pO->SSE[i].Rend]);
   if (pO->SSE[i].type=='H') 
    { fprintf(fp,"  helix from %s%s to %s%s;\n",ch,rsta,ch,rend); }
   if (pO->SSE[i].type=='E') 
    { fprintf(fp,"  strand from %s%s to %s%s;\n",ch,rsta,ch,rend); }

   /*********** COIL next to SSE ***********/
   if (i<(pO->Nsse-1)) 
   { Remove_Space(rsta,pO->Rnum[pO->SSE[i].Rend]);
     Remove_Space(rend,pO->Rnum[pO->SSE[i+1].Rsta]);
     fprintf(fp,"  coil from %s%s to %s%s;\n",ch,rsta,ch,rend); }

  }  /* i loop */

  Remove_Space(rsta,pO->Rnum[pO->SSE[pA->Nsse-1].Rend]);
  Remove_Space(rend,pO->Rnum[pO->Naa]);
  fprintf(fp,"  coil from %s%s to %s%s;\n",ch,rsta,ch,rend);

  /** For ligands, for example "FMN" */
  fprintf(fp,"set colourparts on,\n");
  fprintf(fp,"atomcolour in type FMN rgb 1.0 1.0 1.0;\n");
  fprintf(fp,"ball-and-stick in type FMN;\n");

  fprintf(fp,"set colourparts on,\n");
  fprintf(fp,"atomcolour in type FAD rgb 1.0 1.0 1.0;\n");
  fprintf(fp,"ball-and-stick in type FAD;\n");
  
  fprintf(fp,"end_plot\n");
  fclose(fp);

  return(1);

} /* end of Write_Molscript_Alignment_Each_Protein() */



int Find_nal(q,Nali,nalA)
 int Nali,*nalA;
{
 int i,hit,pos;

 i = 1; pos = -1; 
 while ((pos==-1)&&(i<=Nali))
 {
  if (nalA[i]==q) { pos = i; }
   ++i;  
 }
 return(pos);

} /* end of Find_nal() */




void PrintAlignedString(fp,Ncol,i,Nali,nala,Naa,seq,rend)
 FILE *fp;
 int Ncol;  /* Number of Column */
 int i;     /* Row Number */
 int Nali,*nala,Naa;
 char *seq;
 int *rend; 
{ 
 int j,anum;
 
 *rend = 0;
 for (j=1;j<=Ncol;++j)
 {
  anum = Ncol*i+j;
  if (anum<=Nali)
  {    
        if (nala[anum]<0)   fprintf(fp,"-"); 
   else if (Naa<nala[anum]) fprintf(fp," ");
   else { fprintf(fp,"%c",seq[nala[anum]]);  *rend = nala[anum];}
   }
 }
 
} /* end of PrintAlignedString() */ 






void PrintAlignedString_Sec(fp,Ncol,i,Nali,nala,Naa,seq)
 FILE *fp;
 int Ncol;  /* Number of Column */
 int i;     /* Row Number */
 int Nali,*nala,Naa;
 char *seq;
{
 int j,anum,hon,eon;
 char sym;
 
 hon = eon = 0; 
 for (j=1;j<=Ncol;++j)
 {
  anum = Ncol*i+j;
  if (anum<=Nali)
  {    
        if (nala[anum]<0)   sym = '-'; 
   else if (Naa<nala[anum]) sym = ' ';
   else sym = seq[nala[anum]];
   
   if ((hon==1)&&(sym != 'H')) {hon = 0; fprintf(fp,"</FONT>"); }
   if ((eon==1)&&(sym != 'E')) {eon = 0; fprintf(fp,"</FONT>"); }
   if ((hon==0)&&(sym == 'H')) {hon = 1; fprintf(fp,"<FONT COLOR=\"#A00020\">"); }
   if ((eon==0)&&(sym == 'E')) {eon = 1; fprintf(fp,"<FONT COLOR=\"#00A020\">"); }

   fprintf(fp,"%c",sym); 
  }
 } 
 
if ((hon==1)||(eon==1)) {hon = eon = 0; fprintf(fp,"</FONT>"); }
 
} /* end of PrintAlignedString_Sec() */ 




int Output_Alignment_Information(fname,pA,pB,a)
 char *fname;
 struct PROTEIN *pA,*pB;
 struct AlignNode *a;
{
 FILE *fp;
 char ln[50];
   /*** FILE OPEN ***/ 
  if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to inffile \"%s\"\n",fname); return(1); } }

 Get_Part_Of_Line(ln,pA->compnd,0,40);
 fprintf(fp,"[ALIGN_RANK]  %d\n",a->num);
 fprintf(fp,"[PROTEIN A]   %-8s Naa %3d Nsse %2d \"%s\"\n",pA->cname,pA->Naa,pA->Nsse,ln);
 Get_Part_Of_Line(ln,pB->compnd,0,40);
 fprintf(fp,"[PROTEIN B]   %-8s Naa %3d Nsse %2d \"%s\"\n",pB->cname,pB->Naa,pB->Nsse,ln);
 fprintf(fp,"[ALIGNMENT]   Ncomp_aa %d Ncomp_sse %d\n",a->sim->Ncomp,a->sim->Ncomp_sse);
 fprintf(fp,"[SIMILARITY]  Seq %.1f %%  Sec %.1f %% Exp %.1f %% CRMS %.2f A  DRMS %.2f A\n",
               a->sim->SeqID,a->sim->SecID,a->sim->ExpID,a->sim->CRMS,a->sim->DRMS);
 fprintf(fp,"[SCORE]       ScSSE %.1f ScEnv %.1lf ScDis %.1f Rdis %.1f (%%) Rsse %.1f (%%)\n",
                          a->sim->ScSSE,a->sim->ScEnv,a->sim->ScDis,a->sim->Rdis,a->sim->Rsse);
 fprintf(fp,"[RELIABILITY] Superfamily %5.1f %%  Fold %5.1f %%\n",a->sim->RelS,a->sim->RelO);
 if (fp!= stdout) fclose(fp);

} /* end of Output_Alingmnet_Informatoin() */
  



int Output_Oneline_Information(fname,pA,pB,an)
 char *fname;
 struct PROTEIN *pA,*pB;
 struct AlignNode *an;
{
 FILE *fp;
 char ln[50];
 
 if ((fname[0]=='\0')||(fname[0]=='-')) fp = stdout;
  else 
   { fp = fopen(fname,"w");  
     if (fp==NULL) {printf("#WARNING:Can't write to inffile \"%s\"\n",fname); return(1); } }

 fprintf(fp,"#[proA] [proB] [NaaA] [NaaB] [Ncomp] [ScSSE] [ScEnv] [ScDis] [SqID] [CRMS] [Rdis(ave)]\n");
 fprintf(fp,"%5s %5s %4d %4d %4d %.1lf %.1lf %.1lf %5.1lf %5.2lf %5.2lf <<\n",
        pA->cname, pB->cname, pA->Naa, pB->Naa, an->sim->Ncomp,
        an->sim->ScSSE,an->sim->ScEnv,an->sim->ScDis, an->sim->SeqID, an->sim->CRMS,an->sim->Rdis);

 if (fp!= stdout) fclose(fp);

} /* end of Output_Oneline_Informatoin() */




void Set_SameSeq(pA,pB,an,sameseq,Ncol)
 struct PROTEIN *pA,*pB;
 struct AlignNode *an;
 char *sameseq;
 int Ncol;
{
  int i,max;
 
 max = an->Nali+Ncol+1;
 for (i=0;i<max;++i) sameseq[i] = ' ';
 
  for (i=1;i<= an->Nali;++i)
   if ((an->nalA[i]>0)&&(an->nalB[i]>0))
    if (numAA[pA->Seq[an->nalA[i]]]==numAA[pB->Seq[an->nalB[i]]]) sameseq[i] = '*';

} /* end of Set_SameSeq() */
