/*

 <OutLibL.c>

 for reading and allocation linked-list structure library
 
 coded by Takeshi Kawabata.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h" 
#include "protein.h" 
#include "StrLibL.h" 
#include "Amino.h" 
#include "LlistAlign.h" 
#include "ScoreDIS.h" 


/*** FUNCTIONS (GLOBAL) ***/ 
void Output_Library_with_Compnd();
void Output_Library_Detailed();
void Output_Library_Sec_Symbol();
void Output_Library_ClustalW();
void Cal_NstaNend_CompTime();
void Cal_MaxNaa_from_NstaNend();

/*** FUNCTIONS (LOCAL) ***/ 
static void Dotted_Truncate_String();
static void Symbol_Out_Sec_Line();
static void Symbol_Out_Line_Grid();


void Output_Library_with_Compnd(head,Nbest,type)
 struct AlignNode *head; 
 int Nbest;
 char type;
{
 int p;
 char ln1[60];
 struct AlignNode *bn;

 printf("[BEST_SCORE_RANKING]\n");
   printf("%-3s %-6s %-6s %-6s %-5s %-5s %-5s  %-50s\n",
   "rk","entry","start","end","Rdis","Zsc","SqID","MOLECULAR NAME");

 p = 0;
 bn = head;
 while ((bn->next != NULL) && (p<Nbest))
  {
    bn = bn->next;
    Dotted_Truncate_String(ln1,50,bn->pro->compnd);
     printf("%-3d %-6s %6s %6s %5.1f %6.2f %5.1f \"%s\"\n",
       p+1,bn->pro->cname,bn->pro->Rnum[bn->staB],bn->pro->Rnum[bn->endB],
      bn->sim->Rdis,bn->Zsc,bn->sim->SeqID,ln1);
    ++p;
 }

 printf("\n");

} /* end of Output_Library_with_Compnd() */



void Output_Library_Detailed(head,Nbest,RankType)
 struct AlignNode *head; 
 int Nbest;
 char RankType;
{
 int p;
 char ln[60];
 struct AlignNode *bn;

 printf("[BEST_SCORE_RANKING_WITH_DETAILED_INFORMATION]\n");
   printf("%-3s %-6s %-4s %-4s %-5s %-4s %-5s %-6s %-5s %-5s %-5s %-5s %-5s %-20s\n",
   "rk","entry","Naa","Ncmp","SqID","rms","Ssse","Rsse","Sdis","Rdis","Zsc","RelS","RelO","TAXONOMY");

 bn = head;
 p  = 0;
 while ((bn->next !=NULL) && (p<Nbest))
  {
   bn = bn->next; 
   Dotted_Truncate_String(ln,25,bn->sim->Tax);  
   if (RankType == 'D')
   { bn->sim->RelS = Reliability_Zsc(bn->Zsc,'S');
     bn->sim->RelO = Reliability_Zsc(bn->Zsc,'O'); }
   printf("%-3d %-6s %4d %4d %5.1lf %4.1f %5.0f %5.1f %6.0f %5.1f %6.2f %5.1f %5.1f [%s]\n",
            p+1,bn->pro->cname, 
            bn->pro->Naa, bn->sim->Ncomp, bn->sim->SeqID, bn->sim->CRMS,
            bn->sim->ScSSE,bn->sim->Rsse, bn->sim->ScDis,bn->sim->Rdis, 
            bn->Zsc,bn->sim->RelS,bn->sim->RelO,ln); 
   ++p;
 }

 printf("\n");

} /* end of Output_Library_Detailed() */





void Output_Library_Sec_Symbol(pr,head,Nbest)
 struct PROTEIN *pr;    /* Query Protein */
 struct AlignNode *head; 
 int Nbest;
{
 int i,j,k,p,a,b,Ncol;
 char *seq;
 struct AlignNode *bn;
 
 Ncol = 60;

 printf("[BEST_RANKING_WITH_ONE_LINE_SECONDARY_STRUCTURE]\n");
 printf("%-3s %-6s %-5s|%-5s|", "rk","entry","staQ","staL"); 
 Symbol_Out_Line_Grid(pr,pr->Naa,Ncol);
 printf("|%5s|%5s\n","endL","endQ");
 printf("%-3s(%-6s)%5s|%5s|", "","query",pr->Rnum[1],"---"); 
 Symbol_Out_Sec_Line(pr->Sec,pr->Naa,Ncol);
 printf("|%-5s|%-5s\n","---",pr->Rnum[pr->Naa]); 
 seq = (char *)malloc(sizeof(char)*(pr->Naa+1));

 bn = head;
 p = 0;
 while ((bn->next !=NULL)&&(p<Nbest)) 
 {
  bn = bn->next;  
  printf("%-3d %-6s %5s|%5s|",p+1,bn->sim->Name,pr->Rnum[bn->staA],bn->pro->Rnum[bn->staB]);

  for (i=1;i<=pr->Naa;++i) seq[i] = '-';
  
  for (k=1;k<=bn->Nali;++k)
  { 
    a = bn->nalA[k]; b = bn->nalB[k];
    if ((a>0)&&(b>0))    seq[a] = bn->pro->Sec[b];   
  } 
 
  Symbol_Out_Sec_Line(seq,pr->Naa,Ncol);
  printf("|%-5s|%-5s\n",bn->pro->Rnum[bn->endB],pr->Rnum[bn->endA]);
  ++p;

 } /* while */

 /* printf("done \n");  fflush(stdout); */
 free(seq);

} /* end of Output_Library_Sec_Symbol() */





void Output_Library_ClustalW(pr,head,Nbest,Otype)
 struct PROTEIN *pr;
 struct AlignNode *head; 
 int Nbest;
 char Otype; /* 'S':output secondary structure */
{
 int n,i,j,k,p,a,b,e,max;
 int **QtoL; 
 char ln[32],sameseq[61]; 
 int  AAcount[61][21],AAalign[61],Nmost,most_a;
 struct AlignNode *bn;

 /*** Preparing QtoL ***/
 QtoL = (int **)malloc(sizeof(int *) * Nbest);
 for (i=0;i<Nbest;++i)
  QtoL[i] = (int *)malloc(sizeof(int) * (pr->Naa+1));

 bn = head;
 p = 0;
 while ((bn->next != NULL) && (p<Nbest))
  { 
    bn = bn->next;
    for (a=0;a<=pr->Naa;++a) QtoL[p][a] = -1;

    for (i=1;i<=bn->Nali;++i)
    {  a = bn->nalA[i]; 
       b = bn->nalB[i];
       if (a>0) QtoL[p][a] = b; }
    ++p;
  }
     

/*** Output ***/
 printf("[CLUSTALW_STYLE_ALIGNMENT]\n\n");
 printf("CLUSTAL W (1.82) multiple sequence alignment\n");
 printf("\n\n");
 
 max = pr->Naa/60;  

 for (n=0;n<=max;++n)
 { 
  /* Output SecStr */
  if (Otype == 'S')
  {
   printf("%-16s","[SecStrNum]");
   for (k=1;k<=60;++k)
   { a = 60*n+k; 
     if (a<= pr ->Naa) printf("%c",pr->SseSeq[a]); }
   printf("\n");
   
   printf("%-10s%5s:","[SecStr]",pr->Rnum[60*n+1]);
   for (k=1;k<=60;++k)
   { a = 60*n+k; 
     if (a<= pr ->Naa) { e = a; printf("%c",pr->Sec[e]); } }
   printf(":%s\n",pr->Rnum[e]);
  } 

 /* Initialize AAcount */
  for (k=1;k<=60;++k) 
   { AAalign[k] = 0;
     for (a=0;a<20;++a) AAcount[k][a] = 0;  }

 /* Output Query */
  printf("%-16s","QUERY");
  for (k=1;k<=60;++k)
  { a = 60*n+k; 
    if (a<= pr ->Naa)
      { printf("%c",pr->Seq[a]); 
        ++AAcount[k][numAA[pr->Seq[a]]]; ++AAalign[k];}  
  }
  printf("\n");


  /* Output Library */
  bn = head;
  p = 0;
  while ((bn->next !=NULL) && (p<Nbest))
  {
   bn = bn->next; 
   Get_Part_Of_Line(ln,bn->sim->Name,0,16);
   printf("%-16s",ln);
   for (k=1;k<=60;++k)
   {
    a = 60*n+k; 
    if (a<= pr ->Naa)
    { 
     b = QtoL[p][a];
     if (b>0)  
     { printf("%c",bn->pro->Seq[b]); 
       ++AAcount[k][numAA[bn->pro->Seq[b]]]; ++AAalign[k];}  
     else printf("-"); 
     fflush(stdout);
    } 
   } /* k */
  printf("\n"); 
  ++p;
 
  } /* while (bn) */
 
  /* Set sameseq  */
  for (k=1;k<=60;++k)
  {
    Nmost = 0; most_a = 0;
    for (a=0;a<20;++a)
    { if (AAcount[k][a]>Nmost) { most_a = a; Nmost = AAcount[k][a]; } } 

          if (Nmost ==      AAalign[k]) sameseq[k] = '*'; 
     else if (Nmost >= 0.75*AAalign[k]) sameseq[k] = '+'; 
     else if (Nmost >= 0.5*AAalign[k])  sameseq[k] = '.'; 
     else sameseq[k] = ' ';
     
     if (AAalign[k] <= 0.5*(Nbest+1)) sameseq[k] = ' ';
    
     /*
     printf("k %d Nmost %d most_a %d AAalign %d %c\n",
      k,Nmost,most_a,AAalign[k],sameseq[k]);
     */  
  }

  printf("%-16s"," ");
  for (k=1;k<=60;++k) 
    if ((60*n+k)<= pr ->Naa) printf("%c",sameseq[k]);
  printf("\n\n");


 } /* p */


 for (i=0;i<Nbest;++i) free(QtoL[i]);
 free(QtoL);

} /* end of Output_Library_ClustalW() */






void Dotted_Truncate_String(outstr,Lout,instr)
 char *outstr,*instr;
 int Lout;
{
 /*
 if instr = "ABC" and Lout = 5,     then outstr = "ABC";
 if instr = "ABCDEFG" and Lout = 5, then outstr = "ABC..";
 */

 int Lin;
 Lin = strlen(instr);

 if (Lout >= Lin) sprintf(outstr,"%s",instr);
 else
 {
  Get_Part_Of_Line(outstr,instr,0,Lout-1);
  outstr[Lout-1] = '.';
  outstr[Lout-2] = '.';
  }

} /* end of Dotted_Truncate_String() */




void Symbol_Out_Sec_Line(seq,Nseq,Nline)
  char *seq;  /* from 1 to Nseq */
  int  Nseq;
  int  Nline;
{
 int i,j,sj,ej,H,E,C,G;
 double unit;
 char sym;

 unit = (double)Nseq/Nline; 
 for (i=1;i<=Nline;++i)
 {
   sj = unit * (i-1); 
   if (sj < 1) sj = 1;
   ej = unit * i; 
   if (ej > Nseq) ej = Nseq;
   H = E = C = G = 0;
   for (j=sj;j<ej;++j)
   {
          if (seq[j]=='H') ++H; 
     else if (seq[j]=='E') ++E; 
     else if (seq[j]=='-') ++G; 
     else  ++C;
    }
        if ((H>=E)&&(H>=C)&&(H>=G)) sym = 'H';
   else if ((E>=H)&&(E>=C)&&(E>=G)) sym = 'E';
   else if ((C>=H)&&(C>=E)&&(C>=G)) sym = 'c';
   else if ((G>=H)&&(G>=E)&&(G>=C)) sym = '-';
   else                             sym = '?';
   printf("%c",sym);
  } 

} /* end of Symbol_Out_Sec_Line() */





void Symbol_Out_Line_Grid(pr,Nseq,Nline)
  struct PROTEIN *pr;
  int  Nseq;
  int  Nline;
{
 int i,j,k,jj,sj,ej,L;
 double unit;
 char sym,s10,s50,s100,buff[100];
 int *numseq;

 /*** Make numseq from protein's RNUM string ***/ 
 numseq = (int *)malloc(sizeof(int)*(pr->Naa+1));
 for (i=1;i<=pr->Naa;++i)
 {
   L = strlen(pr->Rnum[i]);
   k = 0; 
   for (j=0;j<L;++j) 
   { sym = pr->Rnum[i][j]; 
     if ((48<= sym) && (sym <= 57)) { buff[k] = sym; ++k;}  }
   buff[k] = '\0';
   numseq[i] = atoi(buff); 
 }
 
 /*** Output grid  ***/ 
 
 unit = (double)Nseq/Nline; 
 for (i=0;i<Nline;++i)
 {
   sj = unit * i; 
   ej = unit * (i+1); 
   sym = s10 = s50 = s100 = ' ';
   for (j=sj;j<ej;++j)
   {
     jj = numseq[j+1];
     if ((jj%10)==0)  s10  = '.';
     if ((jj%50)==0)  s50  = ':';
     if ((jj%100)==0) s100 = '+';
    }
        if (s100 !=  ' ') sym = s100; 
   else if (s50 !=  ' ')  sym =  s50; 
   else if (s10 !=  ' ')  sym =  s10; 
  
    printf("%c",sym);
  } 

 free(numseq);

} /* end of Symbol_Out_Line_Grid() */






void Cal_NstaNend_CompTime(head,Nsta,Nend,bunshi,bunbo)
 struct LibNode *head;
 int *Nsta,*Nend;
 int bunshi,bunbo;
{
 double Ttotal,Taccu,Raccu,D,minDs,minDe,Vsta,Vend,inf;
 struct LibNode *ln,*mn;
 int k,Nlib,Ncomb;

 /***
  This is function for dividing job of all-vs-all comparison.
  It assumes that a computational costs depends on proA.Naa x proB.Naa.
  ***/

 if (bunshi >= bunbo)
 { printf("#ERROR  : bunshi %d >= bunbo %d\n",bunshi,bunbo);
   exit(1); }

 inf = 10000000000000000.0;
 
 /** Calulate Nlib,Ncomb,Ttotal **/

 ln = head; Nlib = Ncomb = 0; Ttotal = 0.0; 
 while (ln->next != NULL) 
 {
   mn = ln;
   ln = ln->next;
   ++Nlib;
   while (mn->next !=NULL)
   {
    mn = mn->next;
    Ttotal += (double)ln->pro->Naa * mn->pro->Naa;
    ++Ncomb;
   }
 }
 
 printf("#Nlib %d Ncomb %d\n",Nlib,Ncomb);

 
 /** Estimate Nsta and Nend **/
 minDs = minDe = inf;

 Vsta = (double)bunshi/(double)bunbo;
 Vend = (double)(bunshi+1.0)/(double)bunbo;

 ln = head; k = 0; Taccu = 0.0;
 while (ln->next != NULL) 
 {
   ln = ln->next;
   mn = ln;
   while (mn->next !=NULL)
   {
    mn = mn->next;
    Taccu += (double)ln->pro->Naa * mn->pro->Naa;
    Raccu = Taccu/Ttotal;
    
    D = (Raccu - Vsta)*(Raccu - Vsta);
    if (D < minDs) { *Nsta = k; minDs = D;}
    D = (Raccu - Vend)*(Raccu - Vend);
    if (D < minDe) { *Nend = k; minDe = D;}
    ++k;
  }
 }

 if (bunshi == 0)         *Nsta = 0;
 if (bunshi == (bunbo-1)) *Nend = Ncomb;

} /* end of Cal_NstaNend_CompTime() */


void Cal_MaxNaa_from_NstaNend(head,Nsta,Nend,MaxNaaA,MaxNaaB)
 struct LibNode *head;
 int Nsta,Nend;
 int *MaxNaaA,*MaxNaaB;
{
 struct LibNode *ln,*mn;
 int ncomb,rangein;

 ln = head; ncomb = 0; 
 *MaxNaaA = *MaxNaaB = 0;

 while (ln->next != NULL) 
 { 
   mn = ln;
   ln = ln->next;
   while (mn->next !=NULL)
   { 
    mn = mn->next;
    if (ncomb==Nsta) rangein = 1;
    if (rangein==1)
    {
      if (ln->pro->Naa >= *MaxNaaA) *MaxNaaA = ln->pro->Naa;
      if (mn->pro->Naa >= *MaxNaaB) *MaxNaaB = mn->pro->Naa;
     }
 
    ++ncomb;
    if (ncomb==Nend) rangein = 0;
  } 
 }  
 
 /*
  printf("#MaxNaaA %d MaxNaaB %d\n",*MaxNaaA,*MaxNaaB); 
 */

} /* end of Cal_MaxNaa_from_NstaNend() */

