/*
 
 <PairAli.c>
 
 Procedures for pairwise alignments
 including Sub-optimal alignments

 coded by Takeshi Kawabata.
 
*/
   

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include "IO.h"
#include "GotohDPdis.h"
#include "DiffStr.h"
#include "BetaTopo.h"
#include "ScoreENV.h"
#include "ScoreDIS.h"
#include "ScoreSSE.h"
#include "SecStrEle.h"
#include "BuildUpL.h"
#include "RMS.h"
#include "OutAln.h"
#include "LlistIO.h"
#include "CalZsc.h"
#include "main.h"
#include "protein.h"
#include "LlistAlign.h"

/***** FUNCTIONS (GLOBAL)  ****/
void Pairwise_3D_Alignment(); 
void Malloc_AliSSEmap();
void Free_AliSSEmap();
void Malloc_Alimap();
void Free_Alimap();


/***** FUNCTIONS (LOCAL)  ****/
static void Mask_AliSSEmap();
static void Mask_Alimap();
static void Show_Node_Test();
static void Mask_Alimap_by_SSEAli();



void Pairwise_3D_Alignment(pA,pB,dmpA,dmpB,Ahead,Shead,inalifile,inSSEalifile)
  struct PROTEIN *pA,*pB;     /* Protein  */ 
  struct MAP2D   *dmpA,*dmpB; /* Distance Map */
  struct AlignNode *Ahead;    /* Head for Residue alignment */
  struct AlignNode *Shead;    /* Head for SSE alignment     */ 
  char   *inalifile;          /* Input Residue Alignment file */
  char   *inSSEalifile;       /* Input SSE Alignment File */
{
 float SelfScSSEA,SelfScSSEB,minScSSE,Rsse;
 float SelfScDisA,SelfScDisB,minScDis,Rdis;
 int Nadd,Nround,Nmal;
 struct AlignNode subhead,TmpA;
 struct AlignNode *an,*bn,*sn;

 /*** Initialize ***/
 Shead->next = NULL;

 /************************/
 /*** Normal Alignment ***/
 /************************/


 if (PAR.SubOpt != 'T')
 {
   Nmal = pA->Naa + pB->Naa;
   Ahead->Nali = Ahead->MaxNali = Ahead->num = 0;
   Ahead->next = (struct AlignNode *)malloc(sizeof(struct AlignNode)); 
   Ahead->next->prev = Ahead;
   Ahead->next->next = NULL;  
   Malloc_AlignNode(Ahead->next,Nmal,'S');
   an = Ahead->next; 
   an->Nali = 0; /* JR */
   if (inalifile[0] != '\0')
     Read_CASP_style_Alignment(inalifile,pA,pB,&(an->Nali),an->nalA,an->nalB);   
  else
  { 
   /*** 1st PHASE  : BuildUp using SseScore ***/
   if (inSSEalifile[0]!= '\0')
   {
    printf("maxnali %d\n",pA->Nsse+pB->Nsse);
    Make_AlignNode_LList(Shead,1,(pA->Nsse+pB->Nsse));
    Read_SSE_Alignment(pA,pB,Shead->next,inSSEalifile);
    }
   else
   {
    if (PAR.SseAli == 'T')
    { Build_Up_LL(pA,pB,Shead);
      /* an->sim->ScSSE     = Shead->next->value;  */
      an->sim->ScSSE     = ScoreSSE(pA,pB,Shead->next);
      an->sim->Ncomp_sse = Shead->next->Nali;  
      an->sn = Shead->next; } 
    else {an->sim->ScSSE = 0.0; an->sim->Ncomp_sse = 0;}
   }


   /***  2nd PHASE  : DP using EnvScore ***/
   if (PAR.EnvAli == 'T')
    an->sim->ScEnv = GotohDP(pA,pB,an,Shead->next);

   /***  3rd PHASE  : Iterative DP using DisScore ***/
   if (PAR.DisAli == 'T')
     an->sim->ScDis = GotohDP_Dscore_Iterative(pA,pB,an,dmpA,dmpB,Shead->next);
  
    }
  }

 /****************************/ 
 /*** SubOptimal Alignment ***/
 /****************************/ 

 else if (PAR.SubOpt == 'T')
 {
  Ahead->Nali = Ahead->MaxNali = Ahead->num = 0;
  Ahead->next = NULL;
  /*** Initialize ***/
  Malloc_AlignNode(&TmpA,(pA->Naa+pB->Naa),'S');
  Malloc_AliSSEmap(pA->Nsse,pB->Nsse);
  Malloc_Alimap(pA->Naa,pB->Naa);
  SelfScSSEA = SelfScoreSSE(pA);
  SelfScSSEB = SelfScoreSSE(pB);
  if (SelfScSSEA<SelfScSSEB) minScSSE = SelfScSSEA; else minScSSE = SelfScSSEB;
  /* printf("minScSSE %f\n",minScSSE); fflush(stdout); */

  SelfScDisA = SelfScoreDistanceS(pA,dmpA);
  SelfScDisB = SelfScoreDistanceS(pB,dmpB);
  if (SelfScDisA<SelfScDisB) minScDis = SelfScDisA; else minScDis = SelfScDisB;
  
  /*** 1st PHASE  : Repeatng "BuildUp" with masking previous alignments  ***/
  Nround = 0; 
  do{ 
  ++Nround;
  Nadd = 0;
  subhead.next = NULL;
  /* printf("[ROUND %d]\n",Nround); */
  Build_Up_LL(pA,pB,&subhead);
  Rsse = 100.0*subhead.next->value/minScSSE;
  if ((Rsse >= PAR.Rsse_sub)||(Nround==1)) 
  { ++Nadd;
    Add_AlignNode_To_Tail(Shead,subhead.next);
    Mask_AliSSEmap(subhead.next); 
   }

   Free_AlignNode_Llist(&subhead);
 
  } while (Nadd>0); 


  /*** Iterating 2nd and 3rd phase for different SSE alignments ***/

  sn = Shead;
  Nadd = 0;
  Nround = 0;
  while (sn->next !=NULL)
  {
   ++Nround;
   sn = sn->next;
   TmpA.sim->ScSSE = sn->value;
   
   if (PAR.EnvAli == 'T')
       TmpA.sim->ScEnv = GotohDP(pA,pB,&TmpA,sn);
   if (PAR.DisAli == 'T') {
       TmpA.sim->ScDis = GotohDP_Dscore_Iterative(pA,pB,&TmpA,dmpA,dmpB,sn); 
       TmpA.sim->Rdis = 100.0*TmpA.sim->ScDis/minScDis;
   }
   else   TmpA.Nali=0;
   /* Show_SSE_Ali(sn); 
   printf("Nali %d ScSSE %f ScDis %f Rdis %f\n",TmpA.Nali,
    TmpA.sim->ScSSE,TmpA.sim->ScDis,TmpA.sim->Rdis);  
   */ 
   /* Mask_Alimap(&TmpA);  */
 
   Mask_Alimap_by_SSEAli(pA,pB,sn); 
   if (1 || (TmpA.sim->Rdis >= PAR.Rdis_sub)||(Nround==1))  /* JR for SO SSE alignments */  
     { TmpA.sn = sn;
       Add_AlignNode_To_Tail(Ahead,&TmpA); ++Nadd; }
  } /* while (sn) */

  Set_Value_AlignNode(Ahead,'S');    /* JR  It was  D */
  Merge_Sort_Double_Linked_List(Ahead,'R');

  Free_AliSSEmap(pA->Nsse,pB->Nsse);
  Free_Alimap(pA->Naa,pB->Naa);
  Free_AlignNode(&TmpA);
 }


 /* Renumbering */
 an = Ahead; an->num = 0;
 while (an->next !=NULL) 
  { an->next->num = an->num + 1; 
    an = an->next; 
    Set_Sta_End_Of_Alignment(an->Nali,an->nalA,&(an->staA),&(an->endA));
    Set_Sta_End_Of_Alignment(an->Nali,an->nalB,&(an->staB),&(an->endB));
   }

} /* end of Pairwise_3D_Alignment() */




void Show_Node_Test(head)
 struct AlignNode *head;
{
  struct AlignNode *an;
 
  an = head;
  while (an->next !=NULL)
  {
    an = an->next;
    printf("num %d Nali %d\n", an->num,an->Nali);
   }

}


void Mask_AliSSEmap(sn)
 struct AlignNode *sn;
{
 int i; 
 if (PAR.mal_alissemap == 1)
 {
   for (i=0;i<sn->Nali;++i)
   if ((sn->nalA[i]>0)&&(sn->nalB[i]>0))
     PAR.AliSSEmap[sn->nalA[i]][sn->nalB[i]]  = 1; 
  }

} /* end of  Mask_AliSSEmap() */





void Malloc_AliSSEmap(NsseA,NsseB)
 int NsseA,NsseB;
{
 int i,j;

 PAR.mal_alissemap = 1;

 PAR.AliSSEmap = (char **)malloc(sizeof(char *)*NsseA);
 for (i=0;i<NsseA;++i)
  PAR.AliSSEmap[i] = (char *)malloc(sizeof(char)*NsseB);

 for (i=0;i<NsseA;++i)
  for (j=0;j<NsseB;++j) PAR.AliSSEmap[i][j] = 0;

} /* end of Malloc_AliSSEmap() */




void Free_AliSSEmap(NsseA,NsseB)
 int NsseA,NsseB;
{
 int i;

 if (PAR.mal_alissemap == 1)
 {
  for (i=0;i<NsseA;++i) free(PAR.AliSSEmap[i]);
  free(PAR.AliSSEmap);
 }

 PAR.mal_alissemap = 0;

} /* end of Free_AliSSEmap() */





void Mask_Alimap(an)
 struct AlignNode *an;
{
 int i;
 
 if (PAR.mal_alissemap == 1)
 {
   for (i=1;i<=an->Nali;++i)
   if ((an->nalA[i]>0)&&(an->nalB[i]>0))
     PAR.Alimap[an->nalA[i]][an->nalB[i]]  = 1; 
  }

} /* end of  Mask_Alimap() */




void Mask_Alimap_by_SSEAli(pA,pB,sn)
 struct PROTEIN *pA,*pB; 
 struct AlignNode *sn;
{
 int i,a,b;
 
 if (PAR.mal_alissemap == 1)
 {
   for (i=0;i<sn->Nali;++i)
   if ((sn->nalA[i]>=0)&&(sn->nalB[i]>=0))
   { 
    for (a= pA->SSE[sn->nalA[i]].Rsta; a <= pA->SSE[sn->nalA[i]].Rend; ++a)
     for (b= pB->SSE[sn->nalB[i]].Rsta; b <= pB->SSE[sn->nalB[i]].Rend; ++b)
      PAR.Alimap[a][b]  = 1; 
    } 
  }

} /* end of  Mask_Alimap_by_SSEAli() */





void Malloc_Alimap(NaaA,NaaB)
 int NaaA,NaaB;
{
 int i,j;

 PAR.mal_alimap = 1;

 PAR.Alimap = (char **)malloc(sizeof(char *)*(NaaA+1));
 for (i=0;i<=NaaA;++i)
  PAR.Alimap[i] = (char *)malloc(sizeof(char)*(NaaB+1));

 for (i=0;i<=NaaA;++i)
  for (j=0;j<=NaaB;++j) PAR.Alimap[i][j] = 0;

} /* end of Malloc_Alimap() */




void Free_Alimap(NaaA,NaaB)
 int NaaA,NaaB;
{
 int i;

 if (PAR.mal_alimap == 1)
 {
  for (i=0;i<=NaaA;++i) free(PAR.Alimap[i]);
  free(PAR.Alimap);
 }

 PAR.mal_alimap = 0;

} /* end of Free_Alimap() */




