/*

< RMS.c >

coded by Takeshi Kawabata

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h"
#include "protein.h"
#include "LlistAlign.h"


/*** STRUCTURE OF ATOM POSITION ARRAY ***/
struct ATOMS{
 double **pos;  /* pos[N][3] (malloc later, [N:0..n-1] )  */ 
 int   N; };


/*** FUNCTIONS (GLOBAL) ***/
double CRMS();
void Mult_Mat_Vec();
int  Write_Rotation_Matrix();
double CRMS_Filter();


/*** FUNCTIONS (LOCAL) ***/
static double DotProd();
static double Determinant();
static void Cal_Optimal_Rmatrix();
static double Cal_RMS();
static void Malloc_Atoms();
static void Free_Atoms();
static void Set_G_to_Zero();


static void Rotation();
static void print_matrix6();
static void find_max_abs();
static void equal_matrix();
static void prod_matrix();
static void Making_Omega();
static void Make_Related_Matrix();
static void Set_Eigen_Vector();
static void Get_Matrix_R();
static void Jacobi_Wilkinson();



double CRMS(pA,pB,an,g1,g2,Rmat,AtomType,InvType)
 struct PROTEIN *pA,*pB;
 struct AlignNode *an; 
 double g1[3],g2[3]; /* Center position */
 double Rmat[3][3];  /* Rotation Matrix */
 char AtomType;     /* 'B' CBatom, 'A' CAatom */
 char InvType;      /* 'I': alignment *a is inverse one (AvsB --> BvsA) */
{ 
  int i,j,m,N,a,b;
  double rms;
  struct ATOMS aA,aB;

  Malloc_Atoms(&aA,an->Nali);
  Malloc_Atoms(&aB,an->Nali);

  /*** SET XYZ DATA ***/ 
  N = 0; 
  for (i=1;i<=an->Nali;++i) 
  {
   if (InvType != 'I') { a = an->nalA[i]; b = an->nalB[i];}
                  else { a = an->nalB[i]; b = an->nalA[i];}
   if ((a>=1)&&(b>=1))
   {

    if (AtomType == 'B')
    { for (m=0;m<3;++m) {aA.pos[N][m] = pA->PosCB[a][m]; aB.pos[N][m] = pB->PosCB[b][m];}  }
    else
    { for (m=0;m<3;++m) {aA.pos[N][m] = pA->PosCA[a][m]; aB.pos[N][m] = pB->PosCA[b][m];}  }
    ++N; 
   }
  }

  Set_G_to_Zero(N,&aA,g1);
  Set_G_to_Zero(N,&aB,g2);
 
  if (an->Nali>0)
  {  Cal_Optimal_Rmatrix(N,&aA,&aB,Rmat,'W');
     Rotation(N,&aA,Rmat);
     rms = (double)Cal_RMS(N,&aA,&aB); } 
  else { rms = 0.0; g1[0]=g1[1]=g1[2]=g2[0]=g2[1]=g2[2]=0.0;
         Rmat[0][0]=Rmat[1][1]=Rmat[2][2] = 1.0; 
         Rmat[0][1]=Rmat[0][2]=Rmat[1][0]=Rmat[1][2]=Rmat[2][0]=Rmat[2][1]=0.0; } 

  Free_Atoms(&aB,an->Nali);
  Free_Atoms(&aA,an->Nali);
  
  return(rms);

} /* end of CRMS() */



void Malloc_Atoms(atoms,N)
 struct ATOMS *atoms;
 int N;
{
 int i;
 atoms->N = N; 
 atoms->pos = (double **)malloc(sizeof(double *)*(N+1));
 for (i=0;i<=N;++i)
  atoms->pos[i] = (double *)malloc(sizeof(double)*3);

} /* end of Malloc_Atoms() */


void Free_Atoms(atoms,N)
 struct ATOMS *atoms;
 int N;
{
 int i;
 for (i=N;i>=0;--i) 
  { free(atoms->pos[i]); atoms->pos[i] = NULL;}
 free(atoms->pos); 
 atoms->pos = NULL;
} /* end of Free_Atoms() */




int Write_Rotation_Matrix(fname,pA,pB,Nali,g1,g2,Rmat,rms)
 char *fname;
 struct PROTEIN *pA,*pB;
 int Nali;
 double g1[3],g2[3]; /* Center position */
 double Rmat[3][3];  /* Rotation Matrix */
 double rms;
{ 
  FILE *fp; 
  int i,j;

  fp = fopen(fname,"w"); 
  if (fp==NULL) 
  { printf("#ERROR : Can't write rotation matrix to \"%s\"\n",fname); 
    return(0); }

  /* 
  fprintf(fp,"RMS %.2lf %s (%d) vs %s (%d) Nali %d [Afit=R*(A-Ga)+Gb]\n",
          rms,pA->cname,pA->Naa,pB->cname,pB->Naa,Nali); 
  */
  fprintf(fp,"#%-14s  %s Naa %d\n","PROTEIN_A",pA->cname,pA->Naa);
  fprintf(fp,"#%-14s  %s Naa %d\n","PROTEIN_B",pB->cname,pB->Naa);
  fprintf(fp,"#%-14s  %d\n","NALIGN",Nali);
  fprintf(fp,"#%-14s  %lf\n","RMS",rms);
  fprintf(fp,"#Afit=R*(A-Ga)+Gb\n");
  fprintf(fp,"Ga %8.5lf %8.5lf %8.5lf\n",g1[0],g1[1],g1[2]);
  fprintf(fp,"Gb %8.5lf %8.5lf %8.5lf\n",g2[0],g2[1],g2[2]);
  fprintf(fp,"R0 %8.5lf %8.5lf %8.5lf\n",Rmat[0][0],Rmat[0][1],Rmat[0][2]);
  fprintf(fp,"R1 %8.5lf %8.5lf %8.5lf\n",Rmat[1][0],Rmat[1][1],Rmat[1][2]);
  fprintf(fp,"R2 %8.5lf %8.5lf %8.5lf\n",Rmat[2][0],Rmat[2][1],Rmat[2][2]);
  fclose(fp);
  return(1);

} /* end of Write_Rotation_Matrix() */ 







void Set_G_to_Zero(n,atoms,Ga)
 int n;
 struct ATOMS *atoms;
 double Ga[3];
{
 int i,j;
 double xsum,ysum,zsum;

 xsum = ysum = zsum = 0.0;
 for (i=0;i<n;++i)
  { xsum += atoms->pos[i][0]; ysum += atoms->pos[i][1]; zsum += atoms->pos[i][2]; }

 Ga[0] = xsum/n; Ga[1] = ysum/n; Ga[2] = zsum/n;

 for (i=0;i<n;++i)
 { atoms->pos[i][0] = atoms->pos[i][0] - Ga[0];
   atoms->pos[i][1] = atoms->pos[i][1] - Ga[1];
   atoms->pos[i][2] = atoms->pos[i][2] - Ga[2]; }

} /* end of Set_G_to_Zero() */



void Rotation(n,atoms,mat)
 int n;
 struct ATOMS *atoms;
 double mat[3][3];
{
 double p[3],q[3];
 int i;

 for (i=0;i<n;++i)
  { p[0] = atoms->pos[i][0] ;  
    p[1] = atoms->pos[i][1] ;  
    p[2] = atoms->pos[i][2] ;  
    Mult_Mat_Vec(q,mat,p);
    atoms->pos[i][0] = q[0]; 
    atoms->pos[i][1] = q[1]; 
    atoms->pos[i][2] = q[2]; }

} /* end of Rotation() */




double Cal_RMS(n,aA,aB)
 int n;
 struct ATOMS *aA,*aB;
{ int i;
  double dx,dy,dz;
  double RM;
  double rms;

 RM = 0.0; 
 for (i=0;i<n;++i) 
 { dx = aA->pos[i][0] - aB->pos[i][0]; 
   dy = aA->pos[i][1] - aB->pos[i][1]; 
   dz = aA->pos[i][2] - aB->pos[i][2]; 
   RM += (dx*dx)+(dy*dy)+(dz*dz);
  }
 rms = sqrt(RM/n);
 return(rms);
} /* end of Cal_RMS() */



 
void Cross_Prod(a,b,c)
  double a[3],b[3],c[3];
 {  c[0] = a[1] * b[2] - a[2] * b[1];
    c[1] = a[2] * b[0] - a[0] * b[2];
    c[2] = a[0] * b[1] - a[1] * b[0];
 }

double DotProd(a,b)
  double a[3],b[3];
{ double ret; 
  int i;
  ret = 0.0; 
  for (i=0;i<3;++i)
    { ret += a[i]*b[i];}
  return(ret); 
 }


void pvector(a)
double a[3];
{
printf(" (%3.2lf %3.2lf %3.2lf)\n",a[0],a[1],a[2]);
}




void Mult_Mat_Vec(y,mat,x)
 double y[3],mat[3][3],x[3];
{ y[0] = mat[0][0]*x[0] + mat[0][1]*x[1] + mat[0][2] * x[2];
  y[1] = mat[1][0]*x[0] + mat[1][1]*x[1] + mat[1][2] * x[2];
  y[2] = mat[2][0]*x[0] + mat[2][1]*x[1] + mat[2][2] * x[2]; }



void Mult_Mat_Mat(A,B,C) /* A * B = C */
 double A[3][3],B[3][3],C[3][3];
{
int i,j,k; 

for (i=0;i<3;++i)
  for (j=0;j<3;++j)
   { C[i][j] = 0.0;
     for (k=0;k<3;++k) C[i][j] +=A[i][k]*B[k][j]; }

} /* end of Mult_Mat_Mat() */



double Determinant(A)
 double A[3][3];
{
double ret;

ret = (A[0][0]*A[1][1]*A[2][2] +
       A[0][2]*A[1][0]*A[2][1] + 
       A[0][1]*A[1][2]*A[2][0]) 
     -(A[0][2]*A[1][1]*A[2][0] +
       A[0][1]*A[1][0]*A[2][2] +
       A[0][0]*A[1][2]*A[2][1] );
return(ret);

}



void Transpose_Matrix(A,T) /* T = tA */
double A[3][3],T[3][3];
{ int i,j;
  for (i=0;i<3;++i)
    for (j=0;j<3;++j)
      T[i][j] = A[j][i];
}








void Print_Matrix(A)
 double A[3][3];
{
 int i;

 for (i=0;i<3;++i)
 printf(" (%lf %lf %lf)\n",A[i][0],A[i][1],A[i][2]);

}





void Cal_Optimal_Rmatrix(n,aA,aB,R,type)
 int n;
 struct ATOMS *aA,*aB;
 double R[3][3];
 char type;
{
 int i,j;
 double U[3][3],Ut[3][3];
 double O[6][6],J[6][6];
 double h[3][3],k[3][3],D[3]; /* Eigen Vector,Eigen Value */ 
 double ht[3][3],kt[3][3],Dm[3][3]; /* Eigen Vector,Eigen Value */ 
 double det; /* Determinant of U */

 Make_Related_Matrix(n,aA,aB,U);
 det = Determinant(U);
 Transpose_Matrix(U,Ut);
 Making_Omega(O,U,Ut);
 Jacobi_Wilkinson(O,J);
 Set_Eigen_Vector(det,O,J,h,k,D);
 Get_Matrix_R(U,h,k,R);

} /* end of Cal_Optimal_Rmatrix() */






void Making_Omega(O,U,Ut)
 double O[6][6];
 double U[3][3],Ut[3][3];
{
int i,j;

for (i=0;i<6;++i)
 for (j=0;j<6;++j)
 {
  O[i][j] = 0.0; 
  if ((3<=i)&&(i<=5)&&(0<=j)&&(j<=2)) O[i][j] = Ut[i-3][j]; 

  if ((0<=i)&&(i<=2)&&(3<=j)&&(j<=5)) O[i][j] = U[i][j-3];
 }


} /* end of Making_Omega() */





void Make_Related_Matrix(n,aA,aB,U)
 int n;
 struct ATOMS *aA,*aB;
 double U[3][3];
{
 int i,j,k;

/* printf("---Make_Related_Matrix()----\n"); */

for (i=0;i<3;++i)
 for (j=0;j<3;++j)
  { 
   U[i][j] = 0.0; 
   for (k=0;k<n;++k) U[i][j] += aA->pos[k][i] * aB->pos[k][j];  
   U[i][j] /= (double)(n*n);  
   }
 

} /* end of Make_Related_Matrix() */





void Set_Eigen_Vector(det,U,J,h,k,D)
 double det;
 double U[6][6],J[6][6];
 double h[3][3],k[3][3];
 double D[3];
{
 double dt[6];
 int num[6];
 double buff;
 int i,j,bnum,end;
 double r2;
 double cro[3];
 double dp;


 for (i=0;i<6;++i) {dt[i] = U[i][i]; num[i] = i;}

 end = 0;
 /*----------- BUBBLE SORT of Eigen value --------------- */
 while (end !=1)
 {
   end = 1;
   for (i=0;i<5;++i)
     if(dt[i+1]>dt[i])
       {buff = dt[i+1];  dt[i+1]  = dt[i];  dt[i]  = buff; 
        bnum = num[i+1]; num[i+1] = num[i]; num[i] = bnum; 
        end = 0;}   
 }



 /* ----------- SET EIGEN VECTOR ------------------------- */
 r2 = sqrt(2.0);
 
 for (i=0;i<3;++i)
  for (j=0;j<3;++j)
    {
     h[i][j] = J[j][num[i]]*r2;
     k[i][j] = J[j+3][num[i]]*r2;
    }

  D[0] = dt[0]; D[1] = dt[1]; D[2] = dt[2];

 /* --------- SET h3 SIGN ----------------------- */
 Cross_Prod(h[0],h[1],cro);
 if (DotProd(h[2],cro)<0.0)
  {  /* printf("\n CHANGE h3 SIGN dot = %lf\n",DotProd(h[2],cro)); */ 
     h[2][0] *= -1.0; h[2][1] *= -1.0; h[2][2] *= -1.0; 
     k[2][0] *= -1.0; k[2][1] *= -1.0; k[2][2] *= -1.0; 
   } 


 /* ------------ CASE of SINGULAR -------------- */
 if (fabs(det)<0.00001)
  { /* printf("\n This Matrix is SINGULAR.\n"); */ 
    Cross_Prod(h[0],h[1],h[2]); 
    Cross_Prod(k[0],k[1],k[2]); }

 /* -------------CASE of NEGATIVE DET----------- */

  if (det<0.0)
  {  /* printf("\n This Matrix is negative-determinant.\n"); */
     k[2][0] *=-1.0; k[2][1] *= -1.0; k[2][2] *= -1.0; }


} /* end of Set_Eigen_Vector() */











void Get_Matrix_R(U,h,k,R)
 double U[3][3],h[3][3],k[3][3],R[3][3];
{
 double kt[3][3];

 Transpose_Matrix(k,kt);
 Mult_Mat_Mat(kt,h,R);

} /* end of Get_Matrix_R() */



void Jacobi(A,U)
double A[6][6],U[6][6];
{ 
 double R[6][6],TR[6][6],BUFF[6][6];
 double max,p,q,co,si,co2,si2;
 int i,j,mi,mj,c,cmax;

 printf("int Jacobi(A,U)\n");
   /* --------SETTING U[][] = I --------*/
   
   for (i=0;i<6;++i)
     for (j=0;j<6;++j) if (i==j) U[i][j] = 1.0;  else U[i][j] = 0.0;
    
    find_max_abs(&mi,&mj,&max,A);
    c = 0;
    
    while ((max>0.00000001)&&(c<100))
    {
      /*--------- SET of cos sin -----------*/ 
      p = (A[mj][mj] - A[mi][mi])/2;  q = A[mi][mj];
      co2 = p/sqrt(p*p+q*q); si2 = -q/sqrt(p*p+q*q);
      co = sqrt((co2+1.0)/2.0);
      si = si2/2.0/co;

      /* -------- SET of Rot Matrix R and TR----- */  
      for (i=0;i<6;++i)
	for (j=0;j<6;++j)  if (i==j)  R[i][j] = 1.0;  else R[i][j] =0.0;
	 R[mi][mi] =  co; R[mi][mj] = -si;
         R[mj][mi] =  si; R[mj][mj] =  co;
     
      for (i=0;i<6;++i)
	for (j=0;j<6;++j) TR[j][i] = R[i][j]; 

       prod_matrix(A,R,BUFF);
       prod_matrix(TR,BUFF,A);
       prod_matrix(U,R,BUFF);
       equal_matrix(U,BUFF);

       find_max_abs(&mi,&mj,&max,A);
       ++c;
 
      } /* end of while */

}/* end of Jacobi() */


void Jacobi_Wilkinson(A,U)
 double A[6][6],U[6][6];
{ 
 double R[6][6],TR[6][6],BUFF[6][6];
 double max,p,q,co,si,co2,si2;
 double wa,sa,r,sqrt2;
 int i,j,mi,mj,c,cmax;

 sqrt2 = sqrt(2.0);

 /* --------SETTING U[][] = I --------*/
   
 for (i=0;i<6;++i)
   for (j=0;j<6;++j) if (i==j) U[i][j] = 1.0;  else U[i][j] = 0.0;
    
  find_max_abs(&mi,&mj,&max,A);
  c = 0;
    
  while ((max>0.00000001)&&(c<100))
    {

  /*--------- SET of cos sin -----------*/ 
      
    wa = (A[mi][mi] + A[mj][mj])/2;  
    sa = (A[mi][mi] - A[mj][mj])/2; 
    r = sqrt(sa*sa + A[mi][mj]*A[mi][mj]);
    /* co =  sqrt(1.0+sa/r)/sqrt2;  si =  A[mi][mj]/2.0/r/co;  */
    if (sa>0.0) 
     { co =  sqrt(1.0+sa/r)/sqrt2;  si =  A[mi][mj]/2.0/r/co; }
    else 
     { co =   sqrt(1.0-sa/r)/sqrt2; si = - A[mi][mj]/2.0/r/co; }
      
    /* -------- SET of Rot Matrix R and TR----- */  
    for (i=0;i<6;++i)
      for (j=0;j<6;++j) if (i==j)  R[i][j] = 1.0;  else R[i][j] =0.0;
     
     R[mi][mi] =   co; R[mi][mj] = -si;
     R[mj][mi] =   si; R[mj][mj] = co;

     for (i=0;i<6;++i)
      for (j=0;j<6;++j) TR[j][i] = R[i][j]; 

     prod_matrix(A,R,BUFF);
     prod_matrix(TR,BUFF,A);
     prod_matrix(U,R,BUFF);
     equal_matrix(U,BUFF);

     find_max_abs(&mi,&mj,&max,A);
     ++c;

    } /* end of while */

}/* end of Jacobi_Wilkinson() */


void prod_matrix(A,B,C) /* C = A * B */
 double A[6][6],B[6][6],C[6][6];
{ int i,j,k;
  double sum;
  for (i=0;i<6;++i)
    for (j=0;j<6;++j)
     { sum =0;
       for (k=0;k<6;++k)
	{ sum = sum + A[i][k] * B[k][j];}
       C[i][j] = sum; }
 } /* end of prod_matrix() */


void equal_matrix(A,B)    /* A = B */
 double A[6][6],B[6][6];
{ int i,j;
  for (i=0;i<6;++i)
   {
    for (j=0;j<6;++j)
     { A[i][j] = B[i][j];}
   }
}/* end of equal_matrix() */



void find_max_abs(mi,mj,max,A) /* Max_Abs element A[mi][mj] = max */
 int *mi,*mj;
 double *max,A[6][6];
{ int i,j,p,q;

  *max =0;
  for (i=0;i<6;++i)
    for (j=0;j<6;++j)
      if (i!=j)
       if (fabs(A[i][j])> *max)
	      { *max = fabs(A[i][j]);     p = i; q= j;}  
   *mi =p; *mj = q;
 } /* end of find_max_abs() */

void print_matrix6(A)
 double A[6][6];
{ int i,j;
  printf("\n");
  for (i=0;i<6;++i)
    for (j=0;j<6;++j)
     { printf(" [%d%d]:%3.8lf",i,j,A[i][j]);
       if (j==5) { printf("\n");}          }
  printf("\n");
 }/* end of print_matrix() */

void print_matrix3(A)
 double A[3][3];
{ int i,j;
  printf("\n");
  for (i=0;i<3;++i)
    for (j=0;j<3;++j)
     { printf(" [%d%d]:%3.8lf",i,j,A[i][j]);
       if (j==2)  printf("\n");          }
  printf("\n");
 }/* end of print_matrix() */



double CRMS_Filter(pA,pB,an,AtomType,CRMSthre)
 struct PROTEIN *pA,*pB;
 struct AlignNode *an; 
 char AtomType;     /* 'B' CBatom, 'A' CAatom */
 float CRMSthre;
{ 
  int i,j,N,m,a,b,max_i;
  double rms;
  struct ATOMS aA,aB;
  double g1[3],g2[3]; /* Center position */
  double Rmat[3][3];  /* Rotation Matrix */
  int *aoni;
  double *error,max_e,d[3];

  printf("#CRMS_Filter('%c' thre %f) Nali %d\n",AtomType,CRMSthre,an->Nali);
  Malloc_Atoms(&aA,an->Nali);
  Malloc_Atoms(&aB,an->Nali);
  aoni  = (int *)malloc(sizeof(int)*(an->Nali+1));
  error = (double *)malloc(sizeof(double)*(an->Nali+1));

  max_i = -1; 
  do
  {
    if (max_i>=0) {an->nalA[aoni[max_i]] = an->nalB[aoni[max_i]] = -1;}
    N = 0; 
    for (i=1;i<=an->Nali;++i) 
    {
     a = an->nalA[i]; b = an->nalB[i];
     if ((a>=1)&&(b>=1))
     {
     if (AtomType == 'B')
     { for (m=0;m<3;++m) {aA.pos[N][m] = pA->PosCB[a][m]; aB.pos[N][m] = pB->PosCB[b][m]; } }
     else
     { for (m=0;m<3;++m) {aA.pos[N][m] = pA->PosCA[a][m]; aB.pos[N][m] = pB->PosCA[b][m]; } }
     aoni[N] = i;  
     ++N; 
     }
   } /* i */

   Set_G_to_Zero(N,&aA,g1); Set_G_to_Zero(N,&aB,g2);
   Cal_Optimal_Rmatrix(N,&aA,&aB,Rmat,'W');
   Rotation(N,&aA,Rmat);
   max_e = rms = 0.0; 
    for (i=0;i<N;++i) 
    { for (m=0;m<3;++m) d[m] = aA.pos[i][m] - aB.pos[i][m]; 
      error[i] = (d[0]*d[0])+(d[1]*d[1])+(d[2]*d[2]); 
      if (error[i]>max_e) {max_e = error[i]; max_i = i;} 
      rms += error[i]; }
   rms = sqrt(rms/N);

   /* printf("rms %lf\n",rms); */

   }while (rms > CRMSthre);

  Free_Atoms(&aB,an->Nali); Free_Atoms(&aA,an->Nali);

  free(error);  free(aoni);
  
  return(rms);

} /* end of CRMS_Filter() */

