/*

<RegressL.c>

for regressing random distribution of score value 

This is for linked list.

coded by Takeshi Kawabata.


*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "main.h" 
#include "LlistAlign.h" 

/*** Functions (global) ***/
void Cal_Raw_Zscore_Llist();
void TwoStep_Linear_Fitting_Llist();
void Set_Value_AlignNode();
void Set_SubValue_AlignNode();
void Count_Thre_Value_AlignNode();

/*** Functions (local) ***/
static void Linear_Fitting_Llist();



void Cal_Raw_MeanSD_Llist(head,Mean,SD)
 struct AlignNode *head; 
 double *Mean,*SD;
{ double S,SS;
  int i,NN;
  struct AlignNode *bn; 
  char  *relate;
  

  S = SS = 0.0;  NN = 0; 
  bn = head;
  while (bn->next !=NULL)
   { 
     bn = bn->next;
     S  += bn->value;  
     SS += bn->value*bn->value;  
     ++NN; 
    }

  *Mean  = S/NN;
  *SD    = sqrt(SS/NN - (*Mean)*(*Mean)); 

} /* end of Cal_Raw_MeanSD_Llist() */




void Cal_Raw_Zscore_Llist(head,Mean,SD)
 struct AlignNode *head; 
 double *Mean,*SD;
{ double S,SS,m,sd;
  int i,NN,Nlib;
  struct AlignNode *an;

  /*** 1ST ESTIMATE **/
  an = head;
  S = SS = 0.0;  NN = 0;
  while (an->next !=NULL)  
   { 
     an = an->next; 
     S  += an->value;  
     SS += an->value*an->value;  
     ++NN; 
     an->mark = 0;
   }

  m  = S/NN; sd = sqrt(SS/NN - m*m); 
  
  /* cal Zsc */ 
  an = head;
  while (an->next !=NULL) 
  { an = an->next; 
    an->Zsc = (an->value - m)/sd; 
    if (an->Zsc > 4.0 ) an->mark = 1; }

  /*** 2ND ESTIMATE **/
  S = SS = 0.0;  NN = 0; 
  an = head;
  while (an->next !=NULL) 
  { 
   an = an->next; 
   if (an->mark==0)
   { S  += an->value;  
     SS += an->value * an->value; 
     ++NN; }
  }

  m  = S/NN; sd = sqrt(SS/NN - m*m); 

  /* cal Zsc */ 
  an = head;
  while (an->next !=NULL) 
  { an = an->next;
    an->Zsc = (an->value - m)/sd; }

  *Mean = m;
  *SD   = sd;

} /* end of Cal_Raw_Zscore_Llist() */



void TwoStep_Linear_Fitting_Llist(head,A,B,SD)
  struct AlignNode *head; 
  double *A,*B,*SD;
{ struct AlignNode *bn;

  /** INITIALIZE **/ 
  bn = head;
  while (bn->next!=NULL) 
   { bn->mark = 0; 
     bn = bn->next;} 

  /*** 1ST FITTING ***/ 
  
  Linear_Fitting_Llist(head,A,B,SD);

  bn = head;
  while (bn->next!=NULL) 
   { 
     bn = bn->next;
     bn->Zsc = (bn->value - (*A * bn->subvalue+ *B))/(*SD); 
     if (bn->Zsc > 4.0)  bn->mark = 1;   
    }

  /*** 2ND FITTING ***/ 
  Linear_Fitting_Llist(head,A,B,SD);
 
  bn = head; 
  while (bn->next!=NULL) 
   { bn = bn->next; 
     bn->Zsc = (bn->value - (*A * bn->subvalue+ *B))/(*SD); }

} /* end of TwoStep_Linear_Fitting_Llist() */  





void Linear_Fitting_Llist(head,A,B,SD)  /* Y = A*X + B */
 struct AlignNode *head; 
 double *A,*B,*SD;
{
 int NN;
 double Sx,Sy,Sxx,Syy,Sxy,x,y,Err;
 struct AlignNode *bn;
  
 NN = 0;
 Sx = Sy = Sxx = Syy = Sxy = 0.0; 
 bn = head;
 while (bn->next !=NULL)
 { 
  bn = bn->next; 
  if (bn->mark == 0)
  { x = bn->subvalue; 
    y = bn->value;
    Sx  += x;  Sxx += x*x;
    Sy  += y;  Syy += y*y;
    Sxy += x*y;
    ++NN; }
 }

 *A = ( NN * Sxy - Sx  * Sy)/(NN * Sxx - Sx * Sx);
 *B = (-Sx * Sxy + Sxx * Sy)/(NN * Sxx - Sx * Sx);

 Err =  Syy + (*A)*(*A)*Sxx + NN*(*B)*(*B) 
      - 2*(*A)*Sxy + 2*(*A)*(*B)*Sx - 2*(*B)*Sy;

 *SD = sqrt(Err/NN);

} /* end of Linear_Fitting_Llist() */


void Set_Value_AlignNode(head,vtype)
 struct AlignNode *head;
 char vtype;
{
 struct AlignNode *bn;

 bn = head;
 while (bn->next !=NULL)
 {
  bn = bn->next;
       if (vtype == 'Z')   bn->value = (float)bn->Zsc;
  else if (vtype == 'A')   bn->value = (float)bn->staA;
  else if (vtype == 'B')   bn->value = (float)bn->staB;
  else if (vtype == 'I')   bn->value = (float)bn->sim->SeqID;
  else if (vtype == 'R')   bn->value = (float)bn->sim->Rdis;
  else if (vtype == 'E')   bn->value = (float)bn->sim->ScEnv;
  else if (vtype == 'S')   bn->value = (float)bn->sim->ScSSE;
  else if (vtype == 'D')   bn->value = (float)bn->sim->ScDis;
  else bn->value = (float)bn->num;
 }

} /* end of Set_Value_AlignNode() */




void Set_SubValue_AlignNode(head,vtype,Ttype)
 struct AlignNode *head;
 char vtype; /* Value type */
 char Ttype; /* Transform Type 'Q':x*x,'G':log,  */
{
 struct AlignNode *bn;

 bn = head;
 while (bn->next !=NULL)
 {
  bn = bn->next;
       if (vtype == 'C')   bn->subvalue = (float)bn->sim->Ncomp;
  else if (vtype == 'S')   bn->subvalue = (float)bn->sim->Ncomp_sse;
  else if (vtype == 'L')   bn->subvalue = (float)bn->sim->Naa;
  else if (vtype == 'A')   bn->subvalue = (float)bn->Nali;
  else bn->value = (float)bn->num;

  if (Ttype == 'Q') bn->subvalue = bn->subvalue * bn->subvalue;
  if (Ttype == 'G') 
   { if (bn->subvalue>0.0) 
      bn->subvalue = log((double)(bn->subvalue));
     else bn->subvalue = 0.0;  }
 }

} /* end of Set_SubValue_AlignNode() */



void Count_Thre_Value_AlignNode(head,thre,type,Nrank)
 struct AlignNode *head;
 float thre;
 char  type; /* 'M'ore or 'L'ess */
 int *Nrank;
{
 struct AlignNode *bn;

 bn = head; *Nrank = 0;
 while (bn->next !=NULL)
 {
  bn = bn->next;
  if ((type == 'M') && (bn->value >= thre) ) ++(*Nrank); 
  if ((type == 'L') && (bn->value <= thre) ) ++(*Nrank); 
 }

} /* end of Count_Thre_Value_AlignNode(head,thre,type) */



