/**

<ScoreDIS.c>

Distance Comparison Score, which depends on Sequence Separation   

coded by Takeshi Kawabata.

**/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <math.h>
#include "main.h" 
#include "protein.h" 
#include "Amino.h" 
#include "Vector.h" 
#include "ScoreENV.h" 
#include "ScoreSSE.h" 
#include "IO.h" 
#include "LlistAlign.h" 


float DisROM[21][50][50];

/**** Functions (GLOBAL) **/
void Read_All_Dis_Score();
float ScoreDistanceS();
static void Read_DisROM_Each_Separation();
float DisScore_with_PreAli();
float GapPenalty();
float SelfScoreDistanceS();
float Reliability_Rdis();
float Reliability_Zsc();


void Read_All_Dis_Score(tailstr)
 char *tailstr;
{
 char fname[256];
 int k,s,t;

 /** Initialize **/ 
 for (k=0;k<=20;++k)
  for (s=0;s<50;++s)  
   for (t=0;t<50;++t) DisROM[k][s][t] = 0.0;  

 /** Read Files  **/ 
 for (k=0;k<=20;++k)
 { sprintf(fname,"%s/dis%d-%s.rom",SCORE_DIR,k,tailstr);
   /* printf("Try to read %s\n",fname); */
   Read_DisROM_Each_Separation(fname); } 

} /* end of Read_All_Dis_Score() */



void Read_DisROM_Each_Separation(fname)
 char *fname;
{
 FILE *fp;
 int s,t,L,separation;
 char buffer[128],buff1[64],buff2[64];
 char rom_on;
 float val;

 /* <FORMAT>
 @CLASS DISTANCE
 @TYPE B
 @SEPARATION 15
 @NSTATE 50
 @NMULTI 3
 @SIGMA  100.000000
 @RELATED_ODDS_MATRIX
 2 2 79.235067
 2 3 51.159827
 2 4 27.631932
 2 5 9.900840
 2 6 -19.574507
 */

 fp = fopen(fname,"r");
 if (fp==NULL) { printf("#ERROR:Can't find disscorefile \"%s\"\n",fname); exit(1);}
 rom_on = 0; separation = -1;
 while (feof(fp)==0)
 {
  buffer[0] = '\0';
  fgets(buffer,128,fp);
  L = strlen(buffer);
  if (buffer[L-1]=='\n') buffer[L-1]= '\0';
  if ((buffer[0]!='#')&&(L>1))
  {
   if (buffer[0]=='@') rom_on = 0;

   if (rom_on==1)
   { sscanf(buffer,"%d %d %f",&s,&t,&val);
     if ((s<0)&&(s>=50)) { printf("#ERROR:Read_DisROM: %d is out of range\n",s); exit(1);}
     if ((t<0)&&(t>=50)) { printf("#ERROR:Read_DisROM :%d is out of range\n",t); exit(1);}
     DisROM[separation][s][t] = DisROM[separation][t][s] = val; 
   }

    if (strncmp(buffer,"@SEPARATION",11)==0) 
     { sscanf(buffer,"%s %d",buff1,&separation); 
       if ((separation<0)||(separation>20))
        { printf("#ERROR:Separation %d is out of range\n",separation); exit(1);}
       }
    if (strncmp(buffer,"@RELATED_ODDS_MATRIX",20)==0) rom_on = 1;
   }

  } /* while */
 fclose(fp);


} /* end of Read_DisROM_Each_Sepatation() */

 







float ScoreDistanceS(pA,pB,a,DmapA,DmapB)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a; 
 struct MAP2D *DmapA, *DmapB;
{ int i,j,Ncomp,dA,dB,kA,kB,k,dAB;
  float diff,Sc,disA,disB,disAB;
 
  Sc = 0.0;
 
  if (PAR.DisScType == 'N')  /* Normal */
  {
  for (i=1;i<=a->Nali;++i)
   if ((a->nalA[i] > 0)&&(a->nalB[i] >0))
    for (j=i+1;j<=a->Nali;++j)
     if ((a->nalA[j] > 0)&&(a->nalB[j] >0))
     { 
          kA = a->nalA[j] - a->nalA[i];
          kB = a->nalB[j] - a->nalB[i];
          if (kA>kB) k = kA; else k = kB;
          if (k>=1) 
          {
          if (21<=k) k = 0;
          dA = (int)DmapA->map[a->nalA[i]][a->nalA[j]];
          dB = (int)DmapB->map[a->nalB[i]][a->nalB[j]];
          if ((dA<50)&&(dB<50)) Sc += DisROM[k][dA][dB];
        }
     }   
 }

 
  else if (PAR.DisScType == 'D') /* "DALI" type */
  {  
   for (i=1;i<=a->Nali;++i)
    if ((a->nalA[i] > 0)&&(a->nalB[i] >0))
    { 
     for (j=i;j<=a->Nali;++j)
     if (i==j) {  Sc += 0.2;  }
     else 
     { 
      if ((a->nalA[j] > 0)&&(a->nalB[j] >0))
        { 
         disA  = DmapA->map[a->nalA[i]][a->nalA[j]];
         disB  = DmapB->map[a->nalB[i]][a->nalB[j]];           
         disAB = (disA+disB)/2.0;
         Sc += (0.2 - fabs(disA-disB)/disAB) * exp(-disAB*disAB/400.0); 
         }
      } 
     } /* i loop */  
   }   

   else if (PAR.DisScType == 'S') /* "SSAP" type */
   { 
    for (i=1;i<=a->Nali;++i)
     if ((a->nalA[i] > 0)&&(a->nalB[i] >0))
     { 
      Sc += PAR.Assap/PAR.Bssap;   
      for (j=i+1;j<=a->Nali;++j)
       if ((a->nalA[j] > 0)&&(a->nalB[j] >0))
        Sc += PAR.Assap/(fabs(DmapA->map[a->nalA[i]][a->nalA[j]] - 
                          DmapB->map[a->nalB[i]][a->nalB[j]]) + PAR.Bssap);
      } 
   }
   
  return(Sc); 

} /* end of ScoreDistanceS() */





void ScoreDistanceSPair(pA,pB,a,DmapA,DmapB)
 struct PROTEIN *pA,*pB;
 struct AlignNode *a; 
 struct MAP2D *DmapA, *DmapB;
{ int i,j,Ncomp,dA,dB,kA,kB,k,dAB;
  float diff,Sc,disA,disB,disAB;
  int ia, ja;
  float ScA, ScB;

 
  for (i=1;i<=a->Nali;++i) {
   Sc=0;
   a->scores[i]=0;
   if ((a->nalA[i] > 0)&&(a->nalB[i] >0)) {
    for (j=1;j<=a->Nali;++j)
     if (i != j && (a->nalA[j] > 0)&&(a->nalB[j] >0))
     { 
          kA = abs(a->nalA[j] - a->nalA[i]);
          kB = abs(a->nalB[j] - a->nalB[i]);
          if (kA>kB) k = kA; else k = kB;
          if (k>=1) 
          {
          if (21<=k) k = 0;
          dA = (int)DmapA->map[a->nalA[i]][a->nalA[j]];
          dB = (int)DmapB->map[a->nalB[i]][a->nalB[j]];
          if ((dA<50)&&(dB<50)) Sc += DisROM[k][dA][dB];
        }
     }
    /* self Score */
    ia=a->nalA[i];
    ScA=0;
    for (ja=1;ja<=pA->Naa;++ja)
      if (ia != ja)
      { 
        k  = abs(ja-ia);
        if (21<=k) k = 0;
        dA = (int)DmapA->map[ia][ja];
        if (dA<50) ScA += DisROM[k][dA][dA];
      }  

    ia=a->nalB[i];
    ScB=0;
    for (ja=1;ja<=pB->Naa;++ja)
      if (ia != ja)
      { 
        k  = abs(ja-ia);
        if (21<=k) k = 0;
        dA = (int)DmapB->map[ia][ja];
        if (dA<50) ScB += DisROM[k][dA][dA];
      }      

    if (ScA+ScB!=0)
      a->scores[i]=2*Sc/(ScA+ScB);
    else 
      printf("Error: ScA + ScB = 0.\n");
   }  
   /* a->scores[i]=Sc; */


  }

 
} /* end of ScoreDistanceSPair() */





float GapPenalty(a,Gext,Gopen)
 struct AlignNode *a; 
 float Gext,Gopen;
{
  int i,Nopen,Nexten;
  float pen; 
 
  Nopen = Nexten = 0;

  for (i=1;i<=a->Nali;++i)
   { 
    if (a->nalA[i]==-1)
     if (a->nalA[i+1]==-1) ++Nexten;  else ++Nopen;
    if (a->nalB[i]==-1)
     if (a->nalB[i+1]==-1) ++Nexten;  else ++Nopen;
    }

  pen = Nexten * Gext + Nopen * Gopen;
  return(pen);

} /* end of GapPenalty() */



float DisScore_with_PreAli(pA,pB,na,nb,pn,DmapA,DmapB,sn)
 struct PROTEIN *pA,*pB;
 int  na,nb;
 struct AlignNode *pn;  /* Previous Alignments */
 struct MAP2D *DmapA, *DmapB;
 struct AlignNode *sn;  /* Constraint of SSE Alignment */
{ int j,dA,dB,kA,kB,k,dAB;
  float Sc,disA,disB,disAB,ninf;
  
  Sc = 0.0;
  ninf = -100000000.0;
  /**
      Constraint 
  **/
  
  
 if ((PAR.ConOffset > 0) && (PAR.ConAtoB[na]>0) && (PAR.ConAtoB[na] == nb)) 
   Sc += PAR.ConOffset;  
  
  /*** (na,nb) is within the previous alignment,  return large penalty.  ***/
  if (PAR.SelfAvoidType == 'T')
    { if ((pn->nalA[1]<=nb)&&(nb <= pn->nalA[pn->Nali])) return(ninf);  
      if ((pn->nalB[1]<=na)&&(na <= pn->nalB[pn->Nali])) return(ninf);  }

  /*( If Alimap[na][nb] == 1 (masked), return large penalty. ***/
  if (PAR.mal_alimap == 1)  if (PAR.Alimap[na][nb]==1) return(ninf); 

  if (PAR.DisScType == 'N') /* Normal */
  { 
   for (j=1;j<=pn->Nali;++j)
    if ((pn->nalA[j] > 0)&&(pn->nalB[j] >0))
     if ((pn->nalA[j] !=na) && (pn->nalB[j] != nb)) 
      { 
         kA = pn->nalA[j] - na;
         if (kA <0) kA *= -1;
         kB = pn->nalB[j] - nb;
         if (kB <0) kB *= -1;
         if (kA>kB) k = kA; else k = kB;
         if (k>=1) 
          {
                if (21<=k)  k = 0;
            dA = (int)DmapA->map[na][pn->nalA[j]];
            dB = (int)DmapB->map[nb][pn->nalB[j]];
            if ((dA<50)&&(dB<50)) Sc += DisROM[k][dA][dB];
           }  
        } 
    Sc = 0.5 * Sc;
    }     

 
   else if (PAR.DisScType == 'D') /* "DALI" type */
    { 
    for (j=1;j<=pn->Nali;++j)
     if ((pn->nalA[j] > 0)&&(pn->nalB[j] >0))
      if ((pn->nalA[j] !=na) && (pn->nalB[j] != nb)) 
        { 
          disA  = DmapA->map[na][pn->nalA[j]];
          disB  = DmapB->map[nb][pn->nalB[j]];           
          disAB = (disA+disB)/2.0;
          Sc += (0.2 - fabs(disA-disB)/disAB) * exp(-disAB*disAB/400.0); }
     Sc = 0.5 * Sc;
     Sc += 0.2;  
    }
 
    else if (PAR.DisScType == 'S') /* "SSAP" type */
    {
     for (j=1;j<=pn->Nali;++j)
      if ((pn->nalA[j] > 0)&&(pn->nalB[j] >0))
       if ((pn->nalA[j] !=na) && (pn->nalB[j] != nb)) 
        { Sc += PAR.Assap/(fabs(DmapA->map[na][pn->nalA[j]]-DmapB->map[nb][pn->nalB[j]]) + PAR.Bssap); } 
     
     Sc = 0.5 * Sc;
     Sc += PAR.Assap/PAR.Bssap;  
    }
  

  if (PAR.SseOffsetD > 0.0) Sc += SseOffsetSc(pA,na,pB,nb,sn,'D'); 
  
  return(Sc); 

} /* end of ScoreDistanceS_with_PreAli() */




float SelfScoreDistanceS(pA,DmapA)
 struct PROTEIN *pA;
 struct MAP2D *DmapA;
{ int i,j,Ncomp,dA,k;
  float Sc,disA;
 
  Sc = 0.0;
 
  if (PAR.DisScType == 'N')  /* Normal */
  {

  for (i=1;i<=pA->Naa;++i)
    for (j=i+1;j<=pA->Naa;++j)
     { 
        k  = j-i;
        if (21<=k) k = 0;
        dA = (int)DmapA->map[i][j];
        if (dA<50) Sc += DisROM[k][dA][dA];
     }   
 }

 
  else if (PAR.DisScType == 'D') /* "DALI" type */
  {  
   for (i=1;i<=pA->Naa;++i)
    for (j=i;j<=pA->Naa;++j)
    {  
     if (i==j) {  Sc += 0.2;  }
      else 
     { 
        disA  = DmapA->map[i][j];
        Sc += 0.2 * exp(-disA*disA/400.0); 
      } 
    }   
 }


  else if (PAR.DisScType == 'S') /* "SSAP" type */
   { 
    for (i=1;i<=pA->Naa;++i)
     { 
      Sc += PAR.Assap/PAR.Bssap;   
      for (j=i+1;j<=pA->Naa;++j) Sc += PAR.Assap/PAR.Bssap;
      } 
   }
   
  return(Sc); 

} /* end of SelfScoreDistanceS() */






float Reliability_Rdis(Rdis,type)
 float Rdis;
 char  type; /* 'S'fam, f'O'ld */
{
 float Rel[21];
 int i,ind;
 float retval;
 
 for (i=0;i<21;++i) Rel[i] = 1.0;
 
 /*
 These value was calculated following conditions.
 DATE    : August 6, 2002 
 LIBRARY : 30scop59nm.list (3381 chains) 
 SCORE   : Rdis (averaged value) 
 Width of Rdis bin is 5 (%). 
 */ 

 if (type == 'S')
 { Rel[0]  = 0.003856; /* 0  (-0.000000) */
   Rel[1]  = 0.075562; /* 5  (5.000000) */
   Rel[2]  = 0.183593; /* 10 (10.000000) */
   Rel[3]  = 0.325770; /* 15 (15.000000) */
   Rel[4]  = 0.472321; /* 20 (20.000000) */
   Rel[5]  = 0.594417; /* 25 (25.000000) */
   Rel[6]  = 0.711550; /* 30 (30.000000) */
   Rel[7]  = 0.833670; /* 35 (35.000000) */
   Rel[8]  = 0.916733; /* 40 (40.000000) */
   Rel[9]  = 0.947270; /* 45 (45.000000) */
   Rel[10] = 0.965200; /* 50 (49.990002) */
   Rel[11] = 0.980108; /* 55 (55.000000) */
   Rel[12] = 0.985765; /* 60 (59.990002) */ }
 

 if (type== 'O')
 { Rel[0]  = 0.008515; /* 0 (-0.000000) */
   Rel[1]  = 0.159908; /* 5 (5.000000)  */
   Rel[2]  = 0.364705; /* 10 (10.000000) */
   Rel[3]  = 0.595756; /* 15 (15.000000) */
   Rel[4]  = 0.789474; /* 20 (20.000000) */
   Rel[5]  = 0.906078; /* 25 (25.000000) */
   Rel[6]  = 0.961702; /* 30 (30.000000) */
   Rel[7]  = 0.981070; /* 35 (35.000000) */
   Rel[8]  = 0.989294; /* 40 (40.000000) */
   Rel[9]  = 0.991935; /* 45 (45.000000) */
   Rel[10] = 0.993859; /* 50 (49.990002) */
   Rel[11] = 0.996383; /* 55 (55.000000) */
   Rel[12] = 0.996441; /* 60 (59.990002) */ }
 
      if (Rdis < 0)     retval = Rel[0];
 else if (Rdis >=100.0) retval = 1.0;
 else
 {
   ind = floor(Rdis/5.0); 
   retval = (Rel[ind+1]-Rel[ind])/5.0*(Rdis-ind*5) + Rel[ind];
  }

 return(100.0*retval);

} /* end of Reliability_Rdis() */




float Reliability_Zsc(Zsc,type)
 float Zsc;
 char  type; /* 'S'fam, f'O'ld */
{
 float Rel[100];
 int i,ind,Nmax;
 float retval,Zsc_min,Zsc_max,Wbin;
 
 /*
 These value was calculated following conditions.
 DATE    : August 6, 2002 
 LIBRARY : 30scop59nm.list (3381 chains) 
 SCORE   : Ncomp^2 fitted Zscore using Sdis 
 Width of Zsc bin is 1
 Min of Zsc is -10.
 Reliability for Zsc=99 is assumed as 100.0 %.
 */ 

 Nmax = 100;
 Wbin = 1.0;
 Zsc_min = -10.0;
 Zsc_max = Wbin*(Nmax-1) + Zsc_min;


 if (type == 'S')
 {
  Rel[0]  = 0.003634; /* -10 -10.000000 */
  Rel[1]  = 0.003632; /* -9 -9.000000 */
  Rel[2]  = 0.003630; /* -8 -8.000000 */
  Rel[3]  = 0.003627; /* -7 -7.000000 */
  Rel[4]  = 0.003625; /* -6 -6.000000 */
  Rel[5]  = 0.003621; /* -5 -5.000000 */
  Rel[6]  = 0.003617; /* -4 -4.000000 */
  Rel[7]  = 0.003616; /* -3 -3.000000 */
  Rel[8]  = 0.003626; /* -2 -2.000000 */
  Rel[9]  = 0.003665; /* -1 -1.000000 */
  Rel[10] = 0.007264; /* 0 0.000000 */
  Rel[11] = 0.042075; /* 1 1.000000 */
  Rel[12] = 0.091694; /* 2 2.000000 */
  Rel[13] = 0.156356; /* 3 3.000000 */
  Rel[14] = 0.233317; /* 4 4.000000 */
  Rel[15] = 0.319160; /* 5 5.000000 */
  Rel[16] = 0.410680; /* 6 6.000000 */
  Rel[17] = 0.498716; /* 7 7.000000 */
  Rel[18] = 0.577373; /* 8 8.000000 */
  Rel[19] = 0.643564; /* 9 9.000000 */
  Rel[20] = 0.700164; /* 10 10.000000 */
  Rel[21] = 0.737785; /* 11 11.000000 */
  Rel[22] = 0.771492; /* 12 12.000000 */
  Rel[23] = 0.798257; /* 13 13.000000 */
  Rel[24] = 0.824177; /* 14 14.000000 */
  Rel[25] = 0.846133; /* 15 15.000000 */
  Rel[26] = 0.863153; /* 16 16.000000 */
  Rel[27] = 0.877940; /* 17 17.000000 */
  Rel[28] = 0.892685; /* 18 18.000000 */
  Rel[29] = 0.903900; /* 19 19.000000 */
  Rel[30] = 0.912722; /* 20 20.000000 */
  Rel[31] = 0.919452; /* 21 21.000000 */
  Rel[32] = 0.924374; /* 22 22.000000 */
  Rel[33] = 0.927255; /* 23 23.000000 */
  Rel[34] = 0.928920; /* 24 24.000000 */
  Rel[35] = 0.933532; /* 25 25.000000 */
  Rel[36] = 0.935963; /* 26 26.000000 */
  Rel[37] = 0.939080; /* 27 27.000000 */
  Rel[38] = 0.940135; /* 28 28.000000 */
  Rel[39] = 0.941217; /* 29 29.000000 */
  Rel[40] = 0.942370; /* 30 29.969999 */
  Rel[41] = 0.943065; /* 31 31.000000 */
  Rel[42] = 0.943997; /* 32 32.000000 */
  Rel[43] = 0.944020; /* 33 33.000000 */
  Rel[44] = 0.943700; /* 34 33.990002 */
  Rel[45] = 0.943629; /* 35 34.990002 */
  Rel[46] = 0.943641; /* 36 36.000000 */
  Rel[47] = 0.943803; /* 37 36.990002 */
  Rel[48] = 0.943925; /* 38 38.000000 */
  Rel[49] = 0.941614; /* 39 38.959999 */
 } 


 if (type== 'O')
 {
  Rel[0]  = 0.008065; /* -10 -10.000000 */
  Rel[1]  = 0.008060; /* -9 -9.000000 */
  Rel[2]  = 0.008054; /* -8 -8.000000 */
  Rel[3]  = 0.008047; /* -7 -7.000000 */
  Rel[4]  = 0.008038; /* -6 -6.000000 */
  Rel[5]  = 0.008023; /* -5 -5.000000 */
  Rel[6]  = 0.008006; /* -4 -4.000000 */
  Rel[7]  = 0.007982; /* -3 -3.000000 */
  Rel[8]  = 0.007974; /* -2 -2.000000 */
  Rel[9]  = 0.007994; /* -1 -1.000000 */
  Rel[10] = 0.015284; /* 0 0.000000 */
  Rel[11] = 0.085823; /* 1 1.000000 */
  Rel[12] = 0.179471; /* 2 2.000000 */
  Rel[13] = 0.290815; /* 3 3.000000 */
  Rel[14] = 0.407713; /* 4 4.000000 */
  Rel[15] = 0.516732; /* 5 5.000000 */
  Rel[16] = 0.616091; /* 6 6.000000 */
  Rel[17] = 0.699037; /* 7 7.000000 */
  Rel[18] = 0.761580; /* 8 8.000000 */
  Rel[19] = 0.810265; /* 9 9.000000 */
  Rel[20] = 0.848359; /* 10 10.000000 */
  Rel[21] = 0.875233; /* 11 11.000000 */
  Rel[22] = 0.895777; /* 12 12.000000 */
  Rel[23] = 0.912422; /* 13 13.000000 */
  Rel[24] = 0.924771; /* 14 14.000000 */
  Rel[25] = 0.937042; /* 15 15.000000 */
  Rel[26] = 0.946560; /* 16 16.000000 */
  Rel[27] = 0.954407; /* 17 17.000000 */
  Rel[28] = 0.961353; /* 18 18.000000 */
  Rel[29] = 0.966971; /* 19 19.000000 */
  Rel[30] = 0.971240; /* 20 20.000000 */
  Rel[31] = 0.975150; /* 21 21.000000 */
  Rel[32] = 0.977677; /* 22 22.000000 */
  Rel[33] = 0.978177; /* 23 23.000000 */
  Rel[34] = 0.978445; /* 24 24.000000 */
  Rel[35] = 0.981281; /* 25 25.000000 */
  Rel[36] = 0.983419; /* 26 26.000000 */
  Rel[37] = 0.985075; /* 27 27.000000 */
  Rel[38] = 0.987126; /* 28 28.000000 */
  Rel[39] = 0.987968; /* 29 29.000000 */
  Rel[40] = 0.990576; /* 30 29.969999 */
  Rel[41] = 0.991593; /* 31 31.000000 */
  Rel[42] = 0.993151; /* 32 32.000000 */
  Rel[43] = 0.993215; /* 33 33.000000 */
  Rel[44] = 0.993298; /* 34 33.990002 */
  Rel[45] = 0.994315; /* 35 34.990002 */
  Rel[46] = 0.995012; /* 36 36.000000 */
  Rel[47] = 0.995798; /* 37 36.990002 */
  Rel[48] = 0.997251; /* 38 38.000000 */
  Rel[49] = 0.997138; /* 39 38.959999 */
  }

  for (i=50;i<Nmax;++i) Rel[i] = Rel[49] + (i-49)*(1.0 - Rel[49])/(Nmax-50); 

      if (Zsc <  Zsc_min)  retval = Rel[0];
 else if (Zsc >= Zsc_max)  retval = Rel[Nmax-1];
 else
 {
   ind = (int)floor((Zsc-Zsc_min)/Wbin); 
   retval = (Rel[ind+1]-Rel[ind])/Wbin*(Zsc- ind*Wbin - Zsc_min) + Rel[ind];
  }

 /*
 printf("#Rel_Zsc Zsc %f ind %d retval %f\n",Zsc,ind,retval);
 */ 
 return(100.0*retval);

} /* end of Reliability_Zsc() */


