/*

 <ScoreSSE.c>

 Score function between SSE pairs    

 coded by Takeshi Kawabata.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "main.h" 
#include "protein.h" 
#include "ScoreENV.h" 
#include "SecStrEle.h" 
#include "LlistAlign.h" 

 
/*** Variables (LOCAL) ***/
float LsseROM[2][50][50]; /* Residue Length "L"  0:Helix 1:Sheet   */
float DsseROM[3][50][50]; /* Distance 0:H-H, 1:E-E, 2:H-E,E-H      */ 
float PsseROM[3][36][36]; /* Dihedral Angle 0:H-H, 1:E-E, 2:H-E,E-H */ 
float TsseROM[3][17][17]; /* Bond Angle 0:H-H, 1:E-E, 2:H-E,E-H */ 


/*** Function(GLOBAL) ***/
void Read_All_SSE_Score();
float ScoreSSE();
float ScoreSSE_AddPair();
float SelfScoreSSE();
float SseOffsetSc();
float AllSseOffsetSc();
float SSE_Gap_Penalty();

/*** Function(LOCAL) ***/
static void Read_SseROM();


void Read_All_SSE_Score(tailstr)
 char *tailstr;
{
 char fname[256];
 int s,t,k;
 
 /** INITIALIZE **/
 for (k=0;k<2;++k)
  for (s=0;s<50;++s) for (t=0;t<50;++t) LsseROM[k][s][t] = 0.0;

 for (k=0;k<3;++k)
  for (s=0;s<50;++s) for (t=0;t<50;++t)  DsseROM[k][s][t] = 0.0;

 for (k=0;k<3;++k)
  for (s=0;s<36;++s) for (t=0;t<36;++t)  PsseROM[k][s][t] = 0.0;

 for (k=0;k<3;++k)
  for (s=0;s<17;++s) for (t=0;t<17;++t)  TsseROM[k][s][t] = 0.0;
 
 /** READ SCORE **/ 
 sprintf(fname,"%s/Lh-%s.rom",SCORE_DIR,tailstr);  Read_SseROM(fname);
 sprintf(fname,"%s/Le-%s.rom",SCORE_DIR,tailstr);  Read_SseROM(fname);
 
 sprintf(fname,"%s/Dhh-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);
 sprintf(fname,"%s/Dee-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);
 sprintf(fname,"%s/Dhe-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);

 sprintf(fname,"%s/Phh-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);
 sprintf(fname,"%s/Pee-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);
 sprintf(fname,"%s/Phe-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);

 sprintf(fname,"%s/Thh-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);
 sprintf(fname,"%s/Tee-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);
 sprintf(fname,"%s/The-%s.rom",SCORE_DIR,tailstr); Read_SseROM(fname);

} /* end of Read_All_SSE_Score() */



float ScoreSSE(pA,pB,sn)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn; 
{ int i,j,s,t,u,v,sec;
  int ai,bi,aj,bj; 
  float S;
 
  S = 0.0;
 
  /*** ONE-BODY ***/ 
  for (i=0;i<sn->Nali;++i)
  {
    ai = sn->nalA[i]; bi = sn->nalB[i];
    if (pA->SSE[ai].type == pB->SSE[bi].type)
     {
      s = pA->SSE[ai].L; if (s>49) s = 49; 
      t = pB->SSE[bi].L; if (t>49) t = 49; 
      if (pA->SSE[ai].type == 'H') sec = 0; else sec = 1;
      S += LsseROM[sec][s][t]; }    
   }

  /*** TWO-BODY ***/ 
 
 for (i=0;i<sn->Nali;++i)
  {
   ai = sn->nalA[i]; bi = sn->nalB[i];
   if (pA->SSE[ai].type == pB->SSE[bi].type)
   for (j=i+1;j<sn->Nali;++j)
   { 
    aj = sn->nalA[j]; bj = sn->nalB[j];
    
    if (pA->SSE[aj].type == pB->SSE[bj].type)
   { 
          if ((pA->SSE[ai].type =='H')&&(pA->SSE[aj].type == 'H')) sec = 0; 
     else if ((pA->SSE[ai].type =='E')&&(pA->SSE[aj].type == 'E')) sec = 1; 
     else sec = 2;
    
     s = (int)(pA->sseDi[ai][aj]); 
     if (s<0) s = 0; if (s>49) s = 49;
     t = (int)(pB->sseDi[bi][bj]); 
     if (t<0) t = 0; if (t>49) t = 49;
     S += DsseROM[sec][s][t]; 
     
     s = (int)((pA->ssePh[ai][aj]+180.0)/10.0); 
     if (s<0) s = 0; if (s>35) s = 35;
     t = (int)((pB->ssePh[bi][bj]+180.0)/10.0); 
     if (t<0) t = 0; if (t>35) t = 35;
     S+= PsseROM[sec][s][t];   
 
     s = (int)((pA->sseT1[ai][aj]-5.0)/10.0);
     if (s<0) s = 0; if (s>16) s = 16;
     t = (int)((pB->sseT1[bi][bj]-5.0)/10.0);
     if (t<0) t = 0; if (t>16) t = 16;
     S += TsseROM[sec][s][t]; 
     
     s = (int)((pA->sseT2[ai][aj]-5.0)/10.0);
     if (s<0) s = 0; if (s>16) s = 16;
     t = (int)((pB->sseT2[bi][bj]-5.0)/10.0);
     if (t<0) t = 0; if (t>16) t = 16;
     S += TsseROM[sec][s][t]; 
   }
 
   } /* j */

  } /* i */

  return(S);

} /* end of  ScoreSSE() */







float ScoreSSE_AddPair(pA,pB,sn,newA,newB)
 struct PROTEIN *pA,*pB;
 struct AlignNode *sn; /* SSE Alignment */ 
 int    newA,newB;   /* newly added SSE number */
{ int j,s,t,aj,bj,sec;
  float S;
 
 if ((newA<0)||(pA->Nsse <= newA)) {printf("#ERROR:newA %d is out of range\n",newA); exit(1);}
 if ((newB<0)||(pB->Nsse <= newB)) {printf("#ERROR:newB %d is out of range\n",newA); exit(1);}
 
  S = 0.0;
  s = pA->SSE[newA].L; if (s>49) s = 49; 
  t = pB->SSE[newB].L; if (t>49) t = 49; 
  if (pA->SSE[newA].type == 'H') sec = 0; else sec = 1;      
  S += LsseROM[sec][s][t];     
 
 for (j=0;j<sn->Nali;++j)
 {
  aj = sn->nalA[j]; bj = sn->nalB[j];

  if (pA->SSE[aj].type == pB->SSE[bj].type)
   { 
          if ((pA->SSE[newA].type =='H')&&(pA->SSE[aj].type == 'H')) sec = 0; 
     else if ((pA->SSE[newA].type =='E')&&(pA->SSE[aj].type == 'E')) sec = 1; 
     else sec = 2;
    
 
     s = (int)(pA->sseDi[aj][newA]); 
     if (s<0) s = 0; if (s>49) s = 49;
     t = (int)(pB->sseDi[bj][newB]); 
     if (t<0) t = 0; if (t>49) t = 49;
     S += DsseROM[sec][s][t];    
     

     s = (int)((pA->ssePh[newA][aj]+180)/10.0); 
     if (s<0) s = 0; if (s>35) s = 35;
     t = (int)((pB->ssePh[newB][bj]+180)/10.0); 
     if (t<0) t = 0; if (t>35) t = 35;
     S += PsseROM[sec][s][t];    
    
      s = (int)((pA->sseT1[newA][aj]-5.0)/10.0);
      if (s<0) s = 0; if (s>16) s = 16;
      t = (int)((pB->sseT1[newB][bj]-5.0)/10.0);
      if (t<0) t = 0; if (t>16) t = 16;
      S += TsseROM[sec][s][t];

      s = (int)((pA->sseT2[newA][aj]-5.0)/10.0);
      if (s<0) s = 0; if (s>16) s = 16;
      t = (int)((pB->sseT2[newB][bj]-5.0)/10.0);
      if (t<0) t = 0; if (t>16) t = 16;
      S += TsseROM[sec][s][t];
 
   }

  } /* j loop */

  return(S);

} /* end of ScoreSSE_AddPair() */ 


float SelfScoreSSE(pr)
 struct PROTEIN *pr;
{
 struct AlignNode sn;  
 int i;
 float S;
 Malloc_AlignNode(&sn,pr->Nsse,'-');
 for (i=0;i<pr->Nsse;++i) sn.nalA[i] = sn.nalB[i] = i;
 sn.Nali = pr->Nsse; 
 S = ScoreSSE(pr,pr,&sn);
 Free_AlignNode(&sn);
 return(S);
} /* end of SelfScoreSSE() */



void Read_SseROM(fname)
 char *fname;
{
 FILE *fp;
 int s,t,L,snum;
 char buffer[128],buff1[64],buff2[64];
 char rom_on,type,sectype[8];
 float val;
 /* <FORMAT>
 @CLASS SSE
 @TYPE T
 @SECTYPE he
 @NSTATE 17
 @NMULTI 3
 @SIGMA  100.000000
 @RELATED_ODDS_MATRIX
 0 0 35.859808
 0 1 28.699575
 :
 */

 fp = fopen(fname,"r");
 if (fp==NULL) { printf("#ERROR:Can't find ssescorefile \"%s\"\n",fname); exit(1);}
 rom_on = 0;
 while (feof(fp)==0)
 {
  buffer[0] = '\0';
  fgets(buffer,128,fp);
  L = strlen(buffer);
  if (buffer[L-1]=='\n') buffer[L-1]= '\0';
  if ((buffer[0]!='#')&&(L>1))
  {
   if (buffer[0]=='@') rom_on = 0;
   if (strncmp(buffer,"@TYPE",5)==0)
    { sscanf(buffer,"%s %s",buff1,buff2); type = buff2[0];}
   
   if (strncmp(buffer,"@SECTYPE",8)==0)
    { sscanf(buffer,"%s %s",buff1,sectype); 
           if (strncmp(sectype,"hh",2)==0) snum = 0;
      else if (strncmp(sectype,"ee",2)==0) snum = 1;
      else if (strncmp(sectype,"he",2)==0) snum = 2;
      else if (strncmp(sectype,"h",1)==0)  snum = 0;
      else if (strncmp(sectype,"e",1)==0)  snum = 1; 
      else {printf("#ERROR:SECTYPE %s is not proper\n",sectype); exit(1);}; 
    }

   if (rom_on==1)
   { sscanf(buffer,"%d %d %f",&s,&t,&val);
     if ((type=='L')&&(s<0)&&(s>=50)) {printf("#ERROR:out of range s %d \n",s); exit(1);}
     if ((type=='D')&&(s<0)&&(s>=50)) {printf("#ERROR:out of range s %d \n",s); exit(1);}
     if ((type=='P')&&(s<0)&&(s>=36)) {printf("#ERROR:out of range s %d \n",s); exit(1);}
     if ((type=='T')&&(s<0)&&(s>=17)) {printf("#ERROR:out of range s %d \n",s); exit(1);}
     if ((type=='L')&&(t<0)&&(t>=50)) {printf("#ERROR:out of range t %d \n",s); exit(1);}
     if ((type=='D')&&(t<0)&&(t>=50)) {printf("#ERROR:out of range t %d \n",s); exit(1);}
     if ((type=='P')&&(t<0)&&(t>=36)) {printf("#ERROR:out of range t %d \n",s); exit(1);}
     if ((type=='T')&&(t<0)&&(t>=17)) {printf("#ERROR:out of range t %d \n",s); exit(1);}

          if (type=='L') LsseROM[snum][s][t] = LsseROM[snum][t][s] = val;
     else if (type=='D') DsseROM[snum][s][t] = DsseROM[snum][t][s] = val;
     else if (type=='P') PsseROM[snum][s][t] = PsseROM[snum][t][s] = val;
     else if (type=='T') TsseROM[snum][s][t] = TsseROM[snum][t][s] = val;
   }

    if (strncmp(buffer,"@RELATED_ODDS_MATRIX",20)==0) rom_on = 1;
   }

  } /* while */
 
  fclose(fp);

} /* end of Read_SseROM() */





float SseOffsetSc(pA,i,pB,j,sn,type)
 struct PROTEIN *pA,*pB;
 int i,j;
 struct AlignNode *sn; /* Constraint of SSE alignment */
 char type;
{
 int match,k;
 float Smatch,Snotmatch,sc;
 
 if (type == 'D') 
 { Smatch = PAR.SseOffsetD; Snotmatch = PAR.NonSseOffsetD; } 
 else
 { Smatch = PAR.SseOffsetE; Snotmatch = PAR.NonSseOffsetE; } 

 sc = Snotmatch;

 if ((Smatch>0.0)||(Snotmatch>0.0))
 { 
  match = 0; k = 0; 
 
  while ((match==0)&&(k<sn->Nali))
  {
    if ((pA->SSE[sn->nalA[k]].Rsta<=i)&&(i<=pA->SSE[sn->nalA[k]].Rend)&&
        (pB->SSE[sn->nalB[k]].Rsta<=j)&&(j<=pB->SSE[sn->nalB[k]].Rend) )
        { sc = Smatch; match = 1;}
    else ++k;
   }
 }

 return(sc);

} /* end of SseOffsetSc() */



float AllSseOffsetSc(pA,pB,an,sn,type)
 struct PROTEIN *pA,*pB;
 struct AlignNode *an; /* alignment */
 struct AlignNode *sn; /* Constraint of SSE alignment */
 char type;
{
 int i;
 float sc;

 sc = 0.0;
 for (i=1;i<=an->Nali;++i)
 {
  if ((an->nalA[i]>0)&&(an->nalB[i]>0))
   sc += SseOffsetSc(pA,an->nalA[i],pB,an->nalB[i],sn,type); 
  }
 return(sc);

} /* end of AllSseOffsetSc() */




float SSE_Gap_Penalty(pA,pB,sn)
  struct PROTEIN *pA,*pB;
  struct AlignNode *sn;  /* SSE Alignment */
{ int a,a0,b,b0,i,k,gapA,gapB,Ngap;
  float S;

  /****
   Gap Penalty is assigned as :

   GapExtS * abs( Difference between residues of gapped SSE )

  This function is defined only for alignment for >= 2 sse pairs

  ****/

 if (sn->Nali==1) return(0.0);

  Ngap = 0;
  a0 = sn->nalA[0]; b0 = sn->nalB[0];
  for (k=1;k<sn->Nali;++k)
  {
    a = sn->nalA[k];
    b = sn->nalB[k];
    gapA = gapB = 0;
    if ((a-a0)>1)
     for (i=(a0+1);i<a;++i) gapA += pA->SSE[i].L;

    if ((b-b0)>1)
     for (i=(b0+1);i<b;++i) gapB += pB->SSE[i].L;
    /*
    if ((a-a0)>1)
     gapA += pA->SSE[a].Rsta - pA->SSE[a0].Rend;
    if ((b-b0)>1)
     gapA += pB->SSE[b].Rsta - pB->SSE[b0].Rend;
     */

    Ngap += abs(gapA-gapB);

    /*
    printf("%d a %d b %d gapA %d gapB %d %d\n",k,a,b,gapA,gapB,abs(gapA-gapB));
    */

   a0 =a; b0 = b;

 } /* k */

  S = PAR.GapExtS * Ngap;
  

  return(S);

} /* end of SSE_Gap_Penalty() */

