/*

 <SecStrEle.c>
 
 coded by Takeshi Kawabata.
  
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "main.h"
#include "protein.h"
#include "SecStrEle.h"
#include "Amino.h"
#include "Vector.h"
#include "DiheDis.h"
#include "Jacobi3.h"


static void IncAddVec();
static void IncMultiVec();
static void Cal_Dis_bwn_SSE();
static void Cal_G_SSE();
static void Cal_V_SSE_Inertia();
static void Cal_PbPe_SSE();
static void B_tV();
static float Dis_bwn_Lines();
static float Dis_Seg_and_Pt();
static float Dis_bwn_Segments();

void Set_SSE();
void Set_Dis_bwn_SSE();
void Output_SSE_PDB();


void Set_SSE(pr)
 struct PROTEIN *pr;
{ int i,j,k,m,hon,eon,h0,h1,e0,e1,Nhel,Nshe;
  char buff[16];

  pr->Nsse = hon  = eon = 0; 
  Nhel = Nshe = 0;

  for (i=1;i<= pr->Naa ;++i)
   {
    if ((pr->Sec[i] == 'H')&&(hon == 0)) 
      { hon = 1; pr->SSE[pr->Nsse].Rsta = i; pr->SSE[pr->Nsse].type = 'H';
        ++Nhel; pr->SSE[pr->Nsse].num = Nhel; }
   
    if (( ((i<pr->Naa)&&(pr->Sec[i+1] != 'H'))
          ||(i== pr->Naa)||(pr->Break[i]==1))&&(hon == 1))
    { hon = 0;   pr->SSE[pr->Nsse].Rend = i; 
      if (pr->SSE[pr->Nsse].Rend > pr->SSE[pr->Nsse].Rsta) ++(pr->Nsse); 
      if (pr->Nsse >= pr->MaxNsse) 
       {printf("#ERROR:Nsse %d is over MaxNsse %d\n",pr->Nsse,pr->MaxNsse); exit(1);}
      } 

    if ((pr->Sec[i] == 'E')&&(eon == 0)) 
     { eon = 1; pr->SSE[pr->Nsse].Rsta = i; pr->SSE[pr->Nsse].type = 'E';
     ++Nshe; pr->SSE[pr->Nsse].num = Nshe; }

    if ((((i<pr->Naa)&&(pr->Sec[i+1] != 'E'))
        ||(i== pr->Naa)||(pr->Break[i]==1))&&(eon == 1))
    { eon = 0;   pr->SSE[pr->Nsse].Rend = i; 
      if (pr->SSE[pr->Nsse].Rend > pr->SSE[pr->Nsse].Rsta) ++(pr->Nsse);   
      if (pr->Nsse >= pr->MaxNsse) 
       {printf("#ERROR:Nsse %d is over MaxNsse %d\n",pr->Nsse,pr->MaxNsse); exit(1);}
     } 
   } /* i loop */

 
  for (i=0;i < pr->Nsse ;++i) 
   { 
    Cal_G_SSE(pr,&(pr->SSE[i])); 
    Cal_V_SSE_Inertia(pr,&(pr->SSE[i]));  
    Cal_PbPe_SSE(pr,&(pr->SSE[i])); 
  }


  /* Making Sse sequnece */
  for (i=1;i<= pr->Naa ;++i) pr->SseSeq[i] = ' ';

  for (j=0;j<pr->Nsse;++j)
  {
    sprintf(buff,"%c%d",pr->SSE[j].type,pr->SSE[j].num);
    for (k=0;k<strlen(buff);++k) 
     {  m = pr->SSE[j].Rsta+k;
        if ((1<=m)&&(m<=pr->Naa))pr->SseSeq[m] = buff[k]; }
   }

} /* end of Set_SSE() */



  

void Cal_G_SSE(pr,s)
 struct PROTEIN *pr;
 struct SecStrEle *s;
{  int i;

    s->G[0] =  s->G[1] =  s->G[2] = 0.0; 
    for (i= s->Rsta; i<= s->Rend ; ++i)  
      IncAddVec(3,s->G,pr->PosCA[i]); 
    s->L = s->Rend - s->Rsta + 1;
    IncMultiVec(3,s->G,1.0/s->L); 
}



void Cal_V_SSE_Inertia(pr,s)
 struct PROTEIN *pr;
 struct SecStrEle *s;
{
  double Smat[3][3],Vmat[3][3];
  float IVmat[3][3],D[3],SV[3];
  int i,j,k;

  /******* Cal Inertial Tensor ******/
  for (i=0;i<3;++i)
   for (j=0;j<3;++j) Smat[i][j] = 0.0;
   for (k= s->Rsta; k<= s->Rend ; ++k)
    {
     Sub_Vec(3,D,pr->PosCA[k],s->G);
     Smat[0][0] += D[1]*D[1] + D[2]*D[2];
     Smat[1][1] += D[0]*D[0] + D[2]*D[2];
     Smat[2][2] += D[0]*D[0] + D[1]*D[1];
     for (i=0;i<3;++i)
      for (j=0;j<3;++j) if (i!=j) Smat[i][j] += -D[i]*D[j];
    }

   for (i=0;i<3;++i) 
    for (j=0;j<3;++j)  Smat[i][j] /= s->L;

  /****** Calculate Eigen Vector ******/
  Jacobi_Wilkinson3(Smat,Vmat);
  
 
  /******* Choose the eigen vector with mimimum eigen value *****/  
  for (i=0;i<3;++i)
    for (j=0;j<3;++j)  IVmat[i][j] = Vmat[j][i];

        if ((Smat[0][0]<=Smat[1][1]) && (Smat[0][0]<=Smat[2][2]))
          Equal_Vec(3,s->V,IVmat[0]);
   else if ((Smat[1][1]<=Smat[0][0]) && (Smat[1][1]<=Smat[2][2]))
            Equal_Vec(3,s->V,IVmat[1]);
   else if ((Smat[2][2]<=Smat[1][1]) && (Smat[2][2]<=Smat[0][0]))
            Equal_Vec(3,s->V,IVmat[2]);

  /****** Judge Direction using Rend-Rsta vector  ******/
   Sub_Vec(3,SV,pr->PosCA[s->Rend],pr->PosCA[s->Rsta]);
   Normalize(3,SV);
   if (Dot_Prod(3,s->V,SV) < 0.0 ) IncMultiVec(3,s->V,-1.0);

} /* end of Cal_V_SSE_Inertia() */



void Cal_PbPe_SSE(pr,s)
 struct PROTEIN *pr;
 struct SecStrEle *s; 
{
  float Ve[3],Vb[3],X[3];
  float prodE,prodB; 
  
  Sub_Vec(3,Vb,pr->PosCA[s->Rsta],s->G);
  prodB = Dot_Prod(3,s->V,Vb);
  B_tV(3,s->Pb,   s->G,prodB,s->V);

  Sub_Vec(3,Ve,pr->PosCA[s->Rend],s->G);
  prodE = Dot_Prod(3,s->V,Ve);
  B_tV(3,s->Pe,   s->G,prodE,s->V);
  
  Sub_Vec(3,s->Veb,s->Pe,s->Pb);
  /*
  B_tV(3,s->Pefar,s->G, 2.0, s->Veb);
  */
  B_tV(3,s->Pefar,s->G, 100.0, s->Veb);
  s->M[0] = (s->Pb[0] + s->Pe[0])/2.0;
  s->M[1] = (s->Pb[1] + s->Pe[1])/2.0;
  s->M[2] = (s->Pb[2] + s->Pe[2])/2.0;

} /* end of Cal_PbPe_SSE() */ 




void Set_Dis_bwn_SSE(pr,type)
 struct PROTEIN *pr;
 char type;
{
  int i,j,sd1,sd2;
  float tA[3],tB[3],Dab[3],Dba[3];
  
  for (i=0;i< pr->Nsse; ++i)
   for (j=0;j< pr->Nsse; ++j)
    if (i!=j)
    { 
      if (type=='M')  /* Define using midpoint-midpoint line */
       {
        pr->sseDi[i][j]  = Distance(3,pr->SSE[i].M,pr->SSE[j].M);
        pr->ssePh[i][j]  = Dihedral_Angle(pr->SSE[i].Pe,pr->SSE[i].M,
                                        pr->SSE[j].M,pr->SSE[j].Pe);
        pr->sseT1[i][j]  = Bond_Angle(pr->SSE[i].Pe,pr->SSE[i].M,pr->SSE[j].M); 
        pr->sseT2[i][j]  = Bond_Angle(pr->SSE[i].M,pr->SSE[j].M,pr->SSE[j].Pe); 
       } 
     
      if (type=='C')  /* Define using Closest-points line */
       { 
        pr->sseDi[i][j] = Dis_bwn_Segments(pr->SSE[i].Pb,pr->SSE[i].Pe,pr->SSE[i].Veb,
                                          pr->SSE[j].Pb,pr->SSE[j].Pe,pr->SSE[j].Veb,
                                          tA,tB);
        pr->ssePh[i][j]  = Dihedral_Angle(pr->SSE[i].Pefar,tA,tB,pr->SSE[j].Pefar);
        pr->sseT1[i][j]  = Bond_Angle(pr->SSE[i].Pefar,tA,tB); 
        pr->sseT2[i][j]  = Bond_Angle(tA,tB,pr->SSE[j].Pefar); 
        }
      
           if (pr->ssePh[i][j] < -180.0 ) pr->ssePh[i][j] += 360.0;
      else if (pr->ssePh[i][j] >  180.0 ) pr->ssePh[i][j] -= 360.0;
      
       sd1 = pr->SSE[i].Rsta - pr->SSE[j].Rend - 1;
       sd2 = pr->SSE[j].Rsta - pr->SSE[i].Rend - 1;

   }

  /* 
  for (i=0;i< pr->Nsse; ++i)
  for (j=i+1;j< pr->Nsse; ++j)
   printf("%d %d T1 %lf %lf T2 %lf  %lf\n",i,j,
    pr->sseT1[i][j],pr->sseT2[j][i],pr->sseT2[i][j],pr->sseT1[j][i]);   
 */

} /* end of  Set_Dis_bwn_SSE() */




float Dis_bwn_Segments(beA,enA,veA,beB,enB,veB,tA,tB)
 float beA[3],enA[3],veA[3],beB[3],enB[3],veB[3],tA[3],tB[3]; 
{ float sAl,sBl,Dl,sA,sB,dis,Dmin;
  float Dpl[4],sApl[4],sBpl[4]; 
  int k,mink; 

  Dl = Dis_bwn_Lines(beA,veA,beB,veB,&sAl,&sBl);
  if ((0.0<=sAl)&&(sAl<=1.0)&&(0.0<=sBl)&&(sBl<=1.0))
  { sA = sAl; sB = sBl; }  
  else
  {
   Dpl[0] =  Dis_Seg_and_Pt(beA,veA,beB,&sApl[0]); sBpl[0] = 0.0;
   Dpl[1] =  Dis_Seg_and_Pt(beA,veA,enB,&sApl[1]); sBpl[1] = 1.0;
   Dpl[2] =  Dis_Seg_and_Pt(beB,veB,beA,&sBpl[2]); sApl[2] = 0.0;
   Dpl[3] =  Dis_Seg_and_Pt(beB,veB,enA,&sBpl[3]); sApl[3] = 1.0;
   Dmin = 100000000.0; 
   for (k=0;k<4;++k) if (Dmin>Dpl[k]) {Dmin = Dpl[k]; mink = k;} 
   sA = sApl[mink]; sB = sBpl[mink];
  }

  B_tV(3,tA,beA,sA,veA);
  B_tV(3,tB,beB,sB,veB);
  dis = Distance(3,tA,tB);
  return(dis);

} /* end of Dis_bwn_Segments() */





float Dis_bwn_Lines(beA,veA,beB,veB,sA,sB)
 float beA[3],veA[3],beB[3],veB[3]; 
      /*  be:beginning point,  ve: direction vector */
 float *sA,*sB;
{ float a,b,c,d,e,f,det,dis;
  float tA[3],tB[3]; 
 
  a =   Dot_Prod(3,veA,veA);   b = - Dot_Prod(3,veA,veB); 
  c = - Dot_Prod(3,veB,veA);   d =   Dot_Prod(3,veB,veB); 
  e = - Dot_Prod(3,veA,beA) + Dot_Prod(3,veA,beB); 
  f =   Dot_Prod(3,veB,beA) - Dot_Prod(3,veB,beB); 

  det = a*d - b*c;

  if (det!=0.0)
   { *sA =  ( d*e - b*f)/det; 
     *sB =  (-c*e + a*f)/det; } 
  else *sA = *sB = 0.5;

 B_tV(3,tA,beA,*sA,veA);
 B_tV(3,tB,beB,*sB,veB);
 dis = Distance(3,tA,tB); 
 return(dis);

}/* end of Dis_bwn_Lines() */



float Dis_Seg_and_Pt(be,ve,X,s)
 float be[3],ve[3],X[3],*s;
{
 float T[3],d;

 *s = (Dot_Prod(3,X,ve) - Dot_Prod(3,be,ve))/Dot_Prod(3,ve,ve);
      if ( *s<0.0) *s = 0.0;
 else if ( *s>1.0) *s = 1.0;
 B_tV(3,T,be,*s,ve);
 d = Distance(3,T,X); 
 return(d);
}





void IncAddVec(n,A,B)
 int n;
 float *A,*B;
{ int i;
  for (i=0;i<n;++i) A[i] += B[i]; }


void IncMultiVec(n,A,x)
 int n;
 float *A,x;
{ int i;
  for (i=0;i<n;++i) A[i] *= x; }


void B_tV(n,X,B,t,V)
 int n;
 float *X,*B,t,*V;
{ int i;
  for (i=0;i<n;++i) X[i] = B[i] + t * V[i]; }



void Output_SSE_PDB(pr,fname,ch)
 struct PROTEIN *pr;
 char *fname;
 char ch;
{
 int i,j,a,r,natom;
 float X[3],T;
 char rname[5];
 FILE *fp;

 printf("#Output_SSE_PDB() --> \"%s\"\n",fname);

 fp = fopen(fname,"w");
 if (fp==NULL) {printf("#ERROR:Can't write to ssepdbfile \"%s\"\n",fname); exit(1);}

 fprintf(fp,"REMARK SSE in PDBformat for %s\n",pr->pname);
 fprintf(fp,"REMARK tFactor : 0 for Cend of SSE, 200 for  helix, 100 for sheet\n",pr->pname);
 natom = 0;
 for (i=0;i<pr->Nsse;++i)
  for (j=0;j<pr->SSE[i].L ;++j) 
  {
    X[0] = pr->SSE[i].Pb[0] + pr->SSE[i].Veb[0] * j/pr->SSE[i].L;  
    X[1] = pr->SSE[i].Pb[1] + pr->SSE[i].Veb[1] * j/pr->SSE[i].L;  
    X[2] = pr->SSE[i].Pb[2] + pr->SSE[i].Veb[2] * j/pr->SSE[i].L; 
         if (j == (pr->SSE[i].L-1)) T = 0.0; 
    else if (pr->SSE[i].type=='H')  T = 200.0; 
    else if (pr->SSE[i].type=='E')  T = 100.0; 
    
    r = pr->SSE[i].Rsta + j; 
    Get_TriAmino(pr->Seq[r],rname);
    /*
    fprintf(fp,"ATOM   %5s CA  %3s %c%5s   %8.3lf%8.3lf%8.3lf  1.00%6.2lf\n",
          pr->Rnum[r],rname,ch,pr->Rnum[r],X[0],X[1],X[2],T);
    */  
    ++natom;
    fprintf(fp,"ATOM   %4d  CA  %3s %c%4s    %8.3lf%8.3lf%8.3lf  1.00%6.2lf\n",
          natom,rname,ch,pr->Rnum[r],X[0],X[1],X[2],T);
  }

  fclose(fp);

} /* end of Output_SSE_PDB() */

