/*

 <StrLibL.c>

 for reading and allocation linked-list structure library
 
 coded by Takeshi Kawabata.

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "main.h" 
#include "protein.h" 
#include "StrLibL.h" 
#include "Amino.h" 
#include "LlistAlign.h" 


/*** FUNCTIONS (GLOBAL) ***/ 
void Read_Library_List();
void Check_Library_List();
void Check_Library_List_MaxNaa();


void Read_Library_List(fname,head,Nlib,max_naa)
 char *fname;
 struct LibNode  *head;
 int  *Nlib,*max_naa;
{
 FILE *fp;
 int  L;
 char buffer[128],name[128],tax[128],buff[26],sym;
 struct LibNode *bn;
 
 /*
 <FORMAT>
 [bsspid] [taxonomy]
 1a02N b.2.5 - b.1.1
 1a02F h.1.3
 1a04A c.23.1 - a.4.6
 1a0aA a.38.1
 1a0cA c.1.15
 : 
 1a17- a.118.8
 1a1hA g.37.1 - g.37.1 - g.37.1
 #MAXLENGTH  151
 */
 
 fp = fopen(fname,"r");
 if (fp==NULL){fprintf(stderr,"#ERROR:Can't open Library List File \"%s\"\n",fname); exit(1);}
 
 *Nlib = *max_naa = 0;

 bn = head; 
 while (feof(fp)==0)
 {
  buffer[0] = '\0';
  fgets(buffer,128,fp);
  /* printf("%s",buffer); */
  L = strlen(buffer);

  /* If line is longer than 128, move fp to the end of line */
  if (buffer[L-1]!='\n')
  { sym = ' ';
    while ((feof(fp)==0)&&(sym!='\n')) sym = fgetc(fp); } 
 
  if ((buffer[0]!='#')&&(L>2)) 
  { if (buffer[L-1]=='\n') buffer[L-1] = '\0';
    name[0] = tax[0] = '\0';
    /* 
    sscanf(buffer,"%s %s",name,tax); 
    */
    Get_First_Field(name,tax,buffer);
    bn->next = (struct LibNode*)malloc(sizeof(struct LibNode)); 
    bn->next->prev = bn;
    bn = bn->next; 
    bn->Name = (char *)malloc(sizeof(char) * (strlen(name)+1));
    bn->Tax = (char *)malloc(sizeof(char) * (strlen(tax)+1));
    sprintf(bn->Name,"%s",name); 
    sprintf(bn->Tax,"%s",tax); 
    bn->num = *Nlib; 
    ++(*Nlib);
    bn->next = NULL;

    }
  if (strncmp(buffer,"#MAXLENGTH",10)==0) { sscanf(buffer,"%s %d",buff,max_naa); }

 } 
 
 fclose(fp);

} /* end of Read_Library_List() */


void Check_Library_List(head)
 struct LibNode  *head;
{
 struct LibNode *bn;
 FILE *fp; 
 char Fname[256],dir[8];

 bn = head;

 while (bn->next != NULL)
 {
  bn = bn->next;
 
 sprintf(Fname,"%s.bssp",bn->Name);
 fp = fopen(Fname,"r");
 if (fp==NULL) {
  Get_Part_Of_Line(dir,bn->Name,1,2);
  sprintf(Fname,"%s/%s/%s.bssp",BSSP_DIR,dir,bn->Name);
  fp = fopen(Fname,"r");
  if (fp ==NULL) 
   {fprintf(stderr,"#ERROR: Can't find bsspfile \"%s.bssp\" and \"%s\"\n",bn->Name,Fname);    exit(1);}
  }

 fclose(fp);
 } 
 
} /* end of Check_Liblary_List() */


void Check_Library_List_MaxNaa(head,MaxNline)
 struct LibNode  *head;
 int  *MaxNline;
{
 struct LibNode *bn;
 FILE *fp; 
 char Fname[256],dir[8];
 char Line[1000];
 int Nline;
 bn = head;

 while (bn->next != NULL)
 {
  bn = bn->next;
 
 sprintf(Fname,"%s.bssp",bn->Name);
 fp = fopen(Fname,"r");
 if (fp==NULL) {
  Get_Part_Of_Line(dir,bn->Name,1,2);
  sprintf(Fname,"%s/%s/%s.bssp",BSSP_DIR,dir,bn->Name);
  fp = fopen(Fname,"r");
  if (fp ==NULL) 
   {fprintf(stderr,"#ERROR: Can't find bsspfile \"%s.bssp\" and \"%s\"\n",bn->Name,Fname);    exit(1);}
  }

  Nline = 0;
  while (feof(fp)==0)
  { Line[0] = '\0';
    fgets(Line,999,fp);
    if (Line[0]!='\0') ++Nline; }
  
 if (Nline> *MaxNline) *MaxNline = Nline;

 fclose(fp);
 } 
 
} /* end of Check_Liblary_List_MaxNaa() */


