////////////////////////////////////////////////////////////////////////////////
// File: inner_product.c                                                      //
// Routines:                                                                  //
//    Inner_Product                                                           //
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//  double Inner_Product(double u[], double v[], int n)                       //
//                                                                            //
//  Description:                                                              //
//     Given a Real vector space V with an orthonormal basis, the inner       //
//     product of two vectors in V is the sum of the component-wise products  //
//     i.e. for u = (u[0],...,u[n-1]) and v = (v[0],...,v[n-1]) in V,         //
//               <u,v> =  u[0]*v[0] + ... + u[n-1]*v[n-1].                    //
//                                                                            //
//  Arguments:                                                                //
//     double u[]  Pointer to the first element of the vector u.              //
//     double v[]  Pointer to the first element of the vector v.              //
//     int     n   The number of components of the vectors u and v.           //
//                                                                            //
//  Return Values:                                                            //
//     Inner Product of u and v.                                              //
//                                                                            //
//  Example:                                                                  //
//     #define N                                                              //
//     double u[N], v[N], inner_product;                                      //
//                                                                            //
//     (your code to intialize the vectors u and v)                           //
//     inner_product = Inner_Product(u,v,N);                                  //
//     printf(" <u,v> = %12.6f\n", inner_product);                            //
//           ...                                                              //
////////////////////////////////////////////////////////////////////////////////
//                                                                            //
double Inner_Product(double u[], double v[], int n)
{
   double inner_product = 0.0;
   int i;

   for (i = 0; i < n; i++) inner_product +=  u[i] * v[i];

   return inner_product;
}
