/////////////////////////////////////////////////////////////////////////////
//
//  Lapack/Blas C++ matrix wrapper class  (MS VC6.0 compiler)
//  (C) R. C. Wilson
//  Dec 2002
/////////////////////////////////////////////////////////////////////////////



#include <iostream>
#include <iomanip>
#include <typeinfo>
#include <math.h>

extern "C" {

#include <string.h>
#include <stdlib.h>
#include "f2c.h" 
#include "blaswrap.h"
/*extern "C" dgemm_();
extern "C" sgemm_();
extern "C" cgemm_();
extern "C" zgemm_();
*/
#include "fblaswr.h"
#include "clapack.h"
}

typedef complex singlecomplex;

using namespace std;

#include "cl_complex.h"

#define __max(x,y) (((x)<(y))?(y):(x))
#define __min(x,y) (((x)>(y))?(y):(x))


// use singlecomplex or doublecomplex for clapack interface
// use ccomplex and zcomplex for matrix interface
typedef cl_complex<float> ccomplex;
typedef cl_complex<double> zcomplex;

/*
struct ccomplex : public complex {
	ccomplex& operator=(const float u) { r=u;i=0;return *this; };
};

struct zcomplex : public doublecomplex {
	zcomplex() { r=i=0.0; };
	zcomplex(double u) { r=u;i=0; };
	zcomplex(double u,double v) { r=u;i=v; };
	zcomplex(const zcomplex& u) { r=u.r;i=u.i; };
	zcomplex& operator=(double u) { r=u;i=0;return *this; };
	zcomplex& operator=(zcomplex u) { r=u.r;i=u.i;return *this; };
	zcomplex operator*(zcomplex u) { return zcomplex(r*u.r-i*u.i,r*u.i+i*u.r);
	};
	zcomplex operator*(double u) { return zcomplex(r*u,i*u); };
	friend zcomplex operator*(double u,zcomplex v) { return zcomplex(v.r*u,v.i*u); };
};
*/

#define Vector Matrix

template<class Type> class Matrix {

public:
	Type **A;
	integer r,c;

	Matrix(int a,int b=1);
	Matrix();
	Matrix(const Matrix<Type>& B);
	~Matrix();
	void resize(int a,int b);
	void	dump(ostream&);
	void 	printgraph(ostream&);

	Type&	elem(int i,int j) const { return A[j][i]; }
	Type&  operator()(int i,int j) const { return A[j][i]; }
	Type&  operator()(int i) const { return A[0][i]; }
	Matrix<Type>& operator=(const Matrix<Type>&);
	Matrix<Type>& operator=(Type);
	bool operator!=(Matrix<Type>&);
	bool operator==(Matrix<Type>&);
	bool operator>(Matrix<Type>&);
	bool operator<(Matrix<Type>&);
	Matrix<Type>		operator+(const Matrix<Type>&);
	Matrix<Type>		operator*(const Matrix<Type>&);
//	friend Matrix<Type>	operator*(const Type,Matrix<Type> const &);
	Matrix<Type>		operator-(const Matrix<Type>&);
	Matrix<Type>		operator-();

	Matrix<Type> row(int n);
	Matrix<Type> col(int n);
};



template<class Type> Matrix<Type>::Matrix()
{
	r=c=0;
	A=NULL;
}

template<class Type> Matrix<Type>::Matrix(const Matrix& B)
{
	r=c=0;
	A=NULL;
	*this=B;
}

template<class Type> Matrix<Type>::Matrix(int a,int b)
{
	r=a;
	c=b;
	A=new Type*[c];
	A[0]=new Type[r*c];
	for(int i=1;i<c;i++)
		A[i]=A[0]+i*r;
}

template<class Type> Matrix<Type>::~Matrix()
{
	if( A ) {
		delete[] A[0];
		delete[] A;
	}
	A=NULL;
}

template<class Type> void Matrix<Type>::resize(int a,int b)
{
	if( a==r && b==c ) return;
	if( A ) {
		delete A[0];
		delete A;
	}
	r=a;
	c=b;
	A=new Type*[c];
	A[0]=new Type[r*c];
	for(int i=1;i<c;i++)
		A[i]=A[0]+i*r;
}

////////////////////////////////////////////////////////////////////

template<class Type> Matrix<Type> Matrix<Type>::row(int n)
{
	Matrix<Type> R(1,c);
	int i;
	for(i=0;i<c;i++)
		R(0,i)=A[i][n];

	return R;
}

template<class Type> Matrix<Type> Matrix<Type>::col(int n)
{
	Matrix<Type> R(r,1);
	memcpy(R.A[0],A[n],c*sizeof(Type));

	return R;
}




/////////////////////////  Maths operators  //////////////////////////////////

template<class Type> Matrix<Type>& Matrix<Type>::operator=(const Matrix<Type>& B)
{
	if( B.r!=r || B.c!=c )
		resize(B.r,B.c);
	memcpy(A[0],B.A[0],r*c*sizeof(Type));
	return *this;
}


template<class Type> Matrix<Type>& Matrix<Type>::operator=(Type v)
{
	if( v==0.0 )
		memset(A[0],0,sizeof(Type)*r*c);
	else {
		int i,j;
		for(i=0;i<r;i++)
			for(j=0;j<c;j++)
				A[j][i]=v;
	}

	return *this;
}

template<class Type> Matrix<Type> Matrix<Type>::operator+(const Matrix<Type>& C)
{
	if( r!=C.r || c!=C.c )
		throw("Matrices are different sized in add.");

	Matrix<Type> res=*this;

	integer i,j,d=1;
	Type alpha=1.0;

	if( typeid(Type)==typeid(double) ) 
		for(i=0;i<c;i++)
			daxpy_((integer*)&res.r,(double*)&alpha,(double*)C.A[i],&d,(double*)res.A[i],&d);
	else if( typeid(Type)==typeid(float) ) 
		for(i=0;i<c;i++)
			saxpy_((integer*)&res.r,(float*)&alpha,(float*)C.A[i],&d,(float*)res.A[i],&d);
	else if( typeid(Type)==typeid(ccomplex) ) 
		for(i=0;i<c;i++)
			caxpy_((integer*)&res.r,(complex*)&alpha,(complex*)C.A[i],&d,(complex*)res.A[i],&d);
	else if( typeid(Type)==typeid(zcomplex) ) 
		for(i=0;i<c;i++)
			zaxpy_((integer*)&res.r,(doublecomplex*)&alpha,(doublecomplex*)C.A[i],&d,(doublecomplex*)res.A[i],&d);
	else 
		for(i=0;i<c;i++)
			for(j=0;j<r;j++)
				res(i,j)+=C(i,j);

	return res;
}

template<class Type> Matrix<Type> Matrix<Type>::operator-(Matrix<Type> const & C)
{
	if( r!=C.r || c!=C.c )
		throw("Matrices are different sized in add.");

	Matrix<Type> res=*this;

	integer i,j,d=1;
	Type alpha=-1.0;

	if( typeid(Type)==typeid(double) ) 
		for(i=0;i<c;i++)
			daxpy_((integer*)&res.r,(double*)&alpha,(double*)C.A[i],(integer*)&d,(double*)res.A[i],(integer*)&d);
	else if( typeid(Type)==typeid(float) ) 
		for(i=0;i<c;i++)
			saxpy_((integer*)&res.r,(float*)&alpha,(float*)C.A[i],(integer*)&d,(float*)res.A[i],(integer*)&d);
	else if( typeid(Type)==typeid(ccomplex) ) 
		for(i=0;i<c;i++)
			caxpy_((integer*)&res.r,(complex*)&alpha,(complex*)C.A[i],(integer*)&d,(complex*)res.A[i],(integer*)&d);
	else if( typeid(Type)==typeid(zcomplex) ) 
		for(i=0;i<c;i++)
			zaxpy_((integer*)&res.r,(doublecomplex*)&alpha,(doublecomplex*)C.A[i],(integer*)&d,(doublecomplex*)res.A[i],(integer*)&d);
	else 
		for(i=0;i<r;i++)
			for(j=0;j<c;j++)
{
                                printf("In ope- %lf c=%d r=%d \n",res(i,j),c,r);
                                printf("In ope- %lf \n",C(i,j));
				res(i,j)-=C(i,j);
                        }
	return res;
}

template<class Type> Matrix<Type> Matrix<Type>::operator-()
{
	Matrix<Type> res(r,c);
	memset(res.A[0],0,res.r*res.c*sizeof(Type));

	integer i,j,d=1;
	Type alpha=-1.0;


	if( typeid(Type)==typeid(double) ) 
		for(i=0;i<c;i++)
			daxpy_((integer*)&res.r,(double*)&alpha,(double*)A[i],&d,(double*)res.A[i],&d);
	else if( typeid(Type)==typeid(float) ) 
		for(i=0;i<c;i++)
			saxpy_((integer*)&res.r,(float*)&alpha,(float*)A[i],&d,(float*)res.A[i],&d);
	else if( typeid(Type)==typeid(ccomplex) ) 
		for(i=0;i<c;i++)
			caxpy_((integer*)&res.r,(singlecomplex*)&alpha,(singlecomplex*)A[i],&d,(singlecomplex*)res.A[i],&d);
	else if( typeid(Type)==typeid(zcomplex) ) 
		for(i=0;i<c;i++)
			zaxpy_((integer*)&res.r,(doublecomplex*)&alpha,(doublecomplex*)A[i],&d,(doublecomplex*)res.A[i],&d);
	else 
		for(i=0;i<c;i++)
			for(j=0;j<r;j++)
				res(i,j)=-elem(i,j);

	return res;
}

template<class Type> Matrix<Type> Matrix<Type>::operator*(const Matrix<Type>& C)
{
	if( c!=C.r )
		throw("Matrices are incompatible sizes in multiply.");
	Matrix<Type> res(r,C.c);
	memset(res.A[0],0,res.r*res.c*sizeof(Type));

	int i,j,k;
	Type alpha=1.0;
	Type beta=0.0;

	if( typeid(Type)==typeid(double) ) 
		dgemm_("N","N",(integer*)&res.r,(integer*)&res.c,(integer*)&c,(double*)&alpha,(double*)A[0],(integer*)&r,(double*)C.A[0],(integer*)&C.r,(double*)&beta,(double*)res.A[0],(integer*)&res.r);
	else if( typeid(Type)==typeid(float) ) 
		sgemm_("N","N",(integer*)&res.r,(integer*)&res.c,(integer*)&c,(float*)&alpha,(float*)A[0],(integer*)&r,(float*)C.A[0],(integer*)&C.r,(float*)&beta,(float*)res.A[0],(integer*)&res.r);
	else if( typeid(Type)==typeid(ccomplex) ) 
		cgemm_("N","N",(integer*)&res.r,(integer*)&res.c,(integer*)&c,(singlecomplex*)&alpha,(singlecomplex*)A[0],(integer*)&r,(singlecomplex*)C.A[0],(integer*)&C.r,(singlecomplex*)&beta,(singlecomplex*)res.A[0],(integer*)&res.r);
	else if( typeid(Type)==typeid(zcomplex) ) 
		zgemm_("N","N",(integer*)&res.r,(integer*)&res.c,(integer*)&c,(doublecomplex*)&alpha,(doublecomplex*)A[0],(integer*)&r,(doublecomplex*)C.A[0],(integer*)&C.r,(doublecomplex*)&beta,(doublecomplex*)res.A[0],(integer*)&res.r);
	else 
		for(i=0;i<res.r;i++)
			for(j=0;j<res.c;j++)
				for(k=0;k<c;k++)
					res(i,j)+=elem(i,k)*C(k,j);

	return res;
}

template<class Type> Matrix<Type> operator*(const Type v,const Matrix<Type>& B)
{
	Matrix<Type> res(B.r,B.c);
	int i,j;
	for(i=0;i<res.r;i++)
		for(j=0;j<res.c;j++)
				res(i,j)=v*B(i,j);

	return res;
}

template<class Type> Matrix<Type> Transpose(Matrix<Type>& B)
{
	Matrix<Type> C(B.c,B.r);
	int i,j;
	for(i=0;i<B.r;i++)
		for(j=0;j<B.c;j++) {
			C(j,i)=B(i,j);
		}
	return C;
}

/////////////////////////  Relational  ///////////////////////////////////////

template<class Type> bool Matrix<Type>::operator==(Matrix<Type>& M)
{
	int i;
	for(i=0;i<r*c;i++)
		if( M.A[0][i]!=A[0][i] ) return false;
	return true;
}

template<class Type> bool Matrix<Type>::operator!=(Matrix<Type>& M)
{
	int i;
	for(i=0;i<r*c;i++)
		if( M.A[0][i]!=A[0][i] ) return true;
	return false;
}

template<class Type> bool Matrix<Type>::operator>(Matrix<Type>& M)
{
	int i;
	for(i=0;i<r*c;i++) {
		if( A[0][i]>M.A[0][i] ) return true;
		if( A[0][i]<M.A[0][i] ) return false;
	}
	return false;
}

template<class Type> bool Matrix<Type>::operator<(Matrix<Type>& M)
{
	int i;
	for(i=0;i<r*c;i++) {
		if( A[0][i]<M.A[0][i] ) return true;
		if( A[0][i]>M.A[0][i] ) return false;
	}
	return false;
}


/////////////////////////  Input/Output  ///////////////////////////////////////


template<class Type> ostream& operator<<(ostream &s, Matrix<Type>& A)
{

	s<<A.r<<" "<<A.c<<endl;
    for(int i=0;i<A.r;i++) {
        for(int j=0;j<A.c;j++)
            s <<setw(10)<<setprecision(6)<<" "<<A(i,j);
        s << "\n";
    }
    return s;
}


template<class Type> istream& operator>>(istream &s, Matrix<Type>& A)
{
    int a,b;
    s>>a>>b;

	A.resize(a,b);
    for(int i=0;i<a;i++)
        for(int j=0;j<b;j++)
            s>>A(i,j);
    return s;
}


inline void dump_element(ostream& out,double a)
{
	out<<setw(10)<<setprecision(6)<<a<<" ";
}

inline void dump_element(ostream& out,zcomplex a)
{
	out<<setw(10)<<setprecision(6)<<a.r<<" "<<a.i<<" ";
}

template<class Type> void Matrix<Type>::dump(ostream& out)
{
    for(int i=0;i<r;i++) {
        for(int j=0;j<c;j++)
			dump_element(out,A[j][i]);
        out << "\n";
    }
}

template<class Type> void Matrix<Type>::printgraph(ostream& out)
{
    for(int i=0;i<r;i++) {
        for(int j=0;j<c;j++)
			out<<(A[j][i]==1?'1':'0');
        out << "\n";
    }
}

///////////////////////  Eigendecomposition ////////////////////////////////////

inline Matrix<double> EigenDecomposition(Matrix<double>& A,double* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<double> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	double* workspace=new double[lwork];

	dsyev_("V","U",&cols,C.A[0],&rows,eigenvalues,workspace,&lwork,&info);

	delete[] workspace;

	return C;
}


inline Matrix<float> EigenDecomposition(Matrix<float>& A,float* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<float> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	float* workspace=new float[lwork];

	ssyev_("V","U",&cols,C.A[0],&rows,eigenvalues,workspace,&lwork,&info);

	delete[] workspace;

	return C;
}

inline Matrix<ccomplex> EigenDecomposition(Matrix<ccomplex>& A,float* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<ccomplex> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	ccomplex* workspace=new ccomplex[lwork];
	float* rwork=new float[3*A.c-2];


cheev_("V","U",&cols,(singlecomplex*)C.A[0],&rows,(float*)eigenvalues,(singlecomplex*)workspace,&lwork,(float*)rwork,&info);

	delete[] workspace;
	delete[] rwork;

	return C;
}

inline Matrix<zcomplex> EigenDecomposition(Matrix<zcomplex>& A,double* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<zcomplex> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	zcomplex* workspace=new zcomplex[lwork];
	double* rwork=new double[3*A.c-2];


zheev_("V","U",&cols,(doublecomplex*)C.A[0],&rows,(double*)eigenvalues,(doublecomplex*)workspace,&lwork,(double*)rwork,&info);

	delete[] workspace;
	delete[] rwork;

	return C;
}


// Eigenvalues

inline void Eigenvalues(Matrix<double>& A,double* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<double> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	double* workspace=new double[lwork];

	dsyev_("N","U",&cols,C.A[0],&rows,eigenvalues,workspace,&lwork,&info);

	delete[] workspace;
}


inline void Eigenvalues(Matrix<float>& A,float* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<float> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	float* workspace=new float[lwork];

	ssyev_("N","U",&cols,C.A[0],&rows,eigenvalues,workspace,&lwork,&info);

	delete[] workspace;
}

inline void Eigenvalues(Matrix<ccomplex>& A,float* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<ccomplex> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	ccomplex* workspace=new ccomplex[lwork];
	float* rwork=new float[3*A.c-2];


cheev_("N","U",&cols,(singlecomplex*)C.A[0],&rows,(float*)eigenvalues,(singlecomplex*)workspace,&lwork,(float*)rwork,&info);

	delete[] workspace;
	delete[] rwork;
}

inline void Eigenvalues(Matrix<zcomplex>& A,double* eigenvalues)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");

	Matrix<zcomplex> C=A;

	integer lwork,info=0,cols=A.c,rows=A.r;
	lwork=32;
	lwork=(lwork+2)*A.c;
	zcomplex* workspace=new zcomplex[lwork];
	double* rwork=new double[3*A.c-2];


zheev_("N","U",&cols,(doublecomplex*)C.A[0],&rows,(double*)eigenvalues,(doublecomplex*)workspace,&lwork,(double*)rwork,&info);

	delete[] workspace;
	delete[] rwork;
}

/////////////////////////////  SVD  //////////////////////////////////////


inline Matrix<double> SingularValueDecomposition(Matrix<double>& A,double* singularvalues)
{

	Matrix<double> U(A.r,A.r);
	Matrix<double> V(A.c,A.c);
	Matrix<double> C;
	C=A;


	integer lwork,info=0,rows=A.r,cols=A.c;
	integer mind=min(A.r,A.c);
	int maxd=max(A.r,A.c);
	lwork=2*(4*mind*mind+maxd+9*mind);
	double* workspace=new double[lwork];
	integer* iwork=new integer[8*mind];


dgesdd_("A",&rows,&cols,C.A[0],&rows,singularvalues,U.A[0],&rows,V.A[0],&cols,workspace,&lwork,iwork,&info);

	delete[] workspace;
	delete[] iwork;

	return Transpose(V);
}

/////////////////////////////  Inverse  //////////////////////////////////////


inline double LogDet(Matrix<double>& M)
{
	if( M.r!=M.c )
		throw("Non-square matrix in Det");

	Matrix<double> MI=M;

	integer info,i,rows=MI.r,cols=MI.c;
	dpotrf_("U",&rows,MI.A[0],&cols,&info);
	double det=0.0;
	for(i=0;i<M.r;i++)
		det+=log(MI(i,i));

	return det;
}

inline Matrix<double> InverseSymmetricPositiveDefinite(const Matrix<double>& M)
{
	if( M.r!=M.c )
		throw("Non-square matrix in InverseSymmetricPositiveDefinite");

	Matrix<double> MI=M;
	integer info,rows=MI.r,cols=MI.c;


	dpotrf_("U",&rows,MI.A[0],&cols,&info);
	dpotri_("U",&rows,MI.A[0],&cols,&info);

	int i,j;
	for(j=0;j<MI.c;j++)
		for(i=j+1;i<MI.r;i++)
			MI(i,j)=MI(j,i);

	return MI;
}

inline Matrix<double> ConditionedInverseSymmetric(Matrix<double>& M,double cut)
{
	if( M.r!=M.c )
		throw("Non-square matrix in ConditionedInverseSymmetric");

	Matrix<double> U,E(M.c,M.c);
	double* ev=new double[M.c];

	U=EigenDecomposition(M,ev);

	E=0.0;
	for(int i=0;i<M.c;i++)
		if( fabs(ev[i])<cut ) E(i,i)=0.0;
		else E(i,i)=1.0/ev[i];

	delete[] ev;

	return U*E*Transpose(U);
}

// double non-symmetric matrix A

inline void NSEigenvalues(Matrix<double>& A,double* eigenvaluesr,double* eigenvaluesi)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");
	Matrix<double> C=A;
	integer lwork,info=0,rows=A.r,cols=A.c;
	integer one=1;
	lwork=32;
	lwork=(lwork+2)*A.c;
	double* workspace=new double[lwork];
	
	dgeev_("N","N",&cols,C.A[0],&rows,eigenvaluesr,eigenvaluesi,NULL,&one,NULL,&one,workspace,&lwork,&info);
	
	delete[] workspace;
}

// float non-symmetric matrix A

inline void NSEigenvalues(Matrix<float>& A,float* eigenvaluesr,float* eigenvaluesi)
{
	if( A.c!=A.r )
		throw("Non-square matrix in EigenDecomposition.");
	Matrix<float> C=A;
	integer lwork,info=0,rows=A.r,cols=A.c;
	integer one=1;
	lwork=32;
	lwork=(lwork+2)*A.c;
	float* workspace=new float[lwork];
	
	sgeev_("N","N",&cols,C.A[0],&rows,eigenvaluesr,eigenvaluesi,NULL,&one,NULL,&one,workspace,&lwork,&info);
	
	delete[] workspace;
}


inline Matrix<double> Solve(Matrix<double>& A,Matrix<double>& B)
{
	Matrix<double> X,Y;
	X=B;
	Y=A;
	integer info,one=1,rows=Y.r,solns=X.r;
	integer* ipiv=new integer[A.r];
	dgesv_(&rows,&one,Y.A[0],&rows,ipiv,X.A[0],&solns,&info);
	delete[] ipiv;
	if(info>0) {
		// singular
		double* svs=new double[__max(Y.r,Y.c)];
		double cond=1e-6;
		integer rank,lwork;
		double* work;
		lwork=10*__max(Y.r,Y.c);
		work=new double[lwork];
		Y=A;X=B;
		dgelss_(&rows,&rows,&one,Y.A[0],&rows,X.A[0],&solns,svs,&cond,&rank,work,&lwork,&info);
		delete[] work;
		if( info!=0 ) throw("SVD failed in solve.");
	}
	return X;
}
